\input LiEwebmac

\nocon

@* Handling subsystems of roots.
This file deals with various aspects of subsystems of root systems: finding
fundamental bases for them, classifying their type, and computing
restriction matrices. For the largest part the algorithms treat subsystems
as abstract root systems, but sometimes use is made of the fact that the
roots are chosen from a root system for which certain standardisations hold;
for instance norms of roots can only be $1$,~$2$, or~$3$, and there is a
predefined concept of positivity of roots.

On several occasions the classification of all simple groups tells us that
certain conditions can never occur. If proper progress of an algorithm
depends on such a condition being satisfied and it is easy to test, we will
occasionally include a call of~|error| in the impossible case, which is
directed to programmers of \LiE., rather than to its user; the corresponding
error messages are usually rather terse.

A local variable~|s| is set to the semisimple rank of the group under
consideration whenever we enter one of the non-local functions defined in this
file.

@h "lie.h"
@d local static
@f local static

@c
@< Function prototypes @>@;

local index s; /* the current semisimple rank */

@< Auxiliary functions @>@;

@ The following function implements |Cartan_type|, |fundam|, and |closure|.
The first two of these perform essentially the same computation but deliver
different parts of its outcome, while |closure| is an extension of |fundam|.
In order to deliver different types of results a bit of trickery is needed.
The stated result type is that of |fundam| and |closure|, but there is also
an argument |group* lie_type@;| in which the result for |Cartan_type| can be
recorded. The convention is that |fundam| and |closure| call the function
with |lie_type==NULL|, in which case the computed matrix is delivered as
result, but |Cartan_type| makes |lie_type| point to a sufficiently large
|group| in which the computed type will be recorded, while the call returns
|NULL| in that case. The distinction between |fundam| and |closure| is made
by the boolean parameter |close|, which is however only relevant in the
presence of roots of unequal length.

Internally the function uses |tp| in place of |lie_type|, with |tp!=NULL|,
because the type is needed during the computation, even if it is not required
afterwards. The matrix~|m| is copied at the beginning to avoid side effects to
the parameter, and if |lie_type==NULL|, it is copied again at the end, this
time to conserve memory, since a number of rows may have been discarded by the
algorithm (no rows are ever added).

@c
matrix* Closure(matrix* m, boolean close, group* lie_type)
{ matrix* result; @+ index i,j;
  group* tp=(s=Ssrank(grp), lie_type==NULL ? mkgroup(s) : lie_type);

  tp->toraldim=Lierank(grp); tp->ncomp=0; /* start with maximal torus */
  m=copymatrix(m);
  @< Set |close=false| if all roots of |grp| have the same length @>
  @< Check that all rows of |m| represent roots, and make them positive @>
  { index next;
    for (i=0; i<m->nrows; i=next)
    @< Isolate a connected component of roots, convert it into the positive
       basis of fundamental roots of the smallest subsystem that contains it,
       and that if |close==true| is also additively closed within the root
       system of~|grp|,
       and make |next| point to the root after the component @>
  }
  if (lie_type==NULL)
    return result=copymatrix(m),freemem(m),freemem(tp),result;
  else return freemem(m),(matrix*)NULL; /* |Cartan_type| doesn't need |m| */
}

@ Here is how |Closure| is called when the user asks for |Cartan_type|.

@c
group* Carttype(matrix* m)
{ group* type=mkgroup(s=Ssrank(grp)); /* rank bounds number of components */
  Closure(m,false,type); return type;
}


@ The code here is skipped if |fundam| is being computed, and it
reduces |closure| to |fundam| if all simple components of |grp| have types
$A_n$,~$D_n$, or~$E_n$.

@h <string.h>
@d two_lengths(type) (strchr("BCFG",type)!=NULL)

@< Set |close=false| if all roots of |grp| have the same length @>=
if (close)
  if (type_of(grp)==SIMPGRP) close = two_lengths(grp->s.lietype);
  else
  { for (i=0; i<grp->g.ncomp; i++)
      if (two_lengths(Liecomp(grp,i)->lietype)) break;
    close= i<grp->g.ncomp;
  }


@ This piece of code ensures that the auxiliary functions do not have to
worry about the presence of other things than positive roots. We call
|Unique| to reduce each pair of opposite roots in the input to a single root.

@< Check that all rows of |m| represent roots, and make them positive @>=
{ entry* t;
  for (i=0; i<m->nrows; i++)
  if (!isroot(t=m->elm[i]))
    error("Set of root vectors contains a non-root\n");
  else if (!isposroot(t=m->elm[i]))
    for (j=0; j<m->ncols; j++) t[j]= -t[j]; /* make positive root */
  Unique(m,cmpfn); 
}

@ The main work of |Closure| is done by auxiliary functions. The call to
|isolcomp| gathers a connected component of roots into the positions
from~|i| up to~|next| (it is defined in the file \.{centr}), and |fundam|
reduces it to a fundamental basis (taking due care of filling any empty
space left by decrease of the number of roots). When closure is necessary,
|long_close| makes the modification and |fundam| restores the fundamental
basis property (in this case no roots are removed, so |next| does not change
either). Then |simp_type| determines the type of this fundamental basis,
reordering the roots according to the standard labeling for a diagram of
this type. Since we want the result of |Cartan_type| to be completely
independent of the order of the roots in the input, we sort the resulting
simple types into a fixed order.

@< Isolate... @>=
{ index d,n=0; @+ simpgrp* c;
  next=isolcomp(m,i);
  fundam(m,i,&next);
  if (close) long_close(m,i,next),fundam(m,i,&next);
  c=simp_type(&m->elm[i],d=next-i);
  @< Insert the simple type |c| into |tp|, keeping it in weakly deacreasing
     order, and insert its |d| roots, which are currently at positions |i| up
     to~|next|, at the corresponding place among the previous roots @>
}

@ The order used to sort simple types is given by |grp_less|: it gives
precedence to the family of the type over its rank. The type is inserted
into |tp| by a simple insertion sort, and a count of the total rank of the
types that move up is kept in~|n|. The roots corresponding to the new
component are then moved into place by shifting them down across |n|
previous roots; this is achieved by a call to |cycle_block|.

@d grp_less(x,y)
 ((x)->lietype<(y)->lietype
      || (x)->lietype==(y)->lietype && (x)->lierank<(y)->lierank)

@< Insert the simple type |c| into |tp|... @>=
{ j=tp->ncomp++;
  while(--j>=0 && grp_less(tp->liecomp[j],c))
    n += (tp->liecomp[j+1]=tp->liecomp[j])->lierank;
  tp->liecomp[++j]=c; tp->toraldim -= d;
    /* insert component and remove rank from torus */
  cycle_block(m,i-n,next,n);
    /* move the |d| rows down across |n| previous rows */
}

@* Classification of subsystems of roots.
The function |simp_type| performs the analysis of the set of fundamental
roots corresponding to a connected component of the Dynkin diagram. When it
is called the roots have been preprocessed so that they are linearly
independent, all inner products are non-positive, and they have been divided
into minimal mutually othogonal classes, which have been grouped into sets of
consectutive rows of the matrix. The parameter~|m| of |simp_grp| points to
the first row of the class of roots under consideration, and |n| specifies
the number of roots in that class. The task of |simp_type| is to find the
type of this simple component, which it returns as its result, and also to
reorder the roots within this component according to the standard labeling
for the Dynkin diagram of this simple type.

@< Auxiliary functions @>=
local simpgrp* simp_type(entry** m, entry n)
{ matrix* adjs=mkmatrix(n,3);
  entry** adj=adjs->elm /* |adj[i]| lists up to 3 neighbours of node |i| */
  ,* norm=mkintarray(3*n) /* norms of roots */
  ,* valency=&norm[n] /* valencies in Dynkin diagram */
  ,* p=&valency[n]; /* permutation of |n| */
  simpgrp* result;
  index i,j,k, a_val[4]={-1,-1,-1,-1};
    /* |a_val[i]| is index of a node of valency |i|, if any */

  if (n==0) error("empty input in simp_type\n");
  @< Compute |adj|, |norm|, |valency|, and |a_val| @>
  if (a_val[3]<0)
  @< Treat the case of a linear diagram @> /* no nodes of valency 3 */
  else @< Treat the case of a non-linear diagram @>
  @< Permute the rows @>

  freemem(adjs); freearr(norm);
  return result;
}

@ We start with computing the norms of the roots, their valencies and lists
of adjacencies, and a representative root of each valency that occurs. To
gather this information each inner product between a pair of roots is
computed once. The direction of the two nested loops below is slightly
subtle: there is a connection between the facts that the |i|'s are
decreasing and that we let |j<i|, and one should not be altered unless the
other is as well; the effect of this choice is that once |i| has been
handled in the outer loop, all its adjacencies have been noted, so that
|a_val| can be set without a loop of its own. When this code has been
executed the adjacencies of |i| listed in |adj[i]| are in decreasing order,
and |a_val[i]| is the minimal index among those of nodes with valency~|i|;
however the algorithm does not depend on this.

@< Compute |adj|... @>=
{ for (i=0;i<n;i++) valency[i]=0;
    /* |valency[i]| is also index of next slot in |adj[i]| */
  for (i=n; --i>=0;)
  { norm[i]= Norm(m[i]); /* where |Norm(x)==Inprod(x,x)/2| */
    for (j=i; --j>=0;)
      if (Inprod(m[i],m[j])!=0) /* then valencies increase */
      { if (valency[i]>=3 || valency[j]>=3) error ("valency >3 found\n");
	adj[i][valency[i]++]=j; adj[j][valency[j]++]=i;
	/* update valencies and adjacencies */
      }
    a_val[valency[i]]=i; /* valency of node |i| is now known */
  }
}

@ When a diagram has no branch points (nodes of valency~3) then the analysis
consists mainly of finding the permutation |p| that linearises the diagram
(i.e., nodes |p[i]| and |p[j]| are connected if and only if $\v i-j\v=1$), and
determining the type based on the norms of the end nodes |p[0]| and |p[n-1]|.
This is greatly facilitated by the precomputed arrays |a_val| and |adj|: an
end node has valency~1, so one can be found in |a_val[1]|, and linearisation
can be achieved by tracing subsequent neighbours, avoiding the ones that we
have already encountered. Two small points must be taken care of: a diagram of
type~$A_1$ has no end nodes in the sense above, and we might have started with
the ``wrong'' end node in the case of asymmetric diagrams.

@< Treat the case of a linear diagram @>=
{ index e; /* index of end node (|valency[e]<=1|) */
  if (a_val[0]>=0) p[0]=e=a_val[0]; /* must be type $A_1$ */
  else
  { if (a_val[1]>=0) p[0]=e=a_val[1]; /* other linear types */
    else error("no end node found\n");

    @< Linearise the diagram, assigning the consecutive indices to
       $p[1]\ldots p[n-1]$ @>
    @< If linearisation is incompatible with the standard labeling for this
       diagram, reverse it @>
  }
  @< Classify the linear diagram, based on the norms of the end
     nodes @>
}

@ Here, and in several places below, we wish to find the other neighbour of
some node~|b| of valency~2 when we have just reached it from its
neighbour~|a|. In other words we want to choose the element of
$\{\\{adj}[b][0],\\{adj}[b][1]\}\setminus\{a\}$. The macro |opposite| achieves
this in a slightly tricky way by subscripting |adj[b]| with a boolean value;
since 1 stands for ``true'' in \Cee, the boolean expression tells when to
choose |adj[b][1]|, and this is when |adj[b][0]==a|.

@d opposite(a,b) adj[b][adj[b][0]==(a)]
@< Linearise the diagram...@>=
{ k=p[1]=adj[e][0]; /* the unique neighbour of node |e| */
  for(i=2;i<n;i++) @+ p[i]=k=opposite(p[i-2],k); /* here |k==p[i-1]| */
}

@ Here a slight difficulty is that the proper ordering of the diagram depends
on its type, which we haven't yet completely determined. Yet the wrongly
ordered diagrams are simply described: the rank~2 diagrams whose pair of norms
are $(1,2)$ (i.e., $C_2$ which should be reversed to yield $B_2$) or $(3,1)$
(i.e., $G_2$ reversed), all diagrams of rank~$>2$ whose first two norms are
unequal, and the rank~4 diagram with ranks $(1,1,2,2)$ (i.e., $F_4$ reversed).
For the rank~2 case we use a bit of linear algebra; the pairs that could occur
and should fail this test are $(1,1)$, $(2,2)$, $(3,3)$ (all three of
type~$A_2$), $(2,1)$ and $(1,3)$ (the proper orderings for $B_2$ and $G_2$).

@< If linearisation is incompatible...@>=
if ( n==2 && norm[p[0]]+2*norm[p[1]]==5
  || n>=3 && norm[p[0]]!=norm[p[1]]
  || n==4 && norm[p[1]]<norm[p[2]]
   )
{ for (i=0; i<n-1-i; i++) swap(&p[i],&p[n-1-i]); e=p[0]; }

@ In this section, like in the previous one, we use that all norms lie in
|{1,2,3}|, since we are dealing with a subsystem of root system whose norms
have already been standardised.

@< Classify the linear... @>=
{ entry norm0=norm[p[0]], norm1=norm[p[n-1]];
  if (norm0==norm1) result = mksimpgrp('A',n);
  else if (norm1==3) result=mksimpgrp('G',2);
  else if (norm1==2) result=mksimpgrp('C',n);
  else if (norm0!=2) error("I don't recognize this Cartan Type\n");
  else if (n==4 && norm[p[2]]==1) result=mksimpgrp('F',4);
  else result=mksimpgrp('B',n);
}

@ A non-linear diagram must be of type $D_n$ or $E_n$. The analysis of these
cases starts at the branching node |a_val[3]|. We start by counting the end
nodes that are neighbours of this node, and recording them in an array~|end|.
If there is more than one such end node, the diagram has type~$D_n$, and
otherwise it has type~$E_n$.

@< Treat the case of a non-linear diagram @>=
{ entry* branch=adj[a_val[3]], end[3], end_count=0;
  for (j=2; j>=0; j--)
    if (valency[branch[j]]==1) end[end_count++]=branch[j];
  if (end_count>1) @< Treat the case of a type $D$ diagram @>
  else if (end_count==1) @< Treat the case of a type $E$ diagram @>
  else error("no end node adjacent to valency 3 node\n");
}

@ For diagrams of type $D_n$ we randomly assign two elements of |end| to
|p[n-1]| and |p[n-2]|, the branching node to |p[n-3]|, and its remaining
neighbour (which in case $D_4$ is an end node as well) to |p[n-4]|. Then we
linearise the remainder of the diagram in much the same way as for linear
diagrams. In this case there is no need for any reversal, since we always
start at the far end of the diagram, and the two (or three) short branches of
the diagram are symmetric, so like in the $A_n$ case, there is no reason to
prefer one ordering over another.

@< Treat the case of a type $D$ diagram @>=
{ p[n-1]=end[1]; p[n-2]=end[0]; p[n-3]=a_val[3];
  k=p[n-4]=branch[0]+branch[1]+branch[2]-p[n-1]-p[n-2];
    /* the remaining branch */
  for(i=n-5; i>=0; i--)
  { if (valency[k]!=2) error("unlinear Dn tail.\n");
    p[i]=k=opposite(p[i+2],k);
  }
  result=mksimpgrp('D',n);
}

@ When we come to this code, |end[0]| is the short branch of the diagram,
which will be assigned to |p[1]|. We start with searching for a branch of
length~2, whose nodes will be assigned to |p[2]| and |p[0]| respectively. From
that point on we proceed similarly to the $D_n$ case, except that the long
branch is now linearised in the order of increasing indices.

@< Treat the case of a type $E$ diagram @>=
{ p[3]=a_val[3]; p[1]=end[0];
  for (j=2; j>=0; j--)
    if (valency[branch[j]]==2)
      if (valency[opposite(a_val[3],branch[j])]==1) break;
  if (j<0) error("type E not recognised\n");
  p[2]=branch[j]; p[0]=opposite(p[3],p[2]);
  p[4]=k=branch[0]+branch[1]+branch[2]-p[1]-p[2]; /* remaining branch */
  for(i=5;i<n;i++)
  { if (valency[k]!=2) error("wrong type E system.\n");
    p[i]=k=opposite(p[i-2],k);
  }
  result=mksimpgrp('E',n);
}

@ Before |simp_type| returns, it permutes the rows of~|m| according to the
permutation~|p| that has been computed, i.e., each |m[i]| will be replaced by
|m[p[i]]| (by pointer assignment). The simplest way to do this is to make a
copy of |m| and then perform the assignments in a simple loop. This would
however require dynamically allocating another array of unknown size and
deallocating it afterwards; to avoid this we use a method that requires
constant memory and is only silghtly more complicated. We decompose |p| into
cycles, and reorder each cycle separately. After |m[i]| has been assigned its
new value we put |p[i]=-1| so that we can later recognise elements of cycles
that have already been treated.

@< Permute the rows @>=
for (i=0; i<n; i++) if (p[i]>=0) /* then |p[i]| starts an untreated cycle */
{ entry* mi=m[j=i]; /* record beginning of cycle */
  while (p[j]!=i)
    { k=j; j=p[j]; m[k]=m[j]; p[k]= -1; @+}
      /* assign |m[j]=m[p[j]]| and advance */
  m[j]=mi; p[j]= -1; /* close the cycle */
}

@* Finding a basis of fundamental roots.
The function |fundam| finds a basis of fundamental roots for minimal root
subsystem containing a given set of roots. The roots are those in the
matrix~|roots| ranging from row~|first| up to row~|*last|. When we enter
|fundam|, the containing simple group |grp| and its semisimple rank |s| have
been set, and all roots have been made positive; we are careful never to
reintroduce negative roots. Since the basis may contain fewer roots than
present in the given set, the number of rows may decrease, and |*last| and
|roots->nrows| will be updated to reflect the change; any rows from |*last|
upwards will be shifted down.

@< Function prototypes @>=
local void fundam(matrix* roots, index first, index* last);

@~In order to be able to use functions operating on matrices, like |Unique|, a
local |matrix mm@;| is declared, and its address assigned to |matrix* m@;|,
whose fields are set such that it is an alias for the rows of |roots| under
consideration, and any permutation of the rows of |mm| actually involves those
of |roots|. This is a bit tricky of course, and we must make sure that no
memory management functions can be called with~|m| as argument. Removing rows
from a matrix is done by calling |cycle_block|, which ensures that all its row
pointers, whether above or below |nrows|, are left pointing to distinct rows
(no aliasing among rows); although this is not vital here (since the matrix is
discarded or copied in |Closure| before returning) we stick to this more
prudent approach, as elsewhere in \LiE..

The method used by |fundam| is to repeatedly compare pairs of roots; each time
a positive inner product is found, we replace them by the fundamental basis
for the rank~2 subsystem they generate, that is positive within the full root
system. We repeat this until no more positive inner products are found. Each
time an improvement is made, the sum of the levels of all roots present
decreases, which guarantees termination. A slightly difficult point is when to
sort |m| to remove duplicates (i.e., when to call |Unique|): if this is
done too frequently it will consume most of the computation time, if it is
done too seldomly we risk doing much work double since while improving the
roots, we may create duplicate roots. We have chosen to reduce the matrix
every time we have compared a root with all previous roots and some change has
actually taken place; whenever this happens we lose track of the root which we
were working with, so we should reset the loop variable to its starting value.
Incidentally, it is not true that |root_i| which we are comparing to is
guaranteed to stay the same when changes are made; the decision to go on
comparing with other roots once a change has been made is just a heuristic
to get some more work done before sorting again.

@c
local void fundam(matrix* roots, index first, index* last)
{ index i,j,d; @+ boolean changed;
  entry* t=mkintarray(s); @+ matrix mm,* m=&mm;
  mm.elm=&roots->elm[first]; mm.nrows=*last-first; mm.ncols=roots->ncols;
  for (i=m->nrows-1; i>0; changed ? Unique(m,cmpfn),i=m->nrows-1 : --i)
  { entry* root_i=m->elm[i]; changed=false;
    @< Compare |root_i| with all roots |j| with |j<i|, and if possible improve
       the pair; if any improvement is made, set |changed=true| @>
  }
  cycle_block(roots,first+mm.nrows,roots->nrows,d=*last-first-mm.nrows);
@/*last-=d; roots->nrows-=d;
  freearr(t);
}

@ When |root_i| happens to be equal to another root we encounter, (this can
happen only if it has been altered during an earlier change) we simply remove
that other root; there is no need to set |changed=true| in such a case. For
all roots $\alpha$ we have $\<\alpha,\alpha>=2$, and since we are computing
the Cartan inner product anyway, this test is easier than a full equality
test (but it is not a sufficient condition for equality). The Cartan inner
product is computed with |root_i| in second position since this gives the
right coefficient for adjusting |root_j| by a multiple of |root_i|, which is
the preferred direction of modification.

@< Compare |root_i| with all roots |j| with |j<i|... @>=
for (j=i-1; j>=0; j--)
{ entry* root_j=m->elm[j]; entry c=Cart_inprod(root_j,root_i);
  if (c==2 && eqrow(root_j,root_i,s))
    @< Remove row |root_j| from position |j| in~|m| @>
  else if (c>0)
  { changed=true;
    @< Replace the pair of roots |root_i,root_j| by the fundamental pair in
       their span; set |j=i| if |root_i| changes @>
  }
}

@ When row~|j| is removed all higher rows shift down a place; this includes
row~|i| since |j<i|.

@< Remove row |root_j| from position |j| in~|m| @>=
{ cycle_block(m,j,m->nrows--,1); root_i=m->elm[--i];
@+} /* |root_i| shifts down */

@ Before we add |-c*root_i| to |root_j| we test if the result is actually a positive
root; if not then we try a similar operation with |root_i| and |root_j| interchanged,
which involves a new call to |Cart_inprod|. It may happen that this still does
not produce the desired effect, but only if the roots lengths are different;
in that case some special measures are called for.

@< Replace the pair of roots |root_i,root_j| ... @>=
{ copyrow(root_j,t,s); add_xrow_to(t,-c,root_i,s);
  if (isposroot(t)) copyrow(t,root_j,s);
  else
  { j=i; c=Cart_inprod(root_i,root_j);
    copyrow(root_i,t,s); add_xrow_to(t,-c,root_j,s);
    if (isposroot(t)) copyrow(t,root_i,s);
    else @< Adjust an unwilling pair of unequal length roots @>
  }
}

@ When we come to this module, the roots |root_i| and |root_j| span a
subsystem of rank~2 in which---with positivity defined be the containing root
system---they are both positive but neither is fundamental (since otherwise we
could have improved the other root with a multiple of it); it follows that the
subsystem has type $B_2$~or~$G_2$. We first order the roots by length, then
decide from the length of the long root which type applies. In case of
type~$B_2$, the fundamental system must be |(ln-sh,2*sh-ln)| where |ln| is the
long root and |sh| the short one. In case of type~$G_2$ there are various
possiblities, but a rather blunt argument settles the matter: since there is a
root of length~3, the containing simple root system must be $G_2$, and the
subsystem must be equal to the full root system, and it's fundamental roots
are given by a $2\times2$ identity matrix.

@< Adjust an unwilling pair of unequal length roots @>=
{ index k; @+ entry* ln,* sh; /* the longer and the shorter root */
  if (Norm(root_i)>Norm(root_j))
    ln=root_i, sh=root_j; @+ else ln=root_j, sh=root_i;
  switch (Norm(ln))
  { case 2: subrow(ln,sh,sh,s); /* |sh=ln-sh| */
    add_xrow_to(ln,-2,sh,s); /* |ln=ln-2*sh| */
  break; case 3: /* |grp=@t$G_2$@>| now */
    for (k=0; sh[k]==0; ++k) {} /* find the place of this $G_2$ component */
    sh[k]=1; sh[k+1]=0; ln[k]=0; ln[k+1]=1;
      /* return standard basis of full system */
  break; default: error("problem with norm 1 roots\n");
  }
}

@* Further auxiliary functions.
The function |cycle_block| is used to permute rows of a matrix in a specific
fashion. It operates on a region of rows of |m|, from |m->elm[first]| up to
|m->elm[last]|, shifting the first |amount| rows to the end of the region,
while moving the remaining rows to the beginning of the region. Stated
differently, all the affected rows move backward |amount| places, but if that
moves them out of the region, the new location is wrapped up by |last-first|.

We perform the permutation in separate cycles, starting with the cycle
containing the first element. During traversal of that first cycle we also
record in |min_done| the smallest index of any other element in that cycle, so
that the remaining cycles are those that start with an element with index
|i<min_done|. Although the same code is used for the subsequent cycles, it
will not change |min_done|, nor does inspecting |min_done| cost any extra
time. We traverse the cycles in the direction opposite to that in which the
rows are actually moved, since this can be done slightly more efficiently.

@< Auxiliary functions @>=
static void cycle_block(matrix* m, index first, index last, index amount)
{ entry** row=&m->elm[first]; @+ index modulus=last-first,i,min_done=amount;
  if (amount>0 && modulus>amount) /* otherwise there is nothing to do */
    for (i=0; i<min_done; ++i) /* |min_done| is fixed after first time round */
    { entry* row_i=row[i]; index j=i+amount,old_j=i;
      do /* perform cycle containing |i| */
      { row[old_j]=row[j]; old_j=j;
	if ((j+=amount)>=modulus) j-=modulus; /* wrap downwards */
      } while (j>=min_done || j>i && (min_done=j,true));
      row[old_j]=row_i; /* close the cycle */
      if (j!=i) error("System error cycling.\n");
    }
}

@ The function |long_close| forms the additive closure within the root system
of~|grp| of the positive fundamental set of roots~|m|; it is used only in the
presence of different root lengths.  For each pair of non-perpendicular short
roots |a,b|, it tests if |a-b| is a root (necessarily a long one).  If so, it
either replaces |a|~by~|a-b| or |b|~by~|b-a|, whichever is positive (thus
avoiding any negative roots). Since any new roots formed are long, they need
not be reconsidered. Afterwards |fundam| will be called a second time
by~|Closure|, which is necessary because the property of being a fundamental
basis is not always preserved.

@< Auxiliary functions @>=
local void long_close(matrix* m, index first, index last)
{ index i,j; @+ entry* root_i,* root_j,* t=mkintarray(s);
  for (i=first; i<last; ++i)
  { root_i=m->elm[i]; if (Norm(root_i)>1) continue;
    for (j=i+1; j<last; ++j)
    { root_j=m->elm[j]; if (Norm(root_j)>1) continue;
      subrow(root_i,root_j,t,s);
      if (isroot(t))
	if (isposroot(t)) @+
	{ copyrow(t,root_i,s); break; @+ } /* need not consider more |j|'s */
	else add_xrow_to(root_j,-1,root_i,s);
    }
  }
  freearr(t);
}

@* Calculating a restriction matrix.
The function |Resmat| is called with a matrix representing a sequence
$\alpha[0],\ldots,\alpha[n-1]$ of roots, which should be a fundamental basis
for some closed subsystem of the root system of |grp|, and it returns the
matrix transforming weights for |grp| to weights for its fundamental Lie
subgroup with that subsystem a its root system. If |grp| contains a central
torus, it is contained in that of the subgroup, so the corresponding weight
coordinates can remain unchanged; only the $s\times s$ matrix for the
semisimple part of~|grp| needs any real computation. If the rank of the
subsystem is maximal, i.e., if $n=s$, then the task is straightforward: the
|j|th coordinate of any weight~$\lambda\in\Lambda$ on the basis of
fundamental weights for the subgroup is $\<\lambda,\alpha[j]>$, and the
|i|th row of the result is obtained by computing these numbers for
$\lambda=\omega_i$.

If $n<s$ however, no such simple formula can be given for the last $s-n$
columns, since the weight lattice for the central torus of the fundamental
Lie subgroup has no preferred basis. In fact we have to be more clear what
we really mean by the restriction matrix in this case, since the fundamental
Lie subgroup itself might not be of the kind directly representable by~\LiE.,
i.e., modulo its central torus it might not be simply connected. (When $n=s$
this cannot happen, since the maximal torus of the subgroup coincides with
that of the whole group; all possible weights for this torus occur for the
whole group, and by restriction also for the subgroup, whence the latter must
be simply connected.) The weights in the image of the restriction matrix
will then be for a direct product of a simply connected semisimple group and
a central torus, which product finitely covers the fundamental Lie subgroup;
consequently the restriction matrix will not be surjective.

It would be possible to choose a covering of arbitrarily high degree, since
there are coverings of any degree of the central torus by another one of the
same rank, but we shall choose the covering of minimal degree that admits a
direct product decomposition of the indicated kind. This means that the
weights on the central torus of the subgroup shall be only those that occur
as restrictions of elements of~$\Lambda$ (the restrictions of weights in
$\{\,\lambda\in\Lambda \mid \forall j\colon \<\lambda,\alpha[j]>=0\,\}$ form
only a sublattice of finite index of this set). In other words the weight
lattice on the central torus of the subgroup is a quotient of~$\Lambda$ by
its sublattice of weights that are $\bf Q$-linear combinations of the
|alpha[j]|. The part of the restriction matrix projecting onto this weight
lattice is therefore a surjective map $\Lambda\to\Zed^{s-n}$ that vanishes on
all the |alpha[j]|; since we are free to choose a basis for the weight
lattice on a central torus, all such maps will do equally well.

We construct such a map by starting with the identity on~$\Lambda$, and
successively making it vanish on each of the |alpha[j]|, while decrementing
the rank of its image on each step. Two records of this mapping under
construction are maintained simultaneously: its matrix |res| on the basis of
fundamental weights of~$\Lambda$, and the matrix |img| whose rows are the
images of the roots |alpha[j]|, i.e., |img==alpha*Cartan()*res|.
Modification of the mapping is done by right multiplication by
transformations that are unimodular up to a sign (i.e., their determinant is
$\pm1$, to preserve surjectivity). This translates into identical column
operations being performed on |res| and |img|. At some points we drop a
column, which represents the only part of the mapping at that point that
does not vanish on~|alpha[j]|, and thus decrement the rank of the image.
Although computation of this map is not necessary when $n=s$ we perform it
in any case, since as a byproduct it provides a test for the linear
independence of the given roots~|alpha[j]|.

@f alpha NULL
@c
matrix* Resmat(matrix* roots)
{ index i,j,k,r=Lierank(grp),s=Ssrank(grp), n=roots->nrows;
  vector* root_norms=Simproot_norms(grp);
  entry* norms=root_norms->compon;
    /* needed to compute $\<\lambda,\alpha[i]>$ */
  matrix* root_images=Matmult(roots,Cartan()),* result=mkmatrix(r,r);
  entry** alpha=roots->elm,** img=root_images->elm,** res=result->elm;

  for (i=0; i<r; i++) for (j=0; j<r; j++) res[i][j]= i==j;
    /* initialise |res| to identity */
  for (j=0; j<n; j++) /* traverse the given roots */
  @< Make column~|j| of~|res| represent the linear form
     $\<\,\cdotp,\alpha[j]>$, and modify the further columns up to |s| so
     that they represent a surjective linear map $\Lambda\to\Zed^{s-1-j}$
     whose kernel contains $\alpha[0],\ldots,\alpha[j]$ @>
  freemem(root_norms); freemem(root_images); return result;
}

@ Of the two tasks mentioned in the name of this module we perform that of
modifying the linear map first, since it frees column~|j| of~|res| for the
other task; it is also the most complicated part of this function, even if
it eventually computes only a small part of the result. We use column
operations to make |img| lower triangular, which corresponds to changing to
a basis for $\Zed^s$ for which the span of the first |i| basis vectors
contains the image of the first |i| of the given roots. Although after
step~|j| we overwrite column~|j| of~|res|, we leave |img| in its lower
triangular form. The action required at step~|j| is ensuring that
|img[j][k]==0| for all |k>j|; we use |v| as abbreviation for |img[j]|.  We
start by checking that |alpha[j]| is indeed a root, which implies that
|v!=0|, and then search for its last non-zero entry. If this entry occurs
before position~|j| then the image of |alpha[j]| is contained in the span of
the previous roots, and a linear dependency has been detected. Otherwise we
traverse columns $j,\ldots,k$ of~|img| from right to left, each time
performing columns operation on a pair of adjacent columns until the
rightmost one of the pair has a zero entry in row~|j|. It is convenient to
ensure that that entry is made positive (by possibly negating its column)
before the (other) column operations are performed; for brevity we shall
take the word ``unimodular'' to include such operations of determinant $-1$.

@< Make column~|j| of~|res|... @>=
{ entry* v=img[j], norm=(checkroot(alpha[j]),Norm(alpha[j]));
  for (k=s-1; v[k]==0; k--) {}
  if (k<j)
    error("Given set of roots is not independent; apply closure first.\n");
  @< Ensure that |v[k]>0| @>
  while(--k>=j)
    /* clear |v[k+1]| by unimodular column operations with column~|j| */
  { @< Find a ``unimodular'' $2\times2$ matrix~|u| such that
       $(v[k]\quad v[k+1])*u=(d\quad 0)$ for some $d>0$ @>
    @< Multiply the pairs of columns |k| and |k+1| of |img| and |res|
       to the right by~|u| @>
   }
  for (i=0; i<s; i++) @< Set $res[i][j]\K\<\omega_i,\alpha[j]>$ @>
}

@ All column operations on |img| are performed identically on |res|. From
the order in which we manipulate columns it can easily be show that |res|
has no non-zero entries more than |j| positions above the main diagonal, so
we need not let |i| run from~0 in the second loop.

@< Ensure that |v[k]>0| @>=
if (v[k]<0)
{ for (i=j; i<n; i++) img[i][k]= -img[i][k];
  for (i=k-j; i<s; i++) res[i][k]= -res[i][k];
}

@ To find the proper linear combinations of columns |k| and |k+1|, we first
process the entries |v[k]| and |v[k+1]| in isolation. The method is
precisely the extended Euclidean algorithm, where we are interested not only
in the linear combination that produces the greatest common divisor (which
will replace |v[k]|) but also in the minimal non-trivial combination that
produces zero (which will replace |v[k+1]|). Repeating this for decreasing
values of~|k|, the entry |v[j]| eventually becomes the gcd of all original
|v[k]| for |k>=j|. When we enter this code, |v[k+1]| has usually already
become quite small, often~$1$; we take care that this case is handled
quickly, which is achieved when in initialising~|u| the columns are swapped
with respect to their natural order. The very nature of the Euclidean
algorithm, where one column is modified only by subtracting a multiple of
the other, guarantees that the determinant of the first two rows of~|u| does
not change, and hence it stays ``unimodular''.

@< Find a ``unimodular'' $2\times2$ matrix~|u|... @>=
  entry u[3][2]; @+ index l=0;
  u[0][1]=u[1][0]=1; u[0][0]=u[1][1]=0;
  u[2][1]=v[k]; u[2][0]=v[k+1];
  if (v[k]<0) u[2][1]= -v[k], u[0][1]= -1; /* make |u[2][1]| non-negative */
  do /* subtract column |l| some times into column |1-l| */
  { entry q=u[2][1-l]/u[2][l]; @+ for (i=0; i<3; i++) u[i][1-l]-=q*u[i][l];
  } while (u[2][l=1-l]!=0);
  if (l==0) @+ for (i=0; i<2; i++) swap(&u[i][0],&u[i][1]);

@ Like above, these loops are restricted to those entries that might be
non-zero. The matrix~|u| declared in the previous section remains in scope
here.

@< Multiply the pairs of columns |k| and |k+1|... @>=
{ for (i=j; i<n; i++) /* combine columns |k| and |k+1| */
  { entry img_i_k=img[i][k];
    img[i][k]  =img_i_k*u[0][0]+img[i][k+1]*u[1][0];
    img[i][k+1]=img_i_k*u[0][1]+img[i][k+1]*u[1][1];
  }
  for (i=k-j; i<s; i++)
  { entry res_i_k=res[i][k];
    res[i][k]=res_i_k*u[0][0]+res[i][k+1]*u[1][0];
    res[i][k+1]=res_i_k*u[0][1]+res[i][k+1]*u[1][1];
  }
}

@ The other part of the restriction matrix is just the transpose of the
input matrix~|alpha| when all roots of~|grp| have the same length.
Otherwise, each entry of |alpha[j]| is multiplied the the norm of the
corresponding fundamental root, and the result is divided by the norm
of~|alpha[j]| itself, which ought to produce an integer result since
|alpha[j]| has already been verified to be a root.

@< Set $res[i][j]\K\<\omega_i,\alpha[j]>$ @>=
{ index inpr= norms[i]*alpha[j][i]; /* this is $(\omega_i,\alpha[j])$ */
  if (inpr%norm!=0) error("Supposed root has non-integer Cartan product.\n");
  res[i][j]=inpr/norm; /* this is $\<\omega_i,\alpha[j]>$ */
}

@* Index.
