\input LiEwebmac
\input references

\def\rf#1{\<\,\cdot\,,#1>} % root form
\def\wi{w_i}

@* Root and weight systems.
In this file we shall define functions that provide the basic functionality of
root system and weights, and certain computations in the Weyl group that are
defined in arbitrary Coxeter groups, such as reduction of words and and
determination of the Bruhat order. The two topics are closely connected, since
for most computations the most efficient way of representing Weyl group
elements is by the corresponding element in the orbit of a selected weight,
mostly the half sum of the positive roots~|rho|.

@h "lie.h"
@f local static
@d local static
@f lambda NULL /* generic name for a weight vector */
@f alpha NULL /* generic name for a root */
@f rho NULL /* the weight with $\<\rho,\alpha_i>=1$ for all $i$ */
@f wi NULL /* the |i|-th letter of the Weyl word |w| */
@f delta NULL
@f mu NULL

@( coxeter.h @>=
@< Prototypes of globally visible functions @>@;

@*1 The geometry of root systems.
Before we deal with roots systems in detail, let us first recall which
elementary data have been defined elsewhere (mostly in the compilation unit
\.{grpdata}). We can access the Cartan matrix for the group by calling
|Cartan|, its determinant by |Detcartan| and the ``inverse'' Cartan matrix
by |Icartan|. The matrix of positive roots can be obtained by calling
|Posroots|, and the norms of the simple roots by calling |Simproot_norms|
(where by the norm of |alpha| we mean |(alpha,alpha)/2|). For simple groups
one can also directly call the corresponding functions like |simp_Cartan|
and |simp_proots|.

@< Proto... @>=
entry simp_inprod(entry* x,entry * y,simpgrp* g);
entry Inprod(entry* x,entry * y);
entry simp_norm(entry* alpha, simpgrp* g);
entry Norm(entry* alpha);
entry Cart_inprod(entry* x, entry* alpha);
matrix* Reflection(entry* alpha);

@~ Taking for~$y$ a fundamental root~$\alpha_j$ we have
$(x,\alpha_j)=\<x,\alpha_j>*N_j$, where $N_j=(\alpha_j,\alpha_j)/2$ can be
found in |g->simp_norm|. To compute $\<x,\alpha_j>$ we express |x| in weight
coordinates as |x*cartan(g)| and take its $j$-th entry; therefore we have in
the general case:  $(x,y)=\sum_{i,j} x_i \\{cartan}_{i,j} N_j y_j$. It is
advantageous to perform the summation on~|i| in the inner loop, as it
involves only one multiplication.

@c
entry simp_inprod(entry* x,entry * y,simpgrp* g)
{ index i,j, r=g->lierank;
  entry* norm=(simp_proots(g),g->root_norm->compon);
  entry** cartan=g->cartan->elm; entry result=0;
  for (j=0; j<r; ++j)
  @/{@; entry s=0; for (i=0; i<r; ++i) s += x[i] * cartan[i][j];
    result += s * norm[j]*y[j];
  }
  return result;
}

@ For composite groups we sum the contributions for the simple factors.

@c
entry Inprod(entry* x,entry * y)
{ if (type_of(grp)==SIMPGRP) return simp_inprod(x,y,&grp->s);
  { index i,t=0,result=0;
    for (i=0; i<grp->g.ncomp; ++i)
    @/{@; simpgrp* g=Liecomp(grp,i);
      result+=simp_inprod(&x[t],&y[t],g); t+=g->lierank;
    }
    return result;
  }
}

@ The functions |simp_norm| and |Norm| assume that their argument |alpha| is
a root. Therefore their result, which could be computed as
|Inprod(alpha,alpha)/2|, is more efficiently found by looking up the
corresponding |root_norm| field stored for simple groups.

@c
entry simp_norm(entry* alpha, simpgrp* g)
{ index i,r=g->lierank; entry level=0,result; boolean neg;
  for (i=0; i<r; ++i) level+=alpha[i]; /* compute level of |alpha| */
  neg=level<0;
  if (neg) {@; level= -level; for(i=0; i<r; ++i) alpha[i]= -alpha[i]; }
  i=find_root(alpha,level,g); assert(i>=0); result=g->root_norm->compon[i];
  if (neg) @+ for(i=0; i<r; ++i) alpha[i]= -alpha[i]; /* restore |alpha| */
  return result;
}

@ For composite groups we call |simp_norm| for the first (and only) simple
factor with a non-zero entry.

@c
entry Norm(entry* alpha)
{ if (type_of(grp)==SIMPGRP) return simp_norm(alpha,&grp->s);
  if (grp->g.ncomp==1) return simp_norm(alpha,Liecomp(grp,0));
  { index i,j,t=0;
    for (i=0; i<grp->g.ncomp; ++i)
    { simpgrp* g=Liecomp(grp,i); index r=g->lierank;
      for (j=0; j<r; ++j) @+
	if (alpha[t+j]!=0) return simp_norm(&alpha[t],g);
      t+=r;
    }
    assert(false); return 0; /* zero vector is not a root */
  }
}

@ |Cart_inprod| calculates $\<x,\alpha>=2(x,\alpha)/(\alpha,\alpha)$, where
$\alpha$ is a root, and |x| is expressed on the root base.
@c
entry Cart_inprod(entry* x, entry* alpha)
@+{@; return Inprod(x,alpha)/Norm(alpha); }

@ The function |Reflection| computes the matrix of a reflection in a
root~|alpha|, given by $x\mapsto x-\<x,\alpha>\alpha$, expressed on the weight
basis. It is implemented by making modifications to an identity matrix.
Therefore the function |simp_refl| that performs the actual computation on the
appropriate simple factor acts by modifying a preallocated identity
matrix~|m|. The modification is to be done on a square block on the main
diagonal, whose top-left corner is at |(offset,offset)|. We subtract in~|m|
the componentwise product (also called matrix tensor product) of the row
vector representing~|alpha| and the column vector representing the linear form
$\rf\alpha$.

@c
local matrix* simp_refl(entry* alpha,simpgrp* g,matrix* m, index offset)
{ index i,r=g->lierank; entry* row=mkintarray(2*r),* col=&row[r];
  @< Set |row| to |alpha| and |col| to $\rf\alpha$, both in weight
     coordinates @>
  for (i=0; i<r; ++i) add_xrow_to(&m->elm[offset+i][offset],-col[i],row,r);
  freearr(row); return m;
}

@ Expressing |alpha| in weight coordinates is done by right multiplication by
the Cartan matrix. The linear form $\rf\alpha$ on the weight lattice is just
the coroot $\check\alpha$ corresponding to $\alpha$, expressed on the basis of
fundamental coroots; it is obtained by multiplying the entries of~|alpha| by
the corresponding norms of simple roots, and dividing by the norm of |alpha|
itself.

@< Set |row| to |alpha|... @>=
{ index norm_alpha=simp_norm(alpha,g); entry* norm=g->root_norm->compon;
  mulvecmatelm(alpha,g->cartan->elm,row,r,r); /* |alpha| on weight basis */
  copyrow(alpha,col,r);
  for (i=0; i<r; ++i) col[i]=(col[i]*norm[i])/norm_alpha;
  /* $\rf\alpha$ on coroot basis */
}

@ The function |Reflection| initialises its result to the identity matrix, and
proceeds to search for the simple component supporting the root similarly to
|Norm|.

@c
matrix* Reflection(entry* alpha)
{ index i,j,t=0; matrix* R=mat_id(Lierank(grp));
  if (type_of(grp)==SIMPGRP) return simp_refl(alpha,&grp->s,R,0);
  if (grp->g.ncomp==1) return simp_refl(alpha,Liecomp(grp,0),R,0);
  for (i=0; i<grp->g.ncomp; ++i)
  { simpgrp* g=Liecomp(grp,i); index r=g->lierank;
    for (j=0; j<r; ++j) @+
      if (alpha[t+j]!=0) return simp_refl(&alpha[t],Liecomp(grp,i),R,t);
    t+=r;
  }
  assert(false); return R; /* zero vector is not a root */
}

@*1 Making weights dominant and related functions.
Now follow some simple functions that deal with bringing a weight into the
dominant chamber. The functions |simp_make_dominant|, and |make_dominant|
are useful functions used internally in several places: they transform a
weight into its dominant counterpart, and return the length of the minimal
Weyl group element that affects this transformation. The remaining functions
below respectively implement the \LiE. functions |length|, |dominant|, and
two versions each of~|W_word|, |canonical|, and~|filter_dom|.

@< Proto... @>=
index simp_make_dominant(entry* lambda, simpgrp* g);
index make_dominant(entry* lambda);
vector* Dominant(vector* lambda);
index Length(vector* w);
vector* Wwordv(entry* lambda, index l);
vector* Wwordm(matrix* m);
vector* Canonical(vector* w);
matrix* Canonicals(matrix* m);
matrix* Filter_dom_m(matrix* m);
poly* Filter_dom(poly* p);

@ The central theme for most of these functions, and indeed for the remainder
of this file, is the following simple game to make a weight vector dominant:
while the vector has negative entries, select one, and apply the corresponding
simple reflection. This will change the sign of that entry, making it
positive, but it may also decrease other entries. Nevertheless the game always
terminates, leaving the weight vector dominant, and in fact one always finds a
minimal sequence of simple reflections that achieves this. Better still, the
set of possible sequences obtained by different choices of negative entries is
in bijection with the set of reduced expressions for the unique Weyl group
element of minimal length whose action makes the original weight dominant. If
we arrange the weight coordinates on the corresponding nodes of the Dynkin
diagram, then the effect of applying the simple reflection for a node consists
of reversing the sign of that entry, while adding a positive multiple of the
old value of that entry to its neigbours; the multiplication factor is~|1|
except when the node corresponds to a long root and the neighbour does not, in
which case the factor is the ratio of the norms of the roots (which is
2~or~3). In this representation the game (that can also be played for
arbitrary Coxeter groups, although some real algebraic factors may then be
involved) is called the numbers game ([Moz]).

We have to have some criterion for selecting a negative entry, and choose
the simplest one possible, namely to always select the first negative entry.
This means that after we have done a move we know that there are no negative
entries up to and including the one just played, except possibly for its
neighbours in the Dynkin diagram. By a fortunate property of the labellings
of Dynkin diagrams used in \LiE., neighbours never have numbers more
than~|2| apart, so if we back up the current position by~|2| after each move
and search on from there, we are certain to still always find the first
negative entry (for linear diagrams we could even back up by~|1|, but we do
not make use of this). The effect is that the running time is bounded by the
sum of the length of the Weyl group element involved and the rank of the
group, rather than by their product (we perform reflections in contant time,
avoiding matrix multiplications).

@d backup(i) (i=(i<=2 ? 0 : i-2))

@c
index simp_make_dominant(entry* lambda, simpgrp* g)
{ index i=0,l=0, r=g->lierank;
  do
  { while (lambda[i]>=0) @+ if (++i==r) return l;
      /* find first negative entry */
    simp_w_refl(lambda,i,g); ++l;
    backup(i); /* reflect and count, then back up */
  } while(true);
}

@ For composite groups we proceed similarly, using |w_refl| instead of
|simp_w_refl|.

@c
index make_dominant(entry* lambda)
{ index i=0,l=0; index s=Ssrank(grp); /* do not affect the toral part */
  if (s==0) return 0; /* this trivial case must be taken out */
  do
  {@; while (lambda[i]>=0) if (++i==s) return l;
    w_refl(lambda,i); ++l; backup(i);
  } while(true);
}

@ The functions |dom_length| and |Dominant| are wrappers around
|make_dominant|, isolating repectively the value returned and the side
effect.

@c
local index dom_length(entry* lambda)
{ index l,s=Ssrank(grp); entry* x=mkintarray(s); @+ copyrow(lambda,x,s);
  l=make_dominant(x); freearr(x); return l;
}

vector* Dominant(vector* lambda)
{@; vector* result=copyvector(lambda);
  make_dominant(result->compon); return result;
}

@ The function |Length| computes the length of element represented by a Weyl
word~|w|. We could compute |dom_length(Waction(rho,w))|, but we can also
count while simulating |Waction|.

@c
index Length(vector* w)
{ index i,s=Ssrank(grp),l=0; entry* x=mkintarray(s),wi;
  for (i=0; i<s; ++i) x[i]=1; /* |x=rho| */
  for (i=0; i<w->ncomp; ++i)
    if ((wi=w->compon[i]-1)>=0)
    {@; if (x[wi]>0) ++l; @+ else --l; w_refl(x,wi); }
  freearr(x); return l;
}

@ The function |Wwordv| is similar to |make_dominant|, but it records the
sequence of reflections used to make |lambda|~dominant. Since the length of
the sequence is needed to allocate the result, the computation will
normally have to be performed twice, which is not really a problem since the
computation is quite fast. However, when calling |Wwordv| internally we
often already know the size of the result, for instance because we have just
built up |lambda| by repeated reflections; we allow this size to be
transmitted to~|Wwordv| by means of an additional parameter~|l|, thus at
least avoiding having to do essentially the same computation more than
twice. In cases where the size is not know in advance we make |l| negative,
telling |Wwordv| to compute the length itself by calling~|dom_length|.

@c
vector* Wwordv(entry* lambda, index l)
{ index i=0,j=0,s=Ssrank(grp); entry* x=mkintarray(s);
  vector* result=mkvector(l>=0 ? l : dom_length(lambda));
  copyrow(lambda,x,s); if (s==0) {@; freearr(x); return result; }
  do
  { while (x[i]>=0) @+ if (++i==s) {@; freearr(x); return result; }
    w_refl(x,i); result->compon[j++]= i+1; backup(i);
  } while(true);
}

@ The function |Wwordm| tries to find a $w\in W$ for which |W_action(w)| is a
given matrix~|m|. We first apply~|m| to~|rho|, to obtain a vector~|x| with
|Wwordv(x)==@t$w\inv$@>|. We could now apply |Wwordv| and reverse the
resulting word, but this would not be the canonical expression for~|w|;
therefore we simulate |make_dominant| while applying the reflections found to
a vector~|y| initialised to~|rho|, thus building up $y=m\inv(\rho)$ (under the
assumption that $m$ does indeed represent the action of some $w\in W$), so
that |Wwordv(y)==w|. Then we do apply |Wwordv|, and check that |W_action|
retrieves the matrix~|m| (otherwise the assumption about |m| fails).

@c
vector* Wwordm(matrix* m)
{ index i,l=0,s=Ssrank(grp);
  entry* x=mkintarray(2*s),* y=&x[s];
  vector* result;
  for (i=0; i<s; ++i) y[i]=1; /* |y=rho| */
  mulvecmatelm(y,m->elm,x,s,s); /* |x=rho*@t$m\v_{s\times s}$@>| */
  i=0;
  if (s>0) do
  { while (x[i]>=0) @+ if (++i==s) goto finish;
    w_refl(x,i); w_refl(y,i); ++l;
    backup(i);
  } while(true);
finish: result=Wwordv(y,l); freearr(x);
  @< Check that |W_action(result)==m| @>
  return result;
}

@ Although the rows and columns of~|m| corresponding to the toral part were
not used to compute~|result|, we do inculde them in the check against
|W_action(result)|.

@< Check that |W_action(result)==m| @>=
{ matrix* check=Weyl_mat(result); index r=Lierank(grp);
  for (i=0; i<r; ++i) if (!eqrow(m->elm[i],check->elm[i],r))
    error("Matrix does not correspond to a Weyl group element.\n");
  freemem(check);
}

@ For |Canonical| we proceed similarly to~|Wwordm|, except that, contrary to a
matrix, the inverse of a Weyl word is easily obtained (by reversal), so that
we can directly obtain a weight to apply |Wwordv| to. Although not all entries
of~|w| necesarily have a positive contribution to the length of the final
result, due to the possibility of zeros and of a reducible expression, it is
easy enough to predict the length before |Wwordv| is called.

@c
vector* Canonical(vector* w)
{ index i,l=w->ncomp,s=Ssrank(grp),wi;
  entry* x=mkintarray(s); vector* result;
  for (i=0; i<s; ++i) x[i]=1; /* |x=rho| */
  for (i=l-1; i>=0; --i) /* apply reverse of |w| to |x| */
    if ((wi=w->compon[i]-1)>=0) {@; w_refl(x,wi); if (x[wi]>0) l-=2; }
    @+ else --l;
  result=Wwordv(x,l); freearr(x); return result;
}

@ The extension of |Canonical| to matrices is straightforward.

@c
matrix* Canonicals(matrix* m)
{ index k,n=m->nrows,s=Ssrank(grp); matrix* result=mkmatrix(n,m->ncols);
  entry* x=mkintarray(s);
  for (k=0; k<n; ++k)
  { entry* w= m->elm[k]; index i,l=m->ncols,wi; vector* t;
    for (i=0; i<s; ++i) x[i]=1;
    for (i=l-1; i>=0; --i)
      if ((wi=w[i]-1)<0) --l; @+ else {@; if (x[wi]<0) l-=2; w_refl(x,wi); }
    t=Wwordv(x,l); copyrow(t->compon,result->elm[k],l); freemem(t);
    for (i=l; i<m->ncols; ++i) result->elm[k][i]=0; /* fill out with 0's */
  }
  freearr(x); return result;
}

@ Finally, here are two trivial functions that throw away non-dominant
vectors.

@c
matrix* Filter_dom_m(matrix* m)
{ index i,j,s=Ssrank(grp),count=0; entry* inx=mkintarray(m->nrows);
  for (i=0; i<m->nrows; ++i) /* get indices of dominant rows into |inx| */
  { for (j=0; j<s; ++j) @+ if (m->elm[i][j]<0) break;
    if (j==s) inx[count++]=i;
  }
  { index r=m->ncols; matrix* result=mkmatrix(count,r);
    for (i=0; i<count; ++i) copyrow(m->elm[inx[i]],result->elm[i],r);
    freearr(inx); return result;
  }
}

poly* Filter_dom(poly* p)
{ index i,j,s=Ssrank(grp),count=0; entry* inx=mkintarray(p->nrows);
  for (i=0; i<p->nrows; ++i) /* get indices of dominant rows into |inx| */
  { for (j=0; j<s; ++j) @+ if (p->elm[i][j]<0) break;
    if (j==s) inx[count++]=i;
  }
  if (count==0) {@; freearr(inx); return poly_null(p->ncols); }
  { index r=p->ncols; poly* result=mkpoly(count,r);
    for (i=0; i<count; ++i)
    { copyrow(p->elm[inx[i]],result->elm[i],r);
      result->coef[i]=p->coef[inx[i]]; setshared(result->coef[i]);
    }
    freearr(inx); return result;
  }
}

@* Coxeter group operations.

@*1 Reduction of Weyl words.
We shall implement the functions |reduce|, |l_reduce|, |r_reduce|
and~|lr_reduce|, that perform reduction of words standing for Weyl group
elements or (double) cosets of the Weyl group with respect to parabolic
subgroups; the function |Length| that finds the (reduced) length of a Weyl
word is also given.

@< Proto... @>=
vector* L_red(vector* L,entry* w,index lw);
vector* Reduce(vector* v);
vector* LR_red(vector* L,entry* w,index lw,vector* R);

@ We start with an auxiliary function |fix_vec|, which checks whether all
entries of a vector~|I| lie in the set $S=[1,s]$ of generators of the Weyl
group~$W$. If so, it returns a dominant weight vector~$x$ whose stabiliser
is~$W_I$, namely one with $x[i]=0$ if and only if $i\in I$; the orbit of
such a vector is in bijection with $W_I\orbit W$. The array~|x| is created
by~|mkintarray|, and should be freed by the caller using |freearr|.

@c
local entry* fix_vec (vector* I, index s)
{ index i; entry* x=mkintarray(s);
  for (i=0; i<s; i++) x[i]=1; /* initialise to positive entries */
  for (i=0; i<I->ncomp; i++)
  { entry e=I->compon[i];
    if (e<=0 || e>s) error("Reflection %ld is out of range.\n",(long)e);
    x[e-1]=0; /* now reflection $r_e$ stabilises |x| */
  }
  return x;
}

@ The basic function for obtaining reduced expressions is |L_red|. Given a
subset |L| of~|S| and $w\in W$, it finds a reduced expression for element of
minimal length in the left $W_L$-coset $W_Lw$ (for ordinary reduction |L_red|
is called with $L=\emptyset$); it must do this by \emph{only} removing letters
from the existing word, so the method of |Canonical| cannot be applied here.
The array~|w| will bemail
 destructively modified; its length is passed as final
argument |lw|. We represent the left coset $W_Lw$ by the image $x\cdot w$ of
the vector |x==fix_vec(L,s)| under the action of~$w$. For any fundamental
reflection~$s_i$, whether the coset $W_Lws_i$ is shorter, equal to, or longer
than $W_Lw$, depends on whether $\<x\cdot w,\alpha_i>$ (the $i$-th coefficient
of $x\cdot w$ on the basis of fundamental weights) is negative, zero or
positive.

@c
vector* L_red(vector* L, entry* w, index lw)
{ index i,s=Ssrank(grp),l=lw;
    /* current length of word; decreases during reduction */
  entry* x=fix_vec(L,s),* y=mkintarray(s); /* current vector and temporary */
  for (i=0; i<lw; i++)
    @< Act on~|x| by~|w[i]| and take action according to the length change @>
  freearr(x); freearr(y);
  @< Collect and |return| the |l| non-zero entries of~|w| @>
}

@ When we arrive here $x$ represents the coset $W_Lw_0\cdots w_{i-i}$. If
$\<x,\alpha_{w_i}><0$, then $w_0\cdots w_i$ is not reduced, and $w_i$ can be
cancelled against some previous factor in~$w$ by the exchange condition. If
$\<x,\alpha_{w_i}>=0$, then $w_i$ stabilises the coset, and can be dropped
from the word, which is indicated by setting |w[i]=0|. For consistency we
also allow the case that already |w[i]==0|, in which case $w_i$ is simply
ignored. Finally, if $\<x,\alpha_{w_i}>>0$, the entry $w_i$ is retained for
the moment (it might be cancelled against a further entry later). In the
cases where $w_i$ is not dropped or ignored, we act on~$x$ with $w_i$ by
calling |w_refl|. This can be done before testing the sign of
$\<x,\alpha_{w_i}>=x[w_i]$, if we bear in mind that the call of |w_refl|
reverses that sign.

@< Act on~|x| by~|w... @>=
{ entry wi=w[i]-1; /* internally, reflections are numbered $0,\ldots,s-1$ */
  if (w[i]==0) --l; /* ignore zero entry */
  else if (x[wi]==0) {@; w[i]=0; --l; } /* |wi| stabilises |x|, delete it */
  else if (w_refl(x,wi),x[wi]>0)
    /* act on~$x$ with $w_i$; test if length decreased */
    @< Cancel |wi| against some previous $w_j$ by the exchange condition @>
}

@ Since the word $v=w_0\cdots w_{i-1}$ is reduced (where we skip the entries
that have been set to~|0|), the entry $w_j$ that cancels against |wi| is
unique; it can be described as follows. Defining the reflection
$t=vw_iv\inv$ and the corresponding half-space $H_t=\setof g\in W:
l(g)<l(tg)\endset$, we have $e\in H_t$ but $v\not\in H_t$, and once the
sequence of elements $e$, $w_0$, $w_0w_1$, \dots,~$v$ leaves~$H_t$, it does
not return into $H_t$ any more. The desired index~$j$ is then such that
$w_0\cdots w_j$ is the first element in this sequence outside~$H_t$. When
searching for~$w_j$ we in fact start at the other end. We construct the
sequence $tv=vw_i$, $tw_0\cdots w_{i-2}=vw_iw_{i-1}$, \dots; this sequence
has decreasing lengths until it leaves $H_t$ at the element
$vw_iw_{i-1}\cdots w_j$. To monitor the length changes we take a copy~|y| of
the value of~|x|, which represents the coset $W_Lvw_i$ when we enter this
module. Since the desired word $w_0\cdots w_{j-1}w_{j+1}\cdots w_{i-1}$
is a reduced expression for the distinguished representative of this coset,
successively acting on~|y| by $w_{i-1}$,~\dots,~$w_{j+1}$ decreases the
length at each step, and $w_j$ is recognised as the first entry that does
not decrease the length (in fact, it increases the length). Once found,
$w_j$ is cancelled against |wi| by setting |w[j]=w[i]=0|.

@< Cancel |wi|... @>=
{ index j=i; @+ copyrow(x,y,s); /* make a temporary copy of~|x| */
  do
    if (w[--j]!=0)
      if (y[w[j]-1]<0) w_refl(y,w[j]-1); @+ else break;
  while (true);
  w[j]=w[i]=0; l-=2; /* cancel reflections by exchange condition */
}

@ After we have determined~|l| and performed all cancellations, we pick up
the remaining entries to form the result.

@< Collect and |return|... @>=
{ vector* result=mkvector(l); entry* res=result->compon;
  for (i=0; i<lw; i++) @+ if (w[i]!=0) *res++=w[i];
  return result;
}

@ For convenience a function |Reduce| is supplied that unpacks a vector and
calls |L_red| with $L=\emptyset$.
Unlike |L_red|, we want |Reduce| to be non-destructive, so the contents of
|v| is first copied.

@c
vector* Reduce (vector* v)
{ vector* empty=mkvector(0),* result; index lw=v->ncomp;
  entry* w=mkintarray(lw);
  copyrow(v->compon,w,lw); result=L_red(empty,w,lw);
  freemem(empty); freearr(w); return result;
}

@ For double cosets there is |LR_red| with a similar interface as |L_red|;
again~|w| is destructively modified.
Two-sided reduction is achieved by first invoking one-sided reduction, and
then applying to the reverse of the reduced word so obtained the same
cancelling algorithm. It is easy to see that cancelling entries by
right-reduction does not affect the left-reducedness of a word, so the
resulting word is indeed doubly reduced. There are fewer cases to deal with
here than in |L_red|: once |w| has been left-reduced, its
entries are non-zero, and successive action of its entries to an initially
dominant weigth~|x| cannot decrease the length.

@c
vector* LR_red(vector* L,entry* w,index lw,vector* R)
{ @< Reduce |w| as left $W_L$ coset representative @>
  { index i,l=lw; entry* x=fix_vec(R,Ssrank(grp));
    for (i=lw-1; i>=0; i--) /* see if |wi| can be eliminated */
    { entry wi=w[i]-1;
      if (x[wi]==0) {@; w[i]=0; --l; } /* |wi| stabilises |x|, delete it */
      else w_refl(x,wi); /* we have |x[wi]>0|; the length increases */
    }
    freearr(x);
    @< Collect and |return|... @>
  }
}

@ Left-reducing |w| is of course done by calling |L_red|.

@< Reduce |w| as left $W_L$ coset representative @>=
{@; vector* t=L_red(L,w,lw);
  lw=t->ncomp; copyrow(t->compon,w,lw); freemem(t);
}

@*1 The Bruhat order.
We now come to various function performing computations related to the
Bruhat order: |Bruhat_leq|, which performs a comparison for the Bruhat order,
|Bruhat_desc|, which finds all descendents in the Bruhat order of a given
element, and |Bruhat_desc_rel| which does the same under the further
restriction that the descendents should also be greater than a given other
element in the Bruhat order, which is useful for generating intervals in the
Bruhat order.

@< Proto... @>=
boolean Bh_leq(vector* v,vector* w);
matrix* Bh_desc(entry* w, index l);
matrix* Bh_desc_rel(entry* w, index l, entry* lwb);

@ In all these cases it will be convenient to work in the reflection
representation of the Weyl group rather than with Weyl words. To be more
precise, |W| is represented by the orbit $\rho W$ of the strictly dominant
weight~|rho|, where $x\in\rho W$ represents the unique $w_x\in W$ with
$\rho\cdot w_x=x$.

There are several equivalent definitions of the Bruhat order, see for
instance~[Deod]. We use the following characterisation [Deod,
Theorem~1.1~(II$'$)], that is computationally useful: the relation~`$\leq$' of
the Bruhat order on~$W$ is characterised by the following two properties.
\smallskip
\itemitem{(i)} For $w\in W$ one has $w\leq e \iff w=e$.
\itemitem{(ii)} If $s\in S$ is a simple reflection and $v,w\in W$ are such
  that both $l(vs)<l(v)$ and $l(ws)<l(w)$, then the relations $v\leq w$, \
  $vs\leq w$, and~$vs\leq ws$ are all equivalent.
\smallskip
The second condition can also be formulated as follows: if $s\in S$ and
$v,w\in W$ with $l(ws)<l(w)$, then $v\leq w$ is equivalent to
$\min(v,vs)\leq ws$, where the minimum is to be interpreted in the sense of
lengths. This suggests to following method to determine whether $v\leq w$
for arbitrary $v,w\in W$: find a reduced expression $w=s_1s_2\cdots s_l$,
put $v_l=v$ and inductively $v_{i-1}=\min(v_i,v_is_i)$ for $i=l,\ldots,1$;
then $v_l\leq w_l\iff v_{l-1}\leq w_{l-1}\iff\cdots\iff v_0\leq w_0$, where
$w_i=s_1\cdots s_i$, so finally (using condition~(i) and~$w_0=e$), $v\leq w$
becomes equivalent to~$v_0=e$. The reduced expression for~$w$ can be
constructed incrementally from right to left, while computing at the same
time $w_i$ and $v_i$ for decreasing~$i$; since $w$ is being represented by
$\rho\cdot w$, this can be done in the same simple way as in
|make_dominant| and |Wwordv|.

Since we have the vectors $x=\rho\cdot v$ and $y=\rho\cdot w$ readily
available, it is easy to perform some extra tests along the way that often
allow us to determine the result of the comparison before we actually
determined~$v_0$. A necessary condition for $v\leq w$ is that $y\prec x$,
i.e., that |x-y| is a sum of positive roots; this is so because $w$ can be
transformed into $v$ by a sequence of multiplications by (not necessarily
simple) reflections, each of which decreases the length, and therefore
affects~$y$ by addition of a positive multiple of a positive root. To
monitor the state of the relation $y\prec x$, we set a vector~|delta| equal
to |x-y| in root coordinates, and check that none of its entries is
negative, or becomes so during the computation; in case that does happen, we
have $y\not\prec x$ and therefore $v\not\leq w$. Since at each step we
modify |x|~and~|y| only by multiples of a single simple root~$\alpha_i$,
only one entry~|delta[i]| changes and needs to be tested. We also maintain
$d=\sum_i\delta[i]$; if it becomes~$0$ while each |delta[i]>=0|, then
|delta==0|, and we have reached a point where $v_i=w_i$, so $v\leq w$ holds.
In fact, with these tests we no longer need the final test whether $v_0=e$:
if we reach $w_0$, i.e., if $y$ becomes dominant while still $y\prec x$,
then we must have $y=x$, and hence $d=0$. 

@c
local boolean Bruhat(entry* x,entry* y) /* modifies arguments */
{ index i,s=Ssrank(grp), d; entry* delta=mkintarray(s);
  @< Set |delta| to |x-y| in root coordinates, and $d=\sum_i\delta[i]$;
     |return false| if some |delta[i]<0| @>
  for (i=0; d>0; backup(i))
  { while (y[i]>=0) ++i; /* find first |y[i]<0|; it exists */
    @< Apply reflection $s_i$ to~|y|, and if |x[i]<0| also to~|x|;
       maintain |delta| and |d| @>
    if (delta[i]<0) {@; freearr(delta); return false; }
      /* because we no longer have $y\prec x$ */
  }
  freearr(delta);
  return true; /* |delta==0|, so |x| and |y| have become equal */
}

boolean Bh_leq(vector* v,vector* w)
{ index i,s=Ssrank(grp); entry* x=mkintarray(2*s),* y=&x[s]; boolean result;
  for (i=0; i<s; ++i) x[i]=y[i]=1;
  Waction(x,v); Waction(y,w); /* $x\K\rho\cdot v$ and $y\K\rho\cdot w$ */
  result = Bruhat(x,y); freearr(x); return result;
}

@ The conversion of |x-y| to root coordinates is a straightforward
application of |Icartan| and |Detcartan|.

@< Set |delta| to |x-y|... @>=
{ matrix* ic=Icartan(); index dc=Detcartan(); entry* t=mkintarray(s);
  subrow(x,y,t,s); mulvecmatelm(t,ic->elm,delta,s,s);
  freemem(ic); freearr(t);
  for (d=0, i=0; i<s; ++i) 
    if ((delta[i]/=dc)<0) {@; freearr(delta); return false; } @+
    else d+= delta[i];
}

@ Applying the reflection $s_i$ on a weight vector is done by |w_refl|; the
value of the entry~|i| afterwards tells which multiple of the
root~$\alpha_i$ was added, so maintanance of |delta| consists simply of
adding or subtracting that entry to or from |delta[i]|, and |d| undergoes
the same change.

@< Apply reflection $s_i$ to~|y|... @>=
{ w_refl(y,i); delta[i]-=y[i]; d-=y[i];
  if (x[i]<0) {@; w_refl(x,i); delta[i]+=x[i]; d+=x[i]; }
}

@ The function |Bh_desc| determines the descendents of~|w| in the Bruhat
order. This is remarkably simple if |w| is already reduced: each descendent
has a unique reduced expression obtained from~|w| by removal of a single
letter (the uniqueness follows from the exchange condition). So all that
really has to be done is to test for each of those subwords of~|w| whether
it is reduced.

@c
matrix* Bh_desc(entry* w, index l) /* |w| is reduced of length |l| */
{ entry* inx; index s=Ssrank(grp),n_desc=0;
  if (l==0) return mkmatrix(0,0); /* avoid attempting a $0\times-1$ matrix */
  @< Collect the indices |i| for which
     $w_0,\ldots,\widehat{w_i},\ldots,w_{l-1}$ is reduced into an
     array~|inx| of length~|n_desc| @>
  @< Build and |return| a $\\{n\_desc}\times l-1$ matrix, whose row~|i| is
     $w_0,\ldots,\widehat{w_{inx[i]}},\ldots,w_{l-1}$ @>
}

@ As always we test for reducedness by acting on a vector~|x| initialised
to~|rho|.

@< Collect the indices |i| for which... @>=
{ index i,j; entry* x=(inx=mkintarray(l+s),&inx[l]);
  for (i=0; i<l; ++i)
  { for (j=0; j<s; ++j) x[j]=1; /* |x=rho| */
    for (j=0; j<l; ++j) @+ if (j!=i)
    @/{@; entry wj=w[j]-1; if (x[wj]>0) w_refl(x,wj); else break; }
    if (j==l) inx[n_desc++]=i; /* if reduced, record |i| */
  }
}

@ Filling the matrix of descendents is trivial (if not boring).

@< Build and |return|... @>=
{ matrix* result=mkmatrix(n_desc,l-1); index i,j;
  for (i=0; i<n_desc; ++i)
  @/{@; entry* row= result->elm[i];
    for (j=0; j<l; ++j) if (j!=inx[i]) *row++=w[j];
  }
  freearr(inx); return result;
}

@ The function |Bh_desc_rel| differs from |Bh_desc| only in an extra
condition that it imposes on the indices selected into~|inx|. The
argument~|lwb| is passed as a weight vector (image of~|rho|), while |w| is a
reduced Weyl word of length~|l|, like in |Bh_desc|.

@c
matrix* Bh_desc_rel(entry* w, index l, entry* lwb)
{ entry* inx; index s=Ssrank(grp),n_desc=0;
  if (l==0) return mkmatrix(0,0); /* avoid attempting a $0\times-1$ matrix */
  @< Collect into an array~|inx| of length~|n_desc| the indices |i| such that
     $v=w_0,\ldots,\widehat{w_i},\ldots,w_{l-1}$ is reduced and |lwb<=v| @>
  @< Build and |return|... @>
}

@ When we call |Bruhat|, both arguments are available, but a working copy of
|lwb| is needed each time.

@< Collect into an array~|inx| of length~|n_desc|... @>=
{ index i,j; entry* x=(inx=mkintarray(l+2*s),&inx[l]), *y=&x[s];
  for (i=0; i<l; ++i)
  { for (j=0; j<s; ++j) x[j]=1; /* |x=rho| */
    for (j=0; j<l; ++j) @+ if (j!=i)
    @/{@; entry wj=w[j]-1; if (x[wj]>0) w_refl(x,wj); else break; }
    if (j==l && (copyrow(lwb,y,s),Bruhat(y,x))) inx[n_desc++]=i;
  }
}

@ This function replaces the non-empty subset of~|W| represented by the rows
of~|*m| by the set of their Bruhat descendents above |lwb|; it is used in
|KL|. Both |lwb| and the rows of~|*m| are encoded as images of~|rho|. The
lengths of the elements represented by the rows of~|*m| are known to be
equal, so an upper bound for the new number of rows of~|*m| can be obtained
by multiplying the current number by the length~|l| of the element
represented by the first row. For each row~$r$ of~|*m| we obtain a reduced word
|word| for the inverse of the element of~|W| represented by~$r$ by a call to
|Wwordv|.

@c
local void Bh_descs(matrix** m, entry* lwb, index s)
{ vector* word=Wwordv((*m)->elm[0],-1); entry* w=word->compon;
  index i,j,k,l=word->ncomp,n_desc=0;
  matrix* descs=mkmatrix(((*m)->nrows)*l+1,s);
  entry* x=mkintarray(2*s),* y=&x[s]; /* working arrays */
  for (k=0; ; w=(word=Wwordv((*m)->elm[k],l))->compon)
  { @< Add the descendents of $w\inv$ above |lwb| to |descs| @> @+
    freemem(word);
    if (++k==(*m)->nrows) break;
  }
  descs->nrows=n_desc;
  freearr(x); freemem(*m); *m=Unique(descs,cmpfn);
}

@ This is like |Bh_desc_rel|, but |w| is traversed from right to left, and
the weight~|x| is recorded rather than its index~|i|. We copy |x| to |descs|
before calling |Bruhat|, but only increment |n_desc| if the test succeeds.

@< Add the descendents of $w\inv$ above |lwb| to |descs| @>=
for (i=0; i<l; ++i)
{ for (j=0; j<s; ++j) x[j]=1; /* |x=rho| */
  for (j=l-1; j>=0; --j) @+ if (j!=i)
  @/{@; entry wj=w[j]-1; if (x[wj]>0) w_refl(x,wj); else break; }
  if (j<0 && (copyrow(lwb,y,s),copyrow(x,descs->elm[n_desc],s),Bruhat(y,x)))
    ++n_desc;
}


@*1 Kazhdan-Lusztig polynomials.
Below we implement the computation of the Kazhdan-Lusztig polynomials as
described in~[KaLu]. The implementation is a very basic one that just
follows the original definitions, without attempting any methods to improve
efficiency. Therefore it will only be useful for occasional computations of
Kazhdan-Lusztig polynomials, and (within reasonable limits of time and
memory) is not likely to produce any answers for particularly difficult
cases. Those who require extensive tables of these polynomials should
certainly use more specialised programs than those contained in \LiE.. It
should be noted that, while those programs give a great improvement over the
straightforward approach followed here, the computational complexity of
computing Kazhdan-Lusztig polynomials is still quite large, and to date no
theoretical results are known that would enable a fundamentally more
efficient method of computation.

The main points where we have let simplicity prevail over efficiency are the
following: 
  no attempt is made to reuse previously computed values, not even among the
recursive invocations caused by a single external call;
  we use the polynomial data type of \LiE. to represent the polynomials in
the computation, which is convenient but not very efficient, given that the
polynomials are univariate and usually have very small coefficients.
In our description we shall write these univariate polynomials using~|q| as
indeterminate, like in~[KaLu], although their representation in~\LiE. does
not allow indeterminates to have special names. Is is useful to install
the polynomials $q$ and $q-1$ during out computations.
@c
static poly* q,* q_1;
   /* |q| respectively |q-1|, when initialised */
void make_q (void) 
{ (q=mkpoly(1,1))->elm[0][0]=1; *q->coef=one; setshared(q);
    /* $X^{[1]}$ means |q| */
  q_1= Add_pol_pol(q,poly_one(1),true); setshared(q_1); 
    /* $X^{[1]}-X^{[0]}$ means |q-1| */
}
void clear_q (void) @+
{@; clrshared(q); freepol(q); clrshared(q_1); freepol(q_1); }

@~
@< Proto... @>=
void make_q(void);
void clear_q(void);
poly* Rpoly(vector* x, vector* y);
poly* KLpoly(vector* x, vector* y);

@ We start with computing $R$-polynomials, also defined in~[KaLu], that are
closely related to Kazhdan-Lusztig polynomials. We shall not actually use
them in the computation of Kazhdan-Lusztig polynomials, but they can be used
for checking the correctness of the latter.

As defining equations for the $R$-polynomials $R_{x,y}$ we use a
definition by induction on~$l(y)$ [KaLu~2.0.b,c]:
$$
  \hbox{if $ys<y$ then}\qquad R_{x,y}=
  \cases{R_{xs,ys} & if $xs<x$,\cr (q-1)R_{x,ys}+qR_{xs,ys} & if $x<xs$,\cr}
$$
for $s\in S$, and with $R_{x,x}=1$ and $R_{x,y}=0$ if $x\not\leq y$.
The main work is done by a local function~$R$, which in fact computes
$q^kR_{x,y}$ (this facilitates its recursive definition). The additional
arguments are $d=l(y)-l(x)$ and the rank~$s$ of the Coxeter group. The
function uses recursion for the first term in the second branch of the
formula above, but iteration for the second term, increasing the
exponent~|k| to take account of the accumulated factors~|q|.

@d Bh(x,y)   (copyrow(x,v,s),copyrow(y,w,s),Bruhat(v,w))

@c
local poly* R(entry* x, entry* y, index k, index d, index s)
  /* modifies |x| and |y| */
{ index i; entry* v=mkintarray(2*s),* w=&v[s];
  poly* sum=poly_null(1);
  while (d>0) /* |d==length(y)-length(x)| */
  { if (!Bh(x,y)) {@; freearr(v);return sum;}
    @< Apply any simple reflections~|s| with both $xs<x$ and $ys<y$ @>
    @< Find |i| such that $ys_i<y$ and $x<xs_i$ @>
    @< Add $q^k(q-1)R_{x,ys_i}$ to |sum| and set $y\K ys_i$ @>
    w_refl(x,i); d-=2; ++k;
      /* repeat with $x\K xs_i$, $y\K ys_i$, adjusted~|d|,
         and increased~|k| */
  }
  freearr(v);
  if (d==0 && eqrow(x,y,s)) /* if |x==y|, add final $q^k$ */
  {@; poly* q_k=copypoly(q); q_k->elm[0][0]=k;
    sum= Add_pol_pol(sum,q_k,false);
  }
  return sum;
}

@ The following is another variation of |make_dominant|, where the
reflections applied are limited to those that make both |x|~and~|y| more
dominant.

@< Apply any simple reflections... @>=
for (i=0; ; backup(i))
{ do {@; if (x[i]<0 && y[i]<0) break;} while (++i<s);
  if (i==s) break;
  w_refl(x,i); w_refl(y,i);
}

@ Because all |i| with $ys_i<y$ and $xs_i<x$ have been eliminated, it
suffices to find |i| with $ys_i<y$. Such an~|i| must exist, since we have
checked that |x<=y| and |d!=0|, so that |x<y|, and hence~|y!=e|.

@< Find |i| such that $ys_i<y$ and $x<xs_i$ @>=
for (i=0; y[i]>=0; ++i) {}


@ It is convenient to set $y\K ys_i$ before calling~$R$ recursively to
compute $R_{x,ys_i}$.

@< Add $q^k(q-1)R_{x,ys_i}$ to |sum|... @>=
{ w_refl(y,i); copyrow(x,v,s); copyrow(y,w,s);
  sum=Add_pol_pol(sum,Mul_pol_pol(q_1,R(v,w,k,d-1,s)),false);
}

@ External calls of~|R| come via~|Rpoly| which does the necessary
preparations and supplies the extra arguments.

@c
poly* Rpoly(vector* x,vector* y)
{ index i,s=Ssrank(grp); entry* v=mkintarray(2*s),* w=&v[s]; poly* result;
  for (i=0; i<s; ++i) v[i]=w[i]=1;
  Waction(v,x); Waction(w,y);
  result=R(v,w,0,dom_length(w)-dom_length(v),s); freearr(v); return result;
}


@ Now we come to the computation of the actual Kazhdan-Lusztig
polynomials~$P_{x,y}$; like for~$R$ we actually compute the product
$q^eP_{x,y}$ with a specified power of~$q$. Again |x|~and~|y| are
|rho|-images, and modified during the call (the macro~|P| is used when this
is not desired); we should have |d==l(y)-l(x)|, |s==Ssrank(grp)|, and
|cmpfn==lex_decr|. As defining formula we shall use [KaLu~(2.2.c)].

@d P(x,y,e,d)  (copyrow(x,v,s),copyrow(y,w,s),KL(v,w,e,d,s))

@c
local poly* KL(entry* x, entry* y, index e, index d, index s)
{ index i; entry* v=mkintarray(2*s),* w=&v[s]; poly* result;
  if (d<0 || !Bh(x,y)) {@; freearr(v); return poly_null(1); }
  if (d<3) @+ triv_result: /* return $q^e$ when |l(y)-l(x)<3| */
    {@; poly* qe=mkpoly(1,1); qe->coef[0]=one; qe->elm[0][0]=e;
      freearr(v); return qe;
    }
  @< Try simplifications of |(x,y)| that do not change $P_{x,y}$;
     if this causes |d<3|, |goto triv_result| @>
  for (i=0; y[i]>=0; ++i) {}
    /* find $i$ with $ys_i<y$; it exists since |y!=e| */
  w_refl(y,i); --d; /* put $y\K ys_i$, maintain |d==l(y)-l(x)| */

  @< Compute $q^eP_{x,ys_i}$ in |result|, using formula [KaLu~(2.2.c)] @>
  freearr(v); return result;
}

@ Simplifying the pair $(x,y)$ means in the first place trying to reduce the
distance $d=l(y)-l(x)$, which may be possible using [KaLu~(2.3.g)]: if
there is some~$s$ such that $x<xs$ and $ys<y$, then $P_{x,y}=P_{xs,y}$.
When we have completed such reductions, we can also try to find instances
where the left hand side of our defining formula [KaLu~(2.2.c)] reduces to a
single term; this happens when $xs<x$, $ys<y$ and $x\not\leq ys$, and in
that case we have $P_{x,y}=P_{xs,ys}$. In doing the test whether $x\leq ys$,
it is easiest to temporarily set $y\K ys$ before invoking |Bh|, and to
either restore it or to also set $x\K xs$ afterwards, depending on the
result of the test. Since the replacement $y\K ys$ changes the set of $s'$
for which $ys'<y$ holds, we retry simplifications of the first kind every
time we find a simplification of the second kind.

@< Try simplifications of |(x,y)|... @>=
do
{ for (i=0; ; backup(i))
  { do {@; if (x[i]>0 && y[i]<0) break; } while (++i<s);
    if (i==s) break;
    w_refl(x,i); if(--d<3) goto triv_result;
  }
  for (i=0; i<s; ++i)
    if (y[i]<0) /* then also |x[i]<0|, because of the loop above */
      if (w_refl(y,i),Bh(x,y)) w_refl(y,i); @+ else {@; w_refl(x,i); break; }
} while (i<s); /* continue as long as second simplification was possible */

@ When we reach this module we have found an appropriate $s_i$ to apply
[KaLu~(2.2.c)], and we shall not change $i$ anymore, so we shall henceforth
write $s$ in place of $s_i$. Having already set $y\K ys$, we have
$xs<x<y<ys$. We rewrite the formula substituting our current variable names:
$y$ is replaced by~$x$, $w$~by~$ys$, $v$~by~$y$, and $c$~by~$1$ (since
$xs<x$), giving
$$
 P_{x,ys}=P_{xs,y}+qP_{x,y}-\sum_z\mu(z,y)q^{(l(y)-l(z)+1)/2}P_{x,z},
$$
where the sum is over all $z$ with $x\leq z<y$, $zs<z$, and $l(y)-l(z)$ odd,
and $\mu(z,y)$ denotes the coefficient of $q^{(l(y)-l(z)-1)/2}$ in~$P_{z,y}$
(if non-zero, it is the leading coefficient). The possible values for~|z|
are generated by level, downwards from~$y$, using |Bh_descs|; we go down one
level initially, and two level every following time.


@< Compute $q^eP_{x,ys_i}$ in |result|, using formula [KaLu~(2.2.c)] @>=
{ matrix* level=mkmatrix(1,s); index j,c=1; /* |c==(l(y)-l(z)+1)/2| */
  @< Set |result| to $q^eP_{xs,y} + q^{e+1}P_{x,y}$ @>
  copyrow(y,level->elm[0],s); Bh_descs(&level,x,s);
  --d; /* |d==l(z)-l(x)| */
  do
  { for (j=0; j<level->nrows; ++j) 
    @/{@; entry* z=level->elm[j]; 
      if (z[i]<0) @< Subtract $\mu(z,y)q^{e+c}P_{x,z}$ from |result| @>@+
    }
    if (d<2) {@; freemem(level); break;} /* we cannot go down 2 more levels */
    Bh_descs(&level,x,s); Bh_descs(&level,x,s); d-=2; ++c;
  } while(true);
}

@
@< Set |result| to $q^eP_{xs,y} + q^{e+1}P_{x,y}$ @>=
{ poly* Pxy=P(x,y,e+1,d);
  w_refl(x,i); result=Add_pol_pol(P(x,y,e,d+1),Pxy,false); w_refl(x,i);
}

@
@< Subtract $\mu(z,y)q^{e+c}P_{x,z}$ from |result| @>=
{ poly* Pzy=P(z,y,0,2*c-1);
  if (Pzy->nrows!=0 && Pzy->elm[0][0]==c-1 && Pzy->coef[0]->size!=0)
    /* |mu(z,y)!=0| */
  { bigint* mu=sub(null,Pzy->coef[0]); freepol(Pzy);
    result=Addmul_pol_pol_bin(result,P(x,z,e+c,d),mu);
  }
}

@ Here is the externally callable function for Kazhdan-Lusztig
polynomials, doing the necessary preparations.

@c
poly* KLpoly(vector* x,vector* y)
{ index i,s=Ssrank(grp); entry* v=mkintarray(2*s),* w=v+s; poly* result;
  for (i=0; i<s; ++i) v[i]=w[i]=1;
  Waction(v,x); Waction(w,y);
  result=KL(v,w,0,dom_length(w)-dom_length(v),s); freearr(v); return result;
}


@* References and Index.

\beginrefs

\ref Deod \key Deodhar
 \author V. V. Deodhar
 \title Some characterisations of Bruhat ordering on a Coxeter group
        and determination of the relative M\"obius function
 \journal Inventiones Math. \vol 39 \pages 187-198 \year 1977
\endref


\ref KaLu \key Kazhdan Lusztig
 \author D. Kazhdan and G. Lusztig
 \title Representations of Coxeter groups and Hecke algebras
 \journal Inventiones Math. \vol 53 \year 1979 \pages 165-184
\endref

\ref Moz \key Mozes
 \author S. Mozes
 \title Reflection processes on graphs and Weyl groups
 \journal J. Combinatorial. Theory, Series A \vol 53 \pages 128-142 \year 1990
\endref

\endrefs
