\input LiEwebmac

@* Weight collection and character decomposition.
This file defines functions of two categories: there is a set of general
purpose routines for collecting a set of weights with multiplicities into a
polynomial, and there are routines that utilise this facility for
implementing the \LiE. functions |decomp| and |vdecomp|, which serve to
decompose (virtual) characters.

@h "lie.h"
@d local static
@f local static
@f lambda NULL
@f mu NULL
@f nu NULL

@c
@< Local variables @>@;

@ The external interface is determined by the header file \.{decomp.h}.

@( decomp.h @>=
@< Prototypes of globally visible functions @>@;

@*1 Accumulation of weights.
We shall first consider the routines used to efficiently build up a
polynomial from numerous small contributions in the form of individual
weights with a multiplicity. Each such accumulation process starts with a
call of |wt_init|, which establishes the size~|r| of the the weight vectors
that will be collected. Individual weights can subsequently by contributed
by calling |wt_ins(p,c,neg)|, where |p| points to the entries of the weight,
and |c| is the (big integer) multiplicity of this contribution; the boolean
|neg| when |true| reverses the sign of the contribution (this often avoids
having to create a separate |bigint| for this purpose). When all weights
have been contributed, the accumulated polynomial can be obtained by calling
|wt_collect|, which also frees the internal resources used for the
accumulation process. The relation between these various calls is purely
chronological: the calls to |wt_ins| need not come from the same function
that called |wt_init|, and indeed calls to |wt_ins| will often be set up to
be invoked at all points in a Weyl orbit via the function |Weylloop|.

@< Prototypes... @>=
void wt_init(index r);
void wt_ins(entry* p, bigint* c, boolean neg);
poly* wt_collect(void);

@ As will be clear from the interface, the administration of the values
accumulated uses storage private to the functions defined here, and in fact
it will be impossible to carry out more than one accumulation process at the
same time, and for instance functions using these routines should avoid using
recursion. While the contributions up to any given point can clearly be
represented by a single polynomial, it would be difficult to maintain such a
polynomial efficiently when it becomes large: if one keeps it sorted, much
time is spent to keep it so (for each new term the average time is
proportional to the size of the polynomial), and if one keeps it unsorted
one either has to do a complete traversal to locate possible existing terms
that combine with the new contribution, or choose not to combine such terms
until the end, in which case the size of the polynomial may grow much more
than necessary. Therefore we choose a somewhat different organisation, which
uses both a sorted and an unsorted accumulator polynomial (in fact two
unsorted ones). Each new contribution is first looked up in the |sorted|
accumulator using a binary search; if found, the coefficient of that term is
adjusted, and otherwise the contribution is placed on an unsorted
accumulator, which will eventually be merged with the sorted one. The two
unsorted accumulators |pos_acc| and |neg_acc| serve to handle contributions
of both signs without having to explicitly negate the |bigint| multiplicity
of a contribution.

@< Local variables @>=
static poly* sorted,* pos_acc,* neg_acc;
  /* sorted and pos/neg unsorted accumulators */


@ The accumulators must be free of aliases, since we shall update them {\sl
in situ}. The unsorted accumulators |pos_acc| and |neg_acc| are allocated to
a larger size than is directly necessary, so that some new terms can be
added without the immediate need for reallocation. Therefore the function
|wt_init|, which clears all the accumulators, make a copy of the zero
polynomial for |sorted| and resets the |nrows| field of of |pos_acc| and
|neg_acc| to~|0| after allocating space for a minimum of |ACCMIN| terms in
each.

@d ACCMIN  8  /* minimum size of unsorted accumulators */

@c

void wt_init(index r)
{ sorted=copypoly(poly_null(r));
  pos_acc=mkpoly(ACCMIN,r); neg_acc=mkpoly(ACCMIN,r);
  pos_acc->nrows=0; neg_acc->nrows=0;
}

@ As was indicated above, |wt_ins| decides after a binary search in |sorted|
whether to add the new term to |sorted| or to an unsorted accumulator. The
binary search is performed by |searchterm|, which returns either the index
of the term whose exponent matches |wt|, or a negative value in case no such
term exists.

@c
void wt_ins(entry* wt, bigint* c, boolean neg)
{ if (c->size==0) {@; freemem(c); return; }
  { index i=searchterm(sorted,wt);
    if (i>=0) @< Adjust coefficient of term |i| of |sorted| @>
    else
    @< Push new term on |pos_acc| or |neg_acc|, after emptying that
       accumulator to |sorted| in case it would otherwise get too large @>
  }
}

@ When a coefficient of |sorted| is to be updated directly, the sign
in~|neg| is taken into account by choosing between the functions |add| and
|sub| to incorporate the multiplicity~|c| into that coefficient.

@< Adjust coefficient of term |i| of |sorted| @>=
{ clrshared(sorted->coef[i]);
  sorted->coef[i]= (neg ? sub : add)(sorted->coef[i],c);
  setshared(sorted->coef[i]);
}

@ When the weight of the new contribution is not found in |sorted|, we use
|neg| to choose one of the unsorted accumulators to add a new term to. If
that accumulator is already full, i.e., if its current number |nrows| of
terms is equal to to the number |rowsize| of terms allocated, then its
contents is combined with that of |sorted| by means of |Add_pol_pol|; this
function takes care of sorting the contents of the unsorted accumulator,
combining terms with equal exponents, choosing whether to add to or subtract
from |sorted|, based on~|neg|, and freeing its argument polynomials. The old
contents of |sorted| is not needlessly sorted by |Add_pol_pol|, since the
|SORTED| bit in |sorted->type| is set. After an unsorted accumulator is
emptied into |sorted| a new polynomial is allocated for it; by keeping the
allocated size proportional to that of |sorted| as more and more terms are
collected, the average time needed to process each new contribution is
limited to~$O(\log(\hbox{|sorted->nrows|}))$.

@< Push new term on |pos_acc| or |neg_acc|... @>=
{ poly** acc= neg ? &neg_acc : &pos_acc;
  index i=(*acc)->nrows;
  if (i==(*acc)->rowsize)
  { sorted=Add_pol_pol(sorted,*acc,neg);
    *acc=mkpoly(Max(sorted->nrows,ACCMIN),sorted->ncols); i=0;
  }
  copyrow(wt,(*acc)->elm[i],sorted->ncols); (*acc)->coef[i++]=c; setshared(c);
  (*acc)->nrows=i;
}

@ The final value of the combined accumulators is determined by
|wt_collect|, where again |Add_pol_pol| does the real work. We avoid calling
|Add_pol_pol| with degenerate polynomials with |nrows==0|.

@c
poly* wt_collect(void)
{ if (pos_acc->nrows>0) sorted=Add_pol_pol(sorted,pos_acc,false);
  else freemem(pos_acc);
  if (neg_acc->nrows>0) sorted=Add_pol_pol(sorted,neg_acc,true);
  else freemem(neg_acc);
  {@; poly* result=sorted; sorted=NULL; return result; }
}

@*1 Decomposition of characters.
While the routines above are independent of any interpretation of the
polynomials involved, the ones in the remainder of this compilation unit
deal with character decomposition, and assume an interpretation of
polynomials as dominant character polynomials or decomposition polynomials
with respect to some specific group. There is a pair of functions
|char_init|~and~|char_decomp|, that may be used together with |wt_ins|, in
place of the pair of |wt_init|~and~|wt_collect|; the effect is that the
contributed weights will be collected just as before, but at the end the
collection of all contributed weights will be decomposed as a dominant
character polynomial, rather than just returned as a polynomial. If the
dominant character polynomial is directly available as a polynomial, then
|Decomp| can be used to do just the decomposition, without the need to pass
all the terms through |wt_ins|. Using either of these methods, it is
essential that all contributed weights are dominant, or the algorithm may
fail (checks elsewhere in \LiE. guarantee that this condition is met).
Finally |Vdecomp| also performs decomposition, but allows its argument to be
a virtual dominant character, rather than a real one. The functionality of
|char_init|~and~|char_decomp| is used in implementing the \LiE. function
|branch|, whence we allow that characters are temporarily interpreted with
respect to a group different from the one specified by the global
variable~|grp|; this group is passed as argument to~|char_init|. The
functions |Decomp| and~|Vdecomp| on the other hand, like other mathematical
functions, use the group specified by~|grp|.

@< Prototypes... @>=
void char_init(object g);
poly* char_decomp(void);
poly* Decomp(poly* p);
poly* Vdecomp(poly* p);

@ For the decomposition of dominant characters, as performed by |char_decomp|,
it is necessary to repeatedly select a maximal term in the height ordering
from the character polynomial as it is decomposed. To this end it is
convenient to (temporarily) use the decreasing height sorting criterion for
polynomials, since then the initial term is always an appropriate one. This
means however, that we have to temporarily override the sorting criterion
|cmpfn|, and |defaultgrp| as well, since a height sorting criterion is always
interpreted with respect to the default group. The old values are set aside,
and will be restored upon completion of |char_decomp|.

@< Local variables @>=
local cmpfn_tp sav_cmpfn; /* to store suspended |cmpfn| */
local object sav_dfgrp; /* to store suspended |defaultgrp| */

@ Here is the function that installs the decreasing height sorting criterion
for a specified group~|g|. The precomputation necessary for
using~|height_decr| will be performed automatically at the first sorting
operation. Neither is there need to call |unmark_sorted| here, since the
change of sorting criterion is only temporary and all existing polynomials are
unaffected; the one polynomial we build using the changed sorting criterion
will be explicitly marked properly.

@c

local void set_weight_sorting(object g)
{@; sav_cmpfn=cmpfn; sav_dfgrp=defaultgrp; cmpfn=height_decr; defaultgrp=g;
}

@ At the end of the decomposition, matters are restored; note that unmarking
of |SORTED| in~|result|.

@< Restore the sorting criterion @>=
{@; cmpfn=sav_cmpfn; defaultgrp=sav_dfgrp; clrsorted(result); }

@ If an error occurs during decomposition, we must also restore the sorting
criterion.

@< Make an error stop... @>=
{ cmpfn=sav_cmpfn; defaultgrp=sav_dfgrp;
  error ("Non-virtual decomposition failed.\n");
}


@ The function |decompose_character| does the actual character decomposition
for |char_decomp| and |Decomp|; it takes the dominant character
polynomial~|ch| to decompose as argument. The algorithm it uses is simple: as
long as there are terms left in~|ch|, one of maximal height is selected and
contributed to the decomposition polynomial, and the dominant character of
the corresponding irreducible module is computed and subtracted from |ch|,
which is guaranteed to remove at least the selected term. Since~|ch| is
assumed to be the dominant character of an actual module, the subtractions
should never introduce negative coefficients; if nevertheless the term
selected at some point is found to have a negative coefficient, the
computation is aborted with an error message. The process of repeatedly
contributing terms to the final decomposition polynomial is done by a
separate |wt_init|--|wt_ins|--|wt_collect| loop, which does not overlap with
a similar loop that might have been used to compute~|ch|.
@c
local poly* decompose_character(poly* ch)
{ wt_init(ch->ncols); /* for building result */
  while (ch->coef[0]->size!=0) /* i.e., |while (ch!=0)| */
  { bigint* c=ch->coef[0];
    if (c->size<0)
      @< Make an error stop after restoring |cmpfn| and |defaultgrp| @>
    @< Contribute the initial term of |ch| to the result, and subtract
       its dominant character from |ch| @>
  }
  { poly* result=wt_collect();
    @< Restore the sorting criterion @>
    return result;
  }
}

@ The initial term of |ch| has coefficient~|c| and exponent |ch->elm[0]|.
Subtraction is achieved by calling |Addmul_pol_pol_bin| with negated
coefficient~|c|; this requires copying~|c| first.

@< Contribute the initial term of |ch|... @>=
{ wt_ins(ch->elm[0],c,false); /* contribute weight to result */
  c=copybigint(c,NULL); c->size= -c->size;
  ch=Addmul_pol_pol_bin(ch,Domchar_irr(ch->elm[0],NULL),c);
}


@ The externally callable functions for decomposing actual characters are
now easily defined. Whereas with |char_init| and |char_decomp| the dominant
character polynomial is built up while using the desired decreasing height
ordering, |Decomp| may have to reorder the polynomial before calling
|decompose_character|; this is achieved by calling |Reduce_pol| after
clearing the `sorted' flag and calling |set_weight_sorting|.

@c
void char_init(object g) @+{@; set_weight_sorting(g); wt_init(Lierank(g)); }
poly* char_decomp(void) @+{@; return decompose_character(wt_collect()); }
poly* Decomp(poly* p)
{ poly* q=copypoly(p);@+ freemem(p); clrsorted(q);
  set_weight_sorting(grp);
  return decompose_character(Reduce_pol(q));
}

@ The remaining function |Vdecomp| has an implementation that is quite
independent of that of |Decomp|; the reason it is given here is that it is
an example of the use of |wt_ins| and its relatives, and that its
specification is an extension of that of |Decomp| (although it is not
expected to be as efficient as |Decomp| for actual characters).

The code below is in fact similar to that used in the compilation unit
\.{tensor}.  This can be explained from the fact that the algorithm used for
computing tensor products consists of generating the full character of one
tensorand representation, and processing that together with the highest
weight (or decomposition polynomial) of the other tensorand to find the
decomposition for the product; applied to an arbitrary (virtual)
representation~$V$ as first tensorand and the trivial representation as
second tensorand we see that the second part of the computation obtains the
decomposition of~$V$ (which is the tensor product in this case) from the
full (virtual) character of~$V$, and it is this procedure that |Vdecomp|
uses. Like for tensor products we shall generate the full character from the
dominant character using |Weylloop| and we need static variables |the_g| to
store the current group, and |cur_expon| to access a workspace array of
dynamically determined size.

@< Local variables @>=

local simpgrp* the_g;
local entry* cur_expon; 

@ The function |add_decomp_wt| applies the action of |alt_dom| to a single
weight, and contributes the result with a proper coefficient (unless that
coefficient would be~|0|).

@c
local void add_decomp_wt(entry* mu)
{ index i,r=the_g->lierank; boolean neg;
  for (i=0; i<r; ++i) cur_expon[i]=mu[i]+1;
  neg=simp_make_dominant(cur_expon,the_g)%2!=0; /* apply |alt_dom| action */
  for (i=0; i<r; ++i) if (--(cur_expon[i])<0)
    return; /* subtract $\rho$ and quit unless result is dominant */
  wt_ins(cur_expon,one,neg); /* and insert into result */
}

@ Unlike |Decomp|, the result of |Vdecomp| on polynomials is computed by
linear combination of the results from individual terms. If the group is
simple, |simp_vdecomp_irr| computes the result for a single weight~|lambda|.

@c
local poly* simp_vdecomp_irr(entry* lambda, simpgrp* g)
{ the_g=g; wt_init(g->lierank);
@/Weylloopinit(g); Weylloop(add_decomp_wt,lambda); Weylloopexit();
  return wt_collect();
}

@ For composite groups |vdecomp_irr| does a fairly standard combination of
the decompositions for individual factors. Starting with a single-term
polynomial for the toral part, the virtual decompositions of the parts for
individual simple components are multiplied together as polynomials in
distinct variables, which is what |Disjunct_mul_pol_pol| does.

@c
local poly* vdecomp_irr(entry* lambda)
{ if (type_of(grp)==SIMPGRP) return simp_vdecomp_irr(lambda,&grp->s);
  if (simpgroup(grp)) return simp_vdecomp_irr(lambda,Liecomp(grp,0));
  { poly* result; index i;
    @< Advance |lambda| to start of toral entries,
       and initialise |result| for toral part @>
    for (i=grp->g.ncomp-1; i>=0; --i)
      /* traverse simple components in reverse order */
    { simpgrp* g=Liecomp(grp,i); lambda-=g->lierank;
      result= Disjunct_mul_pol_pol(simp_vdecomp_irr(lambda,g),result);
    }
    return result;
  }
}

@ The toral part of a weight for~|g| starts at index |Ssrank(g)|; the rest
is trivial here.

@< Advance |lambda| to start of toral entries... @>=
{ index td=grp->g.toraldim; @+ lambda+=Ssrank(grp);
@/result=mkpoly(1,td); copyrow(lambda,*result->elm,td); *result->coef=one;
}

@ The linear combination of results from |vdecomp_irr| by |Vdecomp| is
entirely standard.

@c
poly* Vdecomp(poly* p)
{ index i,r=Lierank(grp); poly* result=poly_null(r);
  cur_expon=mkintarray(r); /* large enough */
  for (i=0; i<p->nrows; ++i)
    result=Addmul_pol_pol_bin(result,vdecomp_irr(p->elm[i]),p->coef[i]);
  freearr(cur_expon); return result;
}

@* Index.
