\input LiEwebmac

\def\norm#1{|#1|^2} % we can still use `|' here
\def\muo{\mu_0}
\def\rhoxhat{\hat\rho}
\def\cxlambda{c_\lambda}
\def\posr{\Phi^\plus}
\def\posrxweight{\posr_{\rm wt}}

@* Computation of dominant characters.
This file deals with the implementation of the functions |dim| (for
representations), |dom_weights|, and |dom_char|.

@h "lie.h"
@d local static
@f local static
@f alpha NULL /* generic root */
@f lambda NULL /* generic weight */
@f mu NULL /* another one */
@f nu NULL /* and another */
@f muo NULL
@f rho NULL /* weight with $\<\rho,\alpha_i>=1$ for all $i$ */
@f rho_hat NULL /* dual,
  linear form with |rho_hat(@t$\alpha_i$@>)==det_Cartan| for all $i$ */
@f c_lambda NULL
@f posr NULL /* list of all positive roots */
@f posr_weight NULL /* same, but in weight coordinates */

@c
@< Auxiliary functions @>

@ The external interface is determined by the header file \.{domchar.h}.

@( domchar.h @>=
@< Prototypes of globally visible functions @>@;

@*1 Dimensions of representations.
It will be a nice start to compute the dimension of a representation,
although this is in fact quite independent of the things done later.

@< Proto... @>=
bigint* simp_dim_irr(entry* lambda,simpgrp* g);
bigint* DimIrr(entry* lambda);
bigint* Dim(poly* p);

@ For each irreducible constituent we shall use Weyl's dimension formula:
$$ \dim(V_\lambda)= \prod_{\alpha\in\posr}
    {(\lambda+\rho,\alpha)\over(\rho,\alpha)}
$$
The linear form $(\,\cdotp,\alpha)$ on the weight lattice is obtained by 
multiplying the entries of~|alpha| by the norms of the
corresponding simple roots, which are available from |g->root_norm|.

@c
bigint* simp_dim_irr(entry* lambda, simpgrp* g)
{ index i,j,r=g->lierank;
  bigint* denom=copybigint(one,NULL),* numer=copybigint(one,NULL);
  matrix* posr=simp_proots(g); entry* norm=g->root_norm->compon;
  for (i=0; i<posr->nrows; ++i)
  { entry* alpha=posr->elm[i]; entry den=0,num=0;
    for (j=0; j<r; ++j)
      {@; den += alpha[j]*norm[j]; num += alpha[j]*norm[j]*(lambda[j]+1); }
    denom=mul1(denom,den); numer=mul1(numer,num);
  }
  return quotient(numer,denom);
}

@ For composite groups, we take the product of the dimensions obtained for
the simple components.

@c
bigint* DimIrr(entry* lambda)
{ if (type_of(grp)==SIMPGRP) return simp_dim_irr(lambda,&grp->s);
  if (simpgroup(grp)) return simp_dim_irr(lambda,Liecomp(grp,0));
  { index i,s=0; bigint* res=one;
    for (i=0; i<grp->g.ncomp; ++i)
    @/{@; simpgrp* g=Liecomp(grp,i); @+
      res=mult(res,simp_dim_irr(&lambda[s],g));
      s+=g->lierank;
    }
    return res;
  }
}

@ For composite representations, we sum the dimensions of the irreducible
constituents, with suitable multiplicities. Any non-dominant terms in~$p$
are made dominant before calling |DimIrr|.

@c
bigint* Dim(poly* p)
{ index i,r=p->ncols; bigint* res=null; entry* lambda= mkintarray(r);
  for (i=0; i<p->nrows; ++i)
  @/{@; copyrow (p->elm[i],lambda,r); make_dominant(lambda); 
    res=add(res,mult(p->coef[i],DimIrr(lambda)));
  }
  freearr(lambda); return res;
}

@*1 Weight levels.
Throughout this file the concept of weight levels will be essential.
The level of a positive root above~|0| is defined as the sum of the
coefficents when it is expressed on the basis of simple roots. For weights
we define $\mu\below\lambda$ if |lambda-mu| is a sum of simple roots (a
positive element in the root lattice), and if the sum of its coefficients on
the root basis is~|n|, we say |mu| is |n|~levels below~|lambda|.

Computing levels on the root basis is trivial: it consists of applying the
linear form that is the transpose of the all-one vector $[1,1,\ldots,1]$. To
compute levels on the weight basis we can transform coordinates to the root
basis (scaling by the determinant of the Cartan matrix to ensure integer
coordinates) by multiplying by the matrix |i_Cartan|, and then apply the same
linear form, while compensating for the mentioned scalar factor. It is more
efficient however to transform the linear form to weight coordinates, by
multiplying by the transpose of the matrix |i_Cartan|, which can then be
applied directly to weight vectors. We call the vector encoding this
transformed linear form the level vector; it has the property that its
standard inner product with each row of the Cartan matrix is equal to the
determinant of that matrix. We shall use such a level vector for each simple
component of the current group; its value is computed by the function
|simp_level_vec|, from the matrix returned by~|simp_icart|.

@< Auxiliary functions @>=
local entry* simp_level_vec(simpgrp* g)
{ index i,r=g->lierank; entry* res=mkintarray(r),* ones=mkintarray(r);
  for (i=0; i<r; ++i) ones[i]=1;
  mulmatvecelm(simp_icart(g)->elm,ones,res,r,r); freearr(ones); return res;
}


@*1 The dominant support of a character.
Before we handle the computation of dominant characters themselves, we must
consider the problem of finding the support of that character in case of a
highest weight module, i.e., the set of dominant weights below a given
dominant weight~|lambda|.

The function |simp_domw| computes the set of dominant weights of the
representation of highest weight~|lambda|. The rows of the resulting array
|weights| are weakly decreasing in height.

@< Aux... @>=
local matrix* simp_domw(entry* lambda, simpgrp* g, vector** levels)
{ index r=g->lierank;
  matrix* posroots=Matmult(simp_proots(g),simp_Cartan(g));
  entry* x=mkintarray(r);
  matrix* weights=mkmatrix(EXTEND,r);
  vector* wt_levels;
  @< Set |wt_levels| to a fresh vector of the right size @>
  @< Fill |weights| with |lambda| and all dominant weights below it,
     while storing the first index of each weight level in |wt_levels| @>
  if (levels!=NULL) *levels=wt_levels; @+ else freemem(wt_levels);
  freearr(x); freemem(posroots); return weights;
}

@ The purpose of |wt_levels| is to store the index within |weights| of the
first weight at each level below~|lambda|, so |wt_levels->compon[0]| will
be~|0| since |weights| will start with~|lambda|, |wt_levels->compon[1]| will
be~|1| since there are no other weights at level~|0| below~|lambda|, and so
on. Each position in |wt_levels| is filled with an index, whether or not there
are any weights at that level below~|lambda|; if there are none, than the next
entry of |wt_levels| will have the same value. To facilitate the use of
|wt_levels|, we include an entry for the level whose weights would have to be
negative, and therefore certainly do not exist. The value of this level is
computed by taking the standard inner product of the level vector
with~|lambda|, dividing by the Cartan determinant (rounding downwards to an
integer), and adding~|1|. Since the first entry of |wt_levels| corresponds to
level~|0| below~|lambda|, we actually add~|2| to the result of the division to
obtain the size of~|wt_levels|.

@< Set |wt_levels| to a fresh vector... @>=
{ entry* rho_hat=simp_level_vec(g); index i,lev=0; @+
  for (i=0; i<r; ++i) lev+=lambda[i]*rho_hat[i]; freearr(rho_hat);
  wt_levels=mkvector(lev/simp_detcart(g)+2);
}

@ Although all weights in the full weight system can be obtained from~|lambda|
by successively subtracting simple roots, we need to be more careful when
restricting to dominant weights only. In turns out that each dominant weight
in the weight system can be obtained from another such weight by subtracting
some (not necessarily simple) root, except of course for the starting
point~|lambda|. Therefore for each successive level below~|lambda| we traverse
the root system by increasing root level, and combine each root with all
previously computed elements of~|weights| at the right level, while
maintaining the vector |wt_levels| to locate the relevant region of |weights|
for each level. The levels of the roots are similarly found in |g->level|,
but here entry~|d| gives the first index of a root whose level exceeds~|d|
(there are no roots at level~|0|); we also use that for the roots (unlike for
|weights|) there are no empty levels, so for each new root $\alpha_k$ its
level~|d| increases by at most~|1|.

@< Fill |weights|... @>=
{ index lev,d,k;
  index lev_start=1,end=1; /* start and end of current level */

  copyrow(lambda,weights->elm[0],r); wt_levels->compon[0]=0;
  wt_levels->compon[1]=1;

  for (lev=1; lev<wt_levels->ncomp-1; wt_levels->compon[++lev]=lev_start=end)
    for (k=0,d=1; k<posroots->nrows; ++k) /* try roots by increasing level */
      if (k==g->level->compon[d] && ++d>lev) break;
	/* now root $\alpha_k$ has level |d| */
      else
      @< Subtract $\alpha_k$ from each element of |weights| at level |lev-d|
	 below~|lambda|, and incorporate each new weight thus formed
	 into~|weights| @>
  weights->nrows=end;
}

@ As we traverse the weights~|mu| at the proper level of~|weights|, the
selection of those that remain dominant after subtraction of~$\alpha_k$ is
facilitated by the function |pos_subrow|, which subtracts two rows into a
third one, and either succeeds returning~|true|, or if the result would have
a negative entry, returns~|false| instead. If successful, the result of the
subtraction is written into the workspace array~|x|.

@< Subtract $\alpha_k$ from each... @>=
{ index w; 
    /* index of a dominant weight $\mu\below\lambda$ at level |lev-d| */
  for (w=wt_levels->compon[lev-d]; w<wt_levels->compon[lev-d+1]; ++w)
    if (pos_subrow(weights->elm[w],posroots->elm[k],x,r))
      /* |x=mu-@t$\alpha_k$@>| must be dominant */
      @< If |x| does not occur in |weights| yet, add it as a new row @>
}

@ The dominant weight~|x| just computed may be present in |weights|, but only
at the current level, which runs from the index~|lev_start| to~|end|. Since
this level is not ordered in any particular way, we search it linearly, and if
|x| is found, we avoid adding it another time; this is one of those situations
where a jump is the most convenient control structure to achieve the desired
effect.

@< If |x| does not occur... @>=
{ index i; @+
  for (i=lev_start; i<end; ++i) @+
    if (eqrow(x,weights->elm[i],r)) goto next_wt;
  if (end==weights->rowsize)
    {@; weights->nrows=end; weights=extendmat(weights); }
  copyrow(x,weights->elm[end++],r);
next_wt: {}
}

@ While |simp_domw| does the hard work for computing systems of dominant
weights, the function that actually implements the \LiE. function
|dom_weights| is |Domweights|; it handles composite groups by combining the
results of calls to |simp_domw| for each simple factor. This combination
amounts to taking the Cartesian product of the sets represented by the
matrices, concatenating for each element of the product the vectors for the
individual simple groups, and possibly adding a fixed part for the central
torus. Since this operation is also needed for computing Weyl group orbits for
composite groups, it is performed by a separate and globally known function
|Cartes_prod|. Its arguments are an array~|M| of |n|~matrices, an array
|toral| that prescibes the toral part of all rows, and the rank~|r| of the
composite group (which defines implicitly the size of |toral|). Since the
array~|M| is only used to collect the argument matrices, it is freed
by~|Cartes_prod|.

@< Proto... @>=
matrix* Cartes_prod(matrix** M,index n,entry* toral,index r);
matrix* Domweights(vector* lambda);

@ Given |Cartes_prod|, which will be discussed below, the implementation of
|Domweights| is quite trivial.

@c
matrix* Domweights(vector* lambda)
{ if (type_of(grp)==SIMPGRP) return simp_domw(lambda->compon,&grp->s,NULL);
  if (simpgroup(grp)) return simp_domw(lambda->compon,grp->g.liecomp[0],NULL);
  { index i,j,n=grp->g.ncomp,r=Lierank(grp);
    matrix** M=alloc_array(matrix*,n);
    for (i=0,j=0; i<n; j+=grp->g.liecomp[i++]->lierank)
      M[i]=simp_domw(&lambda->compon[j],grp->g.liecomp[i],NULL);
    return Cartes_prod(M,n,&lambda->compon[j],r);
  }
}

@ The implementation of |Cartes_prod| is straightforward. We maintain the
index~|j| of the next row of the result, and also the digits of its mixed-radix
representation in |pos|, where |pos[i]| serves as row index into |M[i]|. We
use a working vector~|lambda| to prepare each row in, so that parts of a row
that coincide with the previous row need not be retrieved again from the
various |M[i]| (and from |toral|). After copying |lambda| into |result| we
increment |pos| in the mixed-radix system, and leave~|i| at the first entry
of~|pos| that changed (i.e., the one that was incremented rather than reset
to~|0|), as in indication if which part of~|lambda| must be updated; when all
entries of~|pos| are reset, we are done.

@c
matrix* Cartes_prod(matrix** M,index n,entry* toral,index r)
{ index i=0,j=0; entry* pos,* offset,* lambda; matrix* result;
  @< Compute |pos| and |offset|, allocate |result|, and fill the toral part
     of |lambda| @>
  do /* now |i| indicates first simple factor whose entries in |lambda|
	need to be updated */
  { for (;i<n;++i) copyrow(M[i]->elm[pos[i]],&lambda[offset[i]],M[i]->ncols);
      /* update |lambda| */
    copyrow(lambda,result->elm[j++],r);
      /* copy full vector |lambda| to result */
    while (--i>=0 && ++pos[i] == M[i]->nrows) pos[i]=0;
      /* advance |pos| to match |j| */
  } while (i>=0); /* full circle when |pos| returns to all zeroes */
  freearr(pos); {@; while (++i<n) freemem(M[i]);} freearr(M); return(result);
}

@ The initial values |pos[i]| are all zero, and the position |offset[i]| where
the entries from |M[i]| start is the sum of all |M[j]->ncols| for |j<i|;
furthermore the number of rows of the result is the product all
|M[i]->nrows|.

@< Compute |pos| and |offset|... @>=
{ index i, s=0, p=1;
    /* sum of row sizes and product of column sizes of the |M[i]| */
  pos=mkintarray(2*n+r),offset=&pos[n],lambda=&offset[n];
  for (i=0; i<n; ++i)
    {@; pos[i]=0; offset[i]=s; s+=M[i]->ncols; p*=M[i]->nrows; }
  result = mkmatrix(p,r);
  copyrow(toral,&lambda[s],r-s); /* toral part of working vector */
}

@*1 The dominant character multiplicities.
Our next project will be to attach a multiplicity to each dominant weight
$\mu\below\lambda$, thus forming the dominant character polynomial for the
irreducible module with highest weight~|lambda|. As an option we allow
specification of one particular~|muo|, in which case only the multiplicity of
that weight will be returned; for the full character we pass |muo==NULL|. Since
the result type must be the same in both cases, we always return a polynomial,
but if |muo!=NULL|, this will be a polynomial in |0|~indeterminates. Like for
|dom_weights|, we start with the case of simple groups.

@< Prototypes... @>= poly* simp_domchar(entry* lambda,entry* muo,simpgrp* g);

@~We convert the matrix of dominant weights to a polynomial, and then
compute multiplicities by descending level from~|lambda|. In the
initialisation we will set the index |last_row| to point to the final row
of~|dom_w|, or in case |muo!=NULL| to the row containing~|muo|; we stop when
the multiplicity for this row has been determined.

@c
poly* simp_domchar(entry* lambda,entry* muo,simpgrp* g)
{ index k;
  vector* wt_levels; poly* dom_w= mat2pol(simp_domw(lambda,g,&wt_levels));
  @< Compute some constants and allocate work space for |simp_domchar| @>@;

  dom_w->coef[0]=one; /* initialise first multiplicity to 1 */
  for (k=1; k<=last_row; ++k) @< Find multiplicity of |k|-th dominant weight @>
  freemem(wt_levels); @< Free local constants and work space @>
  if (muo==NULL) return dom_w; @+
  else @< Return the coefficient of |muo| in |dom_w| @>
}

@ To compute the dominant weight multiplicities, we use Freudenthal's
recursion formula, which expresses each of them in terms of the
multiplicities of higher weights. We shall write $m_\lambda(\mu)$ for the
multiplicity of |mu| in the irreducible module $V_\lambda$; we have of
course $m_\lambda(\lambda)=1$. The recursion formula states that for
$\mu\prec\lambda$
$$
  m_\lambda(\mu)={2\over\norm{\lambda+\rho}-\norm{\mu+\rho}}
  \sum_{\alpha\in\Phi^+}\sum_{n>0}(\mu+n\alpha,\alpha)m_\lambda(\mu+n\alpha)
$$
Here of course the infinite sum over $n$ breaks off when $\mu+n\alpha$ leaves
the weight system of~$V_\lambda$. With the scaling of the inner product we
use, the individual terms in the denominator need not have integer values
(although their difference does), so we compute the fraction with an extra
factor |det| in numerator and denominator, where |det| is the determinant of
the Cartan matrix. An auxiliary function~|c| will de defined below that
computes $\\{det}*\norm{\mu+\rho}$ for any weight~|mu|; since the value for
|mu==lambda| is needed repeatedly, we precompute it and store it
in~|c_lambda|.

A simplification of the summation above is possible when the weight~|mu| has
a non-trivial stabiliser~$W_\mu$ in the Weyl group (a rather common
circumstance): all positive roots~|alpha| that a lie in the same
$W_\mu$-orbit give the same contribution in the summation. To take advantage
of this, we compute for each |mu| an array |mults| of multiplicities, one
for each positive root, such that for each class of equivalent positive
roots, one of the multiplicities is set to the size of the class, and all
others are set to~|0| (allowing the corresponding computation to be skipped).

@< Compute some constants... @>=
matrix* roots=Matmult(simp_proots(g),simp_Cartan(g));
  /* positive roots on weight basis */
long c_lambda=c(lambda,g), det=simp_detcart(g);
entry* mults=mkintarray(g->roots->nrows); /* multiplicities for roots */

@~@<Free local constants... @>=
freemem(roots); freearr(mults);

@ The difference $\norm{\lambda+\rho}-\norm{\mu+\rho}$ is actually always an
even integer; the reason is that |lambda-mu| lies in the root lattice, so
that the mentioned difference is equal to |(lambda+mu+2 rho,lambda-mu)|
which is congruent modulo~$2$ to $\norm{\lambda-\mu}$, and $\norm x$ is even
for any |x| in the root lattice. Therfore the factor before the summation is
the inverse of an integer, which we shall call~|denominator|; being a
quadratic expression in the coordinates of |lambda|~and~|mu| it will not get
excessively large, so we store it in a |long| value. On the other hand, the
result of the summation can become very large, so we use a |bigint| to
represent it. The value of $m_\lambda(\mu)$ is eventually obtained by
dividing |sum| by |denominator|, where the latter is converted to |bigint|
for the purpose of the division.

@< Find multiplicity... @>=
{ index i; entry* mu=dom_w->elm[k];
  bigint* sum=null; long denominator=(c_lambda-c(mu,g))/(2*det);
  gather_roots(mu,roots->elm,g,mults);
    /* compute |mults| for stabiliser group $W_\mu$ */
  for (i=0; i<roots->nrows; ++i)
    if (mults[i]>0)
    @/{@; entry* alpha=roots->elm[i];
      @< Add $\\{mults}[i]*
	      \sum_{n>0}(\mu+n\alpha,\alpha)*m_\lambda(\mu+n\alpha)$
	 to |sum| @>
    }
  dom_w->coef[k]=quotient(sum,entry2bigint(denominator));
  setshared(dom_w->coef[k]);
}

@ In computing the inner summation we need 2 workspace vectors |x,y| for
storing weights.

@< Compute some constants... @>=
index r=g->lierank; @+
entry* x=mkintarray(2*r),* y=&x[r]; /* 2 arrays of size |r| */

@~@<Free local constants... @>=
freearr(x);

@ Here is the inner summation. To speed up the computation of the values
$(\mu+n\alpha,\alpha)$, which form an arithmetic progression, we compute the
values |(alpha,alpha)| and |(mu,alpha)| first. For this, it is convenient to
recover the expression~|a| of |alpha| in root coordinates from the list of
positive roots; multiplying each of its entries by the norm of the
corresponding simple root we obtain the coordinates of the linear form
$(\,\cdotp,\alpha)$ on the weight lattice. Within the loop, the weight
$\mu+n\alpha$ need not be dominant, so we convert a copy of it using
|simp_make_dominant|, and look the resulting weight up in~|dom_w| using an
auxiliary function |locate| in order to find the previously computed
multiplicity $m_\lambda(\mu+n\alpha)$. In case |locate| indicates that the
weight is not present, by returning~|-1|, we break off the summation.

@< Add $\\{mults}[i]... @>=
{ index k; entry* a=g->roots->elm[i]; entry inp=0,d=0;
  entry* norm=g->root_norm->compon;
  for (k=0; k<r; ++k) /* set |d=(alpha,alpha)| and |inp=(mu,alpha)| */
    {@; d+=alpha[k]*norm[k]*a[k]; inp+=mu[k]*norm[k]*a[k]; }
  copyrow(mu,x,r); /* initialise |x=mu| */
  do
  { add_xrow_to(x,1,alpha,r);
    copyrow(x,y,r); simp_make_dominant(y,g);
    if ((k=locate(y,dom_w->elm,rho_hat,wt_levels->compon,g))<0)
      break; /* not found */
    sum=add(sum,mul1(copybigint(dom_w->coef[k],NULL),mults[i]*(inp+=d)));
  } while(true);
}

@ In order to search weights by level, we need the level vector |rho_hat|.

@< Compute some constants... @>=
entry* rho_hat=simp_level_vec(g);

@~@< Free local constants... @>=
freearr(rho_hat);

@ Recall that |simp_domw| has ordered its weights by decreasing level, and
has set |wt_levels| to indicate the start of each level. With further aid of
the level vector~|rho_hat|, |locate| performs a linear search of the proper
subrange of weights. Since |mu| is dominant, the depth |l| of |mu| below
|lambda| cannot be so great as to make the reference to |level[l+1]| invalid.

@< Auxiliary functions @>=
local index locate(entry* mu,entry** m,entry* rho_hat,entry* level,simpgrp* g)
{ index l=0,i,r=g->lierank; entry* lambda=m[0]; /* the highest weight */
  for (i=0; i<r; ++i) l += (lambda[i]-mu[i])*rho_hat[i];
    /* level of |mu| below |lambda| */
  if ((l/=simp_detcart(g))>=0)
    for (i=level[l]; i<level[l+1]; ++i) @+
      if (eqrow(mu,m[i],r)) return i;
  return -1; /* weight |mu| is not found if we get here */
}

@ The computation of the multiplicities |mults| of all positive roots,
obtained by gathering all such root in the same $W_\mu$-orbit, is done by
|gather_roots|. The roots~|alpha| that are selected as representatives are
those with $\<\alpha,\alpha_j>\geq0$ for all~|j| with $s_j\in W_\mu$; by the
algorithm similar to |dominant| for the subgroup $W_\mu$ each $W_\mu$-orbit
contains a unique such element, which is moreover a positive root whenever
the orbit contains any positive roots. For convenience we pass the matrix
|posr_weight| of positive roots expressed in weight coordinates to
|gather_roots|, so that the value $\<\alpha,\alpha_j>$ is readily available
as |posr_weight[i][j]|, if |i| is the index of~|alpha| in the list of
positive roots. After initialising all multiplicities to~|1| we traverse all
positive roots~|alpha| and all fundamental roots $\alpha_j$; whenever we
find that $\<\alpha,\alpha_j><0$, we move the multiplicity of |alpha| to
that of $s_j(\alpha)$. In order that all multiplicities are eventually
deposited in the representative of the $W_\mu$-orbit, we must ensure that
this root $s_j(\alpha)$ is encountered in the traversal after~|alpha|; this
is obtained by traversing the roots by increasing level.

@< Aux... @>=
local void gather_roots
  (entry* mu, entry** posr_weight, simpgrp* g, entry* mults)
{ index i,j,k,l,r=g->lierank,n=g->roots->nrows;
  entry** posr=g->roots->elm;
  for (i=0; i<n; ++i) mults[i]=1;
  for (l=1; l<g->level->ncomp-1; ++l)
    /* traverse all root levels except the highest */
    for (j=0; j<r; ++j)
      if (mu[j]==0)
	@< For any root |alpha| at level |l| with $\<\alpha,\alpha_j><0$,
	   move its multiplicity to a higher level @>
}

@ Within one level, we may of course skip those roots whose multiplicity has
already been moved to a higher level for some earlier value of~|j|. For a
root |alpha| whose multiplicity is to be moved, we locate $s_j(\alpha)$ by
calling |find_root|, whose search ought to succeed. For this call we need to
have $s_j(\alpha)$ in root coordinates (whereas |posr_weight| is in weight
coordinates); therefore we use the matrix~|posr| of positive roots for the
current group, temporarily change the row for~|alpha| into $s_j(\alpha)$ by
adding $e=-\<\alpha,\alpha_j>$ to the coordinate of~$\alpha_j$, and look it
up at $e$ levels above the current one.  Since |find_root| only inspects the
indicated level, it will not erroneously return the index~|i|, which would
give a match at that time, but is at the wrong level; we must take care
however to shut off interrupts during the time that root |alpha| is modified
in the list of positive roots.

@< For any root |alpha| at level |l|... @>=
for (i=g->level->compon[l-1]; i<g->level->compon[l]; ++i)
  if (mults[i]!=0)
  { entry e=-posr_weight[i][j];
    if (e>0)
    { ignore_intr(); posr[i][j]+=e; k=find_root(posr[i],l+e,g);
      posr[i][j]-=e; allow_intr(); assert(k>=0);
      mults[k]+=mults[i]; mults[i]=0;
    }
  }

@ Computing~|c(mu,g)| is like computing |Inprod(mu+rho,mu+rho)|, but |Inprod|
cannot be used since it uses root coordinates. We obtain
$\\{det}*(\mu+\rho,\mu+\rho)$ as follows: a copy of |mu+rho| is multiplied
by |i_Cartan| giving $\\{det}*(\mu+\rho)$ in root coordinates, each entry is
multiplied by the norm of the corresponding simple root giving the linear
form $\\{det}*(\,\cdotp,\mu+\rho)$ on the weight lattice, which is then
applied to another copy of |mu+rho|. For efficiency we perform the summations
explicitly rather than calling matrix and vector multiplication routines,
ordering them so as to minimise the number of multplications in the inner
loop. 

@< Auxiliary functions @>=
local long c(entry* mu, simpgrp* g)
{ index a,r=g->lierank,i,j;
  entry** ic=simp_icart(g)->elm,* n=g->root_norm->compon;
  for (a=0,j=0; j<r; ++j)
  @/{@; entry s=0;
    for (i=0; i<r; ++i) s+=(mu[i]+1)*ic[i][j];
    a+=s*n[j]*(mu[j]+1);
  }
  return a;
}

@ It remains to show how we handle a value |muo!=NULL|. We use |locate| to
find its row in |dom_w| and set |last_row| to that value. If |muo| does not
occur in in |dom_w|, then we shall have |last_row==-1|, and all computation
will be conveniently skipped.

@< Compute some constants... @>=
index last_row=  muo==NULL ? dom_w->nrows-1
  : locate(muo,dom_w->elm,rho_hat,wt_levels->compon,g);

@ On returning from |simp_domchar| we pick the desired coefficient from the
computed polynomial, and pack it into a new polynomial with
|0|~inderterminates; in case |last_row==-1|, we return a zero polynomial.

@< Return the coefficient of |muo|... @>=
{ poly* mul=poly_null(0); /* polynomial in |0| indeterminates */
  if (last_row<0) return mul;
  mul=copypoly(mul); mul->coef[0]=dom_w->coef[last_row];
  setshared(mul->coef[0]); freemem(dom_w); return mul;
}

@ Like with |dom_weights|, we export a version that can handle composite
groups.

@< Prototypes... @>= poly* Domchar_irr(entry* lambda,entry* muo);

@~The work of combining the dominant characters from the simple factors and
from the toral part is done by |Disjunct_mul_pol_pol|; it also properly
handles that case that the multiplicity of a single weight~|muo| is
requested. We traverse the simple components backwards so that the exceptional
case of the toral part can be combined with the initialisation of~|result|.

@c
poly* Domchar_irr(entry* lambda,entry* muo)
{ index i,j,r=Lierank(grp),s=Ssrank(grp); poly* result;
  if (type_of(grp)==SIMPGRP) return simp_domchar(lambda,muo,&grp->s);
  if (simpgroup(grp)) return simp_domchar(lambda,muo,Liecomp(grp,0));
  @< Initialise |result| to the value for the toral part @>
  for (i=grp->g.ncomp-1,j=s; i>=0; --i)
    /* traverse components in reverse order */
  { simpgrp* g=grp->g.liecomp[i]; @+ j-=g->lierank;
    result= Disjunct_mul_pol_pol
      (simp_domchar(&lambda[j],muo==NULL?NULL:&muo[j],g),result);
  }
  return result;
}

@ If |muo| is not specified, we initialise |result| to $1X^{\lambda_t}$ where
$\lambda_t$ is the toral part of~|lambda|. Otherwise we set |result| to
$1X^{[\,]}$~or~$0X^{[\,]}$ depending on whether the toral part of~|muo|
matches that of~|lambda| or not; in the latter case there is no point in any
further computation, so we return the result immediately.

@< Initialise |result|... @>=
if (muo==NULL)
{@; result=mkpoly(1,r-s);
  copyrow(&lambda[s],result->elm[0],r-s); *result->coef=one;
}
else
  if (eqrow(&lambda[s],&muo[s],r-s)) result=poly_one(0); @+
  else return poly_null(0);

@ Finally we also deal wil computing dominant characters of reducible modules.
@< Prototypes... @>= poly* Domchar_p(poly* p);

@~The implementation is a trivial linear expansion over the decomposition
polynomial~|p|.

@c
poly* Domchar_p(poly* p)
{ index i,r=p->ncols; poly* result=poly_null(r);
  for (i=0; i<p->nrows; ++i)
  { testdom(p->elm[i],grp);
    result=Addmul_pol_pol_bin(result,Domchar_irr(p->elm[i],NULL),p->coef[i]);
  }
  return result;
}

@ For implementing the height ordering criteria for matrices and
polynomials, we need level vectors for possibly composite groups. Therefore
we export the function |Lv| that exports a level vector some group~|g|.

@< Prototypes... @>= entry* Lv(object g);

@~Since level vectors are applied to vectors in weight coordinates, we place
zeros in the entries for the toral part.

@c
entry* Lv(object g)
{ if (type_of(g)==SIMPGRP) return simp_level_vec(&g->s);
  if (simpgroup(g)) return simp_level_vec(Liecomp(g,0));
  { index i,j=0,r=Lierank(g); entry* result=mkintarray(r);
    for(i=0; i<g->g.ncomp; ++i)
    { simpgrp* s=g->g.liecomp[i]; entry* l=simp_level_vec(s);
      copyrow(l,&result[j],s->lierank); freearr(l); j+=s->lierank;
    }
    for(i=0; i<g->g.toraldim; ++i) result[j+i]=0;
    return result;
  }
}

@* Index.
