\input LiEwebmac
%\input references

\def\rf#1{\<\,\cdot\,,#1>} % root form
\def\posr{\Phi^\plus} % set of positive roots

@* Basic data for groups.
In this file we shall define quantities that determine the basic properties
of groups, such as the Cartan matrix. Ideally there should be a minimal
amount of literal data from which everything else is computationally
derived, since this gives the least opportunity for clerical errors. In
practice, some choices have been made that imply duplication or redundancy
of data, motivated for instance by the possibility to take advantage of the
sparseness of the Cartan matrix, or simply to avoid algorithms that are more
complicated than a description of their results in all cases (like for the
function |center|).

@h "lie.h"
@f local static
@d local static
@f alpha NULL /* a generic positive root */
@f posr NULL /* the set of positive roots */

@( grpdata.h @>=
@< Prototypes of globally visible functions @>@;

@ We start with a test for invalid type-rank combinations.
@< Proto...@>=
boolean wronggroup(char lietype,index rank);

@~Although $B_2$ and $C_2$ are isomorphic, and similarly for $A_3$
and~$D_3$, we allow each of these, because they have different labellings of
the Dynkin diagram. We do not allow $E_4$ and~$E_5$ however, although it
would make some sense to identify these with variant labellings of $A_4$
and~$D_5$ repectively (if we did allow them, many case distinctions would
have to be made more elaborate).

@c
boolean wronggroup(char lietype,index rank)
{ return lietype=='T' ? rank<0
       : lietype=='A' ? rank<1
  @|   : lietype=='B' ? rank<2
       : lietype=='C' ? rank<2
       : lietype=='D' ? rank<3
  @|   : lietype=='E' ? rank<6 || rank>8
       : lietype=='F' ? rank!=4
       : rank!=2;
}

@ Here are some frequently used functions.
@< Proto...@>=
boolean simpgroup(object g);
index Lierank(object grp);
index Ssrank(object g);

@~The function |simpgroup| assumes that |type_of(g)==GROUP|, since objects
of type |simpgrp| have to be treated differently from objects of type
|group| that happen to represent a simple group.
@c
boolean simpgroup(object g)
{@; return (g->g.toraldim==0 && g->g.ncomp==1); }

index Lierank(object grp)
{ index i,r; @+ if (type_of(grp)==SIMPGRP) return grp->s.lierank;

  r=grp->g.toraldim; @+
  for (i=0; i<grp->g.ncomp; ++i) r += (Liecomp(grp, i))->lierank;
  return r;
}

index Ssrank(object g) /* Semisimple rank */
{ index i,r=0; @+ if (type_of(g)==SIMPGRP) return g->s.lierank;
  for (i=0; i<g->g.ncomp; ++i) r += (Liecomp(g,i))->lierank;
  return r;
}

@*1 The Cartan matrix and its inverse.
The most basic data describing a group is its Cartan matrix. Starting a
general pattern for defining such functions, we provide a special function
for simple groups, from which a function for general groups is constructed.
Internally in \LiE. both functions are available, but the user can access
only the general version.

@< Proto...@>=
matrix* simp_Cartan(simpgrp* g);
matrix* Cartan(void);
entry simp_detcart(simpgrp* g);
entry Detcartan(void);
matrix* simp_icart(simpgrp* g);
matrix* Icartan(void);

@~ The Cartan matrix is of course specified explicitly for each type of
simple group. In fact, the matrix is always quite close to that of
type~$A_n$, so we just record explicitly the deviations from that matrix.

@c
matrix* simp_Cartan(simpgrp* g)
{ if (g->cartan!=NULL) return g->cartan;
  { entry r=g->lierank; matrix* cartan=g->cartan=mat_null(r,r);
    entry** m=cartan->elm; @+
    setlonglife(cartan); /* make Cartan matrix permanent */
    @< Initialise |m| to Cartan matrix of type $A_r$ @>
    switch (g->lietype)
    {	    case 'B': m[r-2][r-1]= -2;
    break;  case 'C': m[r-1][r-2]= -2;
    break;  case 'D': m[r-3][r-1]=m[r-1][r-3]= -1;
		      m[r-2][r-1]=m[r-1][r-2]=0;
    break;  case 'E': m[0][1]=m[1][0]=m[1][2]=m[2][1]=0;
		      m[0][2]=m[2][0]=m[1][3]=m[3][1]= -1;
    break;  case 'F': m[1][2]= -2;
    break;  case 'G': m[1][0]= -3;
    }
    return cartan;
  }
}

@ Here the pattern is |2|'s on the diagonal, |-1|'s on the off-diagonals, and
|0|'s elsewhere.

@< Initialise |m|... @>=
{@; index i; m[0][0]=2; 
  for (i=1; i<r; ++i) {@; m[i][i]=2; m[i-1][i]=m[i][i-1]= -1; }
}

@ For composite groups we copy the Cartan matrices for the simple factors as
blocks along the main diagonal of a matrix that is |0| elsewhere. A central
torus leads to extra null columns, since the Cartan matrix should transform
from root to weight coordinates.

@c
matrix* Cartan(void)
{ if (type_of(grp)==SIMPGRP) return simp_Cartan(&grp->s);
  if (simpgroup(grp)) return simp_Cartan(Liecomp(grp,0));
  { index i,j, t=0;
    matrix* cartan=mat_null(Ssrank(grp),Lierank(grp));
    for (i=0; i<grp->g.ncomp; ++i)
    { index r=Liecomp(grp,i)->lierank;
      entry** c=simp_Cartan(Liecomp(grp,i))->elm;
      for (j=0; j<r; ++j) copyrow(c[j],&cartan->elm[t+j][t],r);
      t+=r;
    }
    return cartan;
  }
}

@ Although it could be computed from the Cartan matrix, its determinant is
recorded literally for all types of simple groups in |simp_detcart|. For
groups with a central torus the Cartan matrix is not square, so |Detcartan|
takes the determinant of the (square) Cartan matrix of the semisimple part of
the group.

@c
entry simp_detcart(simpgrp* g)
{ char t=g->lietype; index r=g->lierank; @+
  return t=='A' ? r+1
       : t=='B' || t=='C' ? 2
       : t=='D' ? 4
       : t=='E' ? 9-r
       : 1;
}

entry Detcartan(void)
{ if (type_of(grp)==SIMPGRP) return simp_detcart(&grp->s);
  { index i; entry result=1;
    for (i=0; i<grp->g.ncomp; ++i) result *= simp_detcart(Liecomp(grp,i)); 
    return result;
  }
}


@ The adjugated or ``inverse'' matrix of Cartan matrix is also represented
literally. The validity of these data is easily checked by multiplication
with the Cartan matrix.

@c
matrix* simp_icart(simpgrp* g)
{ if (g->icartan) return g->icartan;
  { index i, j, r=g->lierank;
    matrix* icartan=g->icartan=mkmatrix(r,r); entry** m=icartan->elm;
    @+ setlonglife(icartan); /* permanent data */
    switch (g->lietype)
    {	      case 'A':
      for (i=0; i<r; ++i) @+ for (j=0; j<=i; ++j)
	m[i][j]=m[j][i]=(r-i)*(j+1);
    @+ break; case 'B':
      for (i=0; i<r; ++i) @+ for (j=0; j<=i; ++j) m[i][j]=m[j][i]=2*(j+1);
      for (i=0; i<r; ++i) m[r-1][i]=i+1;
    @+ break; case 'C':
      for (i=0; i<r; ++i) @+ for (j=0; j<=i; ++j) m[i][j]=m[j][i]=2*(j+1);
      for (i=0; i<r; ++i) m[i][r-1]=i+1;
    @+ break; case 'D':
      for (i=0; i<r-2; ++i) @+ for (j=0; j<=i; ++j) m[i][j]=m[j][i]=4*(j+1);
      for (i=0; i<r-2; ++i) m[r-1][i]=m[r-2][i]=m[i][r-1]=m[i][r-2]=2*(i+1);
      m[r-1][r-1]=m[r-2][r-2]=r; m[r-1][r-2]=m[r-2][r-1]=r-2;
    @+ break; case 'E':
      m[0][0]=4; m[1][0]=m[0][1]=r-3; m[0][2]=m[2][0]=r-1;
      m[1][1]=r; m[1][2]=m[2][1]=2*r-6; m[2][2]=2*r-2;
      for (i=1; i<r-2; ++i) @+ for (j=0; j<3; ++j) m[r-i][j]=m[j][r-i]=(j+2)*i;
      for (i=1; i<r-2; ++i) @+ for (j=1; j<=i; ++j)
	m[r-i][r-j]=m[r-j][r-i]=(9-r+i)*j;
    @+ break; case 'F':
      for (i=1; i<4; ++i) @+ for (j=1; j<4; ++j) m[r-i][j-1]=i*j;
      m[1][2]=8;
      for (i=0; i<3; ++i) m[0][i]=m[r-i-1][3]=i+2;
      m[0][3]=2;
    @+ break; case 'G': m[0][0]=m[1][1]=2; m[0][1]=1; m[1][0]=3;
    }
    return icartan;
  }
}

@ Fo composite groups these matrices must be brought onto the common
denominator |Detcartan()|.

@c
matrix* Icartan(void)
{ if (simpgroup(grp)) return simp_icart(Liecomp(grp,0));
  { matrix* result=mat_null(Lierank(grp),Ssrank(grp)); entry** m=result->elm;
    index k,t=0;
    entry det=Detcartan(); /* product of determinants of simple factors */
    for (k=0; k<grp->g.ncomp; ++k)
    { simpgrp* g=Liecomp(grp,k);
      index i,j,r=g->lierank;
      entry** a=simp_icart(g)->elm;
      entry f=det/simp_detcart(g); /* multiplication factor */
      for (i=0; i<r; ++i) @+ for (j=0; j<r; ++j) m[t+i][t+j]=f*a[i][j];
      t+=r;
    }
    return result;
  }
}

@*1 Exponents of the Weyl group.
The exponents of the Weyl group provide useful numerical data about it, that
can be used for instance to determine the number of positive roots and the
number of levels they occur in, so we start by defining these. The list of
exponents is stored permanently for each simple group encountered.

@< Proto...@>=
vector* Exponents(object grp);

@ The following function hard-codes the values of the exponents.

@c
local entry* simp_exponents (simpgrp* g)
{ if (g->exponents!=NULL) return g->exponents->compon;
  { static entry
      exp_E[3][7] = {{4,5,7,8,11},{5,7,9,11,13,17},{7,11,13,17,19,23,29}}
    , exp_F4[3] = {5,7,11};
    index i,r=g->lierank; entry* e=(g->exponents=mkvector(r))->compon; @+
    setlonglife(g->exponents); e[0]=1;
    switch (g->lietype)
    {	 case 'A': /* $1,2,3,\ldots,r$ */
      for (i=1; i<r; ++i) e[i]=i+1; @+
    break; case 'B': case 'C': /* $1,3,5,\ldots,2r-1$ */
      for (i=1; i<r; ++i) e[i]=2*i+1; @+
    break; case 'D': /* $1,3,5,\ldots,r-2,r-1,r,\ldots,2r-3$ 
                     or $1,3,5,\ldots,r-1,r-1,\ldots,2r-3$ */
	for (i=0; 2*i+1<r; ++i) {@; e[i]=2*i+1; e[r-i-1]=2*(r-i)-3; }
	if (2*i+1==r) e[i]=r-1; @+
    break; case 'E': copyrow(exp_E[r-6],&e[1],r-1);
    break; case 'F': copyrow(exp_F4,&e[1],3);
    break; case 'G': e[1]=5;
    }
    return e;
  }
}

@ The exponents of the simple factors are concatenated, followed by zeros for
the central torus.

@c
vector* Exponents(object grp)
{ if (type_of(grp)==SIMPGRP)
    {@; simp_exponents(&grp->s); return grp->s.exponents; }
  if (simpgroup(grp))
    {@; simp_exponents(Liecomp(grp,0)); return Liecomp(grp,0)->exponents; }
  { index i,t=0; vector* v=mkvector(Lierank(grp)); entry* e=v->compon;
    { for (i=0; i<grp->g.ncomp; ++i)
      { simpgrp* g=Liecomp(grp,i); index r=g->lierank; @+
	copyrow(simp_exponents(g),&e[t],r); t+=r;
      }
      for (i=0; i<grp->g.toraldim; ++i) e[t+i]=0;
    }
    return v;
  }
}


@*1 Positive roots.
We shall now compute the system of positive roots, and some related data.

@< Proto...@>=
index simp_numproots(simpgrp* g);
index Numproots(object grp);
matrix* simp_proots(simpgrp* g);
matrix* Posroots(object grp);
vector* Highroot(simpgrp* g);
vector* Simproot_norms(object grp);

@~The number of positive roots is equal to the length of the longest element
in the Weyl group, and therefore also to the sum of the exponents. For
simple groups the average value of the exponents is $1+e\over2$ where $e$ is
the last exponent, because of the symmetry present in the list of exponents
and the fact that the first exponent is~|1|; therefore no summation is
required for simple groups.

@c

index simp_numproots(simpgrp* g)
{@; index r=g->lierank; return r*(1+simp_exponents(g)[r-1])/2; }

index Numproots(object grp) /* should really return bigint */
{ if (type_of(grp)==SIMPGRP) return simp_numproots(&grp->s);
  { index i,d=0; @+
    for (i=0; i<grp->g.ncomp; ++i) d += simp_numproots(Liecomp(grp,i));
    return d;
  }
}

@ One of the most basic sets of data for a simple group is the list of
positive roots. We generate them iteratively by level, starting from the
fundamental roots and adding to the roots at a given level one of the
fundamental roots in all possible ways that result in a new root to form the
next level. We shall use the fact that for a fundamental root $\alpha_j$ and
for all positive roots $\alpha\neq\alpha_j$, all roots of the form
$\alpha+n\alpha_j$ are positive, and this ``ladder'' is symmetric with
respect to the fundamental reflection corresponding to~$\alpha_j$. This
means that in order to decide whether $\alpha+\alpha_j$ is a (positive)
root, it suffices to inspect the value $\<\alpha,\alpha_j>$ and previously
generated positive roots. It is practical to also compute the norms of the
roots as they are generated; what we shall call norms are actually the
values $(\alpha,\alpha)/2\in\{1,2,3\}$. The set of positive roots, the
indices where the levels start and the norms of the roots are so frequently
used that we store them permanently once computed. There are no roots at
level~|0|, so the first index of a root with level~|l| will be stored in
|level[l-l]|. The level of the highest root is equal to the last
exponent~$e$ of the group, so we give |level| size |e+1|, from level[0]
indicating the start of level~|1| to |level[e]| indicating the end of
level~|e|.

\def\alphaxwt{\alpha_{\rm wt}}
@f alpha_wt NULL
@c
matrix* simp_proots(simpgrp* g)
{ if (g->roots!=NULL) return g->roots;
  { index r=g->lierank,l,i,last_root;
    entry** cartan=simp_Cartan(g)->elm;
    entry** posr=(g->roots=mkmatrix(simp_numproots(g),r))->elm;
    entry* level=(g->level=mkvector(simp_exponents(g)[r-1]+1))->compon;
    entry* norm=(g->root_norm=mkvector(g->roots->nrows))->compon;
    entry* alpha_wt=mkintarray(r);
      /* space to convert roots to weight coordinates */
    setlonglife(g->roots),
    setlonglife(g->level),
    setlonglife(g->root_norm); /* permanent data */
    @< Initialise the first level of roots @>
    for (l=0; last_root>level[l]; ++l)
    { level[l+1]=last_root; /* set beginning of a new level */
      for (i=level[l]; i<level[l+1]; ++i)
	@< Add all roots obtained from |posr[i]| @>
    }
    freearr(alpha_wt); return g->roots;
  }
}

@ The first level of roots consists of the fundamental roots, so we set the
first |r| rows of |roots| to the identity matrix. Then we set the
corresponding entries of |norm| explicitly according to the type of~|g|.

@< Initialise the first level of roots @>=
{ index i,j; @+ for (i=0; i<r; ++i) @+ for (j=0; j<r; ++j) posr[i][j] = i==j;
  level[0]=0; last_root=r;
  for (i=0; i<r; ++i) norm[i]=1; /* norms are mostly |1| */
  switch (g->lietype) /* here are the exceptions */
  {	    case 'B': @+
    for (i=0; i<r-1; ++i) norm[i]=2; /* $2,2,\ldots,2,1$ */
  @+ break; case 'C': norm[r-1]=2; /* $1,1,\ldots,1,2$ */
  @+ break; case 'F': norm[0]=norm[1]=2; /* $2,2,1,1$ */
  @+ break; case 'G': norm[1]=3; /* $ 1,3$ */
  }
}

@ We try adding all fundamental roots $\alpha_j$ to |alpha|, while avoiding
duplicate roots within the current level.

@< Add all roots obtained from |posr[i]| @>=
{ index j,k; entry* alpha=posr[i]; @+ mulvecmatelm(alpha,cartan,alpha_wt,r,r);
    /* get values $\<\alpha,\alpha_j>$ */
  for (j=0; j<r; ++j) /* try all fundamental roots */
  { entry new_norm; @+
    @< See if $\alpha+\alpha_j$ is a root;
       if so, set |new_norm| to its norm, otherwise |continue| @>
    ++alpha[j]; /* temporarily set $\alpha\K\alpha+\alpha_j$ */
    for (k=level[l+1]; k<last_root; ++k)
      if (eqrow(posr[k],alpha,r)) break;
      /* if already present, don't add it */
    if (k==last_root)
      {@; norm[last_root]=new_norm; copyrow(alpha,posr[last_root++],r); }
    --alpha[j]; /* restore |alpha| */
  }
}

@ Here we analyse the rank~$2$ subsystem spanned by |alpha| and $\alpha_j$.
The most difficult case is when they are perpendicular short roots in type
$B_n$, $C_n$, or~$F_4$; it requires a search in the previous level of roots.

@< See if $\alpha+\alpha_j$ is a root... @>=
{ if (alpha_wt[j]<0) /* then $\alpha+\alpha_j$ is a root; find its norm */
    if (norm[j]==norm[i]) new_norm=norm[j]; /* |alpha_wt[j]==-1| */
    else new_norm=1; /* regardless of |alpha_wt[j]| */
  else if (norm[i]>1 || norm[j]>1) continue; /* both roots must be short now */
  else if (strchr("ADE",g->lietype)!=NULL) continue;
    /* but long roots must exist */
  else if (alpha_wt[j]>0)
    if (g->lietype!='G'||alpha_wt[j]!=1) continue; @+ else new_norm=3;
    /* $[2,1]\to[3,1]$ for $G_2$ */
  else if (alpha[j]==0) continue;
    /* $\alpha-\alpha_j$ should not have a negative entry */
  else
  {@; @< Look up $\alpha-\alpha_j$ in previous level;
       if not found |continue| @> @+ new_norm=2; }
}

@ Since we have tested that |alpha[j]!=0|, we know that |l>0|, so it is
safe to inspect the previous level.

@< Look up $\alpha-\alpha_j$... @>=
{ --alpha[j]; @+
  for (k=level[l-1]; k<level[l]; ++k) @+
    if (eqrow(posr[k],alpha,r)) break;
  ++alpha[j]; @+
  if (k==level[l]) continue;
}

@ For composite groups we combine the rectangular matrices of positive roots
for the simple factors in a ``diagonal'' fashion; any central torus is
ignored.

@c
matrix* Posroots(object grp)
{ if (type_of(grp)==SIMPGRP) return simp_proots(&grp->s);
  if (simpgroup(grp)) return simp_proots(Liecomp(grp,0));
  { index i,j,t1=0,t2=0;
    matrix* result=mat_null(Numproots(grp),Ssrank(grp));
    entry** m=result->elm;
    for (i=0; i<grp->g.ncomp; ++i)
    { matrix* posr=simp_proots(Liecomp(grp,i));
      index r=Liecomp(grp,i)->lierank;
      for (j=0; j<posr->nrows; ++j) copyrow(posr->elm[j],&m[t1+j][t2],r);
      t1+=posr->nrows; t2+=r;
    }
    return result;
  }
}

@ To obtain the highest root, we take the final row of the list of positive
roots.

@c
vector* Highroot(simpgrp* g)
{ matrix* posr=simp_proots(g); index r=g->lierank; vector* high=mkvector(r);
  copyrow(posr->elm[posr->nrows-1],high->compon,r); return high;
}

@ While the norms of all positive roots for a simple group~|g| can be found in
|g->simp_norm| once |simp_proots(g)| has been called for the first time, we
occasionally also need the norms of all the \emph{simple} roots for
composite groups; these are obtained by calling |Simproot_norms|, which
takes care of calling |simp_proots| for all the simple factors involved.

@c
vector* Simproot_norms(object grp)
{ if (type_of(grp)==SIMPGRP)
    {@; simp_proots(&grp->s); return grp->s.root_norm; }
  {@; index i; for (i=0; i<grp->g.ncomp; ++i) simp_proots(Liecomp(grp,i)); }
  if (grp->g.ncomp==1) return Liecomp(grp,0)->root_norm;
  { index i,t=0; vector* result=mkvector(Ssrank(grp));
    for (i=0; i<grp->g.ncomp; ++i)
    { simpgrp* g=Liecomp(grp,i); index r=g->lierank;
      copyrow(g->root_norm->compon,&result->compon[t],r); t+=r;
    }
    return result;
  }
}

@*1 The adjoint representation, and the centre.
Functions are provided for the adjoint representation, its dimension, and
its kernel, which is the centre of the group.

@< Proto... @>=
poly* Adjoint(object grp);
entry Dimgrp(object grp);
matrix* Center(object grp);


@ The adjoint representation of a simple group is irreducible, and can
therefore be represented by a vector; such a vector is computed by
|set_simp_adjoint| based on the value of |Highroot(g)|, which has to be
converted to weight coordinates. For general (composite) groups the adjoint
representation has to be described by its decomposition polynomial, which is
what |Adjoint| returns.

@c
static void set_simp_adjoint(entry* dst,simpgrp* g)
{ index r=g->lierank; vector* high=Highroot(g);
  mulvecmatelm(high->compon,g->cartan->elm,dst,r,r); freemem(high);
}

poly* Adjoint(object grp)
{ index i,j,r=Lierank(grp)
  ,n=type_of(grp)==SIMPGRP ? 1: grp->g.ncomp+(grp->g.toraldim!=0);
  poly* adj= mkpoly(n,r);
  for (i=0; i<n; ++i)
  {@; adj->coef[i]=one; for (j=0; j<r; ++j) adj->elm[i][j]=0; }
  if (type_of(grp)==SIMPGRP) set_simp_adjoint(adj->elm[0],&grp->s);
  else
  { index offs=0; simpgrp* g;
    for (i=0; i<grp->g.ncomp; offs+=g->lierank,++i)
      set_simp_adjoint(&adj->elm[i][offs],g=Liecomp(grp,i));
    if (grp->g.toraldim!=0)
    @/{@; adj->coef[i]=entry2bigint(grp->g.toraldim);
      setshared(adj->coef[i]);
    }
  }
  return adj;
}

@ The function |Dimgrp| computes the dimension of a group, which is equal to
that of its adjoint representation. The computation does not use |Adjoint|
however.

@c
entry Dimgrp(object grp)
@+{@; return Lierank(grp) + 2*Numproots(grp); }

@ The function |Center| returns a matrix whose rows represent semisimple
elements and one parameter subgroups generating the centre of |grp|. For
semisimple groups the centre is essentially the quotient of the coweight
lattice by the coroot lattice, so its elements could be computed using the
matrix |i_Cartan| and |det_Cartan|. To wit, one obtains all non-identity
elements of the centre by reducing modulo~|det_Cartan| the columns
of~|i_Cartan| corresponding to the positions where |high_root| has an
entry~|1|, adding~|det_Cartan| as final entry, and dividing out the |gcd|.
However since the centre is small and in most cases cyclic, it is easier to
avoid the computation, and determine the generators directly from the types
of the simple factors.

@c
matrix* Center(object grp)
{ index i,j,R=Lierank(grp),n_gen;
  @< Set |n_gen| to the number of generators for the centre @>
  { matrix* res=mat_null(n_gen,R+1); entry** m=res->elm; index k=0,s=0;
    for (j=0; j<grp->g.ncomp; ++j)
    { simpgrp* g=Liecomp(grp,j); index n=g->lierank; entry d=simp_detcart(g);
      if (d>1)
      { @< Add generators of centre for factor |g|, possibly changing |d| @>
	m[k++][R]=d; /* insert denominator for last generator, and advance */
      }
      s+=n; /* advance offset into semisimple elements */
    }
    @< Add the centre of the central torus @>
    assert(k==n_gen); return res;
  }
}

@ For simple groups~|g| the centre is trivial if and only if
|simp_detcart==1| (which holds only if |g| has type $E_8$, $F_4$ or~$G_2$);
otherwise the centre is cyclic, except when |g| has type $D_{2n}$, in
which case it is a Klein 4-group requiring 2 generators. 

@< Set |n_gen| to the number of generators for the centre @>=
for (n_gen=grp->g.toraldim,i=0; i<grp->g.ncomp; ++i)
{ simpgrp* g=Liecomp(grp,i);
  if (simp_detcart(g)>1) n_gen+=1+(g->lietype=='D' && g->lierank%2==0);
}

@ All cases except type~$D_n$ are quite trivial.

@< Add generators of centre for factor... @>=
switch (g->lietype)
{      case 'A': @+ for (i=0; i<n; ++i) m[k][s+i]=i+1;
	/* $[1,2,3,\ldots,n]$; $d=n+1$ */ @+
break; case 'B': m[k][s+n-1]=1; /* $[0,0,\ldots,0,1]$; $d=2$ */ @+
break; case 'C': @+ for (i=0; i<n; i+=2) m[k][s+i]=1;
	/* $[1,0,1,0,\ldots]$; $d=2$ */ @+
break; case 'D': @< Set centre for type $D_n$ @> @+
break; case 'E':
  if (n==7)
    {@; m[k][s+1]=m[k][s+4]=m[k][s+6]=1; } /* $[0,1,0,0,1,0,1]$; $d=2$ */
  else {@; m[k][s]=m[k][s+4]=1; m[k][s+2]=m[k][s+5]=2; }
     /* $[1,0,2,0,1,2]$; $d=3$ */
}

@ For a group of type $D_n$ there are 3 non-trivial elements in the centre;
written with denominator $d=4$ they can be represented as
$[0,0,\ldots,0,2,2]$, $[2,0,2,0,\ldots,n\MOD4,(n+2)\MOD4]$, and
$[2,0,2,0,\ldots,(n+2)\MOD4,n\MOD4]$. When $d$ is odd, each of the last two
elements generates the centre; if $r$ is even however, they all have
order~$2$, so we must set |d=2| while halving the entries, and choose two
elements as generators. In the latter case we choose the first element and
the most regular looking one of the other two; also we must not forget to
store~$d$ and advance~$k$ for the first element (this is done further
outward for the last element). A little trick allows us to share some code
between the cases of odd and even~$n$.

@< Set centre for type $D_n$ @>=
{ m[k][s+n-2]=m[k][s+n-1]=1;
  if (n%2==1) @+ for (i=0; i<n; i+=2) m[k][s+i]+=2;
    /* $[2,0,2,\ldots,2,1,3]$; $d=4$ */
  else
  { d=2; m[k++][R]=d; /* $[0,0,\ldots,0,1,1]$; $d=2$ */
    for (i=0; i<n; i+=2) m[k][s+i]=1; /* $[1,0,1,\ldots,1,0]$; $d=2$ */
  }
}
       

@ The centre of a central torus is of course the torus itself. It is
generated by one-parameter subgroups, each of which represents a
factor~${\bf C}^*$ of the torus. The ``denominator'' stays~$0$ here.

@< Add the centre of the central torus @>=
for (i=0; i<grp->g.toraldim; ++i) m[k++][s+i]=1;

@*1 Testing roots.
Finally, we define a few functions that look up and test roots. The function
|find_root| looks up the index of a root in the list of positive roots for a
simple group (returning |-1| if it is not found); it limits its search to
the roots at the specified level. The predicate |isroot| tells whether a
root vector occurs in the root system (allowing for negative roots as well);
|checkroot| also performs this test, but performs an error stop if it fails;
finally |isposroot| tells whether its argument, which is assumed to satisfy
|isroot|, is positive or not.

@< Proto... @>=
index find_root(entry* alpha, entry level, simpgrp* g);
boolean isroot(entry* alpha);
void checkroot(entry* alpha);
boolean isposroot(entry* alpha);

@ For |find_root|, we proceed by a simple linear search; since |g->level|
starts at level~|1| (there are no roots of level~|0|) the bounds for the
search can be found at positions |level-1| and |level| of that vector.


@c
index find_root(entry* alpha, entry level, simpgrp* g)
{ index i,r=g->lierank; matrix* posr=simp_proots(g);
  for (i=g->level->compon[level-1]; i<g->level->compon[level]; ++i)
    if (eqrow(alpha,posr->elm[i],r)) return i;
  return -1; /* not found */
}

@ Apart from calling |find_root|, |simp_isroot| must compute the level and
deal with negative roots.

@c
local boolean simp_isroot(entry* alpha, simpgrp* g)
{ index i,r=g->lierank; entry level=0; boolean neg,result=false;
  for(i=0; i<r; ++i) level+=alpha[i]; /* compute level of |alpha| */
  neg=level<0; /* if |neg| holds, |alpha| can only be a negative root */
  if (neg) {@; level= -level; for(i=0; i<r; ++i) alpha[i]= -alpha[i]; }
  result=find_root(alpha,level,g)>=0;
  if (neg) @+ for(i=0; i<r; ++i) alpha[i]= -alpha[i]; /* restore |alpha| */
  return result;
}

@ For non-simple groups, |alpha| is a root if and only if it is supported by
precisely~|1| simple component, and is a root there. As soon as we find a
non-zero entry for a second component, we can stop, returning |false|, so
the counter |n_parts| of these components below will in fact never obtain a
value exceeding~|1|.

@c
boolean isroot(entry* alpha)
{ index n_parts=0, i,j;
  if (type_of(grp)==SIMPGRP) return simp_isroot(alpha,&grp->s);
  if (grp->g.ncomp==1) return simp_isroot(alpha,Liecomp(grp,0));
  for (i=0; i<grp->g.ncomp; ++i)
  { simpgrp* g=Liecomp(grp,i); index r=g->lierank;
    for (j=0; j<r; ++j) if (alpha[j]!=0)
      if (n_parts>0 || !simp_isroot(alpha,g)) return false; @+
      else {@; ++n_parts; break; }
    alpha+=r;
  }
  return n_parts==1; /* |alpha| is root if supported on 1 simple component */
}

@ In |isposroot| we assume that |alpha| is already certified as a root, so
without calling |isroot| we just inspect the sign of the first non-zero
entry.

@c
void checkroot(entry* alpha)
{@; if (!isroot(alpha))
  {@; printarr(alpha,Ssrank(grp)); error (" is not a root.\n"); }
}

boolean isposroot(entry* alpha)
{ index i,s=Ssrank(grp); @+
  for (i=0; i<s; ++i) @+ if (alpha[i]!=0) return alpha[i]>0;
  assert(false); return false; /* to avoid compiler warnings */
}

@* Index.
