\input LiEwebmac
%\input cwebhmac
\input references

\nomods

\def\chixlambda{{\chi^\lambda}}

@* Symmetric and alternating tensor powers, and plethysms.
This file implements the functions |Adams|, |sym_tensor|, |alt_tensor|, and
|plethysm|; these are the functions that are related to the decompostion of
tensor powers $V^{\tensor n}$ under the action of the symmetric
group~$\Sym{n}$. Apart from the definitions given in the \LiE. manual, these
operations can be described in terms of symmetric functions. The character of
any actual representation~$V$ can be written as a sum of monomial terms of the
form $X^\nu$, with repetition of identical terms used in case of
multiplicities~$>1$. We can then associate each of these |dim(V)| terms with a
separate indeterminate, and evaluate any symmetric function~$f$ in these
indeterminates accordingly. The resulting polynomial is the character of a
(possibly virtual) representation (the representation is actual if~$f$ is
positive on the basis of Schur functions). We shall say that these are the
(virtual) character and representation corresponding to~$f$ for~$V$. Clearly
the representation $V$ itself corresponds to the first elementary symmetric
function $e_1=\sum_ix_i$, and $V^{\tensor n}$ corresponds to~$e_1^n$. The
virtual representation \hbox{|Adams(n,V)|} corresponds to the $n$-th power sum
symmetric function~$p_n$, the representations \hbox{|sym_tensor(n,V)|} and
\hbox{|alt_tensor(n,V)|} correspond respectively to the $n$-th complete and
$n$-th elementary symmetric functions $h_n$ and~$e_n$, and
\hbox{|plethysm(lambda,V)|} corresponds to the Schur function~$s_\lambda$
(notation as in [Macd]). As for the other representation theoretic functions
in \LiE., the results of these functions are expressed as decomposition
polynomials of the indicated (virtual) representations.

@h "lie.h"
@f lambda NULL
@f mu NULL
@f chi_lambda NULL
	/* irreducibe $\Sym{n}$ character parametrised by |lambda| */

@( plethysm.h @>=
@< Prototypes of globally visible functions @>

@ Addition and multiplication of characters can be achieved on the level of
decomposition polynomials by addition and |tensor|, but to obtain all
symmetric functions starting with $e_1$ one needs more than this. As other
basic ingredient of our computations we shall use formation of power sum
symmetric functions $p_n=\sum_ix_i^n$ (for $n>0$), corresponding to Adams'
operation on virtual representations. It has the advantage that it is easy
to compute at the character level, but it means that to express other
symmetric functions we shall need formulae with (integer) denominators.

@< Prototypes... @>=
poly* Adams(index n,poly* p);

@~Adams' operation has the special property, not shared by the other
operations discussed here, that it is additive in the character it is
applied to: |Adams(n,p+q)==Adams(n,p)+Adams(n,q)|. Therefore |Adams| can
meaningfully be applied to virtual characters as well as to actual ones.  As
the other operations are expressed in terms of |Adams|, they too will allow
virtual characters as arguments, although the definition given for
associating characters to symmetric functions above makes no sense if the
initial module~$V$ is virtual. The implementation of~|Adams| is quite
simple, since |Domchar_p| and |Vdecomp| do all the real work: the full
character for the decomposition polynomial~|p| is computed, all exponent
vectors are scaled up by a factor~|n|, and the result is virtually
decomposed by~|Vdecomp|. Note that for this application the algorithm used
in~|Vdecomp| is usually much more efficient than the one used in~|Decomp|
would be.

@c
poly* Adams(index n,poly* p)
{ if (n==1) return p; /* avoid work in this trivial case */
  { index i,j, r=Lierank(grp); poly* dom_ch=Domchar_p(p);
    for (i=0; i<dom_ch->nrows; ++i)
      for (j=0; j<r; j++) dom_ch->elm[i][j] *= n;
    {@; poly* result=Vdecomp(dom_ch); freepol(dom_ch); return result; }
  }
}

@ The $n$-th symmetric and alternating tensor powers of~$V$ are computed in
\LiE. as |sym_tensor(n,V)| and |alt_tensor(n,V)| respectively, but since the
computations are so similiar, these calls are performed internally by the
same function: they are converted to |SAtensor(false,n,V)| respectively
|SAtensor(true,n,V)|.

@< Prototypes... @>=
poly* SAtensor(boolean alt,index n,poly* p);

@~The computation of |SAtensor| is based in to symmetric function identities
([Macd~I, (2.11)])
$$ nh_n=\sum_{i=1}^np_ih_{n-i} \qquad\hbox{and}\qquad
   ne_n=\sum_{i=1}^n-(-1)^ip_ie_{n-i},
$$
where $h_n=\sum_{i_1\leq\cdots\leq i_n}x_{i_1}x_{i_2}\cdots x_{i_n}$
and $e_n=\sum_{i_1<\cdots<i_n}x_{i_1}x_{i_2}\cdots x_{i_n}$ 
respectively are the $n$-th complete and elementary symmetric functions, and
$p_n=\sum_ix_i^n$ is the $n$-th power sum symmetric function as before.
Writing $S^i(V)$ for the $i$-th symmetric power $\\sym\_tensor(i,V)$ of~$V$,
and $\alt^i(V)$ for the $i$-th alternating power $\\alt\_tensor(i,V)$, this
leads to the recursion relations
$$
  S^n(V)={1\over n} \sum_{i=1}^n \\{Adams}(i,V)\tensor S^{n-i}(V)
\qquad\hbox{and}\qquad
  \alt^n(V)={1\over n}\sum_{i=1}^n-(-1)^i\\{Adams}(i,V)\tensor\alt^{n-i}(V).
$$
These are used in a simple loop to compute successive values $S^n(V)$ or
$\alt^n(V)$ for |n| from~|1| up to the desired final value~|m|. Since the
results of |Adams| and the intermediate values are used many times, they are
stored in local arrays |adams|~and~|q|.

@c
poly* SAtensor(boolean alt,index m,poly* p)
{ index n,r=Lierank(grp); poly** adams,** q,* result;
  if (m==0) return poly_one(r); @+ else if (m==1) return p;

  adams=alloc_array(@[poly*@],m+1); @+
  for (n=1; n<=m; ++n) adams[n]=Adams(n,p);
  q=alloc_array(@[poly*@],m+1);
  q[0]=poly_one(r);
  for (n=1; n<=m; ++n)
  { @< Set |q[n]|${}\K
       \sum_{i=1}^n \eps(i)\\{adams}[i]\tensor q[n-i]$
       where $\eps(i)= -(-1)^i$ if |alt==true|, and $\eps(i)=1$ otherwise @>
    @< Divide all the coefficients of |q[n]| by~|n| @>
  }
  result=q[m];
@/{@; for (n=1; n<=m; ++n) freepol(adams[n]); } freearr(adams);
@/{@; for (n=0; n<m; ++n)  freepol(q[n]); } freearr(q);
@/ return result;
}

@ When the code below is executed we have |n>=1|, hence we may initialise
the result to the first term of the summation, which always has $\eps(1)=1$,
and |adams[1]==p|.  The sign $\eps(i)$ is taken into account by passing the
final negation argument to |Add_pol_pol| as |true| if and only if |alt|
holds and |i| is even.

@< Set |q[n]... @>=
{ index i; q[n]=Tensor(p,q[n-1]); /* the initial term of the summation */
  for (i=2; i<=n; ++i) q[n] =
    Add_pol_pol(q[n],Tensor(adams[i],q[n-i]),alt&&i%2==0);
}

@ Dividing the result of the summation by~|n| is slightly laborious, since
we go through its coefficients by hand. The most subtle point is to adjust
the reference counts of the coefficients properly; although as coefficients
of a newly computed value these are likely to have reference count~|1|, we
do not depend on this. Since the division routine |divq| used is destructive
in its first argument, we make a copy of the coefficient if it was in fact
shared.

@< Divide all the coeff... @>=
{ index i; bigint* big_n=entry2bigint(n); @+ setshared(big_n);
  for (i=0; i<q[n]->nrows; ++i)
  { bigint** cc= &q[n]->coef[i]
         ,* c= (clrshared(*cc),isshared(*cc)) ? copybigint(*cc,NULL) : *cc;
    *cc=divq(c,big_n); setshared(*cc);
    @< Test that the remainder~|c| is~0, and discard it @>
  }
  clrshared(big_n); freemem(big_n);
}

@ The original coefficient~|c| should be reduced to~|0| by~|divq|, or
something is seriously wrong.

@< Test that the remainder... @>=
{ if (c->size != 0)
    error("Internal error (SAtensor): remainder from %ld.\n" ,(long)n);
  freemem(c);
}

@ The call |plethysm(lambda,V)| computes the representation corresponding
for~$V$ to the Schur function~$s_\lambda$; it is converted
by the interface routines to the internal call |Plethysm(lambda,l,n,V)|,
where $\lambda=\lambda_1,\ldots,\lambda_l$ and
$n=\vert\lambda\vert=\sum_{i=1}^l\lambda_i$.

@< Prototypes... @>=
poly* Plethysm(entry* lambda,index l,index n,poly* p);

@~The computation of plethysms is based on the relation between Schur
functions $s_\lambda$ and the irreducible characters |chi_lambda|
of~$\Sym{n}$: with $p_\mu=p_{\mu_1}\cdots p_{\mu_m}$, the inner product
$\<s_\lambda,p_\mu>$ is equal the the value |chi_lambda(mu)| of |chi_lambda|
at the conjugacy class~|mu| (see [Macd~I, (7.7)]). Since the functions
$p_\mu$ are mutually orthogonal, span the space of all symmetric functions,
and satisfy $\<\,p_\mu,p_\mu>=n!\big/\hbox{|Classord(mu)|}$, this leads to the
formula
$$ s_\lambda={1\over n!}\sum_{\mu\in{\cal P}_n}
   \hbox{|chi_lambda(mu)|{}|Classord(mu)|}p_\mu.
$$
Here |chi_lambda(mu)| can be computed by |MN_char|, and the representations
corresponding to the $p_\mu$ by |Adams| in combination with |Tensor|.

@c
poly* Plethysm(entry* lambda,index l,index n,poly* p)
{ if (n==0) return poly_one(Lierank(grp)); @+ else if (n==1) return p;

  { index i,j;
    poly* sum= poly_null(Lierank(grp)),**adams=alloc_array(@[poly*@],n+1);
    poly* chi_lambda=MN_char(lambda,l);
    for (i=1; i<=n; ++i) {@; adams[i]=Adams(i,p); setshared(adams[i]); }
    @< For all terms $cX^\mu$ of |chi_lambda|, contribute
       $c*\hbox{|Classord(mu,n)|}*\Tensor_i\\{adams}(\mu_i)$ to |sum| @>
    freemem(chi_lambda);
    setshared(p); /* protect |p|; it coincides with |adams[1]| */
    for (i=1; i<=n; ++i)
      {@; clrshared(adams[i]); freepol(adams[i]); } @+ freearr(adams);
  @/clrshared(p);

    @< Divide |sum| by |fac(n)| @>

    return sum;
  }
}

@ We do not attempt to avoid recomputations of identical (partial) tensor
products, even though these may very well occur for the various
exponents~|mu| in~|chi_lambda|. To do this effectively would require quite a
bit of extra consideration, since we do not know the set of exponents
beforehand; on the other hand the amount of time saved could be significant.
As it is, the code is striaghtforward; the intermediate tensor products are
discarded as soon as they have been used.

@< For all terms $cX^\mu$ of |chi_lambda|... @>=
for (i=0;i<chi_lambda->nrows;i++)
{ entry* mu=chi_lambda->elm[i]; poly* prod=adams[mu[0]],*t;
  for (j=1; j<n && mu[j]>0; ++j)
    {@; t=prod; prod=Tensor(t,adams[mu[j]]); freepol(t); }
  sum= Addmul_pol_pol_bin(sum,prod,mult(chi_lambda->coef[i],Classord(mu,n)));
}

@ The code below is similar to the division by~|n| in |SAtensor|.

@< Divide |sum| by |fac(n)| @>=
{ bigint* fac_n=fac(n); @+ setshared(fac_n); /* used repeatedly */
  for (i=0; i<sum->nrows; ++i)
  { bigint** cc= &sum->coef[i]
         ,* c= (clrshared(*cc),isshared(*cc)) ? copybigint(*cc,NULL) : *cc;
    *cc=divq(c,fac_n); setshared(*cc);
    if (c->size!=0) error("Internal error (plethysm).\n"); @+ else freemem(c);
  }
  clrshared(fac_n); freemem(fac_n);
}

@* Reference and index.

\beginrefs

\ref Macd \key Macdonald
 \author I. G. Macdonald
 \title Symmetric Functions and Hall Polynomials
 \publ Clarendon press, Oxford \year 1979
 \series Oxford Mathematical Monographs
\endref

\endrefs
