\input LiEwebmac
%\input cwebhmac
%\nomods

\input references


\def\lambdaxprime{\lambda'}
\def\Card#1{|#1|}
\def\Part{{\cal P}}
\def\Sym#1{{\bf S}_{#1}}
\def\Sn{{\bf S}_n}
\def\sg{\mathop{\rm sg}}

@* Functions related to symmetric groups.

@h "lie.h"
@f entry int /* don't index */
@f index int
@)
@f kappa NULL
@f lambda NULL
@f mu NULL
@f nu NULL
@f sigma NULL
@f lambda_prime NULL

@( symg.h @>=
@< Prototypes of globally visible functions @>

@*1 Some enumerative functions. We start with the factorial function, which
is used in this compilation unit and in \.{plethysm}.

@< Prototypes... @>=
bigint* fac(index n);
@~This is too simple (|mul1| multiplies a |bigint| by an |unsigned short|).

@c
bigint* fac(index n)
{@; bigint* f=copybigint(one,NULL); while (n>1) f=mul1(f,n--); return f; }

@ The function |n_parts| computes number of partitions of |n|. It is used
only to precompute the dimension of arrays, so it may use machine integer
precision. We allow for~|n| values up to~|121|, which is the largest value
for which |n_parts(n)<@t$2^{31}$@>|; in practice it is unlikely that such
large arrays can be succesfully allocated. The algorithm is simple and not
the fastest possible; it computes the generating function
  $$\prod_{i>0}{1\over1-X^i},$$
where the product is truncated to terms with |i<=n| to determine the
coefficient of $X^n$.

@c
static index n_parts(index n)
{ index i,k,np; entry* c=mkintarray(n+1); /* coefficients */
  if (n>121) error("Too many partitions to generate.\n");
  for (i=0; i<=n; ++i) c[i]=1; /* initialise to ${1\over1-X}$ */
  for (i=2; i<=n; ++i)
    for (k=i; k<=n; ++k) c[k]+=c[k-i]; /* multiply by ${1\over1-X^i}$ */
  np=c[n]; freearr(c); return np;
}

@ The function |n_tableaux| computes number of tableaux of given
shape~|lambda| (which has |l| parts), using the Frame-Robinson-Thrall
hook-length formula; it is used internally, and is also available directly
to the user.

@< Prototypes... @>=
bigint* n_tableaux(entry* lambda, index l);

@~The hook-length formula is evaluated by rows from bottom to top, each time
extending the factorial in the numerator first, then dividing by the hook
lengths for the new row. This avoids very large numbers, while no remainders
are left, since after processing each row we have the hook length formula
for a partition (|lambda| with some initial parts removed), and therefore an
integer number.

@c
bigint* n_tableaux(entry* lambda, index l)
{ index i,j,k=0; entry* h; bigint* res=copybigint(one,NULL);
  do @+ if (--l<=0) return one; @+
  while (lambda[l]==0); /* find last non-zero part */
  h=mkintarray(lambda[0]); @+
  for(j=0; j<lambda[0]; ++j) h[j]=0; /* accumulated column heigths */
  for(i=l; i>=0; --i)
    @< Update |h| and |res| according to hook length formula for row~|i| @>
  freearr(h); return res;
}

@~The auxiliary function |mul1| might return a newly allocated |bigint|, but
|div1| will not.

@< Update |h| and |res| according to hook length formula... @>=
{ entry li=lambda[i]-1;
  for(j=0; j<=li; ++j) res=mul1(res,++k); /* part of factorial */
  for(j=0; j<=li; ++j) div1(res,(++h[j])+li-j); /* divide by hook lengths */
}

@ The function |Classord| computes order of conjugacy class of cycle type
|kappa|. This order is given by
$$ \hbox{|Classord(kappa)|}=
   {\Card\kappa!\over\prod_{k>0}k^{c_k(\kappa)}c_k(\kappa)!}\ ,
$$
where $c_k(\kappa)$ is the number of parts of~|kappa| equal to~|k|.

@< Prototypes... @>=
bigint* Classord(entry* kappa, index l);

@~We consider one part a the time, extending the numerator before the
denominator for exact division.

@c
bigint* Classord(entry* kappa, index l)
{ index prev=0,i=0,j,n=0,k,f=1; bigint* x=copybigint(one,NULL);
  while (i<l && (k=kappa[i++])>0)
  { for (j=0; j<k; ++j) x=mul1(x,++n);
      /* extend $\Card\kappa!$ in numerator */
    div1(x,k); /* contribution to $k^{c_k(\kappa)}$ in denominator */
    if (k!=prev) {@; f=1; prev=k;}  /* this case applies the first time */
    else div1(x,++f); /* contribution to $c_k(\kappa)!$  in denominator */
  }
  return x;
}

@*1 Testing functions. Here we collect functions that test whether data has
the appropriate format for being interpreted as permutation, partition or
tableau; when data has passed through these functions, there is no need for
additional validity tests the actual algorithms.

Since we allow repetitions in permutations, any sequence of number
qualifies as a permutation. For a sequence to be interpreted as partition,
its entries must be non-negative and weakly decreasing; this is tested by
|check_part|. Since we often also want to know the number of which this is a
partition, that number is returned by |check_part| if it finds no errors.

@< Prototypes... @>=
index check_part(entry* lambda, index l);
@~@c
index check_part(entry* lambda, index l)
{ index i,sum=0;
  for (i=0; i<l; ++i)
    if (lambda[i]<0) error("Negative entry in partition.\n");
    else if (i>0 && lambda[i]>lambda[i-1])
      error("Increasing entries in partition.\n");
    else sum+=lambda[i];
  return sum;
}

@ Young tableaux will be represented as ``lattice permutations'': sequences
of positive integers such that the weight of any prefix is a partition. Here
a weight of a multiset of positive integers is the sequence
$(c_1,c_2,c_3,\ldots)$, where $c_i$ is the multiplicity of the number~$i$.
Such lattice permutations encode tableaux by storing at index~|i| the row
number of the entry~|i| in the tableau. The function |check_tabl| tests this
condition and returns the shape of the tableau represented by the lattice
permutation.

@< Prototypes... @>=
vector* check_tabl(vector* v);
@~@c
vector* check_tabl(vector* v)
{ vector* shape; entry* t=v->compon,* sh; index i,d,n=v->ncomp, max=0;
  @< Test for positivity of entries and compute maximum in |max| @>
  shape=mkvector(max); sh=shape->compon; @+ for(i=0; i<max; ++i) sh[i]=0;
  @< Test lattice condition while computing shape in |sh| @>
  return shape;
}

@
@< Test for positivity of entries... @>=
for(i=0; i<n; ++i)
  if( (d=t[i])<=0 ) error ("Non-positive number in tableau.\n");
  else if (d>max) max=d;

@ By the code above, |sh| will be large enough to have an entry |sh[t[i]-1]|.
The lattice condition is tested by checking every time an entry |t[i]>1| is
encountered, that after increasing the length of row number |t[i]|, it does
not exceed the length of the previous row.

@< Test lattice condition... @>=
for(i=0; i<n; ++i)
  if( ++sh[d=t[i]-1], d>0 && sh[d]>sh[d-1] )
    Printf("%ld at position %ld ",(long)(d+1),(long)(i+1)),
    error("violates tableau condition.\n");

@*1 Incrementor functions. For generating permutations, partitions, and
tableaux, we take an incremental approach: functions are defined that
transform a given element into its successor in a fixed (lexicographic or
similar) ordering. This is not always the most efficient method, but it is
reasonably efficient in all cases, the memory consumption is minimal (only
the current element needs to be remembered), and the functions are completely
independent of the ones calling them.

For permutations we allow for repeated entries; the orbit will be generated
efficiently without any special effort, as long as we make sure that we
always generate a permutation distinct from the given one. We use
lexigraphic order: $[1,1,2,2]$, $[1,2,1,2]$, $[1,2,2,1]$, $[2,1,1,2]$,
$[2,1,2,1]$, $[2,2,1,1]$. The value returned tells whether a new permutation
could be geneated.

@< Prototypes... @>=
boolean Nextperm(entry* w, index n);
@~@c
boolean Nextperm(entry* w, index n)
{ index i,j;
  @< Find |(i,j)| with |i<j|, |w[i]<w[j]|, and |i| and |j| maximal;
     |return false| if |w| is decreasing @>
  swap(&w[i],&w[j]);
    /* now we have |w[i]<=w[i+1]>=@t$\ldots$@>>w[j]>=@t$\ldots$@>>=w[n-1]| */
  for (i++,j=n-1; i<j; i++,j--) swap(&w[i],&w[j]); /* reverse $w[i+1:n-1]$ */
  return true;
}

@ The values of |i| and |j| are found by a straightforward linear search
from right to left. In the second case binary search would have been
possible, but since the linear search will succeed in about two steps on the
average, this does not seem worth while.

@< Find |(i,j)| with... @>=
{ if (n<=1) return false;
  for (i=n-2; w[i]>=w[i+1]; --i) /* find last ascent |(i,i+1)| */
    if (i==0) return false; /* there is none */
  for (j=n-1; w[i]>=w[j]; --j) {} /* find maximal |j>i| with |w[i]<w[j]| */
}

@ We choose the reverse lexicographic order for generating partitions:
$[5]$, $[4,1]$, $[3,2]$, $[3,1,1]$, $[2,2,1]$, $[2,1,1,1]$, $[1,1,1,1,1]$.
The functions |Nextpart| transforms the partition~|lambda| into its
successor of length~|l|; like above the result indicates whether a successor
was found.  It is assumed that there is enough space to store all parts of
the partition, but positions |lambda[l]| and beyond are only used if needed
to accommodate additional non-zero parts.

@< Prototypes... @>=
boolean Nextpart(entry* lambda, index l);
@~@c
boolean Nextpart(entry* lambda, index l)
{ index i,avail=0; entry k;
  @< Set |i| equal to the index of the last part of~|lambda| exceeding~|1|,
     and |avail| to the number of parts~|1| following it;
     |return false| if only parts $\leq1$ are found @>
  k=(lambda[i++]-=1); ++avail; /* decrease the last part possible */
  @< Fill the remaining part of~|lambda| with a partition of |avail| with
     parts of maximal size not exceeding~|k| @>
  return true;
}

@ Since the value of~|l| will be used only to determine the end of any
zero-filling required, we may decrease~|l| to the index of the first zero.

@< Set |i| equal to the index of the last part... @>=
{ while (l>0 && lambda[l-1]==0) l--;
  for (i=l-1; i>=0 && lambda[i]==1; --i) ++avail;
  if (i<0) return false; /* quit if 1's only */
}

@ Three kinds of parts are generated here: parts~|k|, a remainder, and zeros.

@< Fill the remaining part of~|lambda|... @>=
{ while (avail>k) avail-=(lambda[i++]=k); /* distribute |avail| */
  lambda[i++]=avail; /* last part may be smaller */
  while (i<l) lambda[i++]=0; /* fill with zeros */
}

@ The final incrementor function is |Nexttableau|; it computes the next
lattice permutation of the same shape (view as Young tableau) into~|t|,
or returns |false| if the final tableau was already reached.

@< Prototypes... @>=
boolean Nexttableau(entry* t, index n);
@~@c
boolean Nexttableau(entry* t, index n)
{ index i,r; index* lambda,* skew;
  @< Set |i| to the index of the last square that can be moved down,
     storing the weight of the suffix starting from there in~|skew|,
     or |return false| @>
  @< Find the smallest possible |r>t[i]| to replace~|t[i]|,
     then extend |t| with a weakly increasing sequence of weight |skew| @>
  return true;
}

@ The search proceeds from right to left through the lattice permutation,
keeping the weight of the remaining prefix in |lambda| and the weight of the
removed suffix in |skew|. The search ends when the suffix is large enough so
that a lexicographically larger lattice permutation of the same shape can be
obtained by changing the suffix; this is a case as soon as an entry is
removed has a strictly larger column number than the one previously removed.
The variable~|c| is used to facilitate this termination condition.

@< Set |i| to the index of the last square that can be moved down... @>=
{ index c=n;  /* sufficiently large starting value */
  if (n==1) return false;
  
  lambda=alloc_array(@[index@],2*n+1); skew=&lambda[n];
    /* |lambda| and |skew| are |1|-based */
  for (i=2*n; i>0; --i) lambda[i]=0; /* clear |lambda| and |skew| */
  for (i=0; i<n; ++i) ++lambda[t[i]]; /* set $\lambda=\mathop{\rm sh}(T)$ */

  for (i=n-1; i>=0; --i)
  { --lambda[r=t[i]]; ++skew[r];
    if (lambda[r]>c) goto found; @+ else c=lambda[r];
  }
  freearr(lambda); return false; /* final tableau */

found: {}
}

@ Once we know that we can increase the lattice permutation by changing the
suffix, we must find the minimal way to do this. For a row number~|r| to be a
candidate for insertion into~|t[i]|, we must have |r>t[i]|, |skew[r]>0|, and
adding~|1| to |lambda[r]| should keep |lambda| a partition. The remaining
squares of~|skew| can be added to~|t| by weakly increasing row number.

@< Find the smallest possible |r>t[i]|... @>=
{ do ++r; while (skew[r]==0 || lambda[r]==lambda[r-1]);
    /* find row for first changing entry */
  t[i++]=r; --skew[r]; /* replace |t[i]|, update skew */
  for (r=1; r<=n; ++r) @+
    while (skew[r]-->0) t[i++]=r; /* distribute remaining squares */
  freearr(lambda);
}

@*1 Generator functions.
The incrementor functions above can be used in a trivial way to generate
matrices with all permutations, partitions ot tableaux of a given kind.
The function |Permutations| generates a matrix with all permutations of a
vector~|v|.

@< Prototypes... @>=
matrix* Permutations(entry* v,index n);
@~@c
matrix* Permutations(entry* v,index n)
{ index N=1; entry* w=mkintarray(n); copyrow(v,w,n); sortrow(w,n);
  @/{@; index i=0,j=n-1; while (i<j) swap(&w[i++],&w[j--]); }
    /* increasing order */
  @< Compute number~|N| of distinct permutations of~|w| @>
  { matrix* result=mkmatrix(N,n); index i=0;
    do copyrow(w,result->elm[i++],n); while (Nextperm(w,n));
    freearr(w); return result;
  }
}

@ The number |N| to be computed here is a multinomial coefficient, computed
by dividing $n!$ by the product of the factorials of the multiplicities of
all values that occur as entries of~|w|.

@< Compute number~|N|... @>=
{ index i=0, mult=1;
  while (++i<n) {@; N*=i+1; if (w[i]>w[i-1]) mult=1; else N /= ++mult; }
}

@ The function |Partitions| generates a matrix with all partitions of~|n|.

@< Prototypes... @>=
matrix* Partitions(index n);
@~@c
matrix* Partitions(index n)
{ matrix* result=mkmatrix(n_parts(n),n);
  if (n>0)
  { entry* lambda=mkintarray(n),** res=result->elm; index i=0,j;
    lambda[0]=n; @+
    for(j=1;j<n;j++) lambda[j]=0; /* initialise |lambda| to $[n,0,0,\ldots]$ */
    do copyrow(lambda,res[i++],n); while(Nextpart(lambda,n));
    freearr(lambda);
  }
  return result;
}

@ The function |Tableaux| generates a matrix with all tableaux of some shape
|lambda|.

@< Prototypes... @>=
matrix* Tableaux(entry* lambda, index l);
@~@c
matrix* Tableaux(entry* lambda, index l)
{ bigint* nt=n_tableaux(lambda,l); index n=check_part(lambda,l);
  matrix* result=mkmatrix(bigint2entry(nt),n);
  entry** res=result->elm,* t=mkintarray(n);
  freemem(nt);
@)
  @< Initialise |t| to the tableau of shape |lambda| filled by rows @>
  {@; index i=0; do copyrow(t,res[i++],n); while(Nexttableau(t,n)); }@/ 
  freearr(t); return result;
}

@
@< Initialise |t|... @>=
{@; index i=0,j,k;
  for (j=1; j<=l; ++j) for (k=lambda[j-1]; k>0; --k) t[i++]=j;
}


@*1 Auxilliary functions for partitions.
Here are a few simple functions dealing with partitions. The first computes
the transpose partition.

@< Prototypes... @>=
vector* Trans_part(entry* lambda, index l);
@~Although efficiency is not crucial here, the fastest way to compute the
transpose partition is to follow the edge of the Young diagram, from bottom
left to top right.

@c
vector* Trans_part(entry* lambda, index l)
{ index i,j=0;
  vector* result=mkvector(l ? lambda[0] : 0); entry* res=result->compon;
  for (i=l-1; i>=0; --i)
    while (j<lambda[i]) res[j++]=i+1;
  return result;
}

@ The function |Sign_part| computes the sign of a permutation of given cycle
type.

@< Prototypes... @>=
entry Sign_part(entry* lambda, index l);
@~@c
entry Sign_part(entry* lambda, index l)
{ index i,s=0; /* count non-zero even parts: */
  for (i=0; i<l && lambda[i]>0; ++i) if (lambda[i]%2==0) s++;
  return s%2 ? -1 : 1;
}

@* The Robinson-Schensted algorithm.
On lattice permutations, the description of the celebrated
Robinson-Schensted algorithm is very simple. One just keeps track of the row
number |r| where the current ``bump'' takes place while scanning the squares
in order of their entries; each time a square in row |r| is encountered its
row number is changed and |r| is also updated. In the lattice
permutation~|P|, the scan is just through successive positions~|i|, and the
entry |P[i]| is the row number that is matched against~|r| (if~|P[i]==0|,
then the entry~|i| has already been removed).

First, we consider the computation of a permutation from a pair of tableaux;
|r| is decreased until |r==0|.

@< Prototypes... @>=
void Robinson_Schensted (entry* P, entry* Q, index n, entry* sigma);
@~@c
void Robinson_Schensted (entry* P, entry* Q, index n, entry* sigma)
{ index j;
  for(j=n-1; j>=0; --j)
  { entry r=Q[j]; index i=n;
    while ( P[--i]!=r || (P[i]= --r)>0 ) {}
    sigma[j]=i+1; /* permutation values start at~|1| */
  }
}

@ Here is the obvious inverse of the function above. Now |sigma| is input,
and |P,Q| are output. For the |j|-th pass we visit the positions
|sigma[j]|\dots|n| while increasing |r| from 1 upwards whenever an entry
matches it; its final value becomes |Q[j]|. The entries of |sigma| must be a
permutation of~|n|; only their range has been tested before calling
|Schensted_Robinson|.

@< Prototypes... @>=
void Schensted_Robinson (entry* sigma, index n, entry* P, entry* Q);
@~@c
void Schensted_Robinson (entry* sigma, index n, entry* P, entry* Q)
{ index j;
  for(j=0;j<n;++j) P[j]=0;
  for(j=0; j<n; ++j)
  { index i=sigma[j]-1,r=1;
    if (P[i]==0) P[i]=r; @+ else error("Not a permutation.\n");
    while (++i<n) if(P[i]==r) P[i]= ++r;
    Q[j]=r;
  }
}

@* Characters.
The following functions will compute the characters of representations of
the symmetric group (these are known to have integral values). Actually,
only the code that computes the irreducible characters by means of the
Murnaghan-Nakayama rule is compiled; the rest is disabled by the
preprocessor. The disabled functions are the way characters were computed in
the previous version of \LiE.; their definition and documentation has been
retained for reference and comparison. Although their performance is
acceptable for small symmetric groups (the ones one is most likely to need
for computing plethysms), the Murnaghan-Nakayama rule is significantly more
efficient, in particular when it comes to character values at classes with
many fixed points and small cycles.

@*1 Permutation characters.
The first of these representations is the one induced from the trivial
representation of the Young subgroup corresponding to |lambda|, or
equivalently, the permutation representation of $\Sn$ on the cosets of this
Young subgroup. The Young subgroup corresponding to |lambda| is the
stabiliser of the colouring of an |n|-element set where the first
$\lambda_1$ elements get colour~$1$, the next $\lambda_2$ elements get
coulour~$2$, etc.; the cosets of this subgroup can be identified with images
of this colouring under permutations of the elements. The value of the
character of this representation at conjugacy class~|mu| then is equal to
the to number of such colourings for which all cycles of some fixed
permutation in this class are uniformly coloured, or equivalently the number
of maps from the parts of~|mu| to those of~|lambda| such that each part
$\lambda_i$ equals the sum of all the parts $\mu_j$ mapping to it.

This number is computed using a simple recursion, where we start with the
largest part~$M=\mu_1$. Clearly it can be mapped only to parts $\lambda_i$ with
$\lambda_i\geq M$, and if several such $\lambda_i$ are equal, we can
compute the contribution for one of them and multiply by their number. Once
we have chosen $\lambda_i$, we replace it by $\lambda_i-M$, and then use
recursion to find the number of colourings for $(\mu_2,\mu_3,\ldots)$. This
function is used only as an auxilliary one, and the calling function will
ensure that |lambda| and |mu| are no aliases, so that it is safe to modify
|lambda| {\sl in situ}.

@c
#if 0
static entry Young_char_val(entry* lambda, entry* mu, index l, index m)
{ index i=0; entry sum=0,li, M=m==0 ? 0 : mu[0]; /* largest part of |mu| */
  if (M==0) return 1; /* stop when |mu| is depleted */
  while (i<l && (li=lambda[i])>=M) /* try parts $\geq M$ of |lambda| */
  { index j; entry c=1; /* number of equal parts */
    while (++i<l && lambda[i]==li) ++c; /* move |i| beyond largest part |li| */
    @< Subtract |M| from |lambda[i-1]| and reorder parts of |lambda| @>
    sum+=c*Young_char_val(lambda,mu+1,l,m-1);
    @< Restore |lambda| after recursive call of |Young_char_val| @>
  }
  return sum;
}
#endif

@ The value of~|lambda[i-1]| is still in |li| when we come here. All the
parts with |li>lambda[j]>li-M| have to be shifted a place down.

@< Subtract |M| from |lambda[i-1]|... @>=
{ li-=M;
  for(j=i; j<l && lambda[j]>li; ++j) lambda[j-1]=lambda[j];
  lambda[j-1]=li;
}

@ Since |i|, |j|, and~|li| are not affected by the recursion, restoring
|lambda| is straightforward.

@< Restore |lambda| after recursive call of |Young_char_val| @>=
{ while (--j>=i) lambda[j]=lambda[j-1];
  lambda[j]=li+M;
}

@*1 Irreducible characters.
The function |Schur_char_val| computes the value of the character
$\chi^\lambda$ of the irredudicible representation of~$\Sn$ corresponding
to~|lambda|, at the conjugacy class~|mu|. Its implementation is a
straightforward use of the Jacobi-Trudi identity, which expresses
$\chi^\lambda$ in terms of the Young characters $Y(\nu)$, or equivalently,
expresses the Schur function $S_\lambda$ as a determinant in terms of the
total symmetric functions $h_i$ (see [Macd, I~(3.4)]). We use the formula
$$ \chi^\lambda=
   \sum_{\sigma\in\Sym l}\sg(\sigma)Y(\lambda-\rho+\sigma(\rho))
$$
where $\rho=[0,1,\ldots,l-1]$, and the argument of~$Y$ is reordered to form
a partition.

@c
#if 0
entry Schur_char_val(entry* lambda, entry* mu, index l, index m)
{ index i; entry sum=0;
  while (l>0 && lambda[l-1]==0) --l; /* get reduced form of~|lambda| */
  @< Handle the cases |l<=1|, and when the transpose of |lambda| works more
     efficiently @>
  @< Traverse those permutations |sigma| with $\lambda[i]-i+\sigma_i\geq0$
     for all~$i$, and sum $\sg(\sigma)$ times the
     corresponding Young character value @>
  return sum;
}
#endif

@ Our method of computation works best for partitions with few parts;
transposition may improve this.

@< Handle the cases |l<=1|... @>=
if (l<=1) return 1; /* trivial character */
if (l>lambda[0]) /* then better work with the transpose partition */
{ vector* tr=Trans_part(lambda,l);
  entry ch=Schur_char_val(tr->compon,mu,lambda[0],m); freemem(tr);
  return Sign_part(mu,m)*ch;
}

@ Permutations are generated by enumerating the sequences
|(pos[0],@t\dots@>,pos[l-1])| with |0<=pos[i]<=i| for all~|i|; such a
sequence encodes the permutation |sigma| that first interchanges entries
|l-1| and |pos[l-1]|, then |l-2| and |pos[l-2]|, etc. This permutation is
stored as $\sigma(\rho)$. During this process the array |lambda_prime| holds
the sequence $\lambda-\rho+\sigma(\rho)$. We start with |pos[i]==i| for
each~|i|, and hence with $\sigma(\rho)=\rho$, so that we can just copy
|lambda| into~|lambda_prime| at the beginning.

@< Traverse those permutations |sigma|... @>=
{ entry* lambda_prime=mkintarray(4*l)
     ,* sigma=lambda_prime+l,* pos=sigma+l,* nu=pos+l;
    /* 4 length-|l| arrays */
  boolean sg=true; /* positive sign */
  copyrow(lambda,lambda_prime,l);
    /* |lambda| might be alias of |mu|, but |lambda_prime| is not */
  for (i=0; i<l; ++i) pos[i]=sigma[i]=i;
    /* |sigma| is the permutation; |pos| records its swap sequence */
  do
  { copyrow(lambda_prime,nu,l); @< Sort |nu| into decreasing order @>
    sum+= sg ? Young_char_val(nu,mu,l,m) : -Young_char_val(nu,mu,l,m);
    @< Replace |sigma| by the next contributing permutation, adapting
       |lambda_prime| and |sg| correspondingly; if no permutations are left
       |break| @>
  } while (true);
  freearr(lambda_prime);
}

@ In traversing the permutations, we keep the later entries of |pos|, and
therefore also of~|p|, fixed, while trying to modify the earlier ones. This
is the most efficient order, for the following reason. If we choose
|p[i]==i| for all $i\geq i_0$, then there is at least one way to validly
choose |p[i]| for $i<i_0$. Now if for some other choice of |p[i]| we have
$\lambda_i-i+\sigma_i\geq0$ for $i\geq i_0$, then the set of values |p[i]|
already in use is systematically smaller than for the initial choice
|p[i]==i|, and the remaining values are systematically larger; therefore in
this case there will also be at least one way to validly choose |p[i]| for
$i<i_0$. Therefore, every time a value for |pos[i]| is installed below,
at least one permutation will be found with that value, and no time is spent
searching in dead-end branches. The code below uses this explicitly by
dividing the search into one for the minimal prefix that can be altered,
followed by one for the next permutation within that prefix.

@< Replace |sigma| by the next... @>=
{ index i=0,j;
  @< Search minimal |i| for which |pos[i]| can be lowered; set |i=l| if not
     found, and otherwise find maximal |j<pos[i]| to replace |pos[i]| @>
  if (i==l) break;
  @< Replace |pos[i]| by |j|, and $pos[i-1],\ldots,pos[0]$ by maximal
     valid values @>
}

@ When the value of |pos[i]| is released to be replaced with another value
later, the corresponding transposition in~|p| must be undone. The inner loop
tries to find a replacement |j| for |pos[i]|; as long as this fails (leaving
|j<0|), the outer loop increases~|i|.

@< Search minimal |i| for which |pos[i]| can be lowered... @>=
do
{ @< Move the old value of |pos[i]| into |j|, and undo its setting @>
  do
    if(--j<0) break; /* tried all positions for this |i| */
  while (lambda_prime[i]+sigma[j]<0);
} while (j<0 && ++i<l);

@ For |pos[i]| we have already found a candiate~|j| when we get here, and
for $i'<i$ there then always exists a valid value~|j| with |0<=j<=@t$i'$@>|,
as we argued above. For that reason, the |for| loop below does not have to
deal with the possibility that |j| becomes negative.

 @< Replace |pos[i]| by |j|, and... @>=
do /* now |j>=0| and |sigma[j]| can move validly to |sigma[i]| */
{ @< Install |j| as |pos[i]|, and make adjustments for this new value @>
  if (--i<0) break;
  for (j=i; lambda_prime[i]+sigma[j]<0; --j) {} /* this leaves |j>=0| */
} while (true);

@ Since at most steps only a small part of~|sigma| changes, we maintain
|lambda_prime| incrementally. Once |pos[i]| is fixed, |sigma[i]| is
determined, and can be added to |lambda_prime[i]|.

@< Install |j| as |pos[i]|... @>=
{ if ((pos[i]=j)<i) {@; swap(&sigma[i],&sigma[j]); sg=!sg;}
  lambda_prime[i]+=sigma[i]; /* this becomes non-negative */
}

@ When |pos[i]| is removed, we reverse the operations above. Since we started
with $\lambda'=\lambda=\lambda-\rho+\sigma(\rho)$ for |sigma| the
identity, the code below also implicitly takes care of subtracting $\rho$
from~|lambda|.

@< Move the old value of |pos[i]| into |j|... @>=
{ lambda_prime[i]-=sigma[i];
  if ((j=pos[i])<i) {@; swap(&sigma[i],&sigma[j]); sg=!sg;}
}

@ Since |nu| is usually already almost sorted, we use insertion sort for
efficiency.

@< Sort |nu|... @>=
{ index i; for (i=1; i<l; ++i) if (nu[i]>nu[i-1]) /* skip most cases */
  { entry nui=nu[i]; index j=i;
    do nu[j]=nu[j-1]; while (--j>0 && nui>nu[j-1]); nu[j]=nui;
  }
}

@ Here is a function that calls |Schur_char_val| repeatedly to obtain the
full character $\chi^\lambda$.

@c
#if 0
matrix* Schur_char(entry* lambda, index l)
{ index i,n=check_part(lambda,l); entry np=n_parts(n);
  matrix* result=mkmatrix(np,n+1); entry** res=result->elm;
  res[0][0]=n; for (i=1; i<n; ++i) res[0][i]=0; i=0;
  while (res[i][n]=Schur_char_val(lambda,res[i],l,n),++i<np)
  @/{@; copyrow(res[i-1],res[i],n); Nextpart(res[i],n); }
  return result;
}
#endif

@*1 The Murnaghan-Nakayama rule.
Next, we give an algorithm for computing the entire irredudicible
character $\chi^\lambda$ of~$\Sn$, using the Murnaghan-Nakayama rule (see
[Zel,~4.17(b)]); the result is represented as a polynomial
$\sum_{\mu\in\Part}\chi^\lambda(\mu)X^\mu$.

@< Prototypes... @>=
poly* MN_char(entry* lambda, index l);

@~The Murnaghan-Nakayama rule states that the value of $\chi^\lambda$ at the
class |mu| can be obtained by finding all ways to deplete the Young diagram
of~|lambda| by successively removing rim hooks whose sizes are the parts
of~|mu| in some fixed order; each such possiblility contributes its sign to
the character value, which is $-1^k$ where $k$ is the number of pairs of
vertically adjacent squares contained in the same rim hook. We order the
parts of~|mu| decreasingly, so that the most difficult hooks are positioned
first, and counting the possibilities for placing the final hooks of
size~|1| into the remaining Young diagram can be done using the hook length
formula.

@c
poly* MN_char(entry* lambda, index l)
{ index n=check_part(lambda,l);
  if (n==0) return poly_one(0); /* the character of $\Sym0$ */
  while (lambda[l-1]==0) --l; /* minimise |l| */
  wt_init(n); /* get ready for accumulating contributions to the character */
  { @< Declare and initialise variables used in the recursive search of
       |MN_char| @>@;
    @< Construct and remove a sequence of rim hooks of sizes
       $\mu_0\geq\cdots\geq\mu_{i-1}$ with $\mu_{i-1}>1$; define the label
       |resume| as continuation point for the search @>
    @< Extend |mu| by |1|'s, and count the number of ways to remove the
       remaining squares by using the hook-length formula; contribute this
       number with the proper sign to the coefficient of~|mu| @>
    if (i>0) goto resume;
    @< Clean up temporary storage of |MN_char| @>
  }
  return wt_collect();
}

@ The sizes of the rim hooks will be stored in the array~|mu|, and their
positions in~|save|; the temporary array |lambda_prime| serves to hold the
partition passed to the hook length formula. While removing hooks from the
Young diagram however, we use a different data structure to represent
partitions: a description of the edges that form the end of a row or column
of the Young diagram. These combine to a path from bottom left to top right,
that is determined by the orientations (horizontal or vertical) of its
successive segments (this is called a ``Maya diagram'' in [DJKMO]).  In this
representation, every possibility to remove a rim hook of size~|r|
corresponds to a pair of edges |r| places apart that are oriented
horizontal-vertical; removing the hook means interchanging the orientations
of these edges. Encoding horizontal edges as~|0| and a vertical edges as~1,
the sum of the values strictly between the edges delimiting a hook gives its
leg length, i.e., its contribution to the number~|k| used in defining the
sign. The length~|d| of the sequence |edge| is taken large enough to
accommodate the largest hook of~|lambda|. If no hook of length~|r| exists
initially, we need not try partitions with a part~|r|, so as an optimisation
we record in the array |candidate| for which sizes hooks are to be
considered.

@< Declare and initialise variables used in the recursive search 
   of |MN_char| @>=
entry* mu=mkintarray(3*n),* save=mu+n,* lambda_prime=save+n;
int i, j, r, d=lambda[0]+l, k=0; /* sum of leg lengths */ 
boolean* edge=alloc_array(@[boolean@],2*d-2),* candidate=edge+d-2;
  /* index |2<=r<d| */
enum {hor, vert}; /* values used for |edge| */

for (i=0; i<n; ++i) mu[i]=0;
@< Initialise |edge| to the edge structure of |lambda| @>
@< For all |r|,
   set |candidate[r]| iff there exists a hook of length~|r| @>

@~@< Clean up temporary storage... @>= @+
{@; freearr(edge); freearr(mu); }

@ The edges are recorded by scanning the parts of |lambda| in reverse order;
every time the length increases one or more horizontal segments are appended,
and for every part one vertical segment is appended.

@< Initialise |edge|... @>=
{ int r=l-1,c=0; /* current column number */
  for (j=0; r>=0; --r)
  { while (c<lambda[r]) {@; edge[j++]=hor; ++c; } /* columns of length |r| */
    edge[j++]=vert; /* row |r|, of length |c==lambda[r]| */
  }
}

@ We try all hook lengths for the original partition~|lambda|; once one is
found we can go on to the next length.

@< For all |r|, set |candidate[r]|... @>=
for (r=2; r<d; ++r)
{ for (j=0; j+r<d; ++j)
    if (edge[j]==hor && edge[j+r]==vert) break;
  candidate[r]= j+r<d;
}

@ We use a basically recursive technique to find all sequences we are
looking for: after placing a rim hook, we perform the same computation for a
modified Young diagram (as represented by its edge sequence). Because \Cee\
makes it inconvenient to have local variables such as |candidate| that are
shared between recursive incarnations, the recursion is broken and replaced
by explicit jumps at the beginning and end, while saving of values in a
stack-like fashion is done ``by hand''. The loop below therefore
conceptually has a number of nested instantiations, each one traversing the
possibilities for one part of~|mu|; the array |mu| functions as a stack
where the currently active values of the loop variable |r| are stored. This
means that logically the recursive starting point is outside the loop, and
each jump to |recurse| starts a new instantiation of the loop. If we had
done that, a mechanism would have been needed to ensure that the new loop
starts at the current value of~|r| in the ``enclosing'' loop, so that only
weakly decreasing sequences |mu| are generated; instead we take advantage of
the fact that we are making a jump instead of a recursive call, and move the
label |recurse| into the loop body, so that the old value of~|r|
automatically becomes the starting value for the new loop.  Therefore, apart
from being nested within itself, this loop has an extra peculiarity: the
nested instantiations are only virtually entered, although they do complete
in the ordinary way.

@< Construct and remove a sequence of rim hooks... @>=
{ i=0; /* index of last entry that was set in~|mu| */
  for (r=d-1; r>1; --r) /* try hooks of size |r| */
    if (candidate[r])
    { recurse: /* recursive starting point */
      @< Find all ways to remove hooks of length~|r| from the |edge|
	 structure, adjusting |mu| and |k| appropriately; instead of a
	 recursive call say `|goto recurse; resume:@;|'; afterwards return
	 values to their original state @>
    }
}

@ Here is the heart of the search procedure. All posible positions~|j| of a
rim hook of size~|r| with respect to the edge sequence are tried, which
amounts to another loop that is instantiated recursively. We maintain the
accumulated leg length~|k|; since it differs by at most~|1| between
successive positions tries, it is most efficient to update it incrementally.
Once a pair of suitable edge orientations is found, they are reversed, the
current values of the loop variables are saved (where the value~|i|
functions as a stack pointer into |mu|~and~|save|), and the recursive jump
is made. After returning, the values are restored to their original values.

@< Find all ways to remove hooks... @>=
{ for (j=1; j<r; ++j) k+=edge[j]; /* leg length of hook first tried */
  for (j=0; j<d-r; ++j)
  { if (edge[j]==hor && edge[j+r]==vert)
    { edge[j]=vert; edge[j+r]=hor; mu[i]=r; save[i++]=j; goto recurse;
    resume: j=save[--i]; r=mu[i]; mu[i]=0; edge[j]=hor; edge[j+r]=vert;
    }
    k+= edge[j+r]-edge[j+1]; /* adjust |k| for hook tried next */
  }
  while (++j<d) k-= edge[j]; /* restore |k| */
}

@ When we have chosen all rim hooks of size |r>1|, we reconstruct the
remaining shape~|lambda_prime| from~|edge|, and count its number~|s| of
squares; then |mu| is extended with |s| ones, and the number of standard
tableaux of this shape, as determined by |n_tableaux|, is added to or
subtracted from (depending on the parity of the accumulated leg length~|k|)
the coefficient of the class |mu| in the character.

@< Extend |mu| by |1|'s, and... @>=
{ int r=l,c=0,s=0; /* size of |lambda_prime| */
  for (j=0; r>0; )
    if (edge[j++]==vert) s+=lambda_prime[--r]=c; @+ else ++c;
    /* build |lambda_prime| from edges */
  for (j=0; j<s; ++j) mu[i++]=1; /* extend |mu| with |s| ones */
  wt_ins(mu,n_tableaux(lambda_prime,l),k%2);
  for (j=0; j<s; ++j) mu[--i]=0; /* remove the ones again */
}

@ Now we give a variant of this computation where we compute only the
character value at a different class~|mu|. The computation is entirely
similar, but there are enough simplifications to justify writing separate
code for this case.

@< Prototypes... @>=
bigint* MN_char_val(entry* lambda, entry* mu, index l, index m);

@~Like before we distinguish parts $\mu_i\geq2$ from the remaining parts of
size~|1|; the number of parts of the former kind is precomputed in~|m2|.

@c
bigint* MN_char_val(entry* lambda, entry* mu, index l, index m)
{ bigint* value=null; index n=check_part(lambda,l),m2;
  if (n==0) return one;
  while (lambda[l-1]==0) --l; @+ while (mu[m-1]==0) --m;
  for (m2=m; m2>0 && mu[m2-1]==1; --m2) {} /* number of parts $\mu_i\geq2$ */
  { entry* save=mkintarray(2*n),* lambda_prime=save+n;
    int i, j, d=lambda[0]+l, k=0; /* sum of leg lengths */ 
    boolean* edge=alloc_array(@[boolean@],2*d);
    enum {hor, vert};

    @< Initialise |edge|... @>
    @< If |mu| has any forbidden parts, |return null| @>
    @< Recursively generate all contributions to the character value @>
    freearr(edge); freearr(save);
  }
  return value;
}

@ We do not need an array |candidate|, like in the previous function, but if
|mu| has any parts for which we would not have had |candidate[mu[i]]|, then
the coefficient computed will be zero.

@< If |mu| has any forbidden parts... @>=
for (i=0; i<m2; ++i)
{ for (j=0; j+mu[i]<d; ++j)
    if (edge[j]==hor && edge[j+mu[i]]==vert) break;
  if (j+mu[i]==d) return null; /* no hook of size |mu[i]| was found */
}

@ The pattern of recursion is slightly different here than in |MN_char|.
Since |mu| is given, we should not generate a contribution to the
result for each sequence of hooks of sizes $\geq2$ generated: this should
only be done when the sequence has reached length~|m2|.

@< Recursively generate all contributions... @>=
{ i=0; /* index into |mu| */
recurse:
  if (i<m2)
    @< Traverse all possibilities to remove a hook of size |r|;
       save the current values using |i| as stack pointer and say
       `|goto recurse; resume:@;|' @>
  else
    @< Compute the number of ways to distribute the remaining |1|-hooks,
       and contribute it with the proper sign to |value| @>
  if (i>0) goto resume;
}

@ This module is very similar to the corresponding one for |MN_char|, but
|mu| is not changed here.

@< Traverse all possibilities to remove a hook of size |r|... @>=
{ int r=mu[i];
  for (j=1; j<r; ++j) k+=edge[j]; /* leg length of hook first tried */
  for (j=0; j+r<d; ++j)
  { if (edge[j]==hor && edge[j+r]==vert)
    { edge[j]=vert; edge[j+r]=hor; save[i++]=j; goto recurse;
    resume: j=save[--i]; r=mu[i]; edge[j]=hor; edge[j+r]=vert;
    }
    k+= edge[j+r]-edge[j+1]; /* adjust |k| for hook tried next */
  }
  while (++j<d) k-= edge[j]; /* restore |k| */
}

@ Finally, this module has also only undergone the obvious simplifications.

@< Compute the number of ways to distribute the remaining |1|-hooks... @>=
{ int r=l,c=0,s=0; /* size of |lambda_prime| */
  for (j=0; r>0; )
    if (edge[j++]==vert) s+=lambda_prime[--r]=c; @+ else ++c;
    /* build |lambda_prime| from edges */
  value= k%2==0 ? add(value,n_tableaux(lambda_prime,l)) 
                : sub(value,n_tableaux(lambda_prime,l)) ;
}

@* References and index.

\font\frak=eufm9

\beginrefs

\ref Macd \key Macdonald
 \author I. G. Macdonald
 \title Symmetric Functions and Hall Polynomials
 \publ Clarendon press, Oxford \year 1979
 \series Oxford Mathematical Monographs
\endref

\ref Zel \key Zelevinsky
 \author  A. V. Zelevinsky
 \title Representations of finite classsical groups, a Hopf algebra approach
 \series Lecture Notes in Mathematics \vol 869
 \publ Springer-Verlag, Berlin\slash New York \year 1981
\endref

\ref DJKMO \key PAths and Diagrams
 \author E.~Date, M.~Jimbo, A.~Kuniba, T.~Miwa, M.~Okado
 \title Paths, Maya Diagrams and representations of
        $\hat{\hbox{\frak sl}}(r,{\bf C})$
 \series Advanced Studies in Pure Mathematics \vol 16 \year 1989
 \inbook Integrable Systems in Quantum Field Theory and Statistical Mechanics
 \pages 149-191
\endref

\endrefs


