\input LiEwebmac

\nomods

@* Computing tensor products.
This file defines routine for computing tensor product decompositions.

@h "lie.h"
@d local static
@f local static
@f lambda NULL
@f mu NULL
@f nu NULL
@f rho NULL
@c
@< Local variables @>@;

@ The external interface is determined by the header file \.{tensor.h}.

@( tensor.h @>=
@< Prototypes of globally visible functions @>@;

@ Three different functions are provided to perform tensor product
decompositions: |Tensor(p,q)| will compute $p\tensor q$ for (possibly
reducible) modules |p|~and~|q|, while |Tensor_coef(p,q,lambda)| will limit
this computation to what is necessary to determine the coefficient of a given
weight |lambda|. Finally |Ptensor| combines several calls to |Tensor| in
order to compute the decomposition of a tensor power $p^{\tensor m}$.

@< Prototypes of globally visible functions @>=

poly* Tensor(poly* p, poly* q);
bigint* Tensor_coef(poly* p, poly* q,vector* lambda);
poly* Ptensor(index p, poly* m);

@ The computation of tensor product decompositions involves traversal of
Weyl group orbits by |Weylloop|. Communication with the function calling
|Weylloop| is done using the local variables |the_g|, which is set to the
simple group whose Weyl group orbits are currently being computed, |lamrho|,
which is set to |lambda+rho| where |lambda| is the highest weight of the
(larger) module whose full character is \emph{not} being generated, and
|cur_mult| which is set to the multiplicity with which the weights found
during the traversal should be contributed. For sake of efficiency, another
vector |cur_expon| is allocated beforehand, serving as working space.

@< Local variables @>=
local simpgrp* the_g;
local entry* lamrho,* cur_expon;
local bigint* cur_mult; 

@ We have to specify the action to be performed at each weight~|mu|. In case
of a full tensor product computation, this is specified by the function
|add_tensor_wt|. It transforms~|lambda+mu| into a dominant weight using
shifted reflections, as |alt_dom| does, keeping track of the parity of the
number of simple reflections needed in |neg|; because |rho| has already been
added to |lambda| into |lamrho|, its suffices to call |simp_make_dominant|
and to subtract |rho| afterwards. In case the weight becomes non-dominant by
this subtraction (i.e., if the result of |simp_make_dominant| contained any
zeros) there is no contribution from~|mu|, so we stop immediately. Otherwise
the resulting weight vector |cur_expon| is handed to |wt_ins| for
accumulation into the result, with the proper sign, 

@c
local void add_tensor_wt(entry* mu)
{ index k,r=the_g->lierank; boolean neg;
  addrow(lamrho,mu,cur_expon,r); /* |cur_expon=mu+lambda+rho| */
  neg=simp_make_dominant(cur_expon,the_g)%2!=0; /* apply |alt_dom| action */
  for (k=0; k<r; ++k) if (--(cur_expon[k])<0)
    return; /* subtract |rho| and quit unless result is dominant */
  wt_ins(cur_expon,cur_mult,neg); /* and insert into result */
}

@ When we are only interested in the coefficient of a single weight in the
result, we use |add_goal_wt| instead of |add_tensor_wt|. If the weight sought
for is~|nu| we set |goal=nu+rho| (this saves us the subtraction of |rho|
that was needed in |add_tensor_wt|), and the multiplicity so far is recorded
in~|totmul|.

@< Local variables @>=
local entry* goal;
local bigint* totmul;

@~The operation of |add_goal_wt| is quite similar to that of
|add_tensor_wt|, but instead of |ins_wt| we call |add| or |sub| as
appropriate.

@c
local void add_goal_wt(entry* mu)
{ index r=the_g->lierank; boolean neg;
  addrow(lamrho,mu,cur_expon,r); /* |cur_expon=mu+lambda+rho| */
  neg=simp_make_dominant(cur_expon,the_g)%2!=0; /* apply |alt_dom| action */
  if (!eqrow(cur_expon,goal,r)) return; /* quit unless |cur_expon==goal| */
  totmul= neg ? sub(totmul,cur_mult) : add(totmul,cur_mult);
    /* incorporate in |totmul| */
}

@ The function |simp_tensor_irr| does the basic computation for irreducible
modules of simple groups, using the Brauer-Klimyk formula. The arguments
|lambda| and |mu| are the highest weights of the tensorand modules, and |nu|
is either the weight whose coefficient in the tensor product decomposition we
are interested in, or |nu==NULL| when we are interested in the full tensor
product decomposition. The specified return type is |(poly*)|, which is
appropriate for the latter case; in the former case we disguise the |bigint|
as the coefficent of a one-term polynomial.

@c
local poly* simp_tensor_irr(entry* lambda,entry* mu,entry* nu,simpgrp* g)
{ poly* result; index i,r=g->lierank;
  the_g=g; testdom(lambda,(object)g); testdom(mu,(object)g);
  cur_expon=mkintarray(r);
  if (nu!=NULL) @< Set |goal=nu+rho| @> @+ else goal=NULL;
  @< Make sure |lambda| represents the larger module @>
  @< Set |lamrho=lambda+rho| @>
  if (nu!=NULL) {@; totmul=null; setshared(null); } @+ else wt_init(r);
@)
  @< Traverse the full character for the highest weight~|mu| @>
@)
  freearr(cur_expon); freearr(lamrho);
  if (nu==NULL) return wt_collect();
  @+ else @< Pack |totmul| into a polynomial, and |return| it @>
}

@
@< Set |goal=nu+rho| @>=
{ goal=mkintarray(r); copyrow(nu,goal,r); /* |goal=nu| */
  for (i=0; i<r; ++i) ++goal[i];  /* |goal+=rho| */
}

@~
@< Set |lamrho=lambda+rho| @>=
{ lamrho=mkintarray(r); copyrow(lambda,lamrho,r); /* |lamrho=lambda| */
  for (i=0; i<r; ++i) ++lamrho[i]; /* |lamrho+=rho| */
}

@ Which module is larger is decided by comparing dimensions, which can be
computed efficiently by Weyl's dimension formula.

@< Make sure |lambda| represents... @>=
{ bigint* deg_lam=simp_dim_irr(lambda,g),* deg_mu=simp_dim_irr(mu,g);
  if (cmp(deg_lam,deg_mu)<0) {@; entry* t=lambda; lambda=mu; mu=t; }
  freemem(deg_lam); freemem(deg_mu);
}

@
@< Traverse the full character for the highest weight~|mu| @>=
{ poly* domchar=simp_domchar(mu,NULL,g); /* compute dominant character */
  Weylloopinit(g);
  for (i=0; i<domchar->nrows; ++i) /* traverse dominant weights in character */
  { cur_mult=domchar->coef[i];
      /* |cur_mult| is alias for the relevant multiplicity (shared) */
    Weylloop(nu ? add_goal_wt : add_tensor_wt,domchar->elm[i]);
  }
  Weylloopexit();
  freemem(domchar); 
}

@
@< Pack |totmul|... @>=
{@; freearr(goal); result=mkpoly(1,0); result->coef[0]=totmul;
  setshared(totmul); return result;
}

@ The function |tensor_irr| performs the computations necessary for
composite groups, where the function |simp_tensor_irr| is called to do the
work for each simple factor.

@c
local poly* tensor_irr(entry* lambda,entry* mu,entry* nu)
{ if (type_of(grp)==SIMPGRP) return simp_tensor_irr(lambda,mu,nu,&grp->s);
  if (simpgroup(grp)) return simp_tensor_irr(lambda,mu,nu,Liecomp(grp,0));
  { poly* result;
    index s=Ssrank(grp),td=grp->g.toraldim; /* size of toral part */
    @< Initalise |result| to tensor product of toral parts,
       or |return| a 0~polynomial if the toral part of |nu| does not match @>
    @< For each simple component of |grp| compute the tensor product of
       the corresponding part of the weights, and multiply the answer into
       |result| by means of |Disjunct_mul_pol_pol| @>
    return result;
  }
}

@ The tensor product for the toral parts of weights is formed by adding the
weights. If |nu!=NULL| and its toral part is not the sum of those of
|lambda| and |mu|, then the requested coefficient will be~0.

@< Initalise |result| to tensor product of toral parts... @>=
{ lambda+=s; mu+=s; /* move to start of toral part */
  if (nu==NULL)
  {@; result=mkpoly(1,td);
    addrow(lambda,mu,*result->elm,td); *result->coef=one;
  }
  else
  { entry* lm=mkintarray(td); boolean correct_weight;
    addrow(lambda,mu,lm,td); nu+=s; /* move to toral part of weight */
    correct_weight=eqrow(lm,nu,td); freearr(lm);
    if (correct_weight) result=poly_one(0); @+ else return poly_null(0);
  }
}

@ Since initialising the toral parts has left |lambda|, |mu|, and possibly
|nu|, pointing beyond the part corresponding to the last simple factor
of~|grp|, it is easiest to traverse these simple factors in reverse order.

@< For each simple component of |grp| compute the tensor product... @>=
{ index i;
  for (i=grp->g.ncomp-1; i>=0; --i)
  { simpgrp* g=Liecomp(grp,i); index d=g->lierank;
    lambda-=d; mu-=d; @+ if (nu!=NULL) nu-=d;
      /* move back to previous component */
    result= Disjunct_mul_pol_pol(simp_tensor_irr(lambda,mu,nu,g),result);
  }
}

@ The logic of |Tensor| is quite simple: it is a bilinear expansion of the
tensor product for all irreducible factors, using |tensor_irr| for each pair
of terms.

@c
poly* Tensor(poly* p,poly* q)
{ index i,j; poly* ans=poly_null(Lierank(grp));
  for (i=0; i<p->nrows; ++i)
    for (j=0; j<q->nrows; ++j)
      ans=Addmul_pol_pol_bin(ans,tensor_irr(p->elm[i],q->elm[j],NULL)
				,mult(p->coef[i],q->coef[j]));
  return ans;
}

@~For |Tensor_coef|, the function that is called when we need only a single
coefficient of the tensor product decomposition, we proceed as in |Tensor|,
but the single-term polynomial returned from~|tensor_irr| is immediately
dismantled, and only its coefficient is used.

@c
bigint* Tensor_coef(poly* p, poly* q,vector* nu)
{ index i,j; bigint* ans=null;
  for (i=0; i<p->nrows; ++i)
    for (j=0; j<q->nrows; ++j)
    { poly* res=tensor_irr(p->elm[i],q->elm[j],nu->compon);
      ans=add(ans,mult(res->coef[0],mult(p->coef[i],q->coef[j])));
      freepol(res);
    }
  return ans;
}

@ An older version of \LiE. used Chinese powering to compute tensor powers,
reducing the number of tensor products by taking the tensor square of
intermedaite results whenever possible. This turns out to be a Very Bad
Idea, since this leads to calling |tensor_irr| with a pairs of large modules
of about the same size as arguments, in which case it must generate a full
character of one of them, which is very time consuming. Repeatedly taking a
tensor product with the original small module is much more efficient, since
only its character needs to be traversed over and over again; the small size
of this character more than compensates for the larger number of times a
tensor product has to be taken. In fact a further optimisation might be
ventured by trying to precompute the dominant character of the original
module instead of recomputing it for every tensor product computation.

@c
poly* Ptensor(index n, poly* p)
{ poly* x,* y;
  if (n==0) return poly_one(Lierank(grp));
  if (n==1) return p;
  x=p; setshared(p); /* protect |p| against |freepol| */
  do {@; y=Tensor(x,p); freepol(x); x=y; } while (--n>1);
  clrshared(p); /* now |p| needs no more protection */
  return x;
}

@* Index.
