\input LiEwebmac

@* Reflection action of the Weyl group.
We shall presently implement the way individual Weyl group elements act on the
weight and root lattices (as opposed to having the full group act, as in
\.{weylloop.w}). The basic case will be to determine the action of the
simple reflections. This could be realised easily using the Cartan matrix, but
the pertinent matrices would be very sparse (the identity minus a row of the
Cartan matrix), so that considerable efficiency can be gained by coding the
details of the action explicitly into the functions. This does mean the
functions have to distinguish cases according to the types of simple groups,
and that the information of the Cartan matrix is duplicated.

@d local static
@f local static

@h "lie.h"
@f lambda NULL
@f alpha NULL
@f rho NULL

@ Here are functions that let the simple reflection~$s_i$ act on a weight
vector (it is used below and in \.{coxeter.w}). We take the $A_n$ action as
default, and record explicitly only the deviations from that case.

@c
void simp_w_refl(entry* w, index i, simpgrp* g)
{ index r=g->lierank; entry wi=w[i];
  if (wi==0) return; /* weight is on reflection hyperplane */
  {@; if (i>0) w[i-1]+=wi; } w[i]= -wi; 
  {@; if (i<r-1) w[i+1]+=wi; } /* $A_n$ action */
  switch (g->lietype)
  {	 case 'B': @+ if (i==r-2) w[i+1]+=wi; @+
  break; case 'C': @+ if (i==r-1) w[i-1]+=wi; @+
  break; case 'D': if (i>=r-3) if (i==r-1) {@; w[i-2]+=wi; w[i-1]-=wi; }
			       else if (i==r-2) w[i+1]-=wi;
                                    else w[i+2]+=wi; @+
  break; case 'E': if (i<4)
    { if (i<2) {@; w[i+1]-=wi; w[i+2]+=wi; } @+ else w[i-2]+=wi;
      if (i==1 || i==2) w[i-1]-=wi;
    } @+
  break; case 'F': @+ if (i==1) w[2]+=wi; @+
  break; case 'G': @+ if (i==1) w[0]+=2*wi;
  }
}

@ We similarly define the action of simple reflections~$s_i$ on vector
expressed on root basis.
@c

local void simp_rt_refl (entry* w,index i,simpgrp* g)
{ index r=g->lierank; entry c= -w[i]+(i>0?w[i-1]:0)+(i<r-1?w[i+1]:0);
  switch (g->lietype)
  {	 case 'B': @+ if (i==r-1) c+=w[i-1];
        /* long root neighbour of short root contributes twice */
  break; case 'C': @+ if (i==r-2) c+=w[i+1]; /* idem */
  break; case 'D': if (i>=r-3)
   /* correct for different adjacencies at end of diagram */
         if (i==r-1) c+=w[i-2]-w[i-1];
           /* attach node $r$ to $r-2$ rather than to $r-1$ */
	 else if (i==r-2) c-=w[i+1]; /* detach $r-1$ from $r$ */
              else c+=w[i+2]; /* and attach $r-2$ to $r$ */
  break; case 'E': if (i<4) /* similar adjustments here */
    {   if (i<2) c+=w[i+2]-w[i+1];
           /* nodes 1,2 neighbour 3,4 respectively rather than 2,3 */
        else c+=w[i-2]; /* for nodes 3,4, add neighbour 1,2, respectively */
        if (i==1 || i==2) c-=w[i-1];
          /* and for nodes 2,3, remove neighbour 1,2, respectively */
     }
  break; case 'F': @+ if (i==2) c+=w[1]; /* like types $B_n$, $C_n$ */
  break; case 'G': @+ if (i==0) c+=2*w[1];
   /* now long root contributes 3 times */
  }
  w[i]=c;
}

@ The function~|w_refl|, which applies a simple reflection to a weight for an
arbitrary group, is called from several other compilation units in \LiE..
@c

void w_refl(entry* lambda, index wi)
{ if (type_of(grp)==SIMPGRP) simp_w_refl(lambda,wi,&grp->s);
  else if (simpgroup(grp)) simp_w_refl(lambda,wi,Liecomp(grp,0));
  else
  { index i,d,offset=0;
    for (i=0; wi>=(d=Liecomp(grp,i)->lierank); ++i) { offset+=d; wi-=d; }
    simp_w_refl(lambda+offset,wi,Liecomp(grp,i));
  }
}


@ For general Weyl group elements~|word|, |Waction(lambda,word)| applies
|word| to the weight vector~|lambda|.

@c
void Waction(entry* lambda, vector* word)
{ index i; entry* w=word->compon;
  for (i=0; i<word->ncomp; ++i) @+
    if (w[i]!=0) w_refl(lambda,w[i]-1);
}

@ Since there is no need to apply simple reflections on the root base
separately, there is no  analogue of |w_refl| for root coordinates.
Therefore the analogue |Wrtaction| of |Waction| expands the corresponding
code into the body of the loop over the Weyl word.

@c

void Wrtaction(entry* alpha, vector* word)
{ index i; entry* w=word->compon;
  for (i=0; i<word->ncomp; ++i) if(w[i]!=0)
    { index wi=w[i]-1;
      if (type_of(grp)==SIMPGRP) simp_rt_refl(alpha,wi,&grp->s);
      else if (simpgroup(grp)) simp_rt_refl(alpha,wi,Liecomp(grp,0));
      else
      { index j,d,offset=0;
	for (j=0; wi>=(d=Liecomp(grp,j)->lierank); ++j)
          {@; offset+=d; wi-=d; }
	simp_rt_refl(alpha+offset,wi,Liecomp(grp,j));
      }
    }
}

@ The following function is used frequently in \.{weylloop.w}.

@c
matrix* simp_Weylmat(vector* word, simpgrp* g)
{ index i,j,r=g->lierank; matrix* res=mkmatrix(r,r);
  entry** m=res->elm,* w=word->compon;
  for (i=0; i<r; ++i) 
  { for (j=0; j<r; ++j) m[i][j]= i==j;
    for (j=0; j<word->ncomp; ++j) @+
      if(w[j]!=0) simp_w_refl(m[i],w[j]-1,g);
  }
  return res;
}

@ To construct the matrices representing Weyl group elements, we simply apply
|Waction| of |Wrtaction| to the rows of the identity matrix.

@c
matrix* Weyl_mat(vector* word)
{ index i,j,r=Lierank(grp); matrix* res=mkmatrix(r,r); entry** m=res->elm;
  for (i=0; i<r; ++i) 
  { for (j=0; j<r; ++j) m[i][j]= i==j;
    Waction(m[i],word);
  }
  return res;
}

matrix* Weyl_rt_mat(vector* word)
{ index i,j,r=Ssrank(grp); matrix* res=mkmatrix(r,r); entry** m=res->elm;
  for (i=0; i<r; ++i)
  { for (j=0; j<r; ++j) m[i][j]= i==j;
    Wrtaction(m[i],word);
  }
  return res;
}

@* Orders of Weyl groups.
The function |simp_worder| basically computes the order of a simple group,
but since it is calculated by successive multiplications, and since for
composite groups the orders of the simple factors need to be multiplied
together, it is slightly more efficient to directly multiply the order
computed order into a number supplied as argument.

The order of the Weyl group of a simple group of rank~$n$ is always divisible
by~$n!$, so we can shorten this function a bit by only specifying the factor
to multiply $n!$ by.

@c
local bigint* simp_worder(bigint* num, simpgrp* g)
{ index i,r=g->lierank; bigint* result=num;

  i=r; @+ while (i>1) result=mul1(result,i--);
  switch (g->lietype)
  {	 case 'A': @+ result=mul1(result,r+1); @+ 
  break; case 'B':
	 case 'C': @+ for (i=0; i<r; ++i) result = mul1(result,2); @+
  break; case 'D': @+ for (i=1; i<r; ++i) result = mul1(result,2); @+
  break; case 'E': @+ result = mul1(result,r==6 ? 72 : r==7 ? 576 : 17280); @+
  break; case 'F': @+ result = mul1(result,48); @+
  break; case 'G': @+ result = mul1(result,6);
  }
  return result;
}

@ From there it is trivial to determine the order of general Weyl groups.

@c
bigint* Worder(object grp)
{ index i; bigint* result=copybigint(one,NULL);
  if (type_of(grp)==SIMPGRP) return simp_worder(result,&grp->s);
  for (i=0; i<grp->g.ncomp; ++i) result = simp_worder(result,Liecomp(grp,i));
  return result;
}

@ Now we compute the order of the Weyl subgroup generated by given subset of
simple reflections; this is achieved by applying |Worder| to the group
computed by~|Carttype| for the given subset of roots.

@c 
bigint* sub_Worder(vector* v)
{ index i,j,s=Ssrank(grp), n=v->ncomp; matrix* roots=mkmatrix(n,s);
  entry** m=roots->elm; group* h; bigint* result;
  if (n==0) {@; freemem(roots); return one; }
  for (i=0; i<n; ++i) /* select rows od an identity matrix */
  {@; entry* mij= *m++,vi=v->compon[i]-1;
    for (j=0; j<s; ++j) *(mij++)= (j==vi);
  }
  h=Carttype(roots); freemem(roots);
  result= Worder((object)h); freemem(h); return(result);
}

@* Storing Weyl group orbits.
We now use the functions above to generate Weyl group orbits into a matrix or
a polynomial; this can (and used to) be done using the dynamic traversal which
is implemented in \.{weylloop.w}, but since we are storing the result anyway,
there is not much advantage to that. Rather we have reimplemented the
functions using only the action of simple reflections, which allows us to
order the weights in the orbit in a more useful way.

@ These functions compute the size of the stabiliser of a \emph{dominant}
weight, which is relatively simple because we just have to inspect the
collection entries that are~|0|. The hard work is actually done by
|sub_Worder|, which analyses the type of the subgroup stabilising our weight,
and returns the corresponding order. That function (like most functions
written for possibly non-simple groups) receives the pertinent group
implicitly via the global variable~|grp|, but we do not wish to perturb the
value of |grp| by calling |simp_worbitsise| (since it will be called for |g|
traversing the simple components of~|grp|, see below); therefore we save and
restore it. This is one of the rare cases where |grp| ever gets to point to a
|simpgrp| object.

@c
local bigint* simp_stabsize(entry* v, simpgrp* g)
{ object sav_grp=grp; vector* I; bigint* result;
  index i,nz=0,r=g->lierank;
  for (i=0; i<r; ++i) @+ if (v[i]==0) nz++; /* count non-zero coordinates */
  if (nz==0) return one;
  if (nz==r) return simp_worder(copybigint(one,NULL),g);
  I=mkvector(nz); 
  for (i=0,nz=0; i<r; ++i) @+ if (v[i]==0) I->compon[nz++]=i+1;
  grp=(object)g; result=sub_Worder(I); grp=sav_grp;
  freemem(I); return result;
}

bigint* simp_worbitsize(entry* w, simpgrp* g)
 /* |w| is assumed to be dominant */
{@; return quotient(simp_worder(copybigint(one,NULL),g),simp_stabsize(w,g)); }

@ Now we can compute the size of the orbit of any weight vector~|w| for any
group~|grp|. Note that only the semisimple part of~|w| is used (nor does
|make_dominant| try to access any other entries).

@c
bigint* Orbitsize(entry* w)
{ index i,d,s=Ssrank(grp); entry* x=mkintarray(s),* y=x; bigint* result=one;
  copyrow(w,x,s); make_dominant(x);
  if (type_of(grp)==SIMPGRP) return simp_worbitsize(x,&grp->s);
  for (i=0; i<grp->g.ncomp; ++i,y+=d)
  @/{@; simpgrp* g=Liecomp(grp,i); d=g->lierank;
    result=mult(result,simp_worbitsize(y,g));
  }
  freearr(x); return result;
}

@ We shall generate Weyl group orbit vectors~|x| in lexicographic order of the
first reduced expression that transforms the dominant orbit element into~|x|.
Generating orbits so explicilty by reflections (as opposed to using
|Weylloop|, as it used to be done in \LiE.) it is worth while to optionally
record the structure of the orbit as it is generated. To this end we supply an
output parameter~|orbit_graph|, through which (if non-|NULL|) a matrix will be
returned whose |@[@][i][j]| entry gives the index in the~|result| of the image
$\\{result}[i]\cdot s_j$.

@c
matrix* Weyl_orbit(entry* v, matrix** orbit_graph)
{ index i,j,k,r=Lierank(grp),s=Ssrank(grp);
  matrix* result; entry** m;
  index level_start=0, level_end=1, cur=1;
  @< Allocate a matrix of appropriate size as |result|, and set its first
     row to the dominant orbit representative; also allocate |*orbit_graph| if
     required @>
  m=result->elm;
  while (level_start<level_end)
  { @< Generate new weights by applying simple reflections to weights in
       |result| in the range from |level_start| to |level_end| @>
    level_start=level_end; level_end=cur;
  }
  return result;
}

@ To predict the size of the orbit is straightforwared, using |Orbitsize|.
There is no need to initialise any entries of |*orbit_graph|, since they will
all be set during the generation of the orbit.

@< Allocate a matrix of appropriate size as |result|... @>=
{ entry* lambda=mkintarray(r); @+
  copyrow(v,lambda,r); make_dominant(lambda);
  result=mkmatrix(bigint2entry(Orbitsize(lambda)),r);
  copyrow(lambda,result->elm[0],r); freearr(lambda);
  if (orbit_graph!=NULL) *orbit_graph=mkmatrix(result->nrows,s);
}

@ For each reflection~$s_i$ we apply it in situ using |w_refl|, then copy the
row to the end if it is new, and restore the row by applying |w_refl| again.
There are two ways in which the image after reflection can be not new: it
can be not lower in the dominance order than the original (the sequence of
reflections applied is not reduced), or it can be equal to a weight previously
generated for another reduced expression, which must be found at the level
currently being generated.

@< Generate new weights... @>=
for (k=level_start; k<level_end; ++k)
  for (i=0; i<s; ++i)
    if (m[k][i]>0) /* only strictly cross walls, and from dominant side */
    { w_refl(m[k],i);
      for (j=level_end; j<cur; ++j)
      @+  if (eqrow(m[k],m[j],s)) break;
      if (orbit_graph!=NULL)
      {@; (*orbit_graph)->elm[k][i]=j; (*orbit_graph)->elm[j][i]=k; }
      if (j==cur)
	{@; assert(cur<result->nrows);
          copyrow(m[k],m[cur++],r);
        }
      w_refl(m[k],i);
    }
    else if (m[k][i]==0 && orbit_graph!=NULL) (*orbit_graph)->elm[k][i]=k;


@ For historic reasons we also provide the functionality of generating Weyl
group orbits in root coordinates. This used to be a simple variation when
orbits were actually generated using |Weylloop|, but when the more definite
order described above was specified for orbit elements, this would require a
somewhat less trivial modification of the above code, since it is more
difficult to see in root coordinates on which side of a wall one is. We prefer
therefore to instead to transform to and from weight coordinates; although
efficiency is not a real issue for a function that is probably hardly ever
used, this is not a bad way to proceed.

@c
matrix* Weyl_root_orbit(entry* v)
{ index i,j,r=Lierank(grp),s=Ssrank(grp);
  entry* x=mkintarray(r); matrix* orbit, *result; entry** m;
  index dc=Detcartan();
  mulvecmatelm(v,Cartan()->elm,x,s,r);
  orbit=Weyl_orbit(x,NULL);
@/result=mkmatrix(orbit->nrows,s); m=result->elm;
  mulmatmatelm(orbit->elm,Icartan()->elm,m,orbit->nrows,s,s);
  freemem(orbit);
  for (i=0; i<result->nrows; ++i) @+
    for (j=0; j<s; ++j) m[i][j]/=dc;
  return result;
}

@ For polynomials one does not really compute an orbit of the polynomial, but
the multiset of orbits of the multiset represented by the polynomial. After
all exponents are made dominant, the polynomial is reduced, so only one orbit
is generated with the proper multiplicity in case certain exponents were in
the same orbit.

@c
poly* Worbit_p(poly* p)
{ index i,k=0,l=0,r=p->ncols; poly* result; entry** res;
  p=copypoly(p); for (i=0; i<p->nrows; ++i) make_dominant(p->elm[i]);
  Reduce_pol(p);
  for (i=0; i<p->nrows; ++i)
    if ((l += bigint2entry(Orbitsize(p->elm[i])))<0)
      error ("That's too large an orbit");
  result=mkpoly(l,p->ncols); res=result->elm;
  for (i=0; i<p->nrows; ++i)
  { index j; matrix* orbit=Weyl_orbit(p->elm[i],NULL); entry** x=orbit->elm;
    for (j=0; j<orbit->nrows; ++j)
    @/{@; result->coef[k]=p->coef[i]; setshared(p->coef[i]);
      copyrow(*x++,res[k++],r);
    }
    freemem(orbit);
  }
  assert(k==result->nrows);
  return result; /* not sorted, but rows are unique */
}

@ And here is finally the corresponding thing using the alternating Weyl group
action; this is what implements the user function |alt_W_sum|.

@c
poly* alt_Wsum(poly* p)
{ index i,k=0,r=p->ncols; poly* result; entry** res,*rho=mkintarray(r);
  p=Alt_dom(p); for (i=0; i<r; ++i) rho[i]=1;
  for (i=0; i<p->nrows; ++i) add_xrow_to(p->elm[i],1,rho,r);
  result=mkpoly(p->nrows*bigint2entry(Worder(grp)),r); res=result->elm;
  for (i=0; i<p->nrows; ++i)
  { index j,l; matrix* orbit=Weyl_orbit(p->elm[i],NULL); entry** x=orbit->elm;
    bigint* c=p->coef[i],* min_c=sub(null,c);
    for (j=0; j<orbit->nrows; ++j)
    { subrow(*x,rho,res[k],r); l=make_dominant(*x++)%2;
      result->coef[k]= l ? min_c : c; setshared(result->coef[k]); ++k;
    }
    freemem(orbit);
  }
  freearr(rho);
  assert(k==result->nrows);
  return result; /* not sorted, but rows are unique */
}

@* Index.