\input LiEwebmac


@* Traversing Weyl orbits.
This file defines routines for traversing orbits under the Weyl group of a
simple group; it is one of the algorithmically more sophisticated parts
of~\LiE..

@h "lie.h"
@d local static
@f local static

@c
@< Local variables @>@;
@< Local functions @>@;

@ The external interface is determined by the header file \.{weylloop.h}.
@( weylloop.h @>=
@< Prototypes of globally visible functions @>

@ Three functions are defined: |Weylloopinit|, which should be called prior to
the orbit traversal to specify the group whose Weyl group is acting,
|Weylloop|, which performs the actual orbit traversal of a given vector, and
|Weylloopexit| which must be called afterwards to clear the settings installed
by |Weylloopinit|. Only one orbit can be traversed at the same time, and only
for a simple group (the mathematical routines that invoke Weyl orbit traversal
handle the more general cases fairly easily while respecting this
restriction). It is possible however the call |Weylloop| several times between
|Weylloopinit| and |Weylloopexit|, to generate orbits for different vectors.
Apart from the vector~|v| whose orbit is to be traversed, another
argument~|action| must be given to |Weylloop|, which is a function pointer
specifying the action be performed at each point of the orbit.

@< Prototypes of globally visible functions @>=
void Weylloopinit(simpgrp* g);
void Weylloop(void (*action)(entry*),entry* v);
void Weylloopexit(void);

@ Many quantities that remain constant during each orbit traversal are set
up by~|Weylloopinit|, and it also allocates some objects of dynamically
determined size for temporary use, so that no such allocations are needed
during the traversal itself. These values and objects are accessed through
static variables of file scope, since in~\Cee\ this is the easiest way to
make them available from within |Weylloop| and auxiliary functions.

Here we define the most essential of these values. Orbits are generated in
terms of weight coordinates. The simple group whose orbits are being traversed
is stored in~|the_g|, and its rank is also stored separately in~|rnk|. For
efficient generation of Weyl group orbits, we shall transform coordinates to a
basis on which the Weyl group, or a substantial subgroup thereof, acts by
permutation of the coordinates and possibly sign changes of individual
coordinates; we shall call this basis the $\eps$-basis. The number of elements
of the $\eps$-basis may exceed |rnk| (although never by more than~|1|), so we
distinguish the space spanned by the $\eps$-basis from the weight space; we
call it the $\eps$-space, and store its dimension in a variable~|eps_dim|. The
linear transformations from weight coordinates to these $\eps$-coordinates,
and back for the generated orbit vectors, are realised by functions accessed
via the pointers |to_e| and |from_e| respectively, which are set in accordance
with |the_g| by~|Weylloopinit|. After transformation, a large part of the
orbit is generated by permutation of $\eps$-coordinates. Usually all |eps_dim|
coordinates are permuted, but in one case (|E6|) only a number of initial
$\eps$-coordinates is permuted, so we reserve a separate variable~|perm_size|
to store this number in. Whether in addition to permutations of entries sign
changes are also used, depends on whether the classical subgroup of the Weyl
group that is being used has type $A$,~$B$, $C$, or~$D$; this is recorded in
the variable~|subtype|.

@< Local variables @>=
local simpgrp* the_g;  /* the simple group in question */
local index rnk;     /* Lie rank of full group */
local index eps_dim; /* dimension of $\eps$-space */
local index perm_size;	/* number of entries to be permuted */
local char subtype;  /* type of classical subgroup used for suborbits */
@)
typedef void (*trans)(entry const* from, entry* to);
local trans to_e, from_e;
  /* group dependent transformations to and from $\eps$-coordinates */

@* Base changes.
We shall first consider the transformation functions that may be assigned to
|to_e| and |from_e|. The elements of the $\eps$-basis will be denoted by
$\eps_1$, $\eps_2$, \dots~$\eps_d$; as usual the fundamental weights will be
denoted by $\omega_1$, \dots~$\omega_n$, and the fundamental roots by
$\alpha_1$, \dots~$\alpha_n$. Here $n$ is the value of~|rnk|, and $d$ that
of~|eps_dim|; in case we have $d>n$, we shall view the weight space
as a subspace or quotient space of the $\eps$-space. The Weyl-group invariant
inner product on the weight space is then extended to the $\eps$-space, and
the $\eps$-basis consists of orthogonal vectors of equal length.

Originally, \LiE. was set up to generate orbits either in weight or in root
coordinates. Later we realised that the capability to use root coordinates was
never used, and the code was correspondingly simplified. For documentation
purposes however we have retained the description below of the transformations
applicable to root coordinates.

The definition of the $\eps$-basis depends on the (simple) type of group
considered, and possibly on whether the orbits are to be generated in weight
or in root coordinates. The basic cases are $A_n$~and~$C_n$. In case~$A_n$
we take $d=n+1$, and consider only permutations of the coordinates. We view
the weight space as a quotient of the $\eps$-space by the line
$\<\eps_1+\eps_2+\cdots+\eps_d>$, such that $\omega_i$ is represented by
$\eps_1+\cdots+\eps_i$ for $1\leq i\leq n$. This shows directly how to
convert from weight coordinates to $\eps$-coordinates; conversely if a
vector~$v$ has $\eps$-coordinates~$x_i$, i.e., if $v=\sum_{i=1}^d x_i\eps_i$,
then its $i$-th weight coordinate is $x_i-x_{i+1}$ for $1\leq i\leq n$. The
$\eps$-coordinates are also called partition coordinates in this case, since
dominant weights correspond to partitions. When orbits are to be generated
in root coordinates, one could of course use the same scheme, but
transformation matrices will be simpler if we use a slightly different
setup. We still take $d=n+1$, but now view the space spanned by the root
vectors as subspace of the $\eps$-space, namely the one given by
$\sum_{i=1}^d x_i=0$ (where again the $x_i$ are $\eps$-coordinates). This
means we shall use a conversion from root coordinates to $\eps$-coordinates
whose image is confined to this subspace (which is not the case for the
conversion from weight coordinates), and then for the conversion back we
need not consider the behaviour outside this subspace, in particular, the
kernel need (and will) not be $\<\eps_1+\eps_2+\cdots+\eps_d>$. The former
conversion is determined by $\alpha_i=\eps_i-\eps_{i+1}$ for $1\leq i\leq
n$, the latter by taking for the $i$-th root coordinate $\sum_{j=1}^i x_j$.

In the case $C_n$ we take $d=n$, and consider both permutations of the
coordinates and sign changes; since the $\eps$-space coincides with the
weight space, weight and root coordinates are now treated on equal footing.
Like for the $A_n$ case we have $\omega_i=\eps_1+\cdots+\eps_i$ and for
$1\leq i<n$ we have $\alpha_i=\eps_i-\eps_{i+1}$; the final root however is
$\alpha_n=2\eps_n$. The transformations back are given by the same formulae
as in the $A_n$ case, except that the $n$-th weight coordinate equals~$x_n$,
and the $n$-th root coordinate ${1\over2}\sum_{j=1}^n x_j$. The cases $B_n$
and~$D_n$ are similar, but for generating orbits in weight coordinates the
$\eps$-basis has to be scaled differently, in order to avoid half-integer
$\eps$-coordinates; precise transitions matrices will be given below.

For the exceptional groups no simple choice of $\eps$-basis will allow all
Weyl group elements to act by permutations of coordinates and sign changes;
however, this can be achieved for a substantial subgroup, so that only a
limited set of cosets has to be traversed by other means. To this end a
suitable root subsystem (of classical type) is chosen, and weights are
first transformed by restriction to these weight coordinates, which are then
further transformed to the corresponding $\eps$-coordinates. The following
root subsystems are chosen, where $R$ stands for minus the highest root:
$$ \tabskip=1pc plus 1fil \def\\#1{\alpha_{#1}}
\halign{\hfil$#$\hfil&$#$\hfil&\hfil$#$\hfil&\hfil$#$\hfil\cr
\omit\it exceptional type\hfil & \omit\it basis of root subsystem\hfil &
\omit\it type of subsystem\hfil & \omit\it index of subgroup in~$W$ \cr
\noalign{\smallskip}
E_6 & \\6,\\5,\\4,\\3,\\2	    & D_5T_1 & 27 \cr
E_7 & R,\\1,\\3,\\4,\\5,\\6,\\7	    & A_7    & 72 \cr
E_8 & R,\\8,\\7,\\6,\\5,\\4,\\3,\\2 & D_8    & 135 \cr
F_4 & R,\\1,\\2,\\3		    & B_4    & 3 \cr
G_2 & R,\\2			    & A_2    & 2 \cr
}
$$
The restriction matrices transforming weights from the exceptional type
system to the classical type subsystem can easily be found using the \LiE.
function~|res_mat|. Since these bases are not fundamental in the sense of the
\LiE. function |fundam|, negative entries enter into the restriction
matrices; they have however been chosen to make the combined transition from
weights in the exceptional system to $\eps$-coordinates simple and easily
computable. For the transitions from and to root coordinates the same
intermediate subsystems are used, with further transitions to
$\eps$-coordinates as for that classical subsystem; only the vector $\eps_6$
for |E6|, which spans the weights for the |T1| factor of |D5 T1|, is scaled
differently than in the weight case, for convenience.

@ Here are the explicit transformation matrices between weight coordinates
and $\eps$-coordinates, for the classical types. The indicated factor is
applied in the transition $\eps\to\omega$, to avoid fractional matrix entries.
$$
\let\vd=\vdots \let\dd=\ddots \let\cd=\cdots
\def\ld{\smash{\lower1ex\hbox{$\vdots$}}}
\tabskip=1em plus 1fil
\halign to \hsize
{&\hfil$\displaystyle#$\hfil & \hfil$\pmatrix#$\hfil \cr
\omit\it group\hfil &
\omit \hfil\it transformation $\omega\to\eps$\hfil &
\omit \it factor\hfil & 
\omit \hfil\it transformation $\eps\to\omega$\hfil &
\omit \it group acting\hfil\cr
\noalign{\nobreak}
%
A_n &
{ 1 &  0  &	& \cd &	    &	0 \cr
  1 &  1  &  0	&     &	    & \ld \cr
  1 &  1  &  1	&  0		  \cr
\vd &	  &	& \dd &	 0	  \cr
  1 &	  & \cd &     &	 1  &	0 \cr }&
1 &
{ 1  &	0  &	 & \cd &  0  \cr
 -1  &	1  &  0	 &     & \ld \cr
  0  & -1  &  1	 &  0	     \cr
     &	0  & \dd & \dd &  0  \cr
 \vd &	   &  0	 & -1  &  1  \cr
  0  & \cd &	 &  0  & -1  \cr }&
W(A_n) \cr
\noalign{\medbreak}
%
B_n &
{ 2  &	0  &	 & \cd &  0  \cr
  2  &	2  &  0	 &     & \ld \cr
 \vd &	   & \dd &  0  \cr
  2  & \cd &  2	 &  2  &  0  \cr
  1  &	1  & \cd &  1  &  1  \cr }&
1 \over2 &
{ 1  &	0  &	 & \cd &  0  \cr
 -1  &	1  &  0	 &     & \ld \cr
  0  & \dd & \dd &  0	     \cr
 \vd &	0  & -1	 &  1  &  0  \cr
  0  & \cd &  0	 & -1  &  2  \cr }&
W(B_n)=W(C_n) \cr
\noalign{\medbreak}
%
C_n &
{ 1  &	0  &	 & \cd &  0  \cr
  1  &	1  &  0	 &     & \ld \cr
 \vd &	   & \dd &  0	     \cr
  1  & \cd &  1	 &  1  &  0  \cr
  1  &	1  & \cd &  1  &  1  \cr }&
1 &
{ 1  &	0  &	 & \cd &  0  \cr
 -1  &	1  &  0	 &     & \ld \cr
  0  & \dd & \dd &  0	     \cr
 \vd &	0  & -1	 &  1  &  0  \cr
  0  & \cd &  0	 & -1  &  1  \cr }&
W(B_n)=W(C_n) \cr
\noalign{\medbreak}
%
D_n &
{ 2  &	0  &	 & \cd &     &	0  \cr
  2  &	2  &  0	 &     &     & \ld \cr
 \vd &	   & \dd &  0		   \cr
  2  &	2  & \cd &  2  &  0  &	0  \cr
  1  &	1  & \cd &  1  &  1  & -1  \cr
  1  &	1  & \cd &  1  &  1  &	1  \cr }&
1\over	2  &
{ 1  &	0  &	 & \cd &     &	0  \cr
 -1  &	1  &  0	 &     &     & \ld \cr
  0  & \dd & \dd &  0		   \cr
 \ld &	0  & -1	 &  1  &  0  &	0  \cr
     &	   &  0	 & -1  &  1  &	1  \cr
  0  & \cd &	 &  0  & -1  &	1  \cr }&
W(D_n) \cr
}
$$
The following functions implement these matrices; the rank is
obtained from the global variable~|rnk|.

@< Local functions @>=

local void w2eAn(entry const* w,entry* e)
{@; index i=rnk; entry sum=0; while (e[i]=sum,--i>=0) sum+=w[i]; }

local void e2wAn(entry const* e,entry* w)
{@; index i; for (i=0; i<rnk; ++i) w[i]=e[i]-e[i+1]; }

local void w2eBn(entry const* w,entry* e)
{@; index i=rnk-1; entry sum=w[i]; while (e[i]=sum,--i>=0) sum+=2*w[i]; }

local void e2wBn(entry const* e,entry* w)
{@; index i; for (i=0; i<rnk-1; ++i) w[i]=(e[i]-e[i+1])/2; w[i]=e[i]; }

local void w2eCn(entry const* w,entry* e)
{@; index i=rnk-1; entry sum=w[i]; while (e[i]=sum,--i>=0) sum+=w[i]; }

local void e2wCn(entry const* e,entry* w)
{@; index i; for (i=0; i<rnk-1; ++i) w[i]=e[i]-e[i+1]; w[i]=e[i]; }

local void w2eDn(entry const* w,entry* e)
{@; index i=rnk-1; entry sum=w[i]-w[i-1]; while(e[i]=sum,--i>=0) sum+=2*w[i]; }

local void e2wDn(entry const* e,entry* w)
{@; index i; for (i=0; i<rnk-1; ++i) w[i]=(e[i]-e[i+1])/2; w[i]=w[i-1]+e[i]; }

@ Next we consider the types |E6|, |E7|, and~|E8|. The matrices
$\omega\to\eps$ below can be obtained by computing |res_mat| for the root
subsystem indicated above, and multiplying the result by the matrix
$\omega\to\eps$ for the (classical) type of that subsystem. For |E6|, the
subsystem has type |D5 T1|, so the matrix to multiply by in that case is a
$6\times6$ matrix extending the matrix given for type |D5| with entries of
the identity matrix.
$$\tabskip=2pt plus 1fil
\def~{\llap{$-$}}
\halign to \hsize
{&\hfil$\displaystyle#$\hfil & \hfil$\pmatrix#$\hfil \cr
\omit\it group\hfil &
\omit \hfil\it transformation $\omega\to\eps$\hfil &
\omit \it factor\hfil & 
\omit \hfil\it transformation $\eps\to\omega$\hfil &
\omit \it group acting\hfil\cr
\noalign{\nobreak}
%
E_6 &
{ 0  &	 0  &	0  &   0  &   0	 &  ~4	\cr
  1  &	 1  &	1  &   1  &   1	 &  ~3	\cr
  1  &	 1  &	1  &   1  &  ~1	 &  ~5	\cr
  2  &	 2  &	2  &   0  &   0	 &  ~6	\cr
  2  &	 2  &	0  &   0  &   0	 &  ~4	\cr
  2  &	 0  &	0  &   0  &   0	 &  ~2	\cr }&
1\over 4 &
{&  ~1	&   0  &   0  &	  0  &	 0  &	2  \cr
 &  ~1	&   0  &   0  &	  0  &	 2  &  ~2  \cr
 &  ~1	&   0  &   0  &	  2  &	~2  &	0  \cr
 &  ~1	&   2  &   2  &	 ~2  &	 0  &	0  \cr
 &   1	&   2  &  ~2  &	  0  &	 0  &	0  \cr
 &  ~1	&   0  &   0  &	  0  &	 0  &	0  \cr }&
W(D_5) \cr
\noalign{\medbreak}
%
E_7 &
{&  ~1	&   1  &   0  &	  0  &	 0  &	0  &   0  &   0	 \cr
 &  ~2	&   0  &   0  &	  0  &	 0  &	0  &   0  &   0	 \cr
 &  ~2	&   1  &   1  &	  0  &	 0  &	0  &   0  &   0	 \cr
 &  ~3	&   1  &   1  &	  1  &	 0  &	0  &   0  &   0	 \cr
 &  ~2	&   1  &   1  &	  1  &	 1  &	0  &   0  &   0	 \cr
 &  ~1	&   1  &   1  &	  1  &	 1  &	1  &   0  &   0	 \cr
 &   0	&   1  &   1  &	  1  &	 1  &	1  &   1  &   0	 \cr }&
1\over 2 &
{&   0	&  ~1  &   0  &	  0  &	 0  &	0  &   0  \cr
 &   2	&  ~1  &   0  &	  0  &	 0  &	0  &   0  \cr
 &  ~2	&  ~1  &   2  &	  0  &	 0  &	0  &   0  \cr
 &   0	&  ~1  &  ~2  &	  2  &	 0  &	0  &   0  \cr
 &   0	&   1  &   0  &	 ~2  &	 2  &	0  &   0  \cr
 &   0	&   1  &   0  &	  0  &	~2  &	2  &   0  \cr
 &   0	&   1  &   0  &	  0  &	 0  &  ~2  &   2  \cr
 &   0	&   1  &   0  &	  0  &	 0  &	0  &  ~2  \cr }&
W(A_7) \cr
\noalign{\medbreak}
%
E_8 &
{&  ~4	&   0  &   0  &	  0  &	 0  &	0  &   0  &   0	 \cr
 &  ~5	&   1  &   1  &	  1  &	 1  &	1  &   1  &   1	 \cr
 &  ~7	&   1  &   1  &	  1  &	 1  &	1  &   1  &  ~1	 \cr
 &  ~10 &   2  &   2  &	  2  &	 2  &	2  &   0  &   0	 \cr
 &  ~8	&   2  &   2  &	  2  &	 2  &	0  &   0  &   0	 \cr
 &  ~6	&   2  &   2  &	  2  &	 0  &	0  &   0  &   0	 \cr
 &  ~4	&   2  &   2  &	  0  &	 0  &	0  &   0  &   0	 \cr
 &  ~2	&   2  &   0  &	  0  &	 0  &	0  &   0  &   0	 \cr }&
1\over 4 &
{&  ~1	&   0  &   0  &	  0  &	 0  &	0  &   0  &   0	 \cr
 &  ~1	&   0  &   0  &	  0  &	 0  &	0  &   0  &   2	 \cr
 &  ~1	&   0  &   0  &	  0  &	 0  &	0  &   2  &  ~2	 \cr
 &  ~1	&   0  &   0  &	  0  &	 0  &	2  &  ~2  &   0	 \cr
 &  ~1	&   0  &   0  &	  0  &	 2  &  ~2  &   0  &   0	 \cr
 &  ~1	&   0  &   0  &	  2  &	~2  &	0  &   0  &   0	 \cr
 &  ~1	&   2  &   2  &	 ~2  &	 0  &	0  &   0  &   0	 \cr
 &   1	&   2  &  ~2  &	  0  &	 0  &	0  &   0  &   0	 \cr }&
W(D_8) \cr
}$$
The function below implement multiplication by these matrices fairly
directly, employing some obvious regularities. The cases |E6| and |E8| are
sufficiently similar to be handled by common functions |w2eE68|
and~|e2wE68|; for |E7| there is a slight variation, and separate functions
|w2eE7| and |e2wE7| are used.

@< Local functions @>=

local void w2eE68(entry const* w,entry* e)
{ index i, u=perm_size;
  entry sum =e[u-1]= w[1]-w[2], sumsum=4*w[0] -sum;
  for (i=2; i<rnk; ++i) sumsum+=(e[u-i]=sum+= 2*w[i]);
  e[rnk==6 ? 5 : 0]= -sumsum;
}

local void e2wE68(entry const* e,entry* w)
{ index i,u= perm_size; entry sum=e[rnk==6 ? 5 : 0];
  for (i=rnk-1; i>=2; --i) {@;sum+=e[u-i]; w[i]=(e[u-i]-e[u-i+1])/2;}
  w[1]=w[2]+e[u-1]; w[0]=(e[u-1]-sum)/4;
}

local void w2eE7(entry const* w,entry* e)
{ index i; entry sum=e[7]=0; @+ for (i=6; i>=2; --i) e[i]=sum+=w[i];
  e[1]=sum+w[0];
  e[0]=e[6]+e[5]+e[4]-e[3]-e[2]-e[1]-2*w[1];
}

local void e2wE7(entry const* e,entry* w)
{ index i; @+ w[0]=e[1]-e[2]; @+ for(i=2; i<7; i++) w[i]=e[i]-e[i+1];
  w[1]=(e[7]+e[6]+e[5]+e[4]-e[3]-e[2]-e[1]-e[0])/2;
}

@ Here are the matrices for the cases |F4| and |G2|.
$$\tabskip=2pt plus 1fil
\def~{\llap{$-$}}
\halign to \hsize
{&\hfil$\displaystyle#$\hfil & \hfil$\pmatrix#$\hfil \cr
\omit\it group\hfil &
\omit \hfil\it transformation $\omega\to\eps$\hfil &
\omit \it factor\hfil & 
\omit \hfil\it transformation $\eps\to\omega$\hfil &
\omit \it group acting\hfil\cr
\noalign{\nobreak}
%
F_4 &
{& ~2 & 2 & 0 & 0 \cr
 & ~4 & 2 & 2 & 0 \cr
 & ~3 & 1 & 1 & 1 \cr
 & ~2 & 0 & 0 & 0 \cr }&
1\over 2 &
{&  0 &	 0 & 0 & ~1 \cr
 &  1 &	 0 & 0 & ~1 \cr
 & ~1 &	 1 & 0 & ~1 \cr
 &  0 & ~1 & 2 & ~1 \cr }&
W(B_4) \cr
\noalign{\medbreak}
%
G_2 &
{ -1 & 0 & 0 \cr -1 & 1 & 0\cr } &
1 &
{& ~1 & 0 \cr & ~1 & 1 \cr & 2 & ~1 \cr }&
W(A_2) \cr
}$$
It is simplest to implement these small matrices directly, without any loops.

@< Local functions @>=

local void w2eF4(entry const* w,entry* e)
{@; e[1]=(e[2]=(e[3]=w[2])+2*w[1])+2*w[0]; e[0]=-2*w[3]-e[1]-e[2]-e[3]; }

local void e2wF4(entry const* e,entry* w)
{@; w[2]=e[3]; w[1]=(e[2]-e[3])/2; w[0]=(e[1]-e[2])/2;
  w[3]=-(e[0]+e[1]+e[2]+e[3])/2;
}

local void w2eG2(entry const* w,entry* e) @+
{@; e[2]=0; e[0]=-w[0]-(e[1]=w[1]); }

local void e2wG2(entry const* e,entry* w) @+
{@; w[0]=e[2]-(w[1]=e[1]-e[2])-e[0]; }

@ The transition from the $\eps$-basis to the coroot basis is given up to a
scalar by the transpose of the transition matrix from the weight basis to the
$\eps$-basis. In the cases where the $\eps$-space coincides with the weight
space, this is true because the $\eps$-basis is a scalar multiple of its
dual basis for the inner product, and the coroot basis is the dual basis of
the weight basis: the transition $\omega\to\eps$ is an isometry, so its
inverse is given by the transpose matrix, with respect to dual bases.
For type~$A_n$ this argument needs some modification, but the
transformations we explicitly gave are clearly transposes of each other.
The case of |E6| is also slightly special, as the $\eps$-space coincides with
the weight space for |D5 T1|, and we cannot use the degenerate inner
product of that weight space. However, if we use the inner product inherited
from the weight space of~|E6|, then the $\eps$-basis will be orthogonal,
and only $\eps_6$ has a length differing from the others; since $\eps_6$ is
invariant under $W(D_5)$, we can still use the transpose of the
$\omega\to\eps$ matrix, provided the transition from coroot basis to
$\eps$-basis is scaled correspondingly in the $\eps_6$-direction. It is easy
to check that in this case, as well as in the others, proper results are
obtained if the transition matrix from coroot basis to $\eps$-basis is
taken to be the transpose of the $\eps\to\omega$~matrix.

Of course we are interested in the transition to and from the root basis,
not the coroot basis, so for types $B_n$, $C_n$, |F4| and~|G2| scaling of
some rows and columns is needed on the transposed matrices. Since the root
system of $B_n$ can be taken to be the coroot system of~$C_n$, and vice
versa, this rescaling results for these two cases in the transpose of the
$\omega\to\eps$ transition matrix of the \emph{opposite} case. For~|F4| the
same rescaling technique is used resulting in the matrices given below.
For~|G2| however this technique would require introduction of a rational
factor $1\over3$; we prefer an {\it ad hoc\/} method that avoids this.	An
embedding of the root lattice in the $\eps$-space is chosen such that it
coincides with the intersection of the integer lattice spanned by
$\{\eps_1,\eps_2,\eps_3\}$ and the plane $\eps_1+\eps_2+\eps_3=0$, and such
that permutation of the $\eps_i$ induces a subgroup of index~$2$ of the Weyl
group (actually a different subgroup than used for the weight coordinates,
since it contains the reflection of~$\alpha_1$ rather than that
of~$\alpha_2$, but this does not bother us). For the matrix $\eps\to\alpha$
we choose some right-inverse of the $\alpha\to\eps$ matrix, with integer
entries.
$$
\let\vd=\vdots \let\dd=\ddots \let\cd=\cdots
\tabskip=1em plus 1fil
\def~{\llap{$-$}}
\halign to \hsize
{&\hfil$\displaystyle#$\hfil & \hfil$\pmatrix#$\hfil \cr
\omit\it group\hfil &
\omit \hfil\it transformation $\alpha\to\eps$\hfil &
\omit \it factor\hfil & 
\omit \hfil\it transformation $\eps\to\alpha$\hfil &
\omit \it group acting\hfil\cr
\noalign{\nobreak}
%
F_4 &
{&  0 &	 2 & ~2 &  0 \cr
 &  0 &	 0 &  2 & ~2 \cr
 &  0 &	 0 &  0 &  2 \cr
 & ~1 & ~1 & ~1 & ~1 \cr }&
1\over 2 &
{& ~1 & ~2 & ~3 & ~2 \cr
 &  1 &	 1 &  1 &  0 \cr
 &  0 &	 1 &  1 &  0 \cr
 &  0 &	 0 &  1 &  0 \cr }&
W(B_4)=W(C_4) \cr
\noalign{\medbreak}
%
G_2 &
{& 0 & 1 & ~1 \cr & ~1 & ~1 & 2 \cr }&
1 &
{& ~1 & ~1 \cr & 1 & 0 \cr & 0 & 0 \cr }&
W(A_2) \cr
}
$$

@ Here are the (now removed) conversions to and from root coordinates, for
classical types.

@< Local functions @>=

#if 0
local void r2eAn (entry const* r,entry* e)
{@; index i; e[0]=r[0]; for (i=1;i<rnk;++i) e[i]=r[i]-r[i-1]; e[i]=-r[i-1]; }

local void e2rAn (entry const* e,entry* r)
{@; index i; entry sum=r[0]=e[0]; for (i=1;i<rnk;++i) r[i]=sum+=e[i]; }

local void r2eBn (entry const* r,entry* e)
{@; index i; e[0]=r[0]; for (i=1;i<rnk;++i) e[i]=r[i]-r[i-1]; }

@)
#define @! e2rBn e2rAn /* these are identical */

local void r2eCn (entry const* r,entry* e)
{@; index i; e[0]=r[0]; for (i=1;i<rnk;++i) e[i]=r[i]-r[i-1];
  e[i-1]+=r[i-1];
}

local void e2rCn (entry const* e,entry* r)
{@; index i; entry sum=r[0]=e[0]; for (i=1;i<rnk;++i) r[i]=sum+=e[i];
  r[i-1]/=2;
}

local void r2eDn (entry const* r,entry* e)
{@; index i; e[0]=r[0]; for (i=1;i<rnk;++i) e[i]=r[i]-r[i-1]; e[i-2]+=r[i-1]; }

local void e2rDn (entry const* e,entry* r)
{ index i; entry sum=r[0]=e[0]; @+ for (i=1;i<rnk;++i) r[i]=sum+=e[i];
  r[i-2]-=r[i-1]/=2;
}
#endif

@ Of the types $E_n$ we treat |E7| separately, like for the weight case.
Here it seems best to write out that case completely, without using any loops
at all.

@< Local functions @>=

#if 0
local void r2eE68 (entry const* r,entry* e)
{ index i=rnk-1,u=perm_size; entry a=r[0]; @+ e[rnk==6 ? 5 : 0]=-a; @/
  e[u-i]=2*r[i]-a; @+ while (--i>1) e[u-i]=2*(r[i]-r[i+1])-a;
  e[u-1]=2*(r[1]-r[2])+a; e[u-2]+=2*r[1];
}

local void e2rE68 (entry const* e,entry* r)
{ index i,u=perm_size; entry sum=0,a=e[rnk==6 ? 5 : 0]; @+ r[0]=-a;
  for (i=rnk-1; i>2; --i) r[i]=(sum+=e[u-i]-a)/2;
  r[1]=(sum+e[u-2]+e[u-1])/4; r[2]=r[1]-(a+e[u-1])/2;
}

local void r2eE7 (entry const* r,entry* e)
{ entry a=r[1]; @+
  e[0]=-a; e[1]=2*r[0]-a; e[2]=2*(r[2]-r[0])-a; e[3]=2*(r[3]-r[2])-a;
  e[4]=a+2*(r[4]-r[3]); e[5]=a+2*(r[5]-r[4]); e[6]=a+2*(r[6]-r[5]);
  e[7]=a-2*r[6];
}

local void e2rE7 (entry const* e,entry* r)
{ entry a=e[0]; @+ r[1]=-a;
  r[3]=(r[2]=(r[0]=(e[1]-a)/2)+(e[2]-a)/2)+(e[3]-a)/2;
  r[6]=(r[5]=(r[4]=r[3]+(e[4]+a)/2)+(e[5]+a)/2)+(e[6]+a)/2;
}
#endif

@ For |F4| and |G2| we also write out the matrix multiplications
explicitly, without loops.

@< Local functions @>=

#if 0
local void r2eF4 (entry const* r,entry* e)
{@; e[0]=-r[3]; e[1]=2*r[0]-r[3]; e[2]=2*(r[1]-r[0])-r[3];
  e[3]=2*(r[2]-r[1])-r[3];
}

local void e2rF4 (entry const* e,entry* r)
{@; r[3]=-e[0];
  r[2]=(r[1]=(r[0]=(e[1]-e[0])/2)+(e[2]-e[0])/2)+(e[3]-e[0])/2;
}

local void r2eG2 (entry const* r,entry* e) @+
{@; e[1]=r[0]+(e[0]=-r[1]); e[2]=-e[0]-e[1]; }

local void e2rG2 (entry const* e,entry* r) @+
{@; r[0]=(r[1]=-e[0])+e[1]; }
#endif

@ We close the section on base changes by the piece of code in
|Weylloopinit| that sets the proper value for the local variables that
depend directly on |the_g|, including |to_e| and~|from_e|.

@< Set |perm_size|,... @>=
{ int i;
  static trans tab [8][2]=
  { { w2eAn, e2wAn }
  , { w2eBn, e2wBn }
  , { w2eCn, e2wCn }
  , { w2eDn, e2wDn }
  , { w2eE68,e2wE68 }
  , { w2eF4, e2wF4 }
  , { w2eG2, e2wG2 }
  , { w2eE7, e2wE7 }
  };

  if (the_g->lietype=='E' && rnk==7) i=7;
  else @< Set |i| to offset of |the_g->lietype| from |'A'| in alphabet @>
  subtype= "ABCDDBAA"[i];
  eps_dim= subtype=='A' ? rnk+1 : rnk;
  perm_size = the_g->lietype=='E' && rnk==6 ? eps_dim-1 : eps_dim;
@/to_e=tab[i][0]; from_e=tab[i][1];
}

@ To map |'A'| to~|0|, |'B'| to~|1|, \dots, |'G'| to~|6|, the simplest
method is to subtract |'A'| from the character code. However, this is not
guaranteed to work for all character sets, since alphabetic characters need
not be contiguous (even though this particular subsequence happens to be
contiguous even in \.{EBCDIC}), so we provide an alternative that is more
cumbersome but which always works.  The preprocessor will select which code
to use, where we assume that at least alphabetic characters are not
reordered, so that contiguity of the range |'A'|--|'G'| can be checked by
computing |'G'-'A'|.  It does not seem worth while to make the code proof
against hypothetical perverse character sets by extending the condition
after \&{\#if} to a completely rigorous one (after all, we are not even
really sure that |'A'!='\0'|).

@< Set |i| to offset of |the_g->lietype|... @>=

#if 'G'-'A'==6
  i=the_g->lietype-'A';
#else
  {@; static char s[]="ABCDEFG"; i= (int)(strchr(s,the_g->lietype)-s);	}
#endif

@* Auxiliaries for generating classical orbits. We consider next how to
actually generate orbits for the Weyl (sub)groups of classical type, after
converion to a suitable basis. We must start each such orbit in a standard
point: this is because of the way permutations of the entries will be
generated, and to allow comparison of different suborbits in the case of
exceptional groups. The latter use also requires that in cases where the
weight space is viewed as a quotient of the $\eps$-space, vectors with the
same image modulo $\<\eps_1+\cdots+\eps_d>$ are made equal. The function
|normalform| changes the vector~|w|, which is expressed on the $\eps$-basis,
to the canonical representative of its orbit. In all cases this canonical
representative has weakly increasing entries. For |subtype=='A'|, a multiple
of $\eps_1+\cdots+\eps_d$ is moreover added to make the initial entry~$0$,
making a unique representative for the image in the quotient space. For
|subtype!='A'| sign changes may also be applied to the entries, albeit only in
pairs if |subtype=='D'|. This means that if either |subtype!='D'| or there is
at least one entry~|0|, then all entries can be made non-negative; in the
remaining case we may be forced to retain one negative entry, in which case we
choose for this an entry with minimal absolute value. All cases with
|subtype!='A'| are handled at once, by first stripping off all minus signs,
while counting them, then sorting the entries into increasing order, and
finally, if |subtype=='D'| and there was an odd number of minus signs
originally, reattaching this sign to the first entry (it might be~|0|, in
which case nothing happens).

@d is_odd(x) ((x&1)!=0)
@c
local void normalform(entry* w)
{ index i, parity=0;
  if (subtype=='A')
  { @< Sort the entries of |w| into weakly increasing order @>
    if (w[0]!=0)
      /* then add constant to entries to make |w[0]==0| */
      for(i=perm_size-1; i>=0; --i) w[i]-=w[0];
  }
  else
  { for(i=0; i<perm_size; ++i) @+
      if (w[i]<0) {@;w[i]*=-1; ++parity;} /* strip signs */
    @< Sort the entries of |w|... @>
    if (subtype=='D' && is_odd(parity)) w[0]*=-1;
  }
}

@ Sorting the entries of a vector is done using the function |sortrow|
defined in the compilation unit \.{sorting}; this however returns a weakly
decreasing vector, so we reverse the result afterwards.

@< Sort the entries of |w|... @>=
{@; index i=0, j=perm_size-1;
  sortrow(w,perm_size); while (i<j) swap(&w[i++],&w[j--]);
}

@ Before we go into generating permutations and sign changes, we introduce an
additional file-scope variable |alternate|, which records whether sign
changes are to be performed in pairs, i.e., whether both |subtype=='D'| and
none of the entries is~|0|.

@< Local variables @>=
local boolean alternate; /* should signs be changed in pairs (for $D_n$) */

@ When |subtype!='A'|, the permutations will be generated in an outer loop,
and the sign changes in an inner one. This is much more convenient than the
opposite arrangement, since the permutation subgroup acts on the
representatives of the orbits under the sign-change subgroup (namely the
vectors with non-negative entries), but the sign-change subgroup does not
act on the representatives of the orbits under the permutation subgroup
(namely the weakly decreasing vectors)%
\footnote*{There is a close connection with the fact that the sign-change
  subgroup is normal in the classical Weyl group, but the permutation
  subgroup is not.}.
This statement is not quite true when |alternate==true|, since if there is
an odd number of minus signs they cannot all be cancelled, and there is no
systematic way to place for the remaining one that is invariant under
permutations (to see this, suppose all entries have equal absolute values).
Instead, we choose to always place such a remaining minus sign on the first
entry while generating permutations, taking care to withdraw it from the
permutation process itself.

The code below generates successive permutations used for Weyl orbit
traversal. Most of the work is done by the function |Nextperm| defined in
the compilation unit \.{symg}, which changes a vector into its
lexicographical successor in its orbit under permutation of the entries, and
returns a boolean value telling whether any new permutation could be found
at all. What remains to do here is to set aside the sign of the initial
entry if necessary, and to reattach it afterwards. The value left in the
vector after all permutations have been found will not be used, so we do no
effort to restore the initial state when performing a~|break|.

@< Generate the next permutation into~|w|... @>=
{ boolean minus= alternate && w[0]<0;
    /* if this holds, |w[0]| is unique negative entry */
  if (minus) w[0]*=-1;
  if (!Nextperm(w,perm_size)) break;
  if (minus) w[0]*=-1;
}

@ Since generating sign changes is part of the inner loop for Weyl orbit
traversal, we make it more efficient by preparing a list of the indices of
the non-zero entries of~|w|, in an array~|inx|; the list is terminated by
the value~|-1|.

@< List indices of non-zero entries... @>=
{@; index i,j=0;
  {@; for (i=0; i<perm_size; ++i) if (w[i]!=0) inx[j++]=i;} inx[j]= -1;
}

@ We wish to change a minimal number of signs each time, namely~|1|, or if
|alternate| holds,~|2|. In the latter case we always change the sign of
|w[0]|, leaving us with the former case for the remaining entries.
Traversing all sign combinations by changing one sign each time means
traversing a Hamiltonian circuit in a hypercube graph. We do this by
maintaining a counter |signcount| of the number of previously performed sign
changes, and locating the final |0|-bit in its binary representation; the
position of that bit determines (using~|inx|) the index of the sign to
change. When that position is out of range, we are done with sign changes,
and we close the Hamiltonian circuit by changing the final sign (unless
there were no non-zero entries at all). Strictly speaking we are limiting
the possible orbit size by using a machine integer for |signcount|, but to
overflow this counter requires an extremely long computation for a group of
rank at least~|33|. In fact, computing the tensor square of the spin
representation of |B33| will lead to an error (after days of computation),
but this was still deemed more acceptable than slowing down \emph{all}
computations by using |bigint| arithmetic.

@< Generate the next sign combination...  @>=
  { index i=0; unsigned long bits=signcount++;
    if (alternate) {@;	w[0]*=-1; ++i; }
      /* |inx[i]==i| for all |i<perm_size| here */
    while (inx[i]>=0 && is_odd(bits)) {@; bits>>=1; ++i; }
    if (inx[i]<0)
      {@; if (--i>=0) w[inx[i]]*=-1; break;} /* done; restore |w| and quit */
    w[inx[i]]*=-1; /* otherwise change this one sign */
  }

@* Finding suborbit representatives in case of exceptional groups.
So far we have dealt with orbits under a Weyl group of classical type; for
exceptional groups however these only constitute part of the desired orbits,
and the full orbits must be obtained by combining a number of these
suborbits. The method used to do this is to first establish a system of
representatives of the cosets for the classical type subgroup of the Weyl
group (this depends only on the exceptional group considered), and then to
act with each of these on the vector whose orbit is to be generated,
establishing a relatively small set of vectors whose suborbits will combine
to the full orbit; finally, this list is thinned out so that it contains
no more than one element in any suborbit.

Since there are only five rather finite systems of coset representatives to
deal with, it would be possible to compile them explicitly into the source
code of \LiE., however, since the largest system (for |E8|) still contains
135 group elements, it is worth while to use a more compact way to represent
these systems. Actually it is permissible to describe a set that represents
a few cosets more than once, if this proves to be more efficient, because we
will filter repetitions from the set of suborbit representatives anyway; we
shall use this freedom in case of~|E6|. The form chosen for describing the
set of coset representatives is $\setof c^ix: 0\leq i<n, x\in X \endset$,
where $c\in W$, $n\in\Nat$, and $X\subset W$ are the explicitly given data;
it turns out that a fairly efficient representation of this form is possible
in all cases. In fact, the element~$c$ will always be a Coxeter element,
i.e., a product of all distinct simple reflections, in some order. The
values of $c,n,X$ are set up by |Weylloopinit|, and stored as follows:	the
matrix representing the action of $c$ is stored in |cox|, the number~$n$ of
powers of~$c$ used is stored in |cox_order|, the cardinality of~$X$ in
|X_order|, and the matrices representing the actions of the elements of~$X$
in |X_elt[i]| for |0<=i<=X_order|. The list of distinct suborbit
representatives will be stored in |suborbit_reps|. Sufficient space for
these objects, as well as for two working vectors |temp1| and |temp2| will
be allocated by |Weylloopinit|, so that no allocation or deallocation is
needed in~|Weylloop|.

@< Local variables @>=
local index cox_order, X_order;
local matrix* cox,* X_elt[9]; /* matrices of size |[eps_dim]|{}|[eps_dim]| */
local matrix* suborbit_reps; /* size |[cox_order*X_order]|{}|[eps_dim]| */
local vector* temp1,* temp2;  /* scratch space vectors of length |eps_dim| */

@ The function |tabulate_suborbits| fills |suborbit_reps| with suborbit
representatives, and calls |Unique| afterwards to remove any duplicates.

@c

local void tabulate_suborbits(entry* v)
{ index i, k=0;
  entry** p=suborbit_reps->elm; /* next row of |suborbit_reps| */
  entry* cur=temp1->compon,* alt=temp2->compon; /* will be swapped repeatedly */
  copyrow(v,cur,rnk); /* make working copy of |v|, which will be modified */

  do /* effectively |for (k=0; k<cox_order; ++k)|,
	but with exit in middle of loop body */
  { for (i=0; i<X_order; ++i) /* elements of $X$ */
    { mulvecmatelm(cur,X_elt[i]->elm,alt,rnk,rnk); /* |alt=cur*X_elt[i]| */
      (*to_e)(alt,*p); normalform(*p++);
	/* store converted vector after normalising it */
    }
    if (++k>=cox_order) break;
      /* note that if |cox_order==1|, |cox| is never used */
    @< Multiply |cur| by |cox| @>
  } while(true);
  suborbit_reps->nrows=cox_order*X_order;
  Unique(suborbit_reps,cmpfn); /* remove duplicate suborbits */
}

@ The function |mulvecmatelm| can be used to multiply a vector by a matrix
without the overhead of storage allocation, but it requires its source and
destination vectors to be non-overlapping; to have the result as |cur|
afterwards without having to copy the vector, we swap the pointers |cur| and
|alt|.

@< Multiply |cur| by |cox| @>=
{@; entry* t=cur; mulvecmatelm(cur,cox->elm,alt,rnk,rnk); cur=alt; alt=t; }

@* The main functions.
Finally we come to the functions that form the interface to the rest of
\LiE.: |Weylloopinit|, |Weylloop|, and |Weylloopexit|. The first of these,
which is in fact the largest, sets the value of |the_g|,
and initialises various other values that depend on these data only. For
|suborbit_reps|, which also depends on the vector whose orbit is traversed,
memory is allocated, but not initialised.

@c
void Weylloopinit(simpgrp* g)
{ the_g=g;  rnk=g->lierank;
    /* save values used by other functions */
  @< Set |perm_size|, |eps_dim|, |subtype|, |to_e| and |from_e|
     according to |the_g| @>

  temp1=mkvector(eps_dim); temp2=mkvector(eps_dim); /* scratch vectors */
  X_elt[0]=mat_id(rnk);

  if (strchr("ABCD",g->lietype)!=NULL) {@; cox_order=X_order=1; cox=NULL; }
  else
  @< Set |X_order|, |cox|, and |X_elt[i]| for |0<i<X_order|,
     for exceptional groups @>
  suborbit_reps=mkmatrix(cox_order*X_order, eps_dim);
}


@ The objects created by |Weylloopinit| are freed afterwards by
|Weylloopexit|.

@c

void Weylloopexit(void)
{ freemem(temp1); freemem(temp2); freemem(suborbit_reps);
  @/{@; index i; for (i=0; i<X_order; ++i) freemem(X_elt[i]); }
  if (cox!=NULL) freemem(cox);
}

@ We shall want to use several explicit Weyl group elements, expressed as
Weyl words, to pass as argument to |simp_Weylmat|. To that end the have to
be stored in a |vector| structure, and for the sake of readability we do not
want to write separate assignments for each entry, so we use a macro
|set_vec| instead. The natural thing to do would be to pass the
entries as a brace-enclosed list to the macro, to serve as initialiser for a
local static variable which could then be copied into the desired vector;
however for stupid lexical reasons, \Cee\ makes this impossible (the commas
in the initialiser are not shielded by the braces in the macro invocation,
and supplying additional parentheses to shield them makes the initialisation
illegal). Since all entries needed are integers in the range |1|--|8|, we
therefore choose to pass them as characters of an argument string~|s|, and
to convert them to integers while copying to the vector~|v|.

Only when |g->lietype=='E'| does the set $X\subset W$ (which is used in
addition to the Coxeter element |cox| to generate suborbit representatives)
contain more than just the identity element. If so, an additional special
Weyl word |gen| will be used in determining~$X$.

@d set_vec(v,s)
  @/{@; char* p=s; int i=0; do v->compon[i++]=*p++-'0'; while (*p!='\0');
    v->ncomp=i;
  }

@< Set |X_order|,... @>=
{ matrix* tmpmat; index i;
  vector* coxw=mkvector(rnk); /* some Coxeter word */
  if (g->lietype=='E')
  { vector* gen=mkvector(15); /* large enough for largest use */
    if (rnk==6) @< Case of |E6| @> @+
    else if (rnk==7) @< Case of |E7| @> @+
    else @< Case of |E8| @>
    freemem(gen);
  }
  else /* type |F4| or |G2| */
  { if (g->lietype=='F') {@; cox_order=3; set_vec(coxw,"1234")@; }
    else {@; cox_order=2; set_vec(coxw,"12")@; }
    cox=simp_Weylmat(coxw,the_g); X_order=1;
  }
  freemem(coxw);
}

@ For |E6| the parameters we use to generate suborbit representatives are
$c=[6,2,5,4,3,1]$, $n=12$, and $X=\{ e, [2,4,3,1], [1,5,4,2,4,3,1] \}$. This
means that $n\vert X\vert$ (i.e., |cox_order*X_order|) is 36, more than
maximal number 27 of suborbits, which is the index of $W(D_5)$ in $W(E_6)$;
however the amount of time wasted by this fact is negligible.

@< Case of |E6| @>=
{ X_order=3; cox_order=12;
@/set_vec(coxw,"625431")@; cox=simp_Weylmat(coxw,the_g);
@/set_vec(gen,"2431")@; X_elt[1]=simp_Weylmat(gen,the_g);
@/set_vec(gen,"1452431")@; X_elt[2]=simp_Weylmat(gen,the_g);
}

@ For |E7| the parameters we use are $c=[1,2,3,4,5,6,7]$, $n=18$, and
$X=\{e, x, x^2, x^5 \}$, where $x=[7,6,5,4,3,2,1,5,3,4,2]$. The corresponding
$4\times18=72$ elements precisely represent the cosets of $W(A_7)$ in
$W(E_7)$.

@< Case of |E7| @>=
{ X_order=4; cox_order=18;
@/set_vec(coxw,"1234567")@; cox=simp_Weylmat(coxw,the_g);
@/set_vec(gen,"76543215342")@; X_elt[1]=simp_Weylmat(gen,the_g);
  X_elt[2]=Matmult(X_elt[1],X_elt[1]); /* $x^2$ */
  X_elt[3]=Matmult(tmpmat=Matmult(X_elt[2],X_elt[2]),X_elt[1]); /* $x^5$ */
  freemem(tmpmat);
}

@  For |E8| we use the parameters  $c=[7,5,3,2,8,6,4,1]$, $n=15$, and
$X=\{ e, x_1, x_1^3, x_1^4, x_2, x_2^2, x_2^3, x_2^4, x_2^5 \}$, where
$x_1=[7,4,3,2,4,5,6,1,3,4,5,2,4,3,1]$ and $x_2=[8,6,7,5,6,3,4,5,2,4,3,1]$.
We thus obtain $9\times15=135$ elements, which precisely represent the
cosets of $W(D_8)$ in $W(E_8)$.

@< Case of |E8| @>=
{ X_order=9; cox_order=15;
  set_vec(coxw,"75328641")@; cox=simp_Weylmat(coxw,the_g);
  set_vec(gen,"743245613452431")@;
  X_elt[1]=simp_Weylmat(gen,the_g); /* $x_1$ */
  tmpmat=Matmult(X_elt[1],X_elt[1]);
  X_elt[2]=Matmult(tmpmat,X_elt[1]);  /* $x_1^3$ */
  X_elt[3]=Matmult(tmpmat,tmpmat);     /* $x_1^4$ */
  freemem(tmpmat);
  set_vec(gen,"867563452431")@;
  X_elt[4]=simp_Weylmat(gen,the_g); /* $x_2$ */
  for (i=5; i<9; ++i) X_elt[i]=Matmult(X_elt[4],X_elt[i-1]);
    /* $x_2^2$, $x_2^3$, $x_2^4$, $x_2^5$ */
}

@ And now we come to the function all our preparations were intended for,
|Weylloop|, which runs through all elements of the orbit of~|v|, each time
calling a parameter function |action|. Actually, we have already defined
some bits and pieces of it, and it remains to put everything together in a
fairly straightforward manner. The test for |subtype=='A'| was moved out of
the loop traversing permutations, thereby doubling that loop, to gain just a
little bit of extra speed; such efficiency considerations are justified here,
because in heavy computations in \LiE. a significant fraction of the overall
time may be spent inside this function. The traversal of permutations itself
is also simplified slightly when |subtype=='A'|, since it is not necessary
to consider the possibility that |alternate| holds.

@c
void Weylloop(void(* action )(entry*),entry* v)
{ entry* tmp=temp1->compon;
  index k,* inx=NULL; /* indices of non-zero entries in |w| */

  tabulate_suborbits(v);

  if (subtype!='A') inx=alloc_array(@[index@],perm_size+1);
    /* one extra for the terminating |-1| */
  for (k=0; k<suborbit_reps->nrows; ++k)
    /* traverse suborbit representatives */
  { entry* w=suborbit_reps->elm[k]; /* it is safe to modify this vector */
    if (subtype=='A')
      do {@; (*from_e)(w,tmp); (*action)(tmp); } while (Nextperm(w,perm_size));
    else
    { alternate= subtype=='D' && w[0]!=0; /* whether type $D_n$ and no 0's */
      do /* traverse permutations */
      { unsigned long signcount=0;
	@< List indices of non-zero entries of~|w| into~|inx| @>
	do /* traverse sign combinations */
	{ (*from_e)(w,tmp); (*action)(tmp);
	  @< Generate the next sign combination into~|w|,
	     or |break| if none are left @>
	} while(true);
	@< Generate the next permutation into~|w|,
	   or |break| if none are left @>
      } while(true);
    }
  }
  if (inx!=NULL) freearr(inx);
}

@* Index.
