\input LiEwebmac

@* Determination of types.
This file deals with the functions to determine the type of an 
expression.

@h "lie.h"

@< Prototypes of globally visible functions and global variables @>=
objtype eval_type(symblst s);
void init_loop_gettype(void);

@ We shall hold bindings between a variable name and a type on a
stack~|type_stack|. If a function has multiple |return| statements, their
types are collected on |return_type| for comparison.

@d RETURNSTACK_LEN 200
@d VAR_LEN 1000

@< Local variables @>=
static struct {@; strtype name; objtype type; }	type_stack[VAR_LEN];
static objtype return_type[RETURNSTACK_LEN];
static int return_pt, var_pt; 

@
@( gettype.h @>=
@< Prototypes of globally visible functions and global variables @>

@ 
@c
@< Local variables @>@;
@< Auxiliary functions @>@;

extern fobject seqfun;

@ Initially, and after each error, the arrays |type_stack| and |return_type|
are emptied.

@c
void init_loop_gettype(void) @+{@; var_pt=return_pt=0; }

@ The main function of this compilation unit is |eval_type|. It computes the
type of a symbol, with a possible argument list; this is done by a
more-or-less recursive tree traversal (although looping in case of actually
recursive functions is avoided), where the types of the arguments are
determined before that of the symbol itself.

@c
objtype eval_type(symblst s)
{
  symblst t=s->arglst; strtype name=s->a.label->name;
  labeltp label0=label;
  label=s->a.label;
@/@< Compute the types of the arguments @>
  label=label0; /* restore |label| after handling arguments */
  @< Compute the type if the types of the arguments are known @>
}

@ We call |eval_type| recursively for the arguments of~|s|, unless the type
of these arguments is already known. This should succeed in determining the
type of the argument, unless some recursive application is encountered, in
which case type evaluation stops to prevent looping, and |UNKNOWN| is
returned. Normally this terminates the current call of |eval_type| (so the
remaining arguments will not be typechecked), returning |UNKNOWN|. However,
for built-in functions whose name starts with `\..', the argument type is
left undefined and we continue with the remaining arguments; this prevents
|UNKNOWN| from propagating outwards through for instance conditional
statements and statement sequences. Whenever the argument of a |return|
statement is found, its type is added to the array |return_type|; these
types are tested to all be equal to the return type for the function when
the function body is completely traversed.

@< Compute the types of the arguments @>=
for (; t!=NULL; t=t->next)
{ if (t->type==UNKNOWN)
  { if (eval_type(t)==UNKNOWN && name_tab[name][0]!='.') return UNKNOWN;
    if (t->a.label!=NULL && t->a.label!=label_null 
	&& t->a.label->name==return_name)
      { if (return_pt==RETURNSTACK_LEN) error("Too many happy returns:\n");
	return_type[return_pt++]=t->type;
      }
  }
}

@ Once the argument types are known, we sucessively test a number of cases,
to see how determination of the type of~|s| should proceed. The case of a
statement sequence is particularly simple, since the result type is just that
of the final statement of the sequence.

@< Compute the type if the types of the arguments are known @>=
{
  symblst foundsym; strtype name=s->a.label->name;

  if (s->type!=UNKNOWN) return s->type;
  if (name==0) error("System error: eval_type0: no_name.\n");
  if (name==seq_name)
  @/{@; s->class=MAP; s->data.f=seqfun; return s->type=last_type(s->arglst); }
  if (s->class==BLOCK) @< Find type of block @>
  @< Handle previously assigned variables @>
  @< Locate function definition for |name|, and assign it to |foundsym|;
     modify |*s| accordingly @>
  if (name==assign_name || name==assign_loc_name)
    @< Handle assignment statement @>
  if (name==if_name) @< Find type of |if| statement and |return| it @>
  if (s->type==ARGTYPE)
    @< Handle built-in function with result type |ARGTYPE| @>
  if (s->class==FUNCTION) @< Handle function call @>
  return s->type; /* if |s->class==FUNCTION_COPIED|, this is |UNKNOWN| */
}

@ The function |last_type| determines the last type in a list, or |VOID| in
case of an empty list.

@< Auxiliary functions @>=
static objtype last_type(symblst s) 
{ if (s==NULL) return VOID;
  while (s->next!=NULL) s=s->next;
  return s->type;
}

@* The different cases. We now consider the various cases where |gettype|
has to work to find the type of an expression given its various attributes,
and the types of any arguments it may have.

@ The first case is a block; we mark the position on the variable stack,
increase |block_depth|, and push any arguments of the block (which have been
given names in the parser using |add_dollar_names|), after which we call
|eval_type| recursively to obtain the type of the expression which
constitutes the block. That type is copied into |s->type|, and afterwards the
stack is released.

@< Find type of block @>=
{ int var_pt0=var_pt; ++block_depth; push_type(s->arglst);
@/s->type=eval_type(s->data.expr);
@/var_pt=var_pt0; --block_depth;
  return s->type;
}

@ The function |add_var| places a type binding for a local variable (or
function parameter) on |type_stack|. The function |push_type| does the same
for all pairs |(formal,type)| in a list of nodes (in practice always an
argument list).

@< Auxiliary functions @>=

static void add_var(strtype name,objtype type)
{
  if (var_pt>=VAR_LEN)
    error("Number of variables (%ld) exceeded.\n",(long)VAR_LEN);
  type_stack[var_pt].name=name;
  type_stack[var_pt++].type=type;
}

static void push_type(symblst list)
{@;
  for ( ; list!=NULL; list=list->next) add_var(list->formal,list->type);
}

@ If a node is not a block and has no arguments, it is a variable, or maybe
a function with no arguments. The symbol is looked up in the tables, and if
it was a parameterless function this yields |ERROR|, in which case we go on
to the other cases; otherwise the type of the variable is returned and the
|class| is set to |DUMMY|.

@< Handle previously assigned variables @>=
{ objtype type;
  if (s->arglst==NULL && (type=type_var(name))!=ERROR)
  @/{@; s->class=DUMMY; return s->type=type; }
}

@ A variable is first searched in |type_stack|, from top to bottom, where
variables reside that were previously encountered by |gettype|; then
by |srchsym| in the stack |topsym| of local variables under execution by
|getvalue|, and this also inspects the definitions in |top_definitions|.

@< Auxiliary functions @>=
static objtype type_var(strtype name) 
{ int i=var_pt-1;
  for (i=var_pt-1; i>=0; --i)
    if (type_stack[i].name==name) return type_stack[i].type;
  { symblst foundsym=srchsym(topsym,name,NULL);
    if (foundsym==NULL || foundsym->class!=VALUE) return ERROR;
    return foundsym->type;
  } 
}


@ A function is first searched in the list of permanent bindings attached
to the hash table entry for |name|, and afterwards in the list |topsym| of
temporary bindings; if both searches fail, we cannot identify |name|, and call
|error_not_foundsym|. After passing a test |check_function_args| explained
below, the |type| and |class| of the matching binding are copied into |*s|
by |assignsym|. This is how most nodes obtain a |class!=DUMMY| (the parser
only set this for blocks). In case |class==OPERATOR| (which means the
binding is a built-in function whose arguments are evaluated before the call)
the binding is saved in |s->data.expr|, so that it need not be looked up
again by |eval_value|. Note that in case the binding found is that of a
function already being expanded, then |foundsym->class==FUNCTION_COPIED|,
which value is copied into |s->class|, which will prevent further analysis
of the function body by |eval_type|. Eventually |foundsym->class| will be
set back to |FUNCTION|, but |s->class| will remain |FUNCTION_COPIED| until
it is encountered by |eval_value|, which will call |eval_type| again to
expand and type-check the funciton body one level further, and this
continues as deep as the recursion actually proceeds.

@< Locate function definition for |name|... @>=
{ if ((foundsym=srchsym(symbol_tab[name],name,s->arglst))==NULL)
    foundsym=srchsym(topsym, name, s->arglst);
  if (foundsym==NULL) error_not_foundsym(s);
  check_function_args(s,foundsym);
  assignsym(s,foundsym); /* set |s->class| and |s->type| */

  if (s->class==OPERATOR) s->data.expr=foundsym; /* save binding found */
}

@ The function |check_function_args| checks that there are no built-in
functions in the argument list; this is necessary for operations like
\\{map} and \\{make}.

@< Aux... @>=
static void check_function_args(symblst sym,symblst foundsym)
{ symblst f,s;
  for (f=foundsym->arglst, s=sym->arglst; f!=NULL && s!=NULL;
       f=f->next,s=s->next)
    if (isfuntype(f) && s->class != BLOCK)
      error("%s is not a user defined function. \n"
	   ,name_tab[s->a.label->name]);
}

@ For an assignment the result type is always |VOID|, but we must keep track
of the type of the variable on |type_stack|.

@< Handle assignment statement @>=
{ symblst lhs=s->arglst, rhs=lhs->next;
  strtype varname=lhs->a.label->name; /* name of left hand side */
  if (rhs->type==VOID)
  error("You cannot assign a %s value.\n", code_obj_type(VOID));
  if (name==assign_loc_name)
  { if (block_depth==0) error
	("'loc %s= .. ' can only be used inside functions/blocks.\n"
	,name_tab[varname]);
    add_var(varname, rhs->type);
  }
  else check_var_type(varname,rhs->type);
  return s->type=VOID;
}

@ In the function |check_var_type| we compare the old and new type of a
variable.

@< Auxiliary functions @>= 

static void check_var_type(strtype name,objtype type)
{
  int i; objtype old_type;
  if (type==UNKNOWN) return; /* assume all is well in this case */
  for (i=var_pt-1; i>=0; --i)
    if (type_stack[i].name==name)
      {@; old_type=type_stack[i].type; goto found; }
  { symblst foundsym=srchsym(topsym,name,NULL);
    add_var(name,type); /* add local variable, whether or not found */
    if (foundsym==NULL) return;
    old_type=foundsym->type;
  }
found:
  if (block_depth==0)
    {@; if (i>=0) type_stack[i].type=type; } /* change type */
  else
    if (!eq_types(type,old_type))
      error("Illegal type change for %s: %s -> %s,\n"
	   ,name_tab[name],code_obj_type(old_type),code_obj_type(type));
}

@ In |else|-less |if| statement has type |VOID|; if there are two branches
their types must match.

@< Find type of |if| statement and |return| it @>=
{ symblst then_part=s->arglst->next, else_part=then_part->next;
  if (else_part==NULL) return s->type=VOID;
  if (!eq_types(then_part->type,else_part->type)
   && then_part->type!=UNKNOWN && else_part->type!=UNKNOWN)
    error("Type mismatch between branches of if-clause: %s and %s.\n"
	 ,code_obj_type(then_part->type), code_obj_type(else_part->type));
  return s->type=
    then_part->type!=UNKNOWN ? then_part->type : else_part->type;
}

@ Certain built-in functions have their return type specified as |ARGTYPE|;
this means that the result type is not determined by the binding alone, but
depends explicitly on the types of the actual arguments. This is the case
for conditional and loop clauses, |break| and |return| statements, and
for the sequence operator. Except for the conditional clauses, which have
already been treated, the result type is that of the final argument
expression, so we call |last_type| here. 

@< Handle built-in function with result type |ARGTYPE| @>=
{@; if ((s->type=last_type(s->arglst))!=UNKNOWN) return s->type; }

@ We know deal with user defined functions.

@< Handle function call @>=
{ int var_pt0=var_pt, return_pt0=return_pt;
  strtype fun_name_old=fun_name;
  @< Push arguments on stack @>
  @< Load and traverse function body @>
  @< Check return types @>
  @< Pop arguments from stack @>
  return s->type;
}

@ When entering a funciton body, the formal names are first assigned to the
actual arguments by |assignnames| and the resulting bindings are then places
on the stack by~|push_type|. The function name is saved in the global
variable |fun_name| for error reporting.

@< Push arguments on stack @>=
{ fun_name=foundsym->a.name;
  ++block_depth;
  assignnames(s, foundsym->arglst);
  push_type(s->arglst);
}

@ After the function body has been traversed, any additions to |type_stack|
and |return_type| are discarded by resetting |var_pt| and |return_pt| to the
values saved before pushing the function arguments; also |fun_name| is
restored to its old value.

@< Pop arguments from stack @>=
{@; var_pt=var_pt0;
  return_pt=return_pt0;
  fun_name=fun_name_old;
  --block_depth;
}

@ The function body is loaded into |tree_ar|, and |s->data.expr| is made to
point to its root node. During typechecking of the function body we
temporarily set the |class| field of the binding to |FUNCTION_COPIED|, to
prevent looping of the type checker, as explained above. When |eval_type|
returns, the |class| of the function body is changed to |BLOCK|, since
|eval_value| makes no distinction between user-defined function calls and
blocks with arguments.

@< Load and traverse function body @>=
{ s->data.expr=load_function_body(foundsym->data.func);
  foundsym->class=FUNCTION_COPIED;
  s->type= eval_type(s->data.expr);
  foundsym->class=FUNCTION;
  s->class=BLOCK;
}

@ When we are done with a function body we test that its type and that of
any |return| statements it contains match. Some of these types may be
|UNKNOWN|, in which it is assumed (without checking) that these will in fact
return the proper type; however not all exits must give |UNKNOWN| type,
since then apparently the recursive function cannot terminate.

@< Check return types @>=

{ int i;
  for (i=return_pt0; i<return_pt; ++i)
    if (!eq_types(return_type[i],s->type)
     && return_type[i]!=UNKNOWN && s->type!=UNKNOWN) break;
    else if (s->type==UNKNOWN) s->type=return_type[i];
  if (i<return_pt)
  { Printf("Inconsistent return types for %s:\n",name_tab[name]);
    for (i=return_pt0; i<return_pt; ++i)
      Printf("return nr.%d has type: %s,\n"
	    , i-return_pt+1, code_obj_type(return_type[i]));
    error("function body has type: %s\n",code_obj_type(s->type));
  }
  if (s->type==UNKNOWN)
    error("The result of function %s has indeterminate type.\n"
	 ,name_tab[name]);
}

@* Index. 

