\input LiEwebmac

@* Introduction. This file describes the lexical scanner of \LiE.. It was
installed instead of the previously used \.{lex}-generated scanner since
that one was often causing portability trouble; in particular it avoids
incompatibility problems between \.{lex} and \.{flex}, and interference with
the \.{readline} library. Parser and scanner represent tokens by small
numeric codes, recorded on the header file \.{parser.h}.

@h "lie.h"
@h "parser.h" /* contains token codes for transmission to the parser */
@)
@d local static
@f local static

@( lexer.h @>=
@< External declarations @>
@< Public prototypes @>

@
@c
@< Variable declarations @>@;

@ We keep track of the line number during scanning, for indicating errors.

@< Variable... @>=

int lineno=1;

@~@< Ext... @>= extern int lineno;

@* Tables recording identifiers.
All distinct identifiers and operator symbols are stored as null-terminated
strings in dynamically allocated memory. Since these remain in existence
until \LiE. terminates, no provision is made for reclaiming this memory. For
efficiency, substantial blocks of memory are allocated at once, from which
the individual strings are taken; whenever more space is needed than remains
in the block, the remainder is simply forgotten and a new block is
allocated. In practice this will only lead to a very limited waste of
memory. Due to this simple regime, two quantities suffice to record the
current status of this character memory allocation: a pointer |char_point|
pointing to the first free byte that can be assigned to a new string, and an
integer |chars_left| telling how many of the successive bytes remain to be
allocated.

@< Variable... @>=
local char* char_point=NULL;
local int chars_left=0;

@ When a identifier |str| of length |n| has to be stored, we call
|store(str,n)|, it will allocate |n+1| bytes and copy |str| including the
terminating null byte into it.

@d STRBLSIZE 500 /* size of block allocated for strings */

@c
local char* store(char* s,int len)
{ int n= ++len>STRBLSIZE ? len : STRBLSIZE; /* size of new block if needed */
  char* result=char_point;
  if (len>chars_left)
     result=char_point=alloc_array(@[char@],chars_left=n);
  char_point+=len; chars_left-=len; return strncpy(result,s,len);
}

@ The hash code of an identifier is determined by interpreting its characters
as digits of a number in base~256, and then taking the number modulo
|hash_mod|.

@d hash_mod 1997 /* a somewhat odd number */

@c
local short hash(char* s) /* |s| is a non-empty string */
{ char c; long r=*s++; while ((c=*s++)!='\0') r=(c+(r<<8))%hash_mod;
  return (short)r;
}

@ For each identifier that is stored a pointer to it is added to the array
|name_tab|. The hash table |hash_tab| contains numbers that are an index into
|name_tab|, except of course for the free slots in the table, which are
represented by~|-1|. The counter |hash_count| counts the total number of
(distinct) identifiers.

@< Var... @>=
char* name_tab[hash_mod]; /* actual names of identifiers */
local short hash_tab[hash_mod];
  /* non-negative entries point into |name_tab| array */
local short hash_count=0;
  /* number of hash enties (should stay |@[@]<hash_mod|) */

@~@< Ext... @>= extern char* name_tab[];

@ We must set the hash table to empty slots initially.

@< Init... @>=
{@; int i; for(i=0; i<hash_mod; ++i) hash_tab[i]=-1; }

@ When an identifier has been isolated in the input stream, a call to |match|
returns its index in |name_tab|, possibly after adding the identifier to the
tables. This small number is henceforth used to characterise the identifier.
The argument |copy| indicates whether a copy of |str| should be made it it
is added to the tables; this in not needed for instance when the actual
argument is a string denotation.

@< Public prototypes @>= strtype match(char* str, boolean copy);

@~Searching the hash table starts at the hash key and proceeds linearly
through the table until either a match or an empty slot is found; in the
latter case no matching entry is present, and we add one in the place of the
empty slot. This method is valid because we never remove an entry from the
hash table. When we have just filled the last empty slot in the table, the
program is terminated with a fatal error message; thus the possibility of
the search looping forever through the table is avoided without requiring a
test in the |while| loop.

@c
strtype match(char* str, boolean copy)
{ short i,h=hash(str);
  while ((i=hash_tab[h])>=0)
    if (strcmp(name_tab[i],str)==0) return i; /* identifier found in table */
    else if (++h==hash_mod) h=0; /* move past occupied slot */
  if (hash_count>=hash_mod-1) fatal("Hash table overflow");
  hash_tab[h]= hash_count; /* this number represents the new identifier */
  name_tab[hash_count]=copy ? store(str,(int)strlen(str)): str;
  return hash_count++;
}

@ Here we install a few macros that temporarily define the interface to the
rest of lie; they override functions previously defined in \.{sym}. Since the
hash table used to be (ab)used as symbol table as well, we must define one
here as a substitute.

@< Var... @>=
symblst symbol_tab[hash_mod];

@~@< Ext... @>= extern symblst symbol_tab[];

@~@c 

short creatmapname(char *a)
{ char *s=strcat(strcpy(alloc_array(@[char@],strlen(a)+2),"."),a);
  short result=match(s,true); free(s); return result;
}

void clear_symbol_tab(void)
{@; int i; for (i=0; i<hash_mod; ++i) symbol_tab[i]=NULL; }

@~@< Init... @>=
clear_symbol_tab();

@~@< Public proto... @>=
short creatmapname(char *a);
void clear_symbol_tab(void);

@ Since tokens can be arbitrarily long, we use a ``flexible'' buffer to
store their characters in before calling |match|. The variable |tok_buf|
points to the buffer, |tok_len| indicates how many characters are in use,
and |tok_buf_size| records the number currently allocated.

@< Var... @>=
local char* tok_buf; /* points to currently recorded string */
local int   tok_len,tok_buf_size;
  /* used respectively allocated size of buffer */

@ Starting rather arbitrarily with a buffer of 8~bytes, its size will double
at most one or two times for most programs.

@< Initialise scanner @>= tok_buf=alloc_array(@[char@],tok_buf_size=8);

@ To add characters to the buffer, we call |app_c|.
@c
local void app_c(char c)
{ if (tok_len==tok_buf_size)
  { char* p=tok_buf; @+
    tok_buf=strncpy(alloc_array(@[char@],tok_buf_size*=2),tok_buf,tok_len);
    freearr(p);
  }
  tok_buf[tok_len++]=c;
}

@* Lexical functions and states.
We introduce a number of auxiliary functions that perform part of the
lexical analysis. We also define some static variables that determine a
lexical state; this serves to enable some identifiers like `\.{help}' to be
interpreted as keywords only at the beginning of a command, to inhibit
interpretation of newlines as command terminators in places where the
grammatical structure would obviously make them illegal, and to be able to
cater for commands with a deviant lexical structure.

Part of the lexical state serves merely to enable look ahead into the input
stream. The character under current scrutiny during lexical analysis is kept
in the variable~|c| (admittedly a shortish name for a variable with file
scope, but it is rather ubiquitous in the code below); it is advanced to the
next position by the macro |shift|. While outside the lexical analyser, the
first character of the next token is normally not shifted in, so that for
instance no attempt is made yet to read the character after an newline.  For
scanning certain tokens however it is necessary to inspect the character
beyond the end of the token, and this character has to be backed up before
leaving the lexical analyser; we cannot use |ungetc| for this purpose, as it
will not work well in conjunction with |fetch_char|, so we provide our own
1-character buffering facility, using |next_c| and a flag |backed_up|. The
flag |start_of_command|, which is raised before a new command is scanned,
and cleared after any token is read, controls the recognition of
semi-keywords like `\.{help}'. Two mechanisms are available to prevent
interpretation of a newline as the end of a command: setting
|ignore_nl=true| will cause the next newline to be ignored, while making
|nesting>0|, as happens between any pair of grouping symbols such as `\.\{'
and~`\.\}', also disables interpretation of newlines.  The flag |short_ints|
controls the conversion of a following integer denotation:  normally these
are made into a |NUMBER| token whose value is given by a |bigint|, but if
the flag |short_ints| is set, e.g., immediately after a group type
indication such as `\.E' is scanned, they are made into an |INT| token whose
value is a |long|.

@d shift (c= backed_up ? backed_up=false, next_c : fetch_char())
@d unput(c) (backed_up=true,next_c=c)
@< Var...@>=
local int c, next_c;
local boolean backed_up=false; /* whether we should use |next_c| next */
local boolean start_of_command;
local enum { normal, help, edit, learn, ended } command_type;
local boolean ignore_nl; /* whether |'\n'| is treated as space */
local int nesting; /* number of pending opening symbols */
local boolean short_ints; /* whether an integer will be read as an |index| */

@ At the very beginning |init_scanner| is called; the function |init_command|
is called each time before a new command is to be scanned. The function
|clear_input| is called by |error| to remove the remainder of the input line.

@< Public prototypes @>=
void init_scanner(void); void init_command(void); void clear_input(void);
@~@c
void init_scanner(void) @+{@; @< Initialise scanner @> }
void init_command(void)
{ start_of_command=true; command_type=normal; ignore_nl=false; nesting=0;
  short_ints=false; fetch_line(promptlabel);
}
void clear_input(void)
{ if (command_type!=ended) @+ do shift; while(c!='\n' && c!=EOF); @+ }

@ Skipping spaces is a rather common activity during scanning; it is
performed by |skip_space|. When it is called, the first potential space
character has been shifted in, and when it returns a non-space character is
left in shifted-in position. Processing of newline-escaping backslashes and
skipping comments is also performed by |skip_space|.

@c
local void skip_space(void)
{ do
  { if (c=='\\') ignore_nl=true;
    else if (c=='\n')
    {  ++lineno;
      if (ignore_nl || nesting>0)
      {@; ignore_nl=false; fetch_line(PROMPT2); }
      else break; /* stop at non-ignored newline */
    }
    else if (c=='#')
    { do shift; while (c!='#' && c!='\n' && c!=EOF);
      if (c!='#') continue;
	/* |'\n'| requires action, |EOF| must not be skipped */
    }
    else if (!isspace(c)) break;
    shift;
  } while(true);
}

@ Keywords are identified by sequence number in the order by which they are
entered into the table. Position 0 is reserved for the empty string. When we
use this, we shall use the constant |empty_str|.

@< Ext... @>=
enum{ empty_str=0 };

@~
@d first_semi_keyword HELP /* from here only keyword at start of command */
@d first_type_id (HELP-6) /* the next 6 are type identifiers */
@d keyword_limit array_size(keyword) /* from here on ordinary identifiers */

@< Var... @>=

char* keyword[]=
@/{ "",
  "if", "then", "else", "fi",
  "for", "to", "downto", "in", "row",  "while", "do", "od",
  "loc", "return", "break",
@|"make", "iapply", "vapply", "mapply", 
  "on", "off", "setdefault", "savestate", "restorestate",
@|"mat", "vec", "int", "grp", "tex", "pol",
@|
  "help", "type",
  "edit", "read", "write", "monfil",
  "learn", "exec", "quit", "exit",
  "listvars", "listfuns", "listops",
  "maxnodes", "maxobjects"
@/};

@~@< Ext... @>=extern char* keyword[];

@ Here we enter the keywords of the language.

@< Initialise scanner... @>=
{@; int i; for (i=0; i<keyword_limit; i++) match(keyword[i],false);
}

@* The main scanning routine.
Now we come to the scanner proper, which has the silly name |yylex|, as is
dictated by the program ``\.{bison}'' (formerly ``\.{yacc}'', yet two other
silly names) that is used to generate the parser.

When |command_type!=normal| the behaviour of |yylex| is altered. The value
|ended| means that a token |ENTER| has just been passed to the parser, which
is always immediately followed by the terminating token~|0|; this double
token approach is needed because it is not possible to explicitly match the
terminating token in grammar rules of \.{yacc}, and because we may not
assume that the lookahead symbol is a valid one when a rule is reduced.  For
the other non-|normal| values of |command_type| we do start with skipping
spaces and comments, but use non-standard lexical rules for what follows.
For the case |command_type==normal| we essentially perform case analysis on
the first character, but an initial capital letter requires the character
following it to be inspected, because of the special treatment of tokens
like |'A'| and|'X'|. The flags |ignore_nl| and |short_ints| are intended to
affect only the token immediately following the one that sets them. Since
the former only affects |skip_space| we may reset it after that call, but
the latter is needed in scanning numbers, so it is reset at the end; those
tokens that set |short_ints| jump to |finish| to avoid immediate resetting.
The flag |start_of_command| is also reset whenever a token has been scanned.

@< Public proto... @>=int yylex(void);

@~@c
int yylex(void)
{ short code;
  if (command_type==ended) return 0;
restart: shift; skip_space(); ignore_nl=false;
  if (c==EOF || c=='\n')
  @/{@; if (c==EOF && cur_in==stdin) stop=true;
    command_type=ended; return ENTER; 
  }
  if (command_type!=normal) @< Perform a greedy scan @>
  else
  { if (isupper(c)) @< Maybe scan a simple group @>
    if (isalpha(c)) @< Scan an identifier or a keyword @>
    else if (isdigit(c)) @< Scan a number @>
    else @< Scan a token starting with a non alpha-numeric character @>
  }
  short_ints=false; /* flag lasts for one token only */
finish: start_of_command=false; return code;
}

@ Here we take out the case of a single upper case letter not followed by
another letter; this leads to an indication of a simple group or torus,
the `\.X' introducing a polynomial exponent (which may also be written
as~`\.Y'), or an error message for illegal identifiers like~|L261|.

@< Maybe scan a simple group @>=
{ char type=c; int next=shift; @+ unput(next); c=type;
  if (!isalpha(next) && next!='_')
    /* else fall through for multi-letter identifier */
  { if (strchr("ABCDEFGT",type)!=NULL)
      {@; code=GROUPID; yylval.sub=type; short_ints=true; }
    else if (type=='X' || type=='Y') code=X;
    else error ("Lone `%c' not legal as identifier.\n",type);
    goto finish;
  }
}

@ When group denotations and exponent indicators have been dealt with,
everything that looks like an identifier is either that or a keyword. In
any case we start with looking it up in the table, and then the numeric
value of the code returned will allow us to discriminate the possibilities.
Most keywords are given a unique code, coinciding with their position in the
|name_tab| table, but some sets of keywords are merged into a single code, like
the type identifiers `\.{mat}'--`\.{pol}' that are all mapped
to |TYPEID|, with the actual type being recorded in |yylval.sub|.

@< Scan an identifier or a keyword @>=
{ tok_len=0; @+
  do app_c(c), shift; while(isalpha(c) || isdigit(c) || c=='_'); @+
  app_c('\0'); unput(c);
@/code=match(tok_buf,true);
  if (code>= (start_of_command ? keyword_limit : first_semi_keyword))
    @/{@; yylval.sub=code; code=IDENT; }
  else if (code>=first_type_id && code<first_semi_keyword)
    @/{@; yylval.sub=code-first_type_id; code=TYPEID; }
  else switch(code)
  {      case IF: case FOR: case WHILE: ++nesting;
  break; case FI: case OD: --nesting;
  break; case APPLY: case APPLY+1: case APPLY+2:
    yylval.sub=code-APPLY; code=APPLY;
  break; case LISTCOMMAND: case LISTCOMMAND+1: case LISTCOMMAND+2:
    yylval.sub=code-LISTCOMMAND; code=LISTCOMMAND;
  break; case QUIT+1: --code; /* \.{exit} $\to$ \.{quit} */@+
  break; case HELP: yylval.sub=0; command_type=help;
  break; case MAXNODES: case MAXOBJECTS: case ON:
    short_ints=true; goto finish;
  break; case EDIT: case READ: case WRITE: case MONFIL: command_type=edit;
  break; case LEARN: case EXEC: command_type=learn;
  }
}

@ Numbers do not include a sign, and are always scanned as |bigint|s.

@< Scan a number @>=
{ tok_len=0; @+ do app_c(c),shift; while (isdigit(c)); @+
  app_c('\0'); unput(c);
  if (short_ints) {@; code=INT; yylval.val=atol(tok_buf); }
  else {@; code=NUMBER; yylval.num=str2bigint(tok_buf); }
}

@ After splitting off the alphanumeric characters, the scanner plunges into
a large switch on the value of the look-ahead character~|c|.

@< Scan a token... @>=
switch(c)
{
default: printf("Unrecognised character `%c'\n",c); goto restart;
       case '"': @< Scan a string or character denotation @> @+
break; case '?': code=HELP; yylval.sub=1; command_type=help;
break; case '@@': code=QUIT;
break; case ';': code=SEMICOLON; ignore_nl=true;
break; case ',': code=COMMA;
break; case '(': code=LPAR; ++nesting;
break; case ')': code=RPAR; --nesting;
break; case '{': code=LBRACE; ++nesting;
break; case '}': code=RBRACE; --nesting;
break; case '[': code=LBRACK; ++nesting;
break; case ']': code=RBRACK; --nesting;
break; case '|': @+ if (shift=='|') code=OR; @+ else {@; unput(c); code=BAR;}
		 @+ ignore_nl=true;
break; case '&': @+ if (shift!='&') unput(c); @+ code=AND; ignore_nl=true;
break; case '!': @+ if (shift=='=') {@; code=RELOP; yylval.sub=1; }
		 @+ else {@; unput(c); code=NOT; }
break; case '=': @+ if (shift=='=') {@; code=RELOP; yylval.sub=0; }
		 @+ else {@; unput(c); code=BECOMES; }
		 ignore_nl=true;
break; case '<': code=RELOP; yylval.sub= shift=='=' ? 4 : (unput(c),2);
		 ignore_nl=true;
break; case '>': code=RELOP; yylval.sub= shift=='=' ? 5 : (unput(c),3); 
		 ignore_nl=true;
break; case '+': @+ if (shift=='=') code=PLUSAB;
		 @+ else {@; unput(c); code=ADDOP; yylval.sub=0; }
		 ignore_nl=true;
break; case '-': code=ADDOP; yylval.sub=1; ignore_nl=true;
break; case '*': code=TIMES; ignore_nl=true;
break; case '/': code=DIVOP; yylval.sub=0; ignore_nl=true;
break; case '%': code=DIVOP; yylval.sub=1; ignore_nl=true;
break; case '^': code=POWER; ignore_nl=true;
break; case '$': code=DOLLAR; tok_len=0;
		 do app_c(c),shift; while(isdigit(c)); @+
		 app_c('\0'); unput(c);
		 yylval.sub=match(tok_buf,true);
break; case ':': code=COLON; command_type=learn;
}

@ In scanning strings we must look for quote-image symbols and string
breaks; for this reason we always look ahead past the end of the string.

@< Scan a string... @>=
{ tok_len=0;
  do
  { while (shift!='"' && c!='\n' && c!=EOF) app_c(c);
    if (c!='"')
    { printf("Unclosed string denotation on line %d.\n" @+
      "\t(for multi-line strings use string-break.)\n", lineno);
      break;
    }
    if (shift=='"') app_c(c); @+ else skip_space();
  } while (c=='"'); /* continue after quote-image or string-break */
  app_c('\0'); unput(c); yylval.sym=addtekstnode(tok_buf); code=TEXT;
}

@ If the initial token of a command is one of \.{help}, \.?, \.{edit}, 
\.{read}, \.{write}, \.{mapleread}, \.{monfil}, \.{learn}, \.{exec}, and~\.:,
then the remainder of the command is not broken up into tokens in the
ordinary way, but in a much coarser manner. In fact after \.{learn},
\.{exec}, and~\.:, the entire remainder of the line is passed as a single
token; in the other cases white space terminates the token, while for \.{help}
and \.? the character `\.>' is also recognised as a separator, and passed as
an individual token. We must store the string found in some way; since it is
probably being used only in a single command, we do not wish to store it
with the identifiers. If the token is the last one of the command we may
pass a pointer to |tok_buf|, which will still contain the string when the
parser recognises the command. The only case where this is not so is for the
token after \.{help} or \.?; for these we copy the string to the buffer
|subject_name|.

@d subject_name_limit 64
@< Var... @>=
char subject_name[subject_name_limit];

@~@< Perform a greedy scan @>=
if (c=='>' && command_type==help)
@/{@; code= shift!='>' ? unput(c),WRITEHELP : APPENDHELP; command_type=edit; }
else
{ code=ANY_STRING; tok_len=0;
  do app_c(c),shift;
@/while (c!='\n' && c!=EOF &&
         !(command_type!=learn &&
	   (isspace(c) || command_type==help && c=='>'))
	);
  if (command_type==learn) /* then strip trailing spaces */
    while (tok_len>0 && isspace(tok_buf[tok_len-1])) --tok_len;
  app_c('\0'); unput(c);
  if (command_type!=help) yylval.text=tok_buf;
  else
  { sprintf(subject_name,"%.*s",subject_name_limit-1,tok_buf);
    yylval.text=subject_name;
  }
}
  

@* Index.
