\input LiEwebmac

@* General setup. In this file we bring together several elements of the
general structure of \LiE.. Our main output will be to the header file
\.{lie.h} (there is no file \.{lie.c}); this file is included by all regular
\LiE. files, and it will include any other header files that should be visible
throughout \LiE.. In this file we include any type definitions first, before
doinang any other preprocessor directives; this ensures that variable
declarations in other include files will be interpreted properly.

@( lie.h @>=
@< General typedefs @>@;
@p
@< External declarations @>@;

@ Here are all the standard \.{ANSI} header files (output of this module, and
in general all preprocessor commands, will be inserted in the module above).

@h <setjmp.h>
@h <stdio.h>
@h <string.h>
@h <ctype.h>
@h <stdlib.h>
@h <signal.h>
@h <stdarg.h>
@h <limits.h>
@h <assert.h>

@ And here are the header files of some \LiE. modules. For those modules which
do not define a separate header file, there is \.{ansi.h}, which collects all
the (\caps{ANSI} style) headers.


@h "lexer.h"
@h "getl.h"

@h "node.h"
@h "mem.h"
@h "gettype.h"
@h "getvalue.h"
@h "init.h"
@h "sym.h"
@h "main.h"
@h "onoff.h"
@)
@h "ansi.h"


@ Here are some ubiquitous typedefs. The idea is that |entry| is used for
array and matrix entries, and |index| for their indices and bounds. The type
|Short| is used only in a few places; since it does not correspond to a well
defined abstraction, it should be replaced by |short|. The types |digit|
and |twodigits| are used by the unbounded-length integer arithmetic routines.

@d false 0
@d true  1
@d MaxEntry		LONG_MAX
@d MinEntry		LONG_MIN
@d MaxDigit		((1<<15)-1) /* bigints are base $2^{15}$ */
@d max_obj_size		UINT_MAX /* limit for |sizeof(x)| */

@< General typedefs @>=
typedef int boolean;
typedef char* string;
typedef long entry; /* see also |MaxEntry| and |MinEntry| */
typedef long index;
typedef short Short;
typedef unsigned short digit; /* see also |MaxDigit| */
typedef unsigned long twodigits;

@ Next we include header files with specific typedefs for memory management
(note that these definitions do \emph{not} contribute to the `Preprocessor
directives').

@s object int
@s strtype int
@s objtype int
@s objclass int
@s fobject int
@s symbrec int

@< General typedefs @>=
#include "memtype.h"
#include "nodetype.h"
#include "infolrn.h" /* layout of index records for info and learn systems */

@ We add some typedefs that are general, but not related to memory management.
The type |Symbrec| is an overlay for the type |symbrec| defined in
\.{nodetype.h}; it is needed to enable static initialisation of tables, since
initialising union types is problematic.

@< General typedefs @>=

typedef struct {
  objtype type;
  objclass class;
  char *name;
  strtype formal;
  fobject f;
  long arglst;
  long next;
} *Symblst, Symbrec;

typedef struct {@; strtype name; objtype type;} nametype;
typedef struct {@; strtype p[N_PARENTS]; short n; } par_tp;
@)
typedef int cmp_tp;
typedef cmp_tp (*cmpfn_tp) (entry*,entry*,index);


@* Global parameters.
The file \.{lie.h} will also contain values of global compile-time parameters.
When a parameter is needed in one module only, it is better to define that
parameter at the start of the relevant module, to avoid wasting global name
space. We suspect that some parameters below could be moved on these grounds.
We start with the initial values of the memory management parameters.

@d MAXPTRS_DFLT		99999 /* initial size of the object table */
@d GCCRIT		1000
  /* number of spare locations when calling the garbage collector */
@d MAXNODES_DFLT	9999 /* initial size of the program table */

@ And some more specific dimensioning parameters.

@d LINELENGTH  80 /* width of the screen */
@d NPOLY 10 /* Number of stored polynomials */
@d NAMESTACK_LEN   500
@d REPORT_LEN      200

@ Some miscellaneous constants that really belong in per-module header files

@d LMARGIN 5 /* left margin for |Print| function */
@)
@d PROMPT	  "> " /* main prompt */
@d PROMPT2	  "\\ " /* continuation prompt */
@)
@d TO_LOOK 1
@d TO_EDIT 0
@d EXTEND	 8  /* initial size for flexible objects */
@d RANKMAXSUB 8

@ The macros |readmode| and |writemode| are used as arguments to file handling
functions like |fopen|. Since they are use for learn and info files which
should be seekable, a binary mode is specified.

@d readmode "rb"
@d writemode "wb"

@ Here are some defaults for external programs that are called in certain
circumstances.

@d DEFAULT_EDITOR	"emacs"
@d DEFAULT_PAGER	"less"

@ We proceed to some a bit less trivial macros. |Integer| converts its
|(object*)| argument, regardless of its current representation, into an
|entry| (machine integer).


@d array_size(a) (int)(sizeof(a)/sizeof(a[0]))
@d Integer(o)
 (type_of(o)==INTEGER?(o)->i.intval:bigint2entry(&(o)->b))
@d is_int(t) ((t)==INTEGER ||(t)==BIGINT)

@d Max(a,b)  ((a)>=(b)?(a):(b))
@d Min(a,b)  ((a)<=(b)?(a):(b))

@d mul1(num,d)	mul1add(num,(digit)(d),(digit)0)

@ The macro |private_pol| is used to obtain exclusive access to the exponents
of a polynomial (but not its coefficients), under the assumption that anyone
else having access has set |isshared|: it will take a copy of the polynomial
unless the reference count is~|0|.

@d private_pol(p)  ( isshared(p) ? copypoly(p) : p )


@* External definitions.
Some variables are reserved for inter-module communication; their |external|
declarations follow the preprocessor directives in \.{lie.h}.

@s bigint int
@s intcel int
@s vector int
@s matrix int
@s poly int
@s symblst int
@s labeltp int

@< External declarations @>=

extern unsigned long  maxnodes, maxptrs, gccrit, maxenters, maxlabels;

extern strtype seq_name, if_name, dollar_name,
assign_name, assign_loc_name, break_name, return_name, block_name,
setdefault_name;

extern bigint* (*int2bin) (intcel*);
extern intcel* (*bin2int) (bigint*);
extern matrix* (*pol2mat) (poly*);
extern poly* (*mat2pol) (matrix*)
          ,* (*vec2pol) (vector*)
          ,* (*bin2pol) (bigint*)
          ,* (*int2pol) (intcel*); 
extern symblst top_definitions,topsym;

extern bigint *one, *null, *minus_one;
extern intcel *bool_false, *bool_true;

extern boolean isargument, check_return;
extern int nstatic1,nstatic2,nstatic3,nstatic4,nstatic5,
    nstatic6,nstatic7;
extern Symbrec static1[],static2[],static3[],static4[],static5[],
    static6[],static7[];
extern boolean am_monitor, prompt, verbose, runtime, gc_set, bigint_set,
       lprint, parsing;
extern FILE *stderr_out, *monfile;
extern boolean with_Pre_on;
extern int tree_pt, object_pt, block_depth, lex_depth, label_pt, lmargin;

extern labeltp label, label_null; /* needed for error messages . */

extern strtype fun_name;

extern int line; /* line number needed for error messages */

extern index nrefl;

extern boolean alloc_gc;
  /* whether to use |allocmem| rather than |mlalloc| in |creatsym| */



extern nametype var[];
extern objtype return_stack[];

extern object grp, defaultgrp,	repair_obj;

extern cmpfn_tp cmpfn; /* needed to do comparisons */


@* Macros for interfacing functions.
We need to make compiled functions available to the user of \LiE. This is done
by placing certain entries referring to those functions in a table; these
entries are created by a variety of macros, depending on the signature of the
function.

@ The macro call
|C@t${}_i$@>(name,f,result_type,arg_type1,@t\dots@>,arg\_type@t${}_i$@>)|
declares a function with external name |name| (a string) that has $i$ objects
as arguments, and which will call the function |f| (an actual \Cee-function).
The remaining arguments are symbolic indications of the types the result and
argument objects will be supposed to have.

@d C0(name,f,res) {res,OPERATOR,name,1,f,0,0},@[@]

@d C1(name,f,res,arg1) {arg1,DUMMY,NULL,0,NULL,0,0}, {res,OPERATOR,name,2,f,-1,0},@[@] 

@d C2(name,f,res,arg1,arg2)
{arg1,DUMMY,NULL,0,NULL,0,1}, {arg2,DUMMY,NULL,0,NULL,0,0},
{res,OPERATOR,name,3,f,-2,0},@[@]

@d C3(name,f,res,arg1,arg2,arg3)
{arg1,DUMMY,NULL,0,NULL,0,1}, {arg2,DUMMY,NULL,0,NULL,0,1},
{arg3,DUMMY,NULL,0,NULL,0,0}, {res,OPERATOR,name,4,f,-3,0},@[@]

@d C4(name,f,res,arg1,arg2,arg3,arg4)
{arg1,DUMMY,NULL,0,NULL,0,1}, {arg2,DUMMY,NULL,0,NULL,0,1},
{arg3,DUMMY,NULL,0,NULL,0,1}, {arg4,DUMMY,NULL,0,NULL,0,0},
{res,OPERATOR,name,5,f,-4,0},@[@]

@d C5(name,f,res,arg1,arg2,arg3,arg4,arg5)
{arg1,DUMMY,NULL,0,NULL,0,1}, {arg2,DUMMY,NULL,0,NULL,0,1},
{arg3,DUMMY,NULL,0,NULL,0,1}, {arg4,DUMMY,NULL,0,NULL,0,1},
{arg5,DUMMY,NULL,0,NULL,0,0}, {res,OPERATOR,name,6,f,-5,0},@[@]

@ The macros $M_i$ are similar to $C_i$, but are used to install constructions
of type |MAP|, i.e., programming language constructs (for instance loops) that
do not have the usual evaluation rules of their arguments as functions. Here
it is not uncommon that result and argument types are specified vaguely as
|ARGTYPE|.

@d
M0(name,f,res) {res,MAP,name,1,f,0,0},@[@]

@d M1(name,f,res,arg1)
{arg1,DUMMY,NULL,0,NULL,0,0}, {res,MAP,name,2,f,-1,0},@[@]

@d M2(name,f,res,arg1,arg2)  
{arg1,DUMMY,NULL,0,NULL,0,1}, {arg2,DUMMY,NULL,0,NULL,0,0},
{res,MAP,name,3,f,-2,0},@[@]

@d M3(name,f,res,arg1,arg2,arg3)
{arg1,DUMMY,NULL,0,NULL,0,1}, {arg2,DUMMY,NULL,0,NULL,0,1},
{arg3,DUMMY,NULL,0,NULL,0,0}, {res,MAP,name,4,f,-3,0},@[@]

@d M4(name,f,res,arg1,arg2,arg3,arg4)
{arg1,DUMMY,NULL,0,NULL,0,1}, {arg2,DUMMY,NULL,0,NULL,0,1},
{arg3,DUMMY,NULL,0,NULL,0,1}, {arg4,DUMMY,NULL,0,NULL,0,0},
{res,MAP,name,5,f,-4,0},@[@]

@d M5(name,f,res,arg1,arg2,arg3,arg4,arg5)
{arg1,DUMMY,NULL,0,NULL,0,1}, {arg2,DUMMY,NULL,0,NULL,0,1},
{arg3,DUMMY,NULL,0,NULL,0,1}, {arg4,DUMMY,NULL,0,NULL,0,1},
{arg5,DUMMY,NULL,0,NULL,0,0}, {res,MAP,name,6,f,-5,0},@[@]

@ Finally there are the macros $A_i$; these are invoke similarly to $C_i$ and
$M_i$, but do not produce table entries, but rather code that will dynamcally
install a function into the tables.

@d A0(name,f,res)
 push(top_definitions,creatopsym(0,match(name,false),f,res));
@d A1(name,f,res,arg1)
 push(top_definitions,creatopsym(1,match(name,false),f,res,arg1));
@d A2(name,f,res,arg1,arg2)
 push(top_definitions,creatopsym(2,match(name,false),f,res,arg1,arg2));
@d A3(name,f,res,arg1,arg2,arg3)
 push(top_definitions,creatopsym(3,match(name,false),f,res,arg1,arg2,arg3));
@d A4(name,f,res,arg1,arg2,arg3,arg4)
 push(top_definitions
 ,creatopsym(4,match(name,false),f,res,arg1,arg2,arg3,arg4));
@d A5(name,f,res,arg1,arg2,arg3,arg4,arg5)
 push(top_definitions
 ,creatopsym(5,match(name,false),f,res,arg1,arg2,arg3,arg4,arg5));

@d FINISH {0,DUMMY,NULL,0,NULL,0,0}


@* Non-ANSI functions.
Most of \LiE. is written in \caps{ANSI}-\Cee, and should be compiled in a
manner that avoids invoking any additional declarations to those allowed for
in the \caps{ANSI} standard (this is often achieved by supplying an argument
`\.{-ansi}' to the \Cee-compiler); this is necessary in order to avoid name
clashes (in particular, the ubiquitous type identifier~|index| would often
cause problems). Some functionality cannot be achieved however within the
limits of pure \caps{ANSI}-\Cee; this functionality is mostly independent of
the computational functions of \LiE.. We collect these functions in the
file \.{lie.c}, which does \emph{not} declare the standard identifiers used
elsewhere in \LiE., and which should be compiled with more liberal compiler
settings (or otherwise modified to obtain the desired functionality). The
functions defined here assume a \caps{POSIX} environment.

@( non-ANSI.c @>=
#include <unistd.h>  /* needed for |isatty| */
#include <stdio.h> /* needed for |FILE| */

@ We now define the system dependent functions. The function |no_terminal|
determines whether its argument refers to a file that is not attached to a
terminal. The function |sysinit| is a hook for systems where some global
initialisation is required before \LiE. can run properly.

@< External declarations @>=
int no_terminal(FILE* f);

@~@( non-ANSI.c @>=
int no_terminal(FILE* f) {@; return !isatty(fileno(f)); }

void sysinit(void) @+ {}
