\input LiEwebmac

@* Introduction.
This file deals with the functions and data structures used to handle
memory allocation in \LiE..

@h "lie.h"
@c
@< Global variables @>

@ We write two header files \.{memtype.h} and \.{mem.h}, the first
containing all public |typedef| and |enum| declarations, the second
containing all other header information about this module. The reason for
this division is that there are mutual dependencies of prototype declarations
on typedef declarations between this compilation unit and other ones
(notably \.{node}); if all declarations were bundled together, there would
be no way to include the header files so that the necessary typedef
declaration always precede the prototype declarations that require them.

@( memtype.h @>=
@< Public |typedef| and |enum| declarations @>
@~
@( mem.h @>=
@< Public macro definitions @>
@< Prototypes of globally visible functions @>

@*1 Memory tables.
All objects that are dynamically allocated by calls to the memory management
functions are accessible from a single array of |(void*)|~pointers. This table
is itself dynamically allocated so that it can expand, although this only
happens when the user issues the system command \.{maxobjects}. The table is
accessed via the pointer |ptr| and its current size is recorded in |maxptrs|.

For garbage collection purposes there is also a dynamic boolean array
|marked| of the same size as |ptr|, and a variable |gccrit| that gives
the threshold value for the number of objects, above which the garbage
collector will be invoked (but only at convenient times, when \LiE. is not
inside a built-in function).

@< Global variables @>=
long chunks = 0; /* number of objects currently allocated */
static void **ptr;
unsigned long maxptrs=0; /* initialised elsewhere */

static boolean *marked;
unsigned long gccrit;

@ The number |hash_mod| is used for hashing purposes. The hash value of a
pointer |p| is computed by converting it to an unsigned long number, and
simply taking it modulo |hash_mod|. This works best if |hash_mod| is odd.

@d hash(p) ((unsigned long)(p)%hash_mod)
@< Global variables @>=
static unsigned long hash_mod; /* a number slighty less than |maxptrs| */

@ Here is a function that initialises the tables after |maxptrs| has been
set to an appropriate value.

@< Prototypes... @>=
void initmem(void);

@~The value of |GCCRIT| is a system dependent constant that determines how
much |gccrit| will be less than |maxptrs|, in other words how much buffer is
left for memory allocation between potential garbage collecting
opportunities.

@c
void initmem(void)
{ long i;
  ptr = (void**) malloc(sizeof(void*) * maxptrs);
  if (ptr==NULL) fatal("Insufficient memory to allocate object table.\n");
  marked = (boolean*) malloc(sizeof(boolean) * maxptrs);
  if (marked==NULL) fatal("Insufficient memory to allocate mark table.\n");
  for (i=0; i<maxptrs; i++) {@; ptr[i]=NULL; marked[i]=false; }
  gccrit=maxptrs-GCCRIT;
  hash_mod = (maxptrs-2)|0x1; /* slightly less than |maxptrs|, and odd */
}

@ Here is a similar function that is called in order to alter the value
of~|maxptrs|.

@< Prototypes... @>=
void newmem(long newval);

@~We allocate new tables by calling |initmem|, and then copy the information
from the old tables.

@c
void newmem(long newval)
{ unsigned long maxptrs0=maxptrs;
  void** ptr0=ptr; boolean* marked0=marked; /* handles for old values */
  if ((maxptrs=newval)<=GCCRIT)
  @< Protest that |maxptrs| would get to low, and reset it to |maxptrs0| @>
  initmem();
  @< Copy information into the new tables, rehashing @>
  if (!redirected_input)
    Printf("New object table of size %ld.\n",(long)maxptrs);
  free(ptr0); free(marked0); /* release the old tables */
}


@
@< Protest that |maxptrs| would get to low... @>=
{ maxptrs=maxptrs0;
  error("You can't lower maxobjects from %ld to %ld.\n"
       ,(long)maxptrs,(long)newval);
}

@ The call to |initmem| has changed |hash_mod|, so we must recompute the
hash codes when copying pointers into the new table.

@< Copy information into the new tables, rehashing @>=
{ long k;
  for (k=0; k<maxptrs0; k++)
    if (ptr0[k]!=NULL) /* copy all non-null pointers */
    { long h; @+ @< Find an empty slot |ptr[h]| to put |ptr0[k]| in @>
      ptr[h]=ptr0[k]; /* copy pointer to empty slot */
      marked[h]=false; /* make new pointer unmarked */
    }
}

@ Since we might have |maxptrs<maxptrs0|, i.e., the table may have been
made smaller, we have to take account of the possibility that the new
table is too small to accomodate all stored pointers, in which case we
reinstall the old situation and abandon |newmem|.

@< Find an empty slot |ptr[h]|... @>=
{ long i; h=hash(ptr0[k]);
  for (i=0; i<maxptrs; i++) /* find an empty slot; try |maxptrs| times */
    if (ptr[h]==NULL) break; /* found */
    else if (++h>=maxptrs) h=0; /* try next slot, wrapping around */
  if (i==maxptrs)
  { free(ptr); ptr=ptr0; free(marked); marked=marked0;
    maxptrs=maxptrs0; /* reset to old values */
    error("You currently cannot decrease 'maxobjects' below %ld.\n"
         ,chunks);
  }
}

@ In all places when we need to look up a pointer~|p| in the table, it should
always be present if everything is in order; to trap programming errors we
should however consider the possibility that it is not (e.g., a pointer is
being looked up after it has been removed from the table). Two variants of the
lookup function will behave differently in such a situation: |findaddr0| will
return a negative index, while |findaddr| will abort the program after
printing a description of the offender, so that the error may be more easily
found in a post-mortem analysis. In fact we cannot be sure the |p| points to
an object, or even that it did at the time it was inserted into |ptr|, since
it might point to a node (see |creatsym|, |store_function_body|); however it
is our best bet and we are going to die anyway.

@< Prototypes... @>=
long findaddr0(void* p);
long findaddr(void* p);
@~@c
long findaddr0(void* p)
{ if (p==NULL) return -1;
  @< Look up |p|: if an |i| with |ptr[i]==p| exists, |return| it @>
  return -1;
}

long findaddr(void* p)
{ if (p!=NULL)
  @< Look up |p|... @>
  printobjectinfo((object) p);
     /* try to give a description of this stranger */
  fatal(" findaddr: called with unknown address %p %s\n", p, type_tag(p));
  return -1;
}

@ To look up a pointer~|p| in the table we linearly search through the
whole table |ptr|, but we start at index |hash(p)|, which is also where
we started looking for a place to store~|p|, so there is a good chance
that it will be found soon. Note that this technique allows pointers to
be removed from the table without further administration, and without
risking that another pointer will fail to be found later due to the empty
slot left behind; the table might only be organised slightly less than
optimal after the removal.

@< Look up |p|... @>=
{ long i, h = hash(p);
  for (i=0; i<maxptrs; i++)
    if (ptr[h]==p) return h; @+ else if (++h>=maxptrs) h=0;
}

@*1 General object layout.
There is some information that is stored with each individual object,
regardless of its type. This includes an indication |type| of that object
type, and a reference count |nref|. The two other fields |fname| and |line|
are optionally used to record the source file and line that created the
object; they can be stripped away in production versions of \LiE..

@< Common fields @>=
objtype type; reftype nref; int line; char *fname;

@ The least significant byte of the |type| field is one of the following
enumeration constants; the remaining bits are flags. The enumeration values
make it possible to test for related sets like |{INTEGER,VECTOR,MATRIX}| and
|{SIMPGRP,GROUP}| in one masking operation, but this is nowhere used.

@< Public |typedef|... @>=

typedef unsigned short objtype;
typedef unsigned short reftype;
@)
enum
@/{ UNKNOWN = 0x00
@/, TEKST   = 0x11 /* character based type */
  , INTEGER = 0x02, VECTOR  = 0x12, MATRIX  = 0x32 /* integer based types */
  , BIGINT  = 0x03
  , POLY    = 0x14
@/, SIMPGRP = 0x06, GROUP   = 0x16 /* group based types */
  , ERROR   = 0x08
  , ARGTYPE = 0x09
  , VOID    = 0x0A
@/};

@ The type of an object is obtained by invoking the macro |type_of|. The use
of the remaining bits of the |type| field is minimal: there is one bit that,
if set, records that the rows of a matrix or polynomial are known to be
sorted.

@< Public macro... @>=

#define @! TYPEMASK 0xFF
#define @! type_of(obj) (((object)(obj))->any.type & TYPEMASK)
#define @! type_tag(obj) (code_obj_type(type_of(obj)))
@)
#define @! SORTED    0x800   /* whether polynomial or matrix is sorted */
#define @! issorted(obj)   ((((object)(obj))->any.type & SORTED)!=0)
#define @! setsorted(obj)  (((object)(obj))->any.type |= SORTED)
#define @! clrsorted(obj)  (((object)(obj))->any.type &= ~SORTED)

@ The types above have 3-letter representations in \LiE., and the following
table defines the correspondence. Not all types occur in the table, because
not all of them are visible to the user.

@c
static struct {@; objtype num; char str[4];} type_name [] =
{ { UNKNOWN, "unk" }
, { TEKST, "tex" }
, { INTEGER, "int"  }
, { VECTOR, "vec" }
, { MATRIX, "mat" }
, { BIGINT, "bin" }
, { POLY, "pol" }
, { GROUP, "grp" }
, { VOID, "vid" }
, { ARGTYPE, "arg" }
, { GRPDFT, "(g)" }
};

@ The following functions are useful for working with type codes.

@< Prototypes... @>=
objtype type_code(char* name);
char* code_obj_type(objtype t);
boolean real_type(objtype type);

@ The function |type_code| converts string describing a type into the
internal representation of the type.

@c
objtype type_code(char* name)
{ int i;
  for (i=0; i<array_size(type_name); ++i)
    if (strcmp(name,type_name[i].str)==0) return type_name[i].num;
  return ERROR;
}

@ Here is the inverse function |code_obj_type| of |type_code|. It is used by
the function |code_type| define in \.{mem.w} that deals with a more extensive
range of possibilities, used in the langauge interpreter; therefore, if it
does not recognise the type number~|t|, it returns |NULL|, indicating that
further search is required.

@c
char* code_obj_type(objtype t)
{ int i;
  for (i=0; i<array_size(type_name); ++i)
    if (t==type_name[i].num) return type_name[i].str;
  return NULL;
}

@ The predicate |real_type| holds only for true object types, not for
|ERROR| or |VOID|.

@c
boolean real_type(objtype type)
{ return type==INTEGER || type==VECTOR || type==MATRIX
      || type==GROUP || type==BIGINT || type==TEKST || type==POLY;
}

@ The reference count is raised by invoking |setshared| and lowered by
invoking |clrshared| (these names are definitely ill-chosen). A special
value |LONGLIFE| is used for persistent objects; the macros will not change
the reference count of such objects (in fact, if any object is shared
sufficiently often, it will become persistent). To see if the reference
count is non-zero, we can test |isshared| (this actually returns the
reference count itself). For some unclear reason these macros are made to
work safely when given null pointers; we do not know any invocations where
the argument could actually be null.

@< Public macro... @>=

#define @! LONGLIFE USHRT_MAX /* ref count that marks value as persistent */

#define @! refcount(x)   ((object)(x))->any.nref
#define @! isshared(x)   ((x)==NULL ? 0 : refcount(x))
#define @! setshared(x)  ((x)==NULL || refcount(x)==LONGLIFE ? 0 : ++refcount(x))
#define @! clrshared(x)  ((x)==NULL || refcount(x)==LONGLIFE ? 0 : \
                  refcount(x)==0 ? share_error((object)(x)) : --refcount(x))

#define @! setlonglife(x) (refcount(x)=LONGLIFE)
#define @! islonglife(x)  ((x)!=NULL && refcount(x)==LONGLIFE)

@* Allocation and deallocation functions.
The functions that do the actual allocation and deallocation of objects are
called |allocmem| and |freem|, whose interface is like that of |malloc| and
|free|.

@< Prototypes... @>=
void* safe_alloc(size_t size);
void* allocmem(size_t size);
void freem(void* addr);

@ If one needs a local array of dynamically determined size, one can use
|alloc_array| for allocation and |freearr| for deallocation. For the
allocation of complete \LiE.-objects, type-specific allocation functions
will be given that call |allocmem|; such objects are deallocated using
|freemem| (which is a no-op if the reference-count is positive), or by
|freepol| in the case of polynomials (which takes care of freeing the
coefficients as well). Note that |freemem| does not invoke |clrshared|: the
convention is that |freemem| is either called for a local object for which
|setshared| was never invoked (so that the reference count is always~0), or
immediately after |clrshared| is invoked, to free the object if no other
references exist. Between a |setshared| and a matching |clrshared|, the
object is guaranteed not to be freed.

@f type_arg int
@< Public macro... @>=

#define @! alloc_array(@[type_arg@],size) \
  ((type_arg*)safe_alloc((size)*sizeof(type_arg)))
#define @! freearr(addr) (free(addr))
@)
#define @! freemem(addr) (refcount(addr)==0 ? freem(addr),0 : 0)
#define @! freepol(addr) (refcount(addr)==0 ? freep(addr),0 : 0)

@ Here are is the main allocation functions |safe_alloc| and |allocmem|.
@c
void* safe_alloc(size_t size)
{ void* result = malloc(size);
  if (result==NULL)
    error("Can't allocate %lu more bytes\n", (unsigned long)size);
  return result;
}

void* allocmem(size_t size)
{ void* result = safe_alloc(size);
  { long h; @+
    @< Set |h| to the index of an empty slot to store |result| in @>
    ignore_intr(); /* don't interrupt while updating |ptr| */
    ptr[h]=result; ++chunks; allow_intr();
  }
  return result;
}

@ An empty slot to store |result| in is found by a simple linear search,
starting at |hash(result)| and wrapped around. 
@< Set |h| to the index of an empty slot... @>=  
{ long i;
  h=hash(result);
  for (i=0; i<maxptrs; i++)
    if (ptr[h]==NULL) break; @+ else if (++h>=maxptrs) h=0;
  if (i==maxptrs)
    error("Object table overflow (%ld). Try increasing 'maxobjects'.\n"
         ,chunks);
}

@ Here is the counterpart |freem| of |allocmem|.

@c
void freem(void* addr)
{ long i=findaddr(addr); /* locate the pointer; it should be present */
  ignore_intr(); 
  ptr[i]=NULL; /* remove pointer from the table */
  --chunks;
  free(addr); /* the actual release of the memory */
  allow_intr();
}

@ Polynomials are freed by a special function |freep| because the reference
counts of their coefficients must be lowered, and the coefficients
themselves freed if necessary.

@< Prototypes... @>=
void freep(poly* addr);
@~@c
void freep(poly* addr)
{ index j;
  for (j=0; j<addr->nrows; j++)
  @/{@; object c=(object) addr->coef[j];
    assert(isshared(c)); clrshared(c); freemem(c);
  }
  freemem(addr);
}

@* The object types.
So far we have dealed with aspects of objects that is independent of the
actual composition of the various object types (except for the fact that
polynomials contain bigints, which was used for freeing them). Now we shall
deal with the individual object types. However, before we do that, there is
one type that is not really an object type, but is nevertheless used
frequently, namely the (dynamic) integer array that is not packed into a
|vector|. Such arrays are for strictly local use: they are not entered into
the memory tables, and have no reference counts or other common attributes
associated with objects. The allocation and deallocation with |alloc_array|
and |freearr| respectively directly uses the basic memory management
routines, circumventing those of~\LiE..

@< Prototypes... @>=
entry* mkintarray(index n);

@~We impose a limit |max_obj_size| on the number of bytes in any single
object, assuming that this is the maximal size the hardware can handle.

@c
entry* mkintarray(index n)
{ if (n>max_obj_size/sizeof(entry))
    error("Cannot create internal array of %ld entries", (long)n);
  return alloc_array(@[entry@],n);
}

@ Proper objects, like integers, are created by special function calls such
as |mkintcel|. For debugging purposes, we wish to dynamically record the
instance of such a call by which each object was created. To that end, we
define a pair of extra arguments, whose names are always |line| and |file|,
and whose actual arguments are given by the special macros |__LINE__| and
|__FILE__|, which expand to the line number and file name at the point of
call. To avoid having to write these explicitly all the time, we define
functions like |mkintcel| also as macros, whose expansion calls the function
(remember that macros are never recursively invoked in~\Cee) with the
indicated extra arguments. As a consequence we must parenthesise these
function names when they occur in a declaration or heading of the function
definition, in order to prevent macro expansion in these occurrences. Also,
we shall produce declarations of the formal arguments |line| and |file| via
a macro |with_line_and_file|, so that they can be supressed by changing the
definition of this macro; the definitions of the other macros below should
then be changed accordingly. Inside the function definition we invoke
|set_common_fields(x,type)| to initialise the fields common to all object
types; among other things this stores away the |line| and |file| arguments.

@d set_common_fields(x,t)
      ((x)->type=t,(x)->nref=0,(x)->line=line,(x)->fname=file)

@< Public macro... @>=
#define @! with_line_and_file @[@], int line, char* file

#define @! mkintcel(a) mkintcel(a, __LINE__, __FILE__)
#define @! mkbigint(a) mkbigint(a, __LINE__, __FILE__)
#define @! mkvector(a) mkvector(a, __LINE__, __FILE__)
#define @! mkmatrix(a,b) mkmatrix(a,b, __LINE__, __FILE__)
#define @! mkpoly(a,b) mkpoly(a,b, __LINE__, __FILE__)
#define @! mkgroup(a) mkgroup(a, __LINE__, __FILE__)
#define @! mksimpgrp(a,b) mksimpgrp(a,b, __LINE__, __FILE__)
#define @! mktekst(a)  mktekst(a, __LINE__, __FILE__)


@*1 Integers. The simplest object type is \.{int}, which represents a
fixed-size integer (actually, an |entry|), which is packaged into
an~|intcel|. Although their conversion to and from big integers (\.{bin}) is
transparent to the user, they are distinguised within \Cee-functions. The
big integer type |bigint| is the simplest example of a variable size object
type. These are all implemented by a structure containing size information,
and a pointer to the part of variable size. In fact such pointers will
normally point directly after the structure, into memory that was allocated
together with the structure. For |bigint|, the variable size part consists
of an array of ``digits'', where the type |digit| is an unsigned integer type
(so the base of the number system is a convenient power of~2). The field
|data| points to this array, |abs(size)| gives the current number of digits
(which can be~0 in case if the |bigint| value~0), and the sign of |size| is
that of the represented big integer; the |allocsize| field is non-negative
and gives the original allocation size of the object. Thus, if a |bigint|
shrinks, it can grow to again its original size without reallocation.
 

@< Public |typedef|... @>=
typedef struct
{@;
  @< Common fields @>@;
  entry intval;
} intcel;

typedef struct
{@;
  @< Common fields @>@;
  short allocsize, size;
  digit *data;
} bigint;

@ Encapsulating an |entry| into a |intcel| is done by |mkintcel|; big
integers are created by |mkbigint|. Unlike integers, |bigint| objects are
not initialised by the call creating them: calling |mkbigint(n)| will create
room for |n|~``digits'', but leaves these digits themselves undefined.

@< Prototypes... @>=
intcel* (mkintcel)(entry n @+ with_line_and_file);
bigint *(mkbigint)(long size @+ with_line_and_file);

@~The implementation of |mkintcel| is straightforward. The output of
|Printf| and |allocmem| will combine.

@c
intcel* (mkintcel)(entry n @+ with_line_and_file)
{
    intcel *i;
    i = (intcel*) allocmem(sizeof(intcel));
    set_common_fields(i,INTEGER);
    i->intval = n;
    return i;
}

@ As mentioned, memory is allocated simaltaneously for the structure and the
digits. Theoretically there is a slight lack of portability here, since this
only works if the alignment condition for |digit| is no stricter than that
for |bigint|. Since the |bigint| structure ends with a pointer, this is
however very likely to be the case. The address |&result[1]| lies directly
after the structure, and casting it to |(digit*)| is what might be undefined
if we are violating alignment conditions.

@c
bigint* (mkbigint)(long size @+ with_line_and_file)
{
  bigint *result;

  if (size>SHRT_MAX) error("Big integer too big\n");
  result = (bigint*)allocmem(sizeof(bigint)+size*sizeof(digit));
  set_common_fields(result,BIGINT);
  result->allocsize = result->size = size;
  result->data = (digit *)&result[1];
  return result;
}

@ Initialised |bigint| objects can be created by copying an existing one:
|copybigint(src,dst)| copies the number held in |src| into |dst|, or into a
fresh object if |dst| is too small or if |dst==NULL|. Also, if a |bigint|
object must grow beyond its allocated size, it can be reallocated to a
larger space by |extendbigint|.

@< Prototype... @>=
bigint* copybigint(bigint* from, bigint* to);
bigint* extendbigint(bigint* old);

@ The destination of |copybigint| is reserved to hold the new value, so if
it is too small for the value to be copied, it should be deallocated before
a new, larger object is created by |mkbigint|.

@c
bigint* copybigint(bigint* from, bigint* to)
{
  int n = abs(from->size); digit *f, *t;
  if (to==NULL) to = mkbigint(n);
  else if (to->allocsize<n) {@; freemem(to); to=mkbigint(n); }
  to->size = from->size;
  for (f=from->data, t=to->data; n>0; --n) *t++=*f++;
  return to;
}

@ The operations that call |extendbigint| usually do not increase the size
of the object very rapidly (in fact only |mul1add|, which multiplies a
|bigint| by a |digit| and adds another |digit|, uses |extendbigint|).
Therefore we extend the allocated size by small a constant number~|EXTBIG|
of digits.

@d EXTBIG 16
  /* at least one more digit, and few to spare for further multiplications */

@c
bigint *extendbigint(bigint* old)
{ bigint *new;
  if (old->allocsize>SHRT_MAX-EXTBIG) error("Big integer too big\n");
  new = mkbigint(old->allocsize + EXTBIG);
  copybigint(old, new); freemem(old); return new;
}

@*1 Vectors. Vectors, whose type is indicated by \.{vec}, are implemented
similarly to big integers. The |ncomp| field gives the current number of
entries, while the |size| field gives the number that was originally
allocated. The pointer |compon| usually points directly after the |vector|
structure. However, in rare occasions we make it point to an existing
integer array, taking care that no deallocation of the vector is attempted,
thus aliasing the array into a vector for calling routines that require one.

@< Public |typedef|... @>=

typedef struct
{@;
  @< Common fields @>@;
  index ncomp;
  index size;
  entry *compon;
} vector;

@ Vectors can be creade from scratch by |mkvector|, and copied by
|copyvector|.

@< Prototypes... @>=
vector* (mkvector)(index n @+ with_line_and_file);
vector* copyvector(vector *src);

@~Here, like in |mkbigint| we sacrifice a bit of portability to efficiency
of allocation and deallocation.

@c
vector* (mkvector)(index n @+ with_line_and_file)
{
    vector *v;
    if (n > (max_obj_size-sizeof(vector))/sizeof(entry))
	error("Cannot handle a vector with %ld entries.\n",(long)n);
    v=(vector*)allocmem(sizeof(vector)+n*sizeof(entry));
    set_common_fields(v,VECTOR);
    v->size = v->ncomp = n; v->compon = (entry*) &v[1];
    return v;
}

@~
@c

vector* copyvector(vector *src)
{ vector* result=mkvector(src->ncomp);
  copyrow(src->compon,result->compon,src->ncomp);
  return result;
}

@*1 Matrices and polynomials. The types |matrix| and |poly| stand for
matices respectively polynomials, and are indicated in \LiE. by \.{mat}
respectively \.{pol}. They are strongly similar to allow efficient
conversion between the two types, i.e., by reusing the same storage instead
of copying data; these conversions are available to the \LiE. user as the
function \\{support} (from polynomials to matrices) respectively the
operator $X$ (from matrices to polynomials).

The fields |nrows| and |ncols| indicate the number of rows and columns of
the matrix represented by a |matrix| object; for a |poly| they correspond to
the number of terms and the number of exponents within one term. The field
|rowsize| is like the |size| field of vectors: it is for internal use by the
implementation only, and denotes the number of rows in the matrix (or terms
in the polynomial) for which storage was reserved, which may differ from
|nrows| is the matrix has shrunk or does not fully occupy the allocated
space for some other reason. The fields |ncols| and |rowsize| are constant
during the lifetime of the object, but |nrows| may vary from~|0| up to the
maximum |rowsize|. The field |elm| accesses the actual entries (respectively
exponents) through an array of pointers to rows that will be allocated
together with the |matrix| or |poly| structure and the block of entries
thenmselves. The field |coef| of |poly| points to another array of pointers,
namely to the |bigint| coefficients of the terms. In |matrix| objects space
for the same array of pointers is allocated, but it will not be used unless
the |matrix| is converted to a |poly|; the field |null| that matches |coef|
in the layout of |poly| will be set to~|NULL|. Permuting the rows of a
matrix, or the terms of a polynomial, is efficiently possible by permuting
the row pointers accessed via the |elm| field, and in case of polynomials
also the coefficient pointers accessed via the |coef| field.

@< Public |typedef|... @>=

typedef struct
@/{@;
  @< Common fields @>@;
  index nrows, ncols;
  index rowsize;
  entry **elm;
  bigint **null;
} matrix;
@)
typedef struct
@/{@;
  @< Common fields @>@;
  index nrows, ncols;
  index rowsize;
  entry **elm;
  bigint **coef;
} poly;

@ The allocation functions for matrices and polynomials are called (you
guessed it) |mkmatrix| and~|mkpoly|.

@< Prototypes... @>=
matrix* (mkmatrix)(index r,index c @+ with_line_and_file);
poly* (mkpoly)(index r,index c @+ with_line_and_file);

@~As was mentioned above, and similarly to vectors, |matrix| objects in a
single block of memory. Here it is divided into 3~parts: first comes the
|matrix| structure proper, then the array of row pointers, and finally the
block of entries. Like for vectors, the entries themselves are not
initialised.

@c
matrix* (mkmatrix)(index r,index c @+ with_line_and_file)
{
  index i;
  matrix *m;
  size_t size=sizeof(matrix)
    + (r==0 ? 1 : r)*(sizeof(@[bigint*@])+sizeof(@[entry*@]))
    + r*c*sizeof(@[entry@]);

  if (size > max_obj_size)
      error("Cannot handle a %ld by %ld matrix\n",(long)r,(long)c);
  m = (matrix*)allocmem(size);
  set_common_fields(m,MATRIX);
@/m->rowsize = m->nrows = r; m->ncols = c;
  m->null = (bigint**) NULL;
@/
  m->elm=(entry**) &m[1]; /* start of row pointer block */
  m->elm[0]=(entry*) &m->elm[r]; /* start of entry block */
  for (i=1; i<r; ++i) m->elm[i]=&m->elm[i-1][c]; /* remaining row pointers */
  return m;
}

@ The function |mkpoly| is quite similar to |mkmatrix|, but it always
allocates at least one term (with zero coefficient and exponents if 0~terms
were requested) for printing purposes. The array of coefficients comes at
the end of the allocated block.

@c
poly* (mkpoly)(index r,index c @+ with_line_and_file)
{
  index i;
  poly  *p; size_t size;
  boolean  is_null_poly = false;

  if (r == 0) {@; is_null_poly = true; r = 1; } /* avoid empty polynomial */

  size=sizeof(matrix)
    + r*(sizeof(@[bigint*@])+sizeof(@[entry*@])+c*sizeof(@[entry@]));
  
  if (size > max_obj_size)
      error("Cannot handle a %ld by %ld polynomial\n",(long)r,(long)c);
  p = (poly*)allocmem(size);
  set_common_fields(p,POLY);
@/p->rowsize = p->nrows = r; p->ncols = c;
  p->elm=(entry**) &p[1]; /* start of row pointer block */
  p->elm[0]=(entry*) &p->elm[r]; /* start of entry block */
  p->coef = (bigint**) &p->elm[0][r*c]; /* start of coefficient block */
  for (i=1; i<r; ++i)
  {@; p->elm[i]=&p->elm[i-1][c];
    p->coef[i]=NULL;
  }
  if (is_null_poly)
  {@; p->coef[0] = null;
      for (i=0; i<c;i++) p->elm[0][i] = 0;
  }
  return p;
}

@ Like big integers, matrices and polynomials can be copied or extended. In
both cases the value of the copy is the same as that of the original, but in
the case of copying the allocated size is the exact amount needed for the
current value, whereas in the case of extension it is strictly larger than
the allocated size of the original object (the increase in size is about
50\%). Since the latter case is used only to create space for an object to
grow, it automatically frees the original object.

@< Prototypes... @>=
matrix* copymatrix(matrix* old);
poly* copypoly(poly* old);
matrix* extendmat(matrix* old);
poly* extendpoly(poly* old);

@ Copying of matrices is straightforward. For polynomials the coefficients
of the copy will be shared with those of the original.

@c

matrix* copymatrix(matrix* old)
{ index i; matrix* new = mkmatrix(old->nrows,old->ncols);
  for (i=0; i<old->nrows; ++i) copyrow(old->elm[i],new->elm[i],old->ncols);
  return new;
}

poly* copypoly(poly* old)
{ index i; poly* new = mkpoly(old->nrows,old->ncols);
  for (i=0; i<old->nrows; ++i)
  { new->coef[i]=old->coef[i], setshared(new->coef[i]);
    copyrow(old->elm[i],new->elm[i],old->ncols);
  }
  return new;
}

@ When calling |extendmat| or |extendpoly| one should be aware that although
the previously allocated size |rowsize| is used to determine the size of the
object after extension, only |nrows| rows (or terms) of the old object will be
copied into the new object; while it is normally accepatable that |nrows|
differs temporarily from the number of rows (terms) in actual use, it should
be properly updated before extending the object. For extension of polynomials
note that reference counts are increased for the coefficients by |setshared|,
and then decreased by |freepol| unless the polynomial is shared.

@c

matrix* extendmat(matrix* old)
{ index i;
  matrix* new= mkmatrix(3*old->rowsize/2+1, old->ncols);
  for (i=0; i<old->nrows; ++i) copyrow(old->elm[i],new->elm[i],old->ncols);
  new->nrows=old->nrows;
  freemem(old); return new;
}

poly* extendpoly(poly* old)
{ index i; poly* new=mkpoly(3*old->rowsize/2+1,old->ncols);
  for (i=0; i<old->nrows; ++i)
  { new->coef[i]=old->coef[i],setshared(new->coef[i]);
    copyrow(old->elm[i],new->elm[i],old->ncols);
  }
  new->nrows=old->nrows;
  freepol(old); return new;
}

@*1 Groups.
There are two types for groups: |simpgrp| for simple groups, and |group| for
composite groups. Simple groups are treated in a somewhat special manner,
since information about the simple group obtained during previous
computations is stored in the |simpgrp| object, to avoid excessive
recomputations; for the same reason, simple groups are exempt from garbage
collection. The fields |lietype| and |lierank| uniquely characterise the
type of the simple group (and there should be not more than one object
corresponding to any type of simple group), while the remaining fields are
used to store additional data whenever it has been determined: |cartan| and
|icartan| are the Cartan and inverse Cartan (up to a scalar) matrices,
and |roots| to the set of positive roots, listed as a matrix; |exponents| is
the list of exponents of the group, |root_norm| a vector of norms of the
positive roots, |level| a vector of indices at which the levels of positive
roots start. All simple groups are chained via |nextgrp|.

Composite groups are essentially a vector of simple groups, together with
an additional natural number~|toraldim|, the dimension of the central torus.
No precomputed data is stored at the level of composite groups.

@< Public |typedef|... @>=

typedef struct simpgrp_struct simpgrp;
struct simpgrp_struct
{
  @< Common fields @>@;
  char lietype;
  entry lierank;
  matrix* cartan,* icartan,* roots;
  vector* exponents,* level,* root_norm;
  simpgrp* nextgrp;
};
@)
typedef struct
{@;
  @< Common fields @>@;
  index ncomp, toraldim;
  simpgrp **liecomp;
} group;

@ The following macro facilitates selecting a simple component from a group.
The argument |grp| is supposed to have type |object|; thus the macro matches
functions like |Lierank| and |Ssrank|.

@< Public macro... @>=

#define @! Liecomp(grp,i) ((grp)->g.liecomp[i])

@ The head of the chain of simple groups is |simpgrplist|; it starts out
empty.

@< Global var... @>=

static simpgrp* simpgrplist=NULL;

@ Like other object types, simple groups have an allocation function
|mksimpgrp|.

@< Prototypes... @>=
simpgrp* (mksimpgrp)(char type, index rank @+ with_line_and_file);

@~The main difference between |mksimpgrp| and other allocation functions is
that it will not create a new object if a group of the same type and rank is
already present in |simpgrplist|; if this is not the case, then the new
|simpgrp|~object is linked into |simpgrplist|. The fields for precomputed
values are initialised to~|NULL|; when their actual value is computed,
|setlonglife| will be invoked for that value, so that it too is exempt from
garbage collection.

@c

simpgrp* (mksimpgrp)(char type, index rank @+ with_line_and_file)
{
  simpgrp *grp, **loc; 
  @< Lookup group in |simpgrplist|; if found, |return| it, otherwise
     point |loc| to final link @>
  grp = (simpgrp *) allocmem(sizeof(simpgrp));
  set_common_fields(grp,SIMPGRP);
  setlonglife(grp); /* simple group will not be garbage collected */
  grp->lietype = type;
  grp->lierank = rank;
  grp->exponents = grp->level = grp->root_norm = NULL;
  grp->cartan = grp->icartan = grp->roots = NULL;
  grp->nextgrp = NULL; /* this group will be last in |simpgrplist| */
  return *loc = grp; /* add group to end of |simpgrplist| and return it */
}

@ Lookup is by linear search; the number of different simple groups
encountered is probably not very long.

@< Lookup group in |simpgrplist|... @>=
{ for (loc=&simpgrplist; *loc!=NULL; loc=&(*loc)->nextgrp)
    if ((*loc)->lietype==type && (*loc)->lierank==rank) return *loc;
}

@ The allocation functions for general groups is |mkgroup|.

@< Prototypes... @>=
group* (mkgroup)(index ncomp @+ with_line_and_file);

@~This function is entirely analogous to |mkvector|, but we do not test for
overflow of |max_obj_size|.

@c
group* (mkgroup)(index ncomp @+ with_line_and_file)
{
    group *grp;
    grp = (group*) allocmem(sizeof(group)+ncomp*sizeof(simpgrp*));
    set_common_fields(grp,GROUP);
    grp->toraldim = 0;
    grp->ncomp = ncomp;
    grp->liecomp = (simpgrp**)(&grp[1]);
    return grp;
}

@*1 Texts.
Finally there are texts, which are like vectors of characters. The spelling
of |tekst| is Dutch.

@< Public |typedef|... @>=

typedef struct
{@;
  @< Common fields @>@;
  index len;
  char *string;
} tekst;

@ Texts too can be created, using |mktekst|, which is just like |mkgroup|.

@< Prototypes... @>=
tekst* (mktekst)(index n @+ with_line_and_file);

@~Although |tekst| objects have a length indication, they also carry a
terminating |'\0'|~byte.
@c
tekst* (mktekst)(index n @+ with_line_and_file)
{
  tekst *t;
  t = (tekst*)allocmem(sizeof(tekst)+n+1);
  set_common_fields(t,TEKST);
  t->len = n;
  t->string = (char *) (&t[1]);
  t->string[n] = '\0';
  return t;
}

@ Texts can be copied using |copytekst|.
@< Prototypes... @>=
tekst* copytekst(tekst* o);
@~@c
tekst* copytekst(tekst* o)
{
  index n = o->len;
  tekst *result = mktekst(n);
  strncpy(result->string,o->string,n);
  return result;
}

@*1 The type `{\bf object}'.
Finally, we can give the definition of |object|. It is a union of all the
object types we have given above, with one extra variant |any|, wich has
only the common fields, and is the preferred way to access those fields.
The other variants have very short selectors, to reduce the size of
selection expressions (often these varaiant selectors will be followed by
field selectors of the selected structure).

@< Public |typedef|... @>=

typedef union {
  struct {@; @< Common fields @>@; @> } any;
  intcel i;
  bigint b;
  vector v;
  matrix m;
  poly	 pl;
  simpgrp s;
  group g;
  tekst	 t;
} objcel, *object;


@* Expressions.
There are a few functions related to expressions in the program tree that
are similar to the memory management functions for objects.  The functions
are associated with the language interpreter; for more details about the
data structures involved, see the translation unit \.{node}. 

@< Prototypes... @>=
object mkobject(symblst s);
object cpobject(object o);

@ The function |mkobject| reduces to |cpobject|, except for |BLOCK|
expressions.

@c
object mkobject(symblst s)
{@; return s->class == BLOCK ? s->data.val : cpobject(s->data.val); }

@ In all cases except that of groups, which we somehow forgot, an existing
copying function can be used.

@c
object cpobject(object o)
{ switch (type_of(o))
  {
  case INTEGER:return (object)mkintcel(o->i.intval);
  case BIGINT: return (object)copybigint(&o->b,NULL);
  case VECTOR: return (object)copyvector(&o->v);
  case MATRIX: return (object)copymatrix(&o->m);
  case POLY:   return (object)copypoly(&o->pl);
  case GROUP:  @< Return copy of group |&o->g| @>
  case TEKST:  return (object)copytekst(&o->t);
  default: error("Illegal result type %d\n",type_of(o)); return NULL;
  }
}

@ This is about what the function |copygroup| would have looked like, had it
been defined.

@< Return copy of group... @>=
{ group* g=&o->g; index i,n=g->ncomp;
  group* result=mkgroup(n);
  result->toraldim=g->toraldim;
  for (i=0; i<n; i++) result->liecomp[i] = g->liecomp[i];
  return (object) result;
}


@* Garbage collection.
The next part of the program deals with garbage collection. The approach to
garbage collection in \LiE. is a hybrid one, with most of the memory being
deallocated explicitly by calls of |freem|, possibly aided by maintenance
of a reference count using |setshared| and |clrshared|, but with a garbage
collector to clear up the bits and pieces that slip through without
deallocation. The most compelling reason why a garbage collector is really
needed is the way the `control-C' interruption of computation is
implemented:  it performs a |longjmp| in the middle of a computation into
the control loop, which skips over any explicit deallocations that were due
after completion of the computation. Given that in practical computations
large amounts of time can be spent inside single built-in function calls
(quite possibly almost the entire execution of a user comand) this is the
only reasonable approach to interrupt handling (as opposed to for instance
checking the interrupt status between built-in function calls), making a
garbage collector inevitable. Since only proper objects (those entered into
the |ptr| table) are garbage collected, there is a small memory leak for
other dynamic data, such as temporary arrays created with |alloc_array|,
that are in existence at the time of interruption.

The garbage collector is called automatically between function calls if the
memory table is almost full, and also before each prompt for a user command
(this can be switched off by \.{off gc} however); it can also be invoked
explicitly by the user. When garbage is collected, the reference counts are
also recomputed (remember that garbage collection never occurs inside
built-in functions, so this does not interfere with manipulations of the
reference count inside those functions). A drawback of our hybrid approach
is that errors in the explicit memory deallocation regime can thus be masked
by the garbage collector.

@*1 Marking functions.
The function |mark_expression| marks all data values in a portion of an
array of |symbrec| values. It is used in |mark_tree| to mark at once all
values in the current expression tree (whose nodes are stored in a single
array), and also from the garbage collector itself to mark the values
contained in stored function definitions (these are mostly numeric constants).

@< Prototypes... @>=
void mark_expression(symblst s,int n);
@~
@c
void mark_expression(symblst s,int n)
{ int i;
  for (i = 0; i < n; i++)
    if (s[i].class == VALUE) mark(s[i].data.val);
}

@ The basic funcion |mark0| sets the |marked| flag for a pointer in the
table, while graciously ignoring the null pointer.

@c
static void mark0(void* addr)
{@; if (addr!=NULL) marked[findaddr(addr)] = true; }

@ The function |markobj| marks its argument as seen for the purpose of
garbage collection; it returns a boolean value telling whether the value had
already been marked previously. The null pointer and system constants such
as |one|, which are recognised by the predicate |islonglife|, are assumed to
be always marked and will never be removed.
@c
static boolean markobj(object obj) /* true if it was already marked */
{ boolean was_marked = false;
  long i;
  if (obj==NULL || islonglife(obj)) return true;
  @< Find an~|i| for which |ptr[i]==obj|;
     for polynomials recursively mark all coefficients @>
  if (marked[i]) was_marked=true,++obj->i.nref ;
  else marked[i]=true,obj->i.nref=1;
  return was_marked;
}

@ Here the object must be present, so we use |findaddr| instead
of~|findaddr0|.

@< Find an~|i| for which |ptr[i]==obj|... @>=
{ i=findaddr(obj);
  if (type_of(obj)==POLY)
  @/{@; long k;
    for (k=0; k< obj->pl.nrows; k++) markobj((object) (obj->pl.coef[k]));
  }
}

@ The function |mark| does a bit more than |markobj|: the first time a
simple group is encountered, its associated objects such as norm vector and
cartan matrix are also marked; for composite groups this is done for all its
simple components.

@< Prototypes... @>=
void mark(object obj);
@~@c
void mark(object obj)
{ if (markobj(obj)) return;
  switch (type_of(obj))
  { case SIMPGRP:
    markobj((object)obj->s.cartan);
    markobj((object)obj->s.icartan);
    markobj((object)obj->s.roots);
    markobj((object)obj->s.exponents);
    markobj((object)obj->s.level);
    markobj((object)obj->s.root_norm);
  break;
    case GROUP:
    {@; int i;
      for (i=0; i<obj->g.ncomp; ++i) mark((object)Liecomp(obj,i));
    }
  }
}

@*1 The garbage collector.
Now the garbage collector proper. It is called with a single argument that
accesses most current objects.

@< Prototypes... @>=
void gc(void);

@~This is a simple mark-scan garbage collector; the scan phase is simple, but
it consumes more time than the marking phase, in particular when |maxptrs|
(as set by the \.{maxobjects} command) is large.

@c

extern symblst topsym;

void gc(void)
{
  long   i;
  strtype name0 = label->name;
  @< Mark all relevant objects @>
  for (i = 0; i<maxptrs; i++) 
    if (marked[i]) marked[i] = false;
    else if (ptr[i]!=NULL && !islonglife(ptr[i])) freem(ptr[i]);
  label->name = name0; /* restore value */
}

@ Marking is a bit more subtle than scanning. We do not have to clear the
|marked| array, since it is cleared at initialisation and after each garbage
collection. The objects marked are the default group and those default
groups stored by a pending \.{save} operation, the current expression
|topsym| and all expessions accessible through the list~|top_definitions|.
Note the the fact that expression nodes are recorded into the |ptr| table
means that we cannot assume an object layout for values poitned to by elements
of~|ptr|.

@< Mark all relevant objects @>=
{ symblst     v;
  symblst last_v=top_definitions;
  label->name = match("garbage_collection",false); /* indicate current activity */
  mark(defaultgrp);
  mark_defaultgrp_stack();
  mark((object) topsym);
  for (v=top_definitions; v!=NULL; v=v->next)
  { mark0(v);
    if (v->class==VALUE) mark(v->data.val);
    else if (v->class == FUNCTION)
    { funclst f = v->data.func;
      mark0(f);
      mark_expression(f->start_nodes, f->n_nodes);
	/* mark stored function body */
    }
    last_v=v;
  }
  mark_tree();
}

@* Miscellaneous functions.
Sometimes we need to apply an operation to all objects. The function
|for_all_objects| makes this possible even in compilation units where the
object table~|ptr| is not visible.

@< Prototypes... @>=
void for_all_objects(void (*f)(object));

@~@c
void for_all_objects(void (*f)(object))
{@; long i;
  for (i=0; i<maxptrs; i++)
    if (ptr[i]!=NULL) f(ptr[i]);
}

@ This function prints out the internal data of a specific object.

@< Prototypes... @>=
void printobjectinfo(object obj);
@~@c
void printobjectinfo(object obj)
{ if (obj!=NULL)
  { string s=type_tag(obj);
    if (s==NULL) Printf("weird object with type %x.\n",type_of(obj));
    else if (type_of(obj)==VOID) Printf("weird VOID object.\n");
    else
      Printf("object defined in line %d of file %s. Type: %s, #refs=%d\n"
	    ,obj->i.line,obj->i.fname,s,refcount(obj));
  }
  return;
}

@ This function is called when we want to free an object that was not shared.

@c
void share_error(object o)
{ printobjectinfo(o);
  fatal("Reference count underflow");
}



@* Index.
