\input LiEwebmac

@* Overview.
This compilation unit handles the operations on the nodes of the computation
tree.

@h "lie.h"

@
@< Prototypes of global variables @>=
extern symbrec blank_symbol;
extern int tree_pt, label_pt;
extern unsigned long maxnodes, maxlabels;
extern labeltp label_null;

@ Like for the compilation unit \.{mem}, we write two header files, called
\.{nodetype.h} respectively \.{node.h} in the current case. The explanation
for this division is given in \.{mem}.

@( nodetype.h @>=
@< Public |typedef| and |enum| declarations @>
@~
@( node.h @>=
@< Prototypes of globally visible functions @>
@< Prototypes of global variables @>
@< Public macro definitions @>

@
@c
@< Static variables definitions. @>
@< Definitions of global variables @>
@< Routines for creation of computation tree @> 
@< Routines to create new nodes of unknown type @>
@< Routines to add and allocate nodes of certain types @>
@< Creation of nodes without objects @>

@
@< Definitions of global variables @>=
int tree_pt=0, label_pt=0;
unsigned long maxnodes=0, maxlabels=0;
labeltp label_null=&labelrec_null;

@*1 Layout of nodes.
We shall now proceed to describe the way nodes are built up. Since nodes are
used to build up program trees, they are naturally a recursive type. Therefore
we predeclare the node type |symbrec| and the |symblst| pointing to a node or
to an array of them, in terms of |struct symbol| that will be specified later.
By convention we shall use |symblst| only for pointers that actually point to
(the first element of) an array of nodes, and use |(symbrec*)| instead when it
points only to a single node. Another basic type is |strtype|, which is used
to represent strings, but is in fact a index into a global array~|name_tab|,
defined in \.{lexer.w}. We also declare the type |fobject| and its more
specific versions here, as it will be needed in the definition of |struct
symbol|.

@< Public |typedef|... @>=
typedef struct symbol symbrec,* symblst;
typedef unsigned int strtype;
typedef short objclass;
@)
typedef object (*fobject)(), /* generic case */
@/  (*f0object) (void),
@/  (*f1object) (object),
@/  (*f2object) (object,object),
@/  (*f3object) (object,object,object),
@/  (*f4object) (object,object,object,object),
@/  (*f5object) (object,object,object,object,object),
@/  (*f6object) (object,object,object,object,object,object);

@ Some nodes will have a label, accessible through a field of type~|labeltp|;
it records a label and the source file and line of the call that created
the node.

@< Public |typedef|... @>=
typedef struct
{@; strtype name;
  int line;
  char *fname;
} *labeltp, labelrec;

@ The |funcrec| structure represents a function body (?).

@< Public |typedef|... @>=
typedef struct
{ int n_nodes;
  int n_labels;
  int n_args;
  symblst start_nodes;
  symblst start_args;
  labeltp start_labels;
  symbrec* top_expr;
  symbrec* top_arg;
} funcrec, *funclst;

@ The type |atp| is a union of some miscellaneous types that can be used to
store auxiliary information in a node. We put the |(char*)| variant in first
position, so that variables of type~|atp| (most often components of
structures) can be statically initialised by a string, and later converted to
the variant |strtype| by a call of~|match| during dynamic initialisation.

@< Public |typedef|... @>=
typedef union {@; char* init_string; strtype name; labeltp label; int size;}
        atp;

@ Finally, here is the structure of |symbrec| nodes themselves. Their |type|
will be one of |{}|, and their |class| indicates what kind of node this is.

@< Public |typedef|... @>=
struct symbol
{ objtype type;
  objclass class;
  atp a;
  strtype formal;
  union {@; object val; fobject f; symbrec* expr; funclst func; } data;
  symbrec* arglst;
  symbrec* next;
};

typedef object (*symbfobj) (symblst); /* function taking node(s) to object */

@ Here is a blank |symbrec| structure, that can be used to initialise
dynamically created ones; the value |empty_str| is the |strtype| value
encoding~|""|.

@< Definitions of global var... @>=
symbrec blank_symbol={UNKNOWN,DUMMY,{empty_str},empty_str,{NULL},NULL,NULL};

@*1 Initialisation routines.

@< Static variables definitions. @>=
static symblst tree_ar;
static labeltp label_ar;
static labelrec labelrec_null= {0,0,NULL}; /* no label or source line */

@ The following functions are similar to |initmem| and |newmem| in \.{mem}.
@< Prototypes of globally visible fun... @>=
void inittree(void), newtree(void);

@~
@< Routines for creation of computation tree @>= 
void inittree(void)
{ if (maxnodes<2) maxnodes=MAXNODES_DFLT;
  maxlabels=maxnodes/2;
@/tree_ar=alloc_array(@[symbrec@],maxnodes);
  label_ar =alloc_array(@[labelrec@],maxlabels);
}

void newtree(void)
{ if (tree_pt!=0) fatal("System error. newtree.\n");
  freearr(tree_ar); freearr(label_ar); inittree();
  Printf("New tree space with maximum number of nodes: %ld.\n",
  (long) maxnodes);
}

@ The function |creatnode| is local to \.{node}. It initialises a |symbrec|
node pointed to by~|s|, if |name!=empty_str| is given it allocates a new
label and inserts |name| together with the current file name and line number
in the user program (these are held in the global variables |lineno| and
|input_fname|) into the label, and sets the remaining fields to blank values.

@< Routines to create new nodes of unknown type @>=
static void creatnode(symbrec* s,strtype name)
{ *s=blank_symbol;
  if (name==empty_str) s->a.label=label_null;
  else
  { if (label_pt >= maxlabels)
      error("Too big program (maxnodes %ld). Try increasing maxnodes.\n"
	   , (long)maxnodes);
    label_ar[label_pt].name=name; /* name of node */
    label_ar[label_pt].line=lineno; label_ar[label_pt].fname=input_fname;
      /* \LiE. source file and line number */
    s->a.label=&label_ar[label_pt++];
  }
}

@ The function |newnode| allocates a new node from |tree_ar|, and installs
its argument~|a| as label.

@< Prototypes of globally visible fun... @>=
symblst newnode(labeltp a);

@~
@d full_error error("Reduce space full.\n"
                    "Try increasing maxnodes (currently %ld).\n"
		   ,(long)maxnodes)
@< Routines to create new nodes of unknown type @>=
symblst	newnode(labeltp a)
{ if (tree_pt>=maxnodes) full_error;
  creatnode(&tree_ar[tree_pt],0);
  tree_ar[tree_pt].a.label =a; /* use label |a| instead of |label_null| */
  return &tree_ar[tree_pt++];
}

@ The function |addnode| allocates a new node, and after initialising with
the argument |name|, attaches the nodes passed as extra arguments to
|arglst|. The extra arguments are linked through their |next| field, but
must be otherwise fully initialised before the call to~|addnode|.

@< Prototypes of globally visible fun... @>=
symblst addnode(int n, strtype name, ...);

@~
@< Routines to create new nodes of unknown type @>=
symblst addnode(int n, strtype name, ...)
  /* extra arguments have type |(symbrec*)| */
{ if (tree_pt >= maxnodes) full_error;
  creatnode(&tree_ar[tree_pt], name);
  if (n>0) /* then attach the arguments */
  { va_list ap; int i; symblst* t=&tree_ar[tree_pt].arglst;
    va_start(ap, name);
    for (i=0; i<n; ++i)
    {@; *t=va_arg(ap,@[symblst@]); t=&(*t)->next; }
    va_end(ap);
  }
  return &tree_ar[tree_pt++];
}

@ Here are some simple routines that make object-like nodes, i.e., nodes for
which |type| lies in |{INTEGER, BIGINT, VECTOR, MATRIX, POLY, SIMPGRP,
GROUP, TEKST, VOID}| (not all cases are covered). These functions call
|mkintcel| and its relatives, and as a rule just pass their argument to that
function; |addtekstnode| is an exception, as it actually initialises the
characters of the text, unlike |mktekst|. The reference count of the objects
contained in these nodes is set to~1 by means of |setshared|, to reflect the
access from the node.

@< Prototypes of globally visible fun... @>=
symblst addintnode(entry n), addbigintnode(bigint* b), addvecnode(int n)
 , addgroupnode(char,index), addtekstnode(char* t);

@~
@< Routines to add and allocate nodes of certain types @>=
symblst addintnode(entry n)
{ symblst s=addnode(0, empty_str);
  s->class=VALUE;
  s->type=INTEGER;
  s->data.val=(object)mkintcel(n);
  setshared(s->data.val);
  return s;
}

symblst addvecnode(int n)
{ symblst s=addnode(0, empty_str);
  s->class=VALUE;
  s->type=VECTOR;
  s->data.val=(object)mkvector(n);
  setshared(s->data.val);
  return s;
}

symblst addbigintnode(bigint* b)
{ symblst s=addnode(0, empty_str);
  s->class=VALUE;
  s->type=BIGINT;
  s->data.val=(object)b;
  setshared(s->data.val);
  return s;
}

symblst addgroupnode(char lietype,index rank)
{ symblst s=addnode(0,empty_str); s->type=GROUP; s->class=VALUE;
  if (lietype=='T')
  { s->data.val=(object) mkgroup((index) 0); s->data.val->g.toraldim=rank; }
  else
  { s->data.val=(object) mkgroup((index) 1);
    Liecomp(s->data.val,0)=mksimpgrp(lietype, rank);
  }
  setshared(s->data.val); return s;
}

symblst addtekstnode(char* t)
{ symblst s=addnode(0, empty_str);
  s->class=VALUE;
  s->type=TEKST;
  s->data.val=(object)mktekst(strlen(t));
  setshared(s->data.val);
  strcpy(s->data.val->t.string, t);
  return s;
}

@ The following two routines build nodes with |VOID| value (we have
|s->data.val==NULL| from |creatnode|), and given name or label.

@< Prototypes of globally visible fun... @>=
symblst addnamenode(strtype name), addlabelnode(labeltp l)
  , addtypnode(objtype typno),addtypnamenode(objtype typno, strtype name)
  , addprogramnode(symblst prog);
void addstaynode(object o);
@~
@< Routines to add and allocate... @>=
symblst addnamenode(strtype name)
{@; symblst s=addnode(0,name);
  s->class=VALUE;
  s->type=VOID;
  return s;
}

symblst addlabelnode(labeltp l)
{@; symblst s=newnode(l);
  s->class=VALUE;
  s->type=VOID;
  return s;
}

@ Add dummy class nodes. There are no objects created.
@< Creation of nodes without objects @>=
symblst addtypnode(objtype typno)
{@; symblst s=addnode(0, empty_str);
  s->class=DUMMY;
  s->type=typno;
  return s;
}

symblst addtypnamenode(objtype typno, strtype name)
{@; symblst s=addnode(0, name);
  s->class=DUMMY;
  s->type=typno;
  return s;
}

symblst addprogramnode(symblst prog)
{@; symblst s=addnode(1, match(".sequence",false),prog); return s; }

void addstaynode(object o)
{@; symblst s=newnode(NULL);
  s->class=VALUE;
  s->type=type_of(o);
  s->data.val=o;
}

@* Type and class encoding.
The object types are enumerated in the compilation unit \.{mem}.
For expressions, the set of types is extended in various ways.

@< Public |typedef|... @>=
enum
{ FUN     =0x0200   /* whether expression has function type */
, DEFAULT =0x0400   /* whether final argument defaults to group */ 
, ARR     =0x0800   /* Dummy list variable length arg */
};

@ Function types.
  Needed for passing function arguments in repeat, vapply.

@< Public |typedef|... @>=

enum
{ INTFUN=INTEGER|FUN
, VECFUN=VECTOR|FUN
, MATFUN=MATRIX|FUN
, VIDFUN=VOID|FUN
, INTLST=INTEGER|ARR
, VECLST=VECTOR|ARR
, MATLST=MATRIX|ARR
, VIDLST=VOID|ARR
, ARGLST=ARGTYPE|ARR
, GRPDFT=GROUP|DEFAULT
};

@ Here are some macros to manipulate the masks above.

@< Public macro... @>=
#define @! isarrtype(sym)  (((sym)->type&ARR)!=0)
#define @! isfuntype(sym)  (((sym)->type&FUN)!=0)
#define @! isdefault(sym)  ((sym)->next==NULL && ((sym)->type&DEFAULT)!=0)

@ Here is an extension of the function |code_obj_type|, that will in
addition to the types recognised by that function also return for instance
|"fun vec"| for~|VECFUN|, and |"arl"| for~|ARGLST|.

@< Prototypes of globally visible fun... @>=
char* code_type(objtype typeno),* code_class(int classno);

@~
The result is constructed in a small buffer of 8 characters, which is
statically allocated. To allow two calls to |code_type| to appear in one
expression (e.g., as arguments to |printf|), a trick is used to avoid reuse
of the buffer before the constructed string has been seen: successive calls
to |code_type| alternatingly use to different buffers, with the |static|
variable |alt| keeping track of which one is most recently used.

@c
char* code_type(objtype typeno)
{ static char buff[2][8]; static int alt=0; /* alternates between calls */
@)
  char* p=buff[alt=1-alt], *s;
  if ((typeno&FUN)!=0) {@; strcpy(p,"fun "); typeno&=~FUN; p+=4; }
  if ((s=code_obj_type(typeno&TYPEMASK))!=NULL) strcpy(p,s);
  else if ((s=code_obj_type(typeno&(TYPEMASK&~ARR)))!=NULL)
    {@; strcpy(p,s); p[2]='l'; }
  else strcpy(p,"???");
  return buff[alt];
}

@ Possible classes of an object.

@< Public |typedef|... @>=
enum
{ VALUE, OPERATOR, FUNCTION, DUMMY, MAP, FUNCTION_COPIED, ARG, BLOCK=8 };

@ Here is a table for exchanging between interal and external
representations of class names, similar to |type_name| for types.

@< Static var... @>=
static struct {@; objclass num; char* str;} class_name [] =
{ { VALUE, "value" }
, { OPERATOR, "operator" }
, { FUNCTION, "function"  }
, { FUNCTION_COPIED, "function_copied" }
, { MAP, "map" }
, { DUMMY, "dum" }
, { ARG, "arg" }
, { BLOCK, "block" }
};

@ Translation of class codes in printable names.
@c
char* code_class(int classno)
{ static char buff[sizeof("function_copied")+1];
  int i;
  for (i=0; i<array_size(class_name); ++i)
    if (classno==class_name[i].num)
    return sprintf(buff,"%s:",class_name[i].str),buff;
  return "sys error:";
}

@* Moving and changing nodes.
@< Prototypes of globally visible fun... @>=
symblst copynode(symblst t),copytree(symblst s);

@~The function |copy_node| makes a copy of a node into a newly allocated
node, taking account of sharing of any value referenced. The link fields
|next| and |arglst| are not copied.

@c
symblst copynode(symblst t)
{
  symblst s;
  if (tree_pt >= maxnodes) full_error;
  s=tree_ar+tree_pt++;
  assignsym(s, t);
  if (s->class==VALUE)
  { if (!isshared(t->data.val)) t->data.val=NULL;
    setshared(s->data.val);
  }
  s->a.label=t->a.label;
  s->formal=t->formal;
  return s;
}

@ The function |copytree| is a straightforward recursive version
of~|copynode|.

@c
symblst copytree(symblst s)
{
  symblst r;
  if (s==NULL) return NULL;
  r=copynode(s);
  r->arglst=copytree(s->arglst);
  r->next=copytree(s->next);
  if (s->class==BLOCK) r->data.expr=copytree(s->data.expr);
  return r;
}

@ When a function is about to be called, its function body, which is stored
separately, is copied into |tree_ar| for evaluation, which is done by
|load_function_body|.

@< Prototypes of globally visible fun... @>=
symblst load_function_body(funclst func);

@~
Since the tree has many internal pointers in the
|next|, |arglst| and possibly |data.expr| fields, these pointers have to be
adjusted to remain inside the copy.

@d relocate(p,src_base,dst_base) (p=&(dst_base)[(p)-(src_base)])

@c
symblst load_function_body(funclst func)
{ int i,n_nodes=func->n_nodes;
  symbrec* result=func->top_expr;
  symblst src=func->start_nodes, dst=&tree_ar[tree_pt];
  if ((tree_pt+=n_nodes)>maxnodes) full_error;
  for (i=0; i<n_nodes; ++i)
  { symblst node=&dst[i]; @+ *node=src[i]; /* copy entire node */
    if (node->arglst!=NULL) relocate(node->arglst,src,dst);
    if (node->next!=NULL) relocate(node->next,src,dst);
    if (node->class==BLOCK) relocate(node->data.expr,src,dst);
    else if (node->class==VALUE) setshared(node->data.val);
  }
  relocate(result,src,dst);
  return result;
}

@ Here is the corresponding function that stores away a function body,
i.e., copies it from |tree_ar| into separately allocated memory; it is used
when a function definition is processed.

@< Prototypes of globally visible fun... @>=
funclst store_function_body(symblst arg,symblst expr);

@~
@c
funclst store_function_body(symblst arg,symblst expr)
{ funclst result; int i;
  @< Allocate memory for function body @>
  @< Initialise the function body @>
  @< Copy the labels @>
  @< Copy the nodes @>
  return result;
}

@ The function body is allocated in a similar way to objects, with the fixed
size part first, followed by the variable size parts, hoping that alignment
will cause no problems. The use of |allocmem| instead of |malloc| means that
the function body is seen by the garbage collector, and memory will be
reclaimed when the body is no longer accessible.
 
@< Allocate memory for function body @>=
result=(funclst)
allocmem(sizeof(funcrec)+tree_pt * sizeof(symbrec)+label_pt*sizeof(labelrec));

@ Here the allocated memory is partitioned. There are a total of |tree_pt|
nodes, including the arguments, and |label_pt| labels; the division of the
nodes into arguments and other nodes is determined by |arg|. The field
|top_arg| is set to the final argument, if any.


@< Initialise the function body @>=
{ result->start_args=(symbrec*) &result[1];
    /* arguments and nodes follow |funcrec| */
  result->start_labels=(labeltp) &result->start_args[tree_pt];
    /* labels follow nodes */
  result->n_labels=label_pt; /* all current labels are in this function */
  result->n_args= arg==NULL ? 0 : (int)(arg-tree_ar)+1;
    /* |*arg| is last argument, unless |arg==NULL| */
  result->start_nodes= &result->start_args[result->n_args];
    /* nodes follow the arguments */
  result->top_arg= 
    result->n_args==0 ? NULL : &result->start_args[result->n_args-1];
  result->n_nodes=tree_pt-result->n_args;
  result->top_expr=expr;
  relocate(result->top_expr,tree_ar,result->start_args);
}

    
@ Copying labels is simple, no relocation is required.

@< Copy the labels @>=
for (i=0; i<label_pt; ++i) result->start_labels[i]= label_ar[i];

@ Storing the function body is similar to what happens in
|load_function_body|, but in the opposite direction. Here however, we also
have to move the labels to the space reserved, so that their space in
|label_ar| can be liberated; in |load_function_body| there was no
corresponding move of labels since we don't mind that the |labelpt| pointers
in the loaded function body still point to the stored copies, since labels
are ``read only''items.

@< Copy the nodes @>=
{ symblst src=tree_ar, dst=result->start_args;
  for (i=0; i<tree_pt; ++i)
  { symbrec* node=&dst[i]; @+ *node=src[i];
    if (node->next!=NULL) relocate(node->next,src,dst);
    if (node->arglst!=NULL) relocate(node->arglst,src,dst);
    if (node->class==BLOCK) relocate(node->data.expr,src,dst);
    if (node->a.label!=label_null)
      relocate(node->a.label,label_ar,result->start_labels);
  }
}

@ When the current expression tree is to be marked for garbage collection,
the function |mark_tree| is called.

@< Prototypes of globally visible fun... @>=
void mark_tree(void);

@~It suffices to mark all nodes in |tree_ar|; nodes are moved in and out of
that array in a stacklike manner only at function entry and return, and
there is no fragmentation. It should be noted that the garbage collector is
not interested in the nodes themselves, only in the data values accessed
through those with |class==VALUE|.

@c
void mark_tree(void)
@+{@; mark_expression(tree_ar,tree_pt); }

@ When a node has been evaluated, i.e., its |class| changes to |VALUE|, the
function |assign_type_to_node| is called.

@< Prototypes of globally visible fun... @>=
void assign_type_to_node(symblst s);

@~Despite its name, |assign_type_to_node| mainly checks if the type of the
computed value correponds to the precomputed type.

@c
void assign_type_to_node(symblst s)
{
  object t=s->data.val;
  if (t==NULL && s->type== VOID) return;
  if (s->type==UNKNOWN || s->type==ARGTYPE
       || stop_loop!=NULL || stop_fun!=NULL)
  @/{@; s->type= t==NULL ? VOID : type_of(t); return; }
  if (t!=NULL) {
      if (s->type==type_of(t)) return;
      if (coerc_tab(type_of(t),s->type))
	{@; s->type=type_of(t); return; }
  }
  error("Type %s expected - type %s computed.\n",
  code_type(s->type), t?code_type(type_of(t)):"vid");
}

@
@< Prototypes of globally visible fun... @>=
void assignobject(symblst s,object t);
object force_integer(symblst topstack);

@~@c
void assignobject(symblst s,object t)
{@;
    s->data.val=t;
    s->class=VALUE;
    assign_type_to_node(s);
    s->a.label=label_null;
}

object force_integer(symblst topstack)
{
    object i_dex;
    object top_obj=topstack->data.val;
    if (!is_int(type_of(top_obj)))
	error("Integer expected - %s  encountered.\n",code_type(type_of(
		top_obj)));
    if (!isshared(top_obj) && type_of(top_obj)==INTEGER)
    @/{@;
	setshared(top_obj);
	return i_dex=top_obj;
    }
    if (type_of(top_obj)==BIGINT) i_dex=(object) (*bin2int)(&top_obj->b);
	else i_dex=cpobject(top_obj);
    setshared(i_dex);
    topstack->data.val=i_dex;
    assign_type_to_node(topstack);
    return i_dex;
}

@* Index.

