\input LiEwebmac

@* Setting runtime switches. 
This small module deals with global settings controlled by the user, such as
the sorting criterion used for ordering terms in polynomials. In this file the
auxiliary function definitions precede the declaration of some variables,
whose initialisers contain pointers to the former.

@h "lie.h"
@d local static
@f local static

@c
@< Local typedefs @>
@< Early variable declarations @>
@< Auxiliary function definitions @>
@< Late variable declarations @>
@< Globally visible functions @>

@ The following symbolic option values are used. They are negative to
distinguish them from explicit numeric values that the user might provide, for
instance by \.{on 8 lmargin}.

@( onoff.h @>=
#define PLUS -3
#define MINUS  -4
#define OFFCODE -2
#define ONCODE -1


@ When the parser sees a line starting with \.{on} or \.{off} followed by an
option, possibly with an intervening argument, it calls |set_on| respectively
|set_off| with a number for the argument, and a code identifying the option
(as if it were an identifier). Here's how this proceeds. where we use some
variables to be declared later. The basic idea is to look up the option in a
list |mode| which also contains the functions that modify that option, and to
call the appropriate function. Although not all such functions need it, they
are always supplied with the supplied number~|val|.
The return value of |set_on| and |set_off| indicates whether a valid option
was \emph{not} found.

@< Globally visible functions @>=
boolean set_on(entry val, strtype name)
{ int i = 0; size_t n = sizeof(mode)/sizeof(mode[0]);
  for (i=0; i<n; ++i) if (mode[i].a.name == name) break;
  if (i==n) return true; /* not found */
  *mode[i].state = val;
  if (mode[i].fnc_on!=NULL) mode[i].fnc_on(val);
  else error("Cannot switch %s on.\n",name_tab[name]);
  return false;
}

boolean set_off(entry val, strtype name)
{ int i = 0; size_t n = sizeof(mode)/sizeof(mode[0]);
  for (i=0; i<n; ++i) if (mode[i].a.name == name) break;
  if (i==n) return true; /* not found */
  *mode[i].state = val;
  if (mode[i].fnc_off!=NULL) mode[i].fnc_off(val);
  else error("Cannot switch %s off.\n",name_tab[name]);
  return false;
}

@ Here is the type |state_tp| of the state structure maintained, and the type
|mode_tp| of the of the elements of the array~|mode| used above.

@< Local typedefs @>=

typedef void (*switch_func_tp)(entry);
typedef struct
@/{@; int runtime, verbose, bigints, gc, prompt, monitor, lex, degree, height,
   lprint, lmargin; } state_tp;
typedef struct
 {@; atp a; switch_func_tp fnc_on, fnc_off; int* state;} mode_tp;

@ And here is how |state| and |mode| are defined and initialised. This code
should be fixed, since the global variables corresponding to |state| are
already initialised in \.{main.w}, and those values should not be overridden
(as they are in fact) by the values specified implicitly below.

@< Early variable declarations @>=
state_tp state=
{OFFCODE, OFFCODE, ONCODE,ONCODE,ONCODE,OFFCODE,PLUS,OFFCODE,OFFCODE,ONCODE,OFFCODE};

cmpfn_tp cmpfn; /* selected sorting criterion */

@ Here is where we have to skip past the definitions of some functions.

@< Late variable declarations @>=

local mode_tp mode[] =@/
{
@/ {{"runtime"},runtime_on, runtime_off, &state.runtime},
@/ {{"verbose"},verbose_on, verbose_off, &state.verbose},
@/ {{"bigints"},bigint_set_on, bigint_set_off, &state.bigints},
@/ {{"gc"}, gc_set_on, gc_set_off, &state.gc},
@/ {{"prompt"},prompt_on,prompt_off, &state.prompt},
@/ {{"monitor"},monitor_on,monitor_off, &state.monitor},
@/ {{"lex"},cmpfn_lex_on, (switch_func_tp) NULL, &state.lex},
@/ {{"degree"},cmpfn_deg_on, (switch_func_tp) NULL, &state.degree},
@/ {{"height"},cmpfn_height_on, (switch_func_tp) NULL, &state.height},
@/ {{"lprint"},lprint_on,lprint_off, &state.lprint},
@/ {{"lmargin"},lmargin_on,lmargin_off, &state.lmargin}
};


@ Here are some pairs of functions controlling the variables |prompt|,
|lprint|, |gc_set|, and~|lmargin|, all declared in \.{main.w}. This is a bit
silly, since these variables always match the corresponding fields of~|state|
(except that for instance |prompt==true| correspondes to
|state.prompt==ONCODE|); their use in other places should be replaced by those
fields.

@< Auxiliary function... @>=

local void prompt_on(entry unused) @+{@; prompt = true; }
local void prompt_off(entry unused) @+{@; prompt = false; }

local void lprint_on(entry unused) @+{@; lprint = true; }
local void lprint_off(entry unused) @+{@; lprint = false; }

local void gc_set_on(entry unused)  @+{@; gc_set = true; }
local void gc_set_off(entry unused) @+{@; gc_set = false; }

local void lmargin_on(entry argument)  @+{@; lmargin = argument; }
local void lmargin_off(entry unused) @+{@; lmargin = LMARGIN; }

@ The control of |am_monitor|, also declared in \.{main.w}, is slightly less
trivial, since we actually open the file whose name is held in~|monfil|, or
close it again. Calling \.{on runtime} will also start clocking the CPU-time.

@< Auxiliary function... @>=

local void monitor_on(entry unused)
{ if (!am_monitor)
   if ((monfile = fopen(monfil,"a"))==NULL)
     error("Cannot open %s.\n",monfil);
   else am_monitor=true;
}

local void monitor_off(entry unused)
{ if (am_monitor)
  { fprintf(monfile,"\n"); /* make sure last line has ended */
    fclose(monfile); monfile = NULL;
    am_monitor=false;
  }
}

local void runtime_on(entry unused) @+
{@; if (!runtime) registrate_cpu(), runtime=true; }

local void runtime_off(entry unused) @+
{@; if (runtime)  print_runtime(), runtime = false; }

local void verbose_on(entry unused) @+ {@; verbose=true; }

local void verbose_off(entry unused) @+ {@; verbose = false; }

@ The function |unmark_sorted| clears all the |sorted| flags present in matrix
and polynomial type objects, which is necessary when the sorting criterion is
changed.

@< Auxiliary function... @>=

local void unsort(object x) @+
{@; if (type_of(x)==MATRIX || type_of(x)==POLY) clrsorted(x); }
void unmark_sorted(void) @+
{@; for_all_objects(unsort); }

@ These functions maintain three fields of |state| simultaneously, whose value
is one of |OFFCODE|, |PLUS|, |MINUS|.

@< Auxiliary function... @>=

local void cmpfn_deg_on(entry argument)
{ cmpfn = argument==MINUS ?  deg_decr : (argument=PLUS,deg_incr) ;
@/state.degree=argument; state.lex=OFFCODE; state.height=OFFCODE;
  unmark_sorted();
}

local void cmpfn_lex_on(entry argument)
{ cmpfn = argument==MINUS ?  lex_decr : (argument=PLUS,lex_incr) ;
@/state.degree=OFFCODE; state.lex=argument; state.height=OFFCODE;
  unmark_sorted();
}

local void cmpfn_height_on(entry argument)
{ cmpfn = argument==MINUS ? height_decr : (argument=PLUS,height_incr) ;
@/state.degree=OFFCODE; state.lex=OFFCODE; state.height=argument; 
  unmark_sorted();
}

@ The function |rebuild_symbol_tab| links all occurrences of any identifier
into a list starting at the index of |symbol_tab| corresponding to that
identifier (via |match|); it does this by traversing all definitions contained
in the $\\{static}_i$ tables in reverse order. The function is called
initially, and after switching the state of \.{bigints}, since this enables a
different set of integer related built-in functions.

@< Auxiliary function definitions @>=
local void rechain_symbols(symbrec s[],int n)
{ int i; @+
  for (i=n-1;i>=0;i-=s[i].formal)
  @/{@; s[i].next = symbol_tab[s[i].a.name];
    symbol_tab[s[i].a.name]=&s[i];
  }
}

local void rebuild_symbol_tab(void)
{ static boolean first = true; @+
    if (!first) clear_symbol_tab();@+ else first = false;
  rechain_symbols((symblst)static3,nstatic3);
  rechain_symbols((symblst)static4,nstatic4);
  rechain_symbols((symblst)static5,nstatic5);
  rechain_symbols((symblst)static2,nstatic2);
  rechain_symbols((symblst)static1,nstatic1);
  if (bigint_set) rechain_symbols((symblst)static7,nstatic7);
  else rechain_symbols((symblst)static6,nstatic6);
}

@ Initially \.{bigints} are set to \.{on}, and |rebuild_symbol_tab| is called.

@< Auxiliary function... @>=

local void bigint_set_on(entry unused)
{ static boolean start = true;
  if (!bigint_set || start)
  {@; bigint_set = true;
    rebuild_symbol_tab();
    start = false;
  }
}

local void bigint_set_off(entry unused)
{@; if (bigint_set)
  {@; bigint_set = false;
    rebuild_symbol_tab();
  }
}


@

@d STACK_LEN 100

@< Early variable declarations @>=

local struct {@; int pt; state_tp s[STACK_LEN];} mstack; /* state stack */
local struct {@; int pt; object defaultgrp[STACK_LEN];} gstack;
 /* default group stack */

@ The functions |init_state| serves set any necessary variables according to
the initialisation of~|state|.

@< Globally visible... @>=
void init_state(void)
{ int i; size_t n = sizeof(mode)/sizeof(mode[0]);
  for (i=0; i<n; ++i)
  { mode[i].a.name = match(mode[i].a.init_string,false);
    if (*mode[i].state==OFFCODE)
    {@; if (mode[i].fnc_off!=NULL) mode[i].fnc_off(OFFCODE); }
    else {@; if (mode[i].fnc_on!=NULL) mode[i].fnc_on(*mode[i].state); }
  }
  mstack.pt = 0;
  gstack.pt = 0;
}

@ Saying just \.{on} or \.{off} will activate |print_state|.
@< Globally visible... @>=

void print_state(void)
{ int i; size_t n = sizeof mode/sizeof *mode;
  for (i=0; i<n; ++i)
  { if (*mode[i].state == OFFCODE) Printf("off\t");
    else if (*mode[i].state == ONCODE) Printf("on\t");
    else if (*mode[i].state == PLUS) Printf("on\t+\t");
    else if (*mode[i].state == MINUS) Printf("on\t-\t");
    else Printf("on\t%d\t",*mode[i].state);
    Printf("%s\n",name_tab[mode[i].a.name]); 
  }
}

@ Here's how we implement \.{savestate} and \.{restorestate}. It is currently
broken because a call like |set_on_off(OFFCODE,match("lex",false))| during
|restore_state| will fail with an error message. Got to find a better way.

@< Globally visible... @>=

void save_state(void)
{ if (mstack.pt >= STACK_LEN) error("Not possible to save more options.\n");
  mstack.s[mstack.pt++]=state;
  gstack.defaultgrp[gstack.pt++]=defaultgrp;
}

local void set_on_off(entry val,strtype name)
{@; if (val == OFFCODE) set_off(val, name); else set_on(val,name); }

void restore_state(void)
{ int i; size_t n=sizeof mode/sizeof *mode;
  if (mstack.pt==0) error("There are no saved options.\n");
  state=mstack.s[--mstack.pt];
  for (i=0; i<n; ++i) set_on_off(*mode[i].state,mode[i].a.name);
  defaultgrp=gstack.defaultgrp[--gstack.pt];
}


void mark_defaultgrp_stack(void)
{@; int i; for (i=0; i<gstack.pt; ++i) mark(gstack.defaultgrp[i]); }

@* Index.