\input LiEwebmac

@* Symbol tables.
In this compilation unit we treat matters associated the
association of values and expressions to identifiers, by which they behave as
variables or function names. The hash table used to encode identifiers as
small numbers is not defined here, but in the file \.{lexer.w}; it is called
|name_tab|.

@< Prototypes of globally visible functions and global variables @>=
void addsym(strtype name, symblst arg, symblst expr, int class);
void assignsym(symblst s, symblst t);
void push(symblst lst, symblst sym);
symblst push_value(symblst list);
symblst pop_value(symblst topoldsym);
void assignnames(symblst list, symblst name_list);
void add_dollar_names(symblst list);

@
@( sym.h @>=
@< Prototypes of globally visible functions and global variables @>@;
@< Public macro definitions @>@;

@ Here is the general layout of this compilation unit.

@h "lie.h"

@c
@<Search in symbol table@>@;
@<Add a new symbol to the symbol table@>@;


@* Tables of user definitions.
Values or expressions can be associated to identifiers in various ways:
many identifiers have one or more function definitions built into~\LiE., the
user can add new definitions to this collection, as well as introduce
variables by assigning values to identifiers, and there is a runtime stack
on which values of function parameters and local variables are kept. In each
of these cases the binding between an identifier and a value or expression
is held in a |symbrec| structure, somewhat confusingly the same structure
used to build up representations of programs. Here the identifier in the
binding is held in the |a| field for function definitions, or in the
|formal| field for bindings on the runtime stack; the associated expression
or value is in the |data.expr| or |data.val| field. These binding nodes are
arranged into lists through the |next| pointer. The permanent function
definitions for a name are reached through the |symbol_tab|
table (defined in \.{lexer.w}), the temporary ones (for all names)
through |top_definitions|, and the runtime stack is accessed
through~|topsym|.

@ For matching purposes certain types conversions are permissible: integers
and big integers are considered equivalent. The macro |coerc_tab(x,y)|
either returns a pointer to a function converting type~|x| to type~|y|,
or~|NULL| if the conversion is not possible. The macro |eq_types| can be
used instead of the equality operator~`|==|' in cases where these type
conversions are allowed.

@< Public macro... @>=

#define @! coerc_tab(t1,t2)   \
    (t1 == INTEGER && t2 == BIGINT ? (fobject)int2bin : \
     t1 == BIGINT  && t2 == INTEGER ? (fobject)bin2int : \
     (fobject) NULL)

#define @! eq_types(t1,t2) ((t1 == t2) || coerc_tab(t1,t2))


@ When a function definition is being looked up, we have to match the types
of the formal parameter list |pattern| against the types of the actual
arguments |actual|; this is done by the function~|match_arg|. Normally, a match
occurs if the lists have the same length and their types are pairswise equal.
There are a few exceptional cases:
\item{1)}
  An item in |pattern| with type |ARGTYPE| matches any type in |actual|.
\item{2)}
  An item in |actual| with type |UNKNOWN| matches any type in |pattern|.
\item{3)}
  The final item in |pattern| may have a type satisfying |isarrtype|, i.e.,
  one of |{INTLST, VECLST, MATLST, VIDLST, ARGLST}|; in this case the
  corresponding item in |actual| and any ones following it must match the
  base type of that array type, which is respectively |INTEGER|, |VECTOR|,
  |MATRIX|, |VOID|, or |ARGTYPE|.
\item{4)}
  If there is no exact match, but the actual argument can be coerced into
  the one specified by |pattern| according to |coerc_tab|, then this counts
  as a match.
\item{5)}
  If the parameter |deflt| to |match_arg| is |true|, then |actual| may be one
  item shorter than~|pattern|, provided the final item of |pattern| satisfies
  |isdefault| (this happens only if it is~|GRPDFT|).

@<Search in symbol table@>=
static boolean match_arg(symblst pattern, symblst actual, boolean deflt)
{ while (pattern!=NULL && actual!=NULL)
    if (   type_of(actual)!=type_of(pattern)
	&& type_of(pattern)!=ARGTYPE
	&& type_of(actual)!=UNKNOWN
	&& !coerc_tab(type_of(actual),type_of(pattern))
       ) return false;
    else
    {@; actual=actual->next; if (!isarrtype(pattern)) pattern=pattern->next; }
  return actual==NULL
      && (pattern==NULL || isarrtype(pattern) || deflt&&isdefault(pattern));
}

@ The function |srchsym| is the one that actually calls |match_arg|; it
traverses a list~|table| of bindings, matching it against a name and a list
of actual arguments, and returns the first one for which |match_arg| holds, or
|NULL| if no match is found. There are various possibilities for~|table|:
it can be the set of built-in bindings for an identifier from the hash table
(in this case the |name| field will always match), or the set
|top_definitions| of user defined bindings, or the runtime stack~|topsym|.
In the last case if the end of the list is reached without a match,
searching continues in |top_definitions|. Since it may be relevant whether a
match occurred on the stack, a global variable |isargument| is set when this
is the case, and cleared otherwise. If no match in the |table| is found when
|match_arg| is called with |deflt==false|, it is traversed again, calling
|match_arg| with |deflt==true| this time, so that matches with omitted default
group may be found.

@< Prototypes... @>=
symblst srchsym(symblst lst, strtype name, symblst arg);

@~@<Search in symbol table@>=

symblst srchsym(symblst table, strtype name, symblst arg)
{ symblst s;
  if (table==topsym)
    /* then search stack first, and user definitions afterwards */
  { for (s=table ; s!=NULL; s=s->next)
      if (s->formal==name) return isargument=true, s;
      table=top_definitions;
  }

  for (s=table ; s!=NULL; s=s->next)
    if (s->a.name==name && match_arg(s->arglst, arg, false))
      return isargument=false, s;
  for (s=table ; s!=NULL; s=s->next)
    if (s->a.name==name && match_arg(s->arglst, arg, true))
      return isargument=false, s;
  return isargument=false,s; /* |s==NULL|, not found anywhere */
}

@ The simple function |creatsym| builds a blank |symbrec| for storing a
binding for an identifier~|name|; the node is allocated by |allocmem|, and
will therefore be reclaimed by the garbage collector when no longer
accesible.

@< Prototypes... @>=
symbrec* creatsym(strtype a);

@~@c
symbrec* creatsym(strtype a)
{ symbrec* s = (symbrec*) allocmem(sizeof(symbrec));
  *s=blank_symbol; s->a.name = a;
  return s;
}

@ The function |addsym| is called by the parser either for the outer level
expression to be evaluated, or for a new function definition. In the former
case |name| corresponds to |"$"|, |arg==NULL|, |class==VALUE|, and |expr| is
the expression entered; in the latter case |name| gives the name of the
function defined, |arg| is its formal argument list, |expr| its body, and
|class==FUNCTION|. In either case the binding is recorded either in a fresh
|symbrec| created by |creatsym|, or by reusing the node for an existing
matching binding if there is one. When a new value is bound to `\.\$', the
reference count for the old value is removed by |clrshared|, but for the new
value nothing has to be done because the reference from the runtime stack is
replaced by the reference from the symbol~`\.\$'.

@<Add a new symbol to the symbol table@>=
void
addsym(strtype name, symblst arg, symblst expr, int class)
{ symblst s;
  @< Check that no built-in definition is overwritten @>
  s=srchsym(top_definitions, name, arg);
    /* if an old binding matches, overwrite it */
  if (s==NULL)
  @/{@; s=creatsym(name);
    s->next=top_definitions->next;
    top_definitions->next=s;
  }
  s->class=class;
  if (class==FUNCTION)
  { s->data.func=store_function_body(arg, expr);
    s->arglst=s->data.func->top_arg;
    s->type=UNKNOWN;
  }
  else if (class==VALUE) {@; clrshared(s->data.val); assignsym(s,expr); }
}

@ Here is a function that prints a function name with list of argument types
as part of an error message.

@< Prototypes... @>=
void error_prototype(string name, symblst arg);
@~@c
void error_prototype(string name, symblst arg)
{ err_Printf("%s(",name);
  if (arg!=NULL)
    while (err_Printf("%3s",code_obj_type(type_of(arg))),(arg=arg->next)!=NULL)
      err_Printf(", ");
  err_Printf(")");
}

@ The list of built-in bindings for the identifier |name| is obtained via
|symbol_tab[name]|; the new definition should not attempt to supersede
such a binding.

@< Check that no built-in definition is overwritten @>=
{ symblst ref=symbol_tab[name]; /* the permanant bindings */
  if (ref!=NULL && srchsym(ref,name,arg)!=NULL) /* one of them matches */
  { err_Printf("Sorry, there is a primitive definition for ");
    error_prototype(name_tab[name],arg);
    error("\nwhich you cannot redefine.\n");
  }
}

@* Stack handling.
The following 4 routines deal the transport of values, copying, pushing and
popping concerning the stack of symbols. The function |assignsym| copies
the content of |t| to |s|, except de fields |a|, |formal|, |next|, and
|arglst|. The function~|push| (which is only used by the dynamic function
loading macros $A1$--$A5$ defined in \.{lie.w}) places the new symbol after
the list head~|l|.

@c
void assignsym(symblst s, symblst t) /* |s=t| */
{@; s->class=t->class; s->type=t->type; s->data=t->data; }

void push(symblst l, symblst sym)
{@; sym->next = l->next;
   l->next=sym;
 }

symblst push_value(symblst list)
{ symblst* p=&list, q;
  for (q=list; q!=NULL; q=q->next) /* duplicate list */
  {@; *p=copynode(q); (*p)->arglst=q->arglst; p=&(*p)->next; }
  *p=topsym->next; return topsym->next=list; /* insert list after |topsym| */
}

symblst pop_value(symblst topoldsym)
  /* clear sharing in stack down to |topoldsym| */
{ symblst list;
  for (list=topsym->next; list != topoldsym; list=list->next)
    if (list->class==VALUE) clrshared(list->data.val);
  return topoldsym;
}

@* Assigning names (types) to entries of argument lists.  
The function |assignnames| assigns the names of a list |formal| of formal
function parameters to the list |actual| of actual arguments of a function
call; the names are inserted into the |formal| field of the actual
arguments. The list |actual| is either of the same length as |formal| or one
item shorter; in the latter case the list of actula arguments is extended by
the default group.

@c
void assignnames(symblst actual, symblst formal)
{ symblst* p;
  for (p=&actual->arglst; *p!=NULL; p=&(*p)->next, formal=formal->next)
    (*p)->formal=formal->a.label->name;
  if (formal!=NULL)
    @< Extend argument list at |*p| with a node for the default group @>
}

@ The defualt group is found by looking up the symbol with name
|"_gsetdefault"|, which is bound to a parameterless function returning the
default group. From the binding node the |type| and |class| fields are
copied by |assignsym|, and the |formal| field is taken from the final item
of the formal parameter list. The final assignment to |*p->data.expr| is a
trick that also occurs in \.{gettype}: the binding node is stored in the as
yet unused |data| field so that repeated searching at evaluation time will
not be necessary.

@< Extend argument list at |*p|... @>=
{ strtype name=match("_gsetdefault",false);
  symblst foundsym=srchsym(symbol_tab[name],name,NULL);
  *p=addnode(0,name),
  assignsym(*p,foundsym); (*p)->formal=formal->a.label->name; 
  (*p)->data.expr=foundsym;
}

@ The function |assigntype| is used by the parser to associate a type to a
number of consecutive formal parameters of a function (a sequence separated
by commas rather than semicolons). It assigns the type~|t| to all elements
of an argument list, except that if |t==GROUP|, then the final element gets
|GRPDFT| instead. The fact that such a final element might be followed by a
semicolon and further formal parameters appears to do little harm, since
|isdefault| tests that a default parameter is actually the final one in the
entire parameter list; this in turn raises doubts about the necessity of
distinguishing types |GRPDFT| and |GROUP| at all.

@< Prototypes... @>=
void assigntype(objtype t, symblst list);

@~
@c
void assigntype(objtype t, symblst list)
{
  for ( ;list!=NULL; list=list->next)
  @/{@; list->type= list->next==NULL && t==GROUP ? GRPDFT : t;
    list->class=DUMMY;
  }
}

@ Here is another function called by the parser; it serves to bind at most
10 actual arguments of a block to formal arguments called `\.{\$1}',
`\.{\$2}',~\dots.

@c
void add_dollar_names(symblst list)
{ static char *name[]=
    {"$1","$2","$3","$4","$5","$6","$7","$8","$9","$10"};
  int i;
  for (i=0; i<10 && list!=NULL; ++i,list=list->next)
    list->formal=match(name[i],false);
  if (list!=NULL) error("At most 10 arguments to a block allowed.\n");
}

@ The function |error_not_foundsym| reports that a function or operator
application could not be found in the various binding tables. It tries to
analyse the particular situation at hand in order to give a meaningful
indication of the problem. We only call the identifier undefined if no
definition of the identifier exists at all.

@< Prototypes... @>=
void error_not_foundsym(symblst s);

@~@c
void error_not_foundsym(symblst s)
{ strtype id=s->a.label->name; char* name = name_tab[id];
  label->line = s->a.label->line; /* adjust error indication */
  @< Give special message in case of programming language construct @>
  if (name[0] == '.') ++name; /* skip |'.'| */
  @< If a definition for |id| exists, report type mismatch error @>
  error("Identifier %s is not defined. \n",name);
}

@ Many programming language constructs are implemented using function-like
operators, often overloaded, with names like |"_select"| (that particular
one selects a subvalue from a vector, matrix of polynomial). In order to
hide these internal names from the user, we trap the normal error message
for mismatched function arguments for these special cases, giving a more
understandable diagnostic.

@< Give special message in case of programming language construct @>=
{ if (strcmp(name, "_select")==0)
    error("Impossible to select component from %s.\n"
	 , code_type(s->arglst->type));
  if (strcmp(name, ".strucval")==0) error("Array has inconsistent types.\n");
}

@ If an identifier was not found with the given set of arguments, but
another definition of the identifier does exist, we indicate it here; we
give a meaningful message in case of a function used as a variable, but do
not give a separate message for a variable used as a function, because
bindings as variable and function may coexist, and functions may be
parameterless.

@< If a definition for |id| exists, report type mismatch error @>=
{ symblst p;
  for (p=top_definitions; p!=NULL; p=p->next) @+
    if (p->a.name==id) goto found;
  for (p=symbol_tab[id]; p!=NULL; p=p->next) @+
    if (p->a.name==id) goto found;
  if (false) /* i.e., |if (p!=NULL)|; only do next after |goto found|: */
  { found:
    if (s->arglst==NULL) err_Printf("Arguments required for %s.\n",name);
    else
    { err_Printf("Argument types do not match in call. Types are: ");
      error_prototype(name,s->arglst); err_Printf(".\n");
    } 
    err_Printf("Valid argument types are for instance: ");
    error_prototype(name,p->arglst); error(".\n");
  }
}

@*Index.
